/* 
 * files.c -- DHCP server file manipulation *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 */
 //mcli add
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include "arpping.h"
#include "socket.h"
#include "options.h"
#include "packet.h"
#include "serverpacket.h"
#include "pidfile.h"
//mcli add

#include <stdio.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>

#include "debug.h"
#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "leases.h"

//mcli add
#include <signal.h> 
static void exit_server(int retval)
{
	pidfile_delete(server_config.pidfile);
	CLOSE_LOG();
//	printf("in exit_server \n");
	exit(retval);
}
//mcli add

/* on these functions, make sure you datatype matches */
static int read_ip(char *line, void *arg)
{
	struct in_addr *addr = arg;
	//struct in_addr  tempaddr;
	
//	char buf[20];
//	memset(buf,0,20);
	
	//inet_aton(line, &tempaddr);
//	memcpy(buf,line,strlen(line));
//	buf[(strlen(buf)-1)]='\0';
	inet_aton(line, arg);
	
	//memcpy((unsigned char*)arg,(unsigned char*)&tempaddr,4);//mcli
		
	return 1;
}


static int read_str(char *line, void *arg)
{
	char **dest = arg;
	int i;
	
	if (*dest) free(*dest);
	*dest = strdup(line);
	
	/* elimate trailing whitespace */
	for (i = strlen(*dest) - 1; i > 0 && isspace((*dest)[i]); i--);
	(*dest)[i > 0 ? i + 1 : 0] = '\0';
	return 1;
}


static int read_u32(char *line, void *arg)
{
	u_int32_t *dest = arg;
	*dest = strtoul(line, NULL, 0);
	return 1;
}


static int read_yn(char *line, void *arg)
{
	char *dest = arg;
	if (!strcasecmp("yes", line) || !strcmp("1", line) || !strcasecmp("true", line))
		*dest = 1;
	else if (!strcasecmp("no", line) || !strcmp("0", line) || !strcasecmp("false", line))
		*dest = 0;
	else return 0;
	
	return 1;
}


/* read a dhcp option and add it to opt_list */
static int read_opt(char *line, void *arg)
{
	struct option_set **opt_list = arg;
	char *opt, *val;
	char fail;
	struct dhcp_option *option = NULL;
	int length = 0;
	char buffer[255];
	u_int16_t result_u16;
	int16_t result_s16;
	u_int32_t result_u32;
	int32_t result_s32;
	
	int i;
	
	
//	iprintf("in read_opt ssssssssssssssssssssss\n");
	
	if (!(opt = strtok(line, " \t=")))
	{
//		iprintf("11111111111111\n");
		return 0;
	}
	
	for (i = 0; options[i].code; i++)
	{
		if (!strcmp(options[i].name, opt)) 
		{
			option = &(options[i]);
			break;
		}
	}
		
	if (!option) 
		return 0;
	
	do {
		val = strtok(NULL, ", \t");

		if (val) 
		{
			fail = 0;
			length = 0;
			switch (option->flags & TYPE_MASK) {
			case OPTION_IP:
				read_ip(val, buffer);
//				iprintf("break 1\n");
				break;
			case OPTION_IP_PAIR:
				read_ip(val, buffer);
				if ((val = strtok(NULL, ", \t/-")))
					read_ip(val, buffer + 4);
				else fail = 1;
//				iprintf("break 2\n");
				break;
			case OPTION_STRING:
				length = strlen(val);
				if (length > 254) length = 254;
				memcpy(buffer, val, length);
//				iprintf("break 3\n");
				break;
			case OPTION_BOOLEAN:
				if (!read_yn(val, buffer)) fail = 1;
//				iprintf("break 4\n");
				break;
			case OPTION_U8:
				buffer[0] = strtoul(val, NULL, 0);
//				iprintf("break 5\n");
				break;
			case OPTION_U16:
				result_u16 = htons(strtoul(val, NULL, 0));
				memcpy(buffer, &result_u16, 2);
//				iprintf("break 6\n");
				break;
			case OPTION_S16:
				result_s16 = htons(strtol(val, NULL, 0));
				memcpy(buffer, &result_s16, 2);
//				iprintf("break 7\n");
				break;
			case OPTION_U32:
				result_u32 = htonl(strtoul(val, NULL, 0));
				memcpy(buffer, &result_u32, 4);
//				iprintf("break 8\n");
				break;
			case OPTION_S32:
				result_s32 = htonl(strtol(val, NULL, 0));	
				memcpy(buffer, &result_s32, 4);
//				iprintf("break 9\n");
				break;
			default:
//				iprintf("break a\n");
				break;
			}
			length += option_lengths[option->flags & TYPE_MASK];
			if (!fail)
				attach_option(opt_list, option, buffer, length);
		} 
		else 
			fail = 1;
	} while (!fail && option->flags & OPTION_LIST);
	
	return 1;
}


static struct config_keyword keywords[] = {
	/* keyword	handler   variable address		default */
	{"start",	read_ip,  &(server_config.start),	"10.1.1.100"},
	{"end",		read_ip,  &(server_config.end),		"10.1.1.149"},
	{"interface",	read_str, &(server_config.interface),	"eth0"},
	{"option",	read_opt, &(server_config.options),	""},
	{"opt",		read_opt, &(server_config.options),	""},
	{"max_leases",	read_u32, &(server_config.max_leases),	"254"},
	{"remaining",	read_yn,  &(server_config.remaining),	"yes"},
	{"auto_time",	read_u32, &(server_config.auto_time),	"2"},//mcli modify 7200 before
	{"decline_time",read_u32, &(server_config.decline_time),"3600"},
	{"conflict_time",read_u32,&(server_config.conflict_time),"3600"},
	{"offer_time",	read_u32, &(server_config.offer_time),	"60"},
	{"min_lease",	read_u32, &(server_config.min_lease),	"60"},
	{"lease_file",	read_str, &(server_config.lease_file),	"/swap/udhcpd.leases"},
	{"pidfile",	read_str, &(server_config.pidfile),	"/swap/udhcpd.pid"},
	{"notify_file", read_str, &(server_config.notify_file),	""},
	{"siaddr",	read_ip,  &(server_config.siaddr),	"0.0.0.0"},
	{"sname",	read_str, &(server_config.sname),	""},
	{"boot_file",	read_str, &(server_config.boot_file),	""},
	{"",		NULL, 	  NULL,				""}
};

int read_config(char *file)
{
	FILE *in;
	char buffer[80], *token, *line;
	int i;

//	printf("in read config 000\n");
	
	for (i = 0; strlen(keywords[i].keyword); i++)
	{
	//	iprintf("keywords[i].keyword=====%s\n",keywords[i].keyword);
		if (strlen(keywords[i].def))
			keywords[i].handler(keywords[i].def, keywords[i].var);
	}
	
	if (!(in = fopen(file, "r"))) {
		LOG(LOG_ERR, "unable to open config file: %s", file);
//		printf("unable to open config file: %s", file);
		return 0;
	}
	
	memset(buffer,0,80);
	
	while (fgets(buffer, 80, in)) {
		
		if (strchr(buffer, '\n')) *(strchr(buffer, '\n')) = '\0';
		
		if (strchr(buffer, '#')) *(strchr(buffer, '#')) = '\0';
		
		token = buffer + strspn(buffer, " \t");
		
		if (*token == '\0') continue;
		
		line = token + strcspn(token, " \t=");
		
		if (*line == '\0') continue;
		
		*line = '\0';
		line++;
		line = line + strspn(line, " \t=");
		
		if (*line == '\0') continue;
		
//		iprintf("token===%s\n",token);
//		iprintf("line===%s\n",line);
		
		for (i = 0; strlen(keywords[i].keyword); i++)
		{
		//	iprintf("keywords[i].keyword ===%s\n",keywords[i].keyword);
			if (!strcasecmp(token, keywords[i].keyword))
			{
		//		iprintf("token 222===%s\n",token);
		//		iprintf("line 222===%s\n",line);
		//		printf("keywords[i].var addr ===%x\n",(keywords[i].var));
			//	iprintf("keywords[i].keyword 222===%s\n",keywords[i].keyword);
				
				keywords[i].handler(line, keywords[i].var);
			}
		}
				
		memset(buffer,0,80);
			
	}
	
	
	fclose(in);
	
	return 1;
}

/* the dummy var is here so this can be a signal handler */
void write_leases(int dummy)
{
	FILE *fp;
	unsigned int i;
	char buf[255];
	time_t curr = time(0);
	unsigned long lease_time;
	
	dummy = 0;
	
//	printf("before write lease file \n");
	
	if (!(fp = fopen(server_config.lease_file, "w+"))) {
		LOG(LOG_ERR, "Unable to open %s for writing", server_config.lease_file);
		return;
	}
	
	for (i = 0; i < server_config.max_leases; i++) {
		if (leases[i].yiaddr != 0) {
			if (server_config.remaining) {
				if (lease_expired(&(leases[i])))
				{
					lease_time = 0;
					continue;
				}
				else lease_time = leases[i].expires - curr;
			} else lease_time = leases[i].expires;
			lease_time = htonl(lease_time);
			fwrite(leases[i].chaddr, 16, 1, fp);
			fwrite(&(leases[i].yiaddr), 4, 1, fp);
			fwrite(&lease_time, 4, 1, fp);
		}
	}
	fclose(fp);
	
	if (server_config.notify_file) {
		sprintf(buf, "%s %s", server_config.notify_file, server_config.lease_file);
		system(buf);
	}
}


void read_leases(char *file)
{
	FILE *fp;
	unsigned int i = 0;
	time_t curr = time(0);
	struct dhcpOfferedAddr lease;
	
	//iprintf("before read  lease file \n");
	
	if (!(fp = fopen(file, "r"))) {
//mcli		LOG(LOG_ERR, "Unable to open %s for reading", file);
		return;
	}
	
	while (i < server_config.max_leases && (fread(&lease, sizeof lease, 1, fp) == 1)) {
		if (lease.yiaddr >= server_config.start && lease.yiaddr <= server_config.end) {
			leases[i].yiaddr = lease.yiaddr;
			leases[i].expires = ntohl(lease.expires);	
			if (server_config.remaining) leases[i].expires += curr;
			memcpy(leases[i].chaddr, lease.chaddr, sizeof(lease.chaddr));
			i++;
		}
	}
	
	DEBUG(LOG_INFO, "Read %d leases", i);
	
	if (i == server_config.max_leases) {
		if (fgetc(fp) == EOF)
			/* might be helpfull to drop expired leases */
			LOG(LOG_WARNING, "Too many leases while loading %s\n", file);
	}
	fclose(fp);
}

void reconfig()
{
	
	FILE *in;
	char buffer[80], *token, *line;
	int i=0;
	time_t curr = time(0);
	struct dhcpOfferedAddr lease;
	int fd = -1;
	struct ifreq ifr;
	unsigned long timeout_end;
	struct option_set *option;
	struct sockaddr_in *sin;

	if (!(in = fopen("/swap/udhcpd.leases", "w+")))
	{
		printf("Clear the leases file error!\n");
		return ;
	}
	
	fclose(in);
	
	clear_options(&server_config.options);
	
	//memset(&server_config, 0, sizeof(struct server_config_t));

	if (!(in = fopen("/swap/udhcpd.conf", "r"))) {
		printf("Unable to open the config file dhcpd.conf!\n");
		return 0;
	}
	
	memset(buffer,0,80);
	
	while (fgets(buffer, 80, in)) {
		
		if (strchr(buffer, '\n')) *(strchr(buffer, '\n')) = '\0';
		
		if (strchr(buffer, '#')) *(strchr(buffer, '#')) = '\0';
		
		token = buffer + strspn(buffer, " \t");
		
		if (*token == '\0') continue;
		
		line = token + strcspn(token, " \t=");
		
		if (*line == '\0') continue;
		
		*line = '\0';
		line++;
		line = line + strspn(line, " \t=");
		
		if (*line == '\0') continue;
		
		//iprintf("token===%s\n",token);
		//iprintf("line===%s\n",line);
		
		for (i = 0; strlen(keywords[i].keyword); i++)
		{
		//	iprintf("keywords[i].keyword ===%s\n",keywords[i].keyword);
			if (!strcasecmp(token, keywords[i].keyword))
			{
		//		iprintf("token 222===%s\n",token);
		//		iprintf("line 222===%s\n",line);
		//		printf("keywords[i].var addr ===%x\n",(keywords[i].var));
			//	iprintf("keywords[i].keyword 222===%s\n",keywords[i].keyword);
				
				keywords[i].handler(line, keywords[i].var);
			}
		}
				
		memset(buffer,0,80);
			
	}
	fclose(in);
	
	if ((option = find_option(server_config.options, DHCP_LEASE_TIME))) {
		memcpy(&server_config.lease, option->data + 2, 4);
		server_config.lease = ntohl(server_config.lease);
	//	printf("server_config.lease====%d\n",server_config.lease);
	}
	else server_config.lease = LEASE_TIME;
	
	leases = malloc(sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
	
	if(leases==NULL)
		printf(" malloc error\n");                                                                                                                                                                                                                                                                                                                       
		
	memset(leases, 0, sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
	
//	printf("server_config.lease_file====%s\n",server_config.lease_file);
	
	read_leases(server_config.lease_file);

	if((fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) >= 0) {
		ifr.ifr_addr.sa_family = AF_INET;
		strcpy(ifr.ifr_name, server_config.interface);
		if (ioctl(fd, SIOCGIFADDR, &ifr) == 0) {
			sin = (struct sockaddr_in *) &ifr.ifr_addr;
			server_config.server = sin->sin_addr.s_addr;
			DEBUG(LOG_INFO, "%s (server_ip) = %s", ifr.ifr_name, inet_ntoa(sin->sin_addr));
		} else {
			LOG(LOG_ERR, "SIOCGIFADDR failed!");
			exit_server(1);
		}
		if (ioctl(fd, SIOCGIFINDEX, &ifr) == 0) {
			DEBUG(LOG_INFO, "adapter index %d", ifr.ifr_ifindex);
			server_config.ifindex = ifr.ifr_ifindex;
		} else {
			LOG(LOG_ERR, "SIOCGIFINDEX failed!");
			exit_server(1);
		}
		if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0) {
			memcpy(server_config.arp, ifr.ifr_hwaddr.sa_data, 6);
			DEBUG(LOG_INFO, "adapter hardware address %02x:%02x:%02x:%02x:%02x:%02x",
				server_config.arp[0], server_config.arp[1], server_config.arp[2], 
				server_config.arp[3], server_config.arp[4], server_config.arp[5]);
		} else {
			LOG(LOG_ERR, "SIOCGIFHWADDR failed!");
			exit_server(1);
		}
	} else {
		LOG(LOG_ERR, "socket failed!");
		exit_server(1);
	}
	close(fd);

	timeout_end = time(0) + server_config.auto_time;
	
}