/*
 * common.h
 *
 * This file contains definitions useful in all sorts of places.
 *
 * Copyright (C) 1998 Brad M. Garcia <garsh@home.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _DNRD_COMMON_H_
#define _DNRD_COMMON_H_

#include <netinet/in.h>
#include <syslog.h>
#include <semaphore.h>
#include "mcd.h"

#ifndef IFNAMSIZ
#define IFNAMSIZ	16
#endif

// #define DEFAULT_DNS_COPY	1	/* if we want to send a copy of dns query to default link ? */

/* Set the default timeout value for select in seconds */
#ifndef SELECT_TIMEOUT
#define SELECT_TIMEOUT 500000		//usec
#endif
#define RESETUP_SOCKET_INTERVAL	30

#define MAX_LINK           4		/* maximum number support link */
#define MAX_SERV	   2		/* maximum number of DNS servers */
#define MAX_FAIL_SWITCH	   2

#define MAX_DOMAIN_NUM	  50
#define MAX_DOMAIN_LEN	  64
#define MAX_PACKET_LEN		512

struct dnssrv_t {
    int                    sock[MAX_SERV];      /* for communication with server */
    struct sockaddr_in     addr[MAX_SERV];      /* IP address of server */
    char		   ifname[IFNAMSIZ+1]; /* dns server location interface */
    struct sockaddr_in	   saddr;	/* src addr to send query */
    char*                  domain;    /* optional domain to match.  Set to
					 zero for a default server */
    int			   fail_count[MAX_SERV];
    int			   cur_srv;
    int			   send_count;
};

extern const char*         version;   /* the version number for this program */
extern const char*         progname;  /* the name of this program */
extern int                 opt_debug; /* debugging option */
extern const char*         pid_file; /* File containing current daemon's PID */
extern int                 isock;     /* for communication with clients */
extern int                 tcpsock;   /* same as isock, but for tcp requests */
extern struct dnssrv_t     dns_srv[]; /* DNS server information struct */
extern int                 serv_act; /* index into dns_srv for active server */
extern int                 link_cnt;  /* number of DNS servers */
extern struct sockaddr_in  recv_addr; /* address on which we receive queries */
extern struct sockaddr_in  send_addr; /* address on which we receive queries */
extern uid_t               daemonuid; /* to switch to once daemonised */
extern gid_t               daemongid; /* to switch to once daemonised */
extern int                 select_timeout; /* select timeout in seconds */
extern int                 gotterminal;
extern char                master_config[256];
extern sem_t               dnrd_sem;  /* Used for all thread synchronization */
extern int		   dm_cnt;

/* kill any currently running copies of dnrd */
int kill_current();
#ifndef EMBED
/* print messages to stderr or syslog */
void log_msg(int type, const char *fmt, ...);

/* same, but only if debugging is turned on */
void log_debug(const char *fmt, ...);	
#else
#define log_msg(fmt,args...);		do{} while(0);
#define log_debug(fmt,args...);		do{} while(0);

int dump_dnspacket(char *type, unsigned char *packet, int len);
#endif

/* cleanup everything and exit */
void cleanexit(int status);

/* Reads in the domain name as a string, allocates space for the CNAME
   version of it */
char* make_cname(const char *text, const int maxlen);
//void sprintf_cname(const char *cname, int namesize, char *buf, int bufsize);

char *cname2asc(const char *cname); 
/* Dumping DNS packets */


#endif  /* _DNRD_COMMON_H_ */
