#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "policy_domain.h"
#include "common.h"

char	pattern_token[16][32];		/* maxium pattern: a.b.c.d.e.f.g.h.i.j.k.l.m.n.o.p123456789012345678901234567890 */
char	candidate_token[16][32];
char 	candidate_buf[64];
int  	cand_cnt = 0;
char 	pattern_buf[64];
char 	*p;
struct  policy_domain domainlist[MAX_DOMAIN_NUM];
int	dm_cnt = 0;

#define DOMAINLIST_MINCYCLE	5	/* seconds */

int dns_indx( char *ifname )
{
	int j;

	for( j=0; j<MAX_LINK; j++)
	{
		if( strcmp( &dns_srv[j].ifname[0], ifname ) == 0 )
			return j;
	}
	return 0; //default is link 0
}


/* DIFFERENCE:
 *
 *	kernel & dnrd use same "policy_domain.h" but dnrd use  'policy_domain.hash_id' as link index !!
 *
 */
int match_domain(struct policy_domain *pattern, const char *domain, int candidate_len )
{
	int i;
	
	/* 
	printk("cand=%s cnt=%d pattern=%s token=%d \n", domain, cand_cnt, pattern->domain, pattern->len );
	*/
	
	if( pattern->type == PD_TYPE_FQDN )
	{
		if( candidate_len != pattern->len ) return 0;

		if( strncmp(domain, pattern->domain, candidate_len) == 0 )
			return 1;
	}
	else /* token */
	{
		i = 0;
		strncpy(pattern_buf, pattern->domain, sizeof(pattern_buf)-1 );
		if( pattern_buf[0] == '.' )
		{
			pattern_token[i++][0] = '*'; /* GE type */
		}
		p = strtok( pattern_buf, "." );
		while( p )
		{
			if( p[0] ) strncpy( &pattern_token[ i ][0], p , sizeof(pattern_token[0]) );
			else pattern_token[ i ][0] = '*';
			
			i++;
			p = strtok( NULL, "." );
		}
		
		if( pattern->type == PD_TYPE_TOKEN_EQ )
		{
			if( cand_cnt != pattern->len )
				return 0;
			
			for( i=0; i< cand_cnt ; i++)
			{
				if( pattern_token[i][0] == '*' )
					continue;
				if( strcmp( pattern_token[i], candidate_token[i] ) != 0 )
					return 0;
			}
		}
		else if( pattern->type == PD_TYPE_TOKEN_GE )
		{
			int diff = 0;
			
			if( cand_cnt < pattern->len )
				return 0;
		
			diff = cand_cnt - pattern->len;
			
			for(i=0; i<pattern->len; i++)
			{
				if( pattern_token[i][0] == '*' )
					continue;
				if( strcmp( pattern_token[i], candidate_token[i+diff] ) != 0 )
					return 0;
			}
		}
	
	}
	
	return 1;
}

/* find [first] matched domain,   *.com match www.iss.com */

int find_match_domain(const char *domain)
{
	struct policy_domain *link=NULL;
	int i;
	
	if( domain == NULL || domain[0] == '\0' )
		return 0;

	cand_cnt = 0;
	
	//fprintf(stderr,"\n\ndnrd: searching %s\n\n", domain);
	
	strncpy(candidate_buf, domain, sizeof(candidate_buf)-1);
	p = strtok( candidate_buf, "." );
	while( p )
	{
		if( p[0] ) strncpy( &candidate_token[ cand_cnt ][0], p , sizeof(candidate_token[0]) );
		else candidate_token[ cand_cnt ][0] = '*';

		cand_cnt++;
		p = strtok( NULL, "." );
	}

	for(i=0; i<MAX_HASH_SIZE; i++)
	{
		link =  &domainlist[i] ;	
		if( link->domain[0] && match_domain( link, domain, strlen(domain) ) )
		{
			fprintf(stderr,"match link %d for %s \n", link->hash_id, domain );
			return link->hash_id;
		}
	}

	return 0;
}

/*
 *  ifname domain_pattern
 */
int read_domain_list(char *file)
{
	FILE *fr;
	char line[255];
	char domain[64];
	char *p;
	struct policy_domain *link;
	time_t         now;
	static time_t  lastexpire = 0;
	static long lastmodified = 0;
	static long lastsize = 0;
	int data_fd;
	struct stat statbuf;
	
	if (lastexpire == 0) lastexpire = time(NULL);

	now = time(NULL);
	if (now - lastexpire < DOMAINLIST_MINCYCLE)
	{
		return (0);
	}
	if( file==NULL || file[0]=='0')
		return -1;

	data_fd = open(file, O_RDONLY, 0);
	if (data_fd == -1) return -1;
	fstat(data_fd, &statbuf);
	close(data_fd);
	
	lastexpire = now;	/* check freq. > 10 sec */
	
	if( lastsize != statbuf.st_size ||
		lastmodified != statbuf.st_mtime )
	{
		lastsize = statbuf.st_size;
		lastmodified = statbuf.st_mtime;
	}
	else
	{
		return 0;
	}
	
	if( (fr=fopen(file,"rt"))==NULL)
	{
		fprintf(stderr,"error on open %s\n",file);
		return -1;
	}
	
	memset(&domainlist[0], 0, sizeof(domainlist));
	
	while( fgets(line,sizeof(line),fr) )
	{
		if( line[0] == '\r' || line[0] == '\n' ||
			line[0] == '#' || line[0] == '\'')
			continue;
		p = strtok(line,"\r\n\t ");
		if( p == NULL ) continue;
		
		strlcpy( &domainlist[dm_cnt].ifname[0], p, IFNAMSIZ );
		p = strtok(NULL,"\r\n\t ");
		if( p == NULL ) continue;
		strlcpy( domain, p, MAX_DOMAIN_LEN-1 );
		link = &domainlist[dm_cnt];

	/* the following lines is taken from kernel/linux/net/ipv4/netfilter_ipv4/policy_domain.c:add_domain() */
	
	link->hash_id = dns_indx( domainlist[dm_cnt].ifname );  /* we use hash_id as dns srv index */
	strncpy( &link->domain[0], domain, sizeof(link->domain)-1);
	link->len = strlen(domain);
	link->type = PD_TYPE_FQDN;
	
	if( ( domain[0] == '*' && strlen(domain)==1 ) ||  domain[1] == '\0' )
	{
		link->type = PD_TYPE_TOKEN_GE;
		link->len = 1;
	}
	else
	{	
		char *p;
		char buf[64];
		
		if( domain[0] == '.' )
		{
			link->len = 1;
			link->type = PD_TYPE_TOKEN_GE ;
		}
		else 
		{
			link->len = 0;
			link->type =  PD_TYPE_TOKEN_EQ;
		}
		strcpy(buf,domain);
		p = strtok( buf, "." );
		while(p)
		{
			link->len++;
			p = strtok( NULL, "." );
		}
	}

		dm_cnt++;
		fprintf(stderr,"dmlist: %s\t%s\n", &domainlist[dm_cnt-1].ifname[0],
			&domainlist[dm_cnt-1].domain[0] );
	}
	fclose(fr);

	return 0;
}
