#ifndef _POLICY_DOMAIN_H
#define _POLICY_DOMAIN_H

#define MAX_HASH_SIZE		16
#define MAX_ARECORD		256
#define MAX_POLICY_DOMAIN	30

enum PD_TYPE { PD_TYPE_FQDN=0, PD_TYPE_TOKEN_EQ, PD_TYPE_TOKEN_GE };

struct policy_domain
{
	unsigned char hash_id;
	char	domain[64];
	unsigned char	type;	/* PD_TYPE_FQDN, PD_TYPE_TOKEN_EQ, PD_TYPE_TOKE_GE */
	unsigned short	len;	/* len or token count */
#ifdef __KERNEL__
	struct arecord *addr_node;
	struct policy_domain *prev;
	struct policy_domain *next;
#else
	char	ifname[17];
#endif

};

struct arecord
{
	struct policy_domain *domain;
	unsigned long ip;
	unsigned long ttl;
	struct arecord *prev;
	struct arecord *next;
};

#define SO_POLICYDOMAIN 88

typedef struct {
	int op;
	int ans;
	char domain[64];
} policydomain_request;

/* NOTE: I deliberately break the first cut ippolicydomain utility. Nobody uses it. */

#define POLICYDOMAIN_BAD001		0x00000008

#define POLICYDOMAIN_CLEAN_DOMAIN	0x00000009	/* req.index, no arguments */
#define POLICYDOMAIN_MARK_DOMAIN	0x00000010	/* req.index, no arguments */
#define POLICYDOMAIN_FLUSH_DOMAIN	0x00000011	/* req.index, no arguments */
#define POLICYDOMAIN_ADD_DOMAIN		0x00000012	/* add addr to pool */
#define POLICYDOMAIN_DEL_DOMAIN		0x00000013	/* del addr from pool */
#define POLICYDOMAIN_LIST_DOMAIN	0x00000014	/* del addr from pool */
#define POLICYDOMAIN_LIST_ALL		0x00000015	/* del addr from pool */
#define POLICYDOMAIN_HIGH_NR		0x00000016	/* result in req.index */

#ifdef __KERNEL__

struct arecord *get_arecord_pool(void);
struct policy_domain *get_domain_pool(void);
unsigned char domain_hash(const char *domain);
int add_record(const char *domain,unsigned long ip, unsigned long ttl);
struct policy_domain* find_domain(const char *domain);
int find_ip(struct policy_domain *domain, unsigned long addr);
int clean_domain(void);
int match_domain_ip( const char *domain, unsigned long addr );
struct policy_domain* find_match_domain(const char *domain);

#endif

#endif
