#ifndef CONVERT_H
#define CONVERT_H

/* conversion table:
pc			return
"TCP"		PROTOCOL_TCP
"UDP"		PROTOCOL_UDP
"TCPUDP"	PROTOCOL_TCP_AND_UDP
other		PROTOCOL_NONE
*/
char GetIPProtocol(const char *pc);

/* conversion table:
cIPProtocol				return
PROTOCOL_TCP			"TCP"
PROTOCOL_UDP			"UDP"
PROTOCOL_TCP_AND_UDP	"TCPUDP"
other					""
*/
const char *GetIPProtocolString(char cIPProtocol);

/* conversion table:
pc					return
"RIP1"				ROUTING_RIP1
"RIP2"				ROUTING_RIP2
"RIP1_COMPATIBLE"	ROUTING_RIP1_COMPATIBLE
other				ROUTING_DISABLE
*/
char GetRoutingProtocol(const char *pc);

/* conversion table:
cRoutingProtocol		return
ROUTING_RIP1			"RIP1"
ROUTING_RIP2			"RIP2"
ROUTING_RIP1_COMPATIBLE	"RIP1_COMPATIBLE"
ROUTING_DISABLE			"Disable"
other					""
*/
const char *GetRoutingProtocolString(char cRoutingProtocol);

/* conversion table:
pc		return
"WAN"	NET_WAN
"LAN"	NET_LAN
other	NET_ERR
*/
char GetNetInterface(const char *pc);

/* conversion table:
cNetInterface	return
NET_WAN			"WAN"
NET_LAN			"LAN"
other			""
*/
const char *GetNetInterfaceString(char cNetInterface);

/* IP1 - IP2 */
long IPSub(unsigned long ulIP1, unsigned long ulIP2);
/* IP + AddValue */
unsigned long IPAdd(unsigned long ulIP, long lAddValue);

/* Modify IP according Mask & IPRef */
unsigned long IPMask(unsigned long ulIP, unsigned long ulIPRef, unsigned long ulMask);

#define __SWAP(type, a, b) \
{\
	type __swap_temp;\
	__swap_temp = (a);\
	(a) = (b);\
	(b) = __swap_temp;\
}


/* Constant definition. */

#ifdef CONVERT_C
#define G_CONST(varName, varValue) const char *varName = varValue
#else
#define G_CONST(varName, varValue) extern const char *varName
#endif


G_CONST(G_PC_NO_CACHE_HEADER, "Pragma: no-cache\r\nCache-Control: no-cache\r\nExpires: 01 Jan 1970 00:00:00 GMT\r\n");

G_CONST(G_PC_REDIRECT_URL, "RedirectUrl");
G_CONST(G_PC_REDIRECT_TIME_SHOW, "RedirectTimeShow");
G_CONST(G_PC_REDIRECT_TIME_DEFER, "RedirectTimeDefer");
G_CONST(G_PC_REDIRECT_INFO, "RedirectInfo");

G_CONST(G_PC_HOSTNAME, "HostName");
G_CONST(G_PC_USERNAME, "User");
G_CONST(G_PC_PASSWORD, "Pass");
G_CONST(G_PC_ENABLE, "Enable");
G_CONST(G_PC_TIMEOUT, "Timeout");

G_CONST(G_PC_STATE, "State");
G_CONST(G_PC_ID, "ID");
G_CONST(G_PC_TIME, "Time");

G_CONST(G_PC_REMOTEIP, "RemoteIP");
G_CONST(G_PC_IP, "IP");
G_CONST(G_PC_NETMASK, "Netmask");
G_CONST(G_PC_GATEWAY, "Gateway");

G_CONST(G_PC_ACTION, "Action");
G_CONST(G_PC_INDEX, "Index");

G_CONST(G_PC_ETH0, "eth0");
G_CONST(G_PC_ETH1, "eth1");
G_CONST(G_PC_PPP0, "ppp0");
G_CONST(G_PC_PPP1, "ppp1");
G_CONST(G_PC_PPP2, "ppp2");

#endif
