
#include "../CommonDef.h"

static pthread_mutex_t g_ptmFlash = PTHREAD_MUTEX_INITIALIZER;

#ifdef EMBED

#define BUF_SIZE (64*1024)
#define MIN_BIN_IMG_SIZE	500*1024

extern unsigned int g_model_id;

int WriteWinbondFlash_me(UINT32 dest, int len, char *src)

{

    int failed = 0;
    
    WriteWinbondFlash(dest, len, src);

    char *pc = (char *)malloc(len + 4);

    if (pc == NULL)
        fprintf(stderr, "no memory.\n");
    else    
    {
        ReadWinbondFlash(dest, len, pc);
        if (memcmp(pc, src, len) != 0)
	{
            printf("Flash write failed\n");
	    failed = 1;
        }
        free(pc);
    }

    return failed;

}


INT WriteFile2Image(tfooter * image_footer, UINT32 buffer, INT file, INT (*read_func)(INT, VOID *, UINT32),int *len )
{

	UINT32 blockSize,i,src,dest;
	unsigned long long sum;
	UINT32 k;
	UINT32 bytes;
	UINT32 footer_addr=0; //norton initialize it!!
	int failed = 0;

	
	if( (image_footer->base & (FLASH_BLOCK_SIZE-1)) )
	{
		fprintf(stderr,"WriteFile2Image err1\n");
		return -1;	// not block alignment
	}
	// check if word aligment
	if( (image_footer->length & 0x3) )
		image_footer->length=(image_footer->length&(~0x3))+4;

	// check write corrupt
	if( !CorruptCheck( image_footer ))
	{
		// Write program
		i=image_footer->length;
		dest=image_footer->base;
		sum=0;
		printf("image_footer->length=%d , FLASH_BLOCK_SIZE=%d \n",
			i, FLASH_BLOCK_SIZE );
			
		while(i)
		{
			blockSize=FLASH_BLOCK_SIZE;
			if((bytes=read_func(file, (VOID *)buffer, blockSize))<0 )
			{
				fprintf(stderr,"WriteFile2Image err2\n");
				return -1; // read file error
			}
			src=(UINT32)buffer;
			if( i <= blockSize )
			{
				failed += WriteWinbondFlash_me(dest,i,(UCHAR *)src);
				*len+=i;

				for(k=0;k<i/4;k++)
				{
					sum+=inpw(src+k);
				}
				sum = ~((sum&(-1LU))+(sum>>32));
				image_footer->image_checksum=(UINT32)sum; // checksum of image

				sum=0;
				for(k=0;k<sizeof(tfooter)/4-1;k++)
				{
					sum += *((UINT32 *)image_footer+k);
				}
				sum = ~((sum&(-1LU))+(sum>>32));
				image_footer->checksum=(UINT32)sum;		// checksum of footer

				if(((blockSize-i)<sizeof(tfooter)) ) // Create a new block for footer
				{
					//int i=0;
					
					fprintf(stderr,"WinbondFlash 1a write %d bytes at 0x%08lx\n",i,dest+2*blockSize-sizeof(tfooter));
					failed += WriteWinbondFlash_me(dest+2*blockSize-sizeof(tfooter),sizeof(tfooter),(UCHAR *)image_footer);
					fprintf(stderr,"WinbondFlash 1b write %d bytes at 0x%08lx\n",i,dest+2*blockSize-sizeof(tfooter));
					footer_addr=dest+2*blockSize-sizeof(tfooter);
				}
				else	// Create the footer at current block
				{
					//int i=0;
					fprintf(stderr,"WinbondFlash 2a write %d bytes at 0x%08lx\n",i,dest+blockSize-sizeof(tfooter));
					failed += WriteWinbondFlash_me(dest+blockSize-sizeof(tfooter),sizeof(tfooter),(UCHAR *)image_footer);
					fprintf(stderr,"WinbondFlash 2b write %d bytes at 0x%08lx\n",i,dest+blockSize-sizeof(tfooter));

					footer_addr=dest+blockSize-sizeof(tfooter);
				}
				*len+=sizeof(tfooter);

				blockSize=i; // Check if > a block size
			}
			else
			{
				failed += WriteWinbondFlash_me(dest,FLASH_BLOCK_SIZE,(UCHAR *)src);
				fprintf(stderr,"WinbondFlash 3 write %d bytes at 0x%08x\n",i,dest);
				*len+=blockSize;
				for(k=0;k<blockSize/4;k++)
				{
					sum+=inpw(src+k);
				}
			}
			dest+=blockSize;
			i-=blockSize;
		}
	}
	else
	{
		fprintf(stderr,"WriteFile2Image CorruptCheck error\n");
		return -1;
	}
	// image write finished, check image checksum

	sum=0;
	for(k=0;k<image_footer->length/4;k++)
	{
		sum+=inpw( ((image_footer->base)+k) );
	}

	sum = ~((sum&(-1LU))+(sum>>32));

#if 0
	if( ((UINT32)sum) != image_footer->image_checksum )
	{
		fprintf(stderr, "WriteFile2Image image_footer->image_checksum failed\n");
		//ignore this errr ? return -1;// image check sum error
	}
#endif
	// check footer check sum

	sum=0;
	for( i=footer_addr;i<footer_addr+sizeof(tfooter)-4;i+=4)
	{
		sum+=inpw( (i));
	}
	sum = ~((sum&(-1LU))+(sum>>32));

#if 0
	if( ((UINT32)sum) != image_footer->checksum )
	{
		fprintf(stderr,"WriteFile2Image checksum err\n");
		//ignore this error ?  return -1; // footer check sum error
	}
#endif

	return failed;
}

INT WriteFlashRom(tfooter * image_footer, UINT32 buffer, INT file, INT (*read_func)(INT, VOID *, UINT32),int *len )
{

	UINT32 blockSize,i,src,dest;
	unsigned long long sum;
	UINT32 k;
	UINT32 bytes;
	UINT32 footer_addr=0; //norton initialize it!!
	int failed = 0;

	
	if( (image_footer->base & (FLASH_BLOCK_SIZE-1)) )
	{
		fprintf(stderr,"WriteFile2Image err1\n");
		return -1;	// not block alignment
	}
	// check if word aligment
	if( (image_footer->length & 0x3) )
		image_footer->length=(image_footer->length&(~0x3))+4;

	// check write corrupt
	//if( !CorruptCheck( image_footer ))
	{
		// Write program
		i=image_footer->length;
		dest=image_footer->base;
		sum=0;
		printf("image_footer->length=%d , FLASH_BLOCK_SIZE=%d \n",
			i, FLASH_BLOCK_SIZE );
			
		while(i)
		{
			blockSize=FLASH_BLOCK_SIZE;
			if((bytes=read_func(file, (VOID *)buffer, blockSize))<0 )
			{
				fprintf(stderr,"WriteFile2Image err2\n");
				return -1; // read file error
			}
			src=(UINT32)buffer;
			if( i <= blockSize )
			{
				failed += WriteWinbondFlash_me(dest,i,(UCHAR *)src);
				*len+=i;

				for(k=0;k<i/4;k++)
				{
					sum+=inpw(src+k);
				}
				sum = ~((sum&(-1LU))+(sum>>32));
				image_footer->image_checksum=(UINT32)sum; // checksum of image

				sum=0;
				for(k=0;k<sizeof(tfooter)/4-1;k++)
				{
					sum += *((UINT32 *)image_footer+k);
				}
				sum = ~((sum&(-1LU))+(sum>>32));
				image_footer->checksum=(UINT32)sum;		// checksum of footer

				blockSize=i; // Check if > a block size
			}
			else
			{
				//int i=0;
				
				failed += WriteWinbondFlash_me(dest,FLASH_BLOCK_SIZE,(UCHAR *)src);
				fprintf(stderr,"WinbondFlash 3 write %d bytes at 0x%08x\n",i,dest);
				*len+=blockSize;
				 
				for(k=0;k<blockSize/4;k++)
				{
					sum+=inpw(src+k);
				}
			}
			dest+=blockSize;
			i-=blockSize;
		}
	}
	// image write finished, check image checksum

	sum=0;
	for(k=0;k<image_footer->length/4;k++)
	{
		sum+=inpw( ((image_footer->base)+k) );
	}

	sum = ~((sum&(-1LU))+(sum>>32));

	if( ((UINT32)sum) != image_footer->image_checksum )
	{
		fprintf(stderr, "WriteFile2Image image_footer->image_checksum failed\n");
		//ignore this errr ? return -1;// image check sum error
	}
	// check footer check sum

	sum=0;
	for( i=footer_addr;i<footer_addr+sizeof(tfooter)-4;i+=4)
	{
		sum+=inpw( (i));
	}
	sum = ~((sum&(-1LU))+(sum>>32));

	if( ((UINT32)sum) != image_footer->checksum )
	{
		fprintf(stderr,"WriteFile2Image checksum err\n");
		//ignore this error ?  return -1; // footer check sum error
	}

	return failed;
}


int WriteFirmWare(void *filename,int *len)
{
	int fd;
	char *buf;
	tfooter footer;
	tfooter * pfooter=0;
	FirmWare header;
	int ret;
	unsigned long i=0;
	unsigned int totallen;
	int updaterom = 0;

	fd=open(filename,O_RDONLY);
	if(fd<0)
		return -1;

	ret=read(fd,&header,sizeof(header));

	if( header.magic == GENROM_MAGIC )
	{
		updaterom = 1;
	}
	else if( header.magic != GENBIN_MAGIC )
	{
		fprintf(stderr,"Invalid firmware\n");
		close(fd);
		return -2;
	}
	
	if( g_model_id && (header.model != g_model_id) && (header.model != 0xf0f0) )
	{
		fprintf(stderr,"Invalid firmware model\n");
		close(fd);
		return -2;		
	}

	if(ret!=sizeof(header))
	{
		fprintf(stderr,"firmware err1\n");
		close(fd);
		return -1;
	}

	
	buf=malloc(BUF_SIZE);
	if(buf<=0)
	{
		fprintf(stderr,"firmware err2\n");
		close(fd);
		return -1;
	}
	
	//check file corrupt or not
	if( header.length_bin + header.length_img > MIN_BIN_IMG_SIZE )
	{
		unsigned char sum = 0;
		unsigned long sum2 = 0;
		int readn = 0;
		
		lseek(fd, sizeof(header), SEEK_SET);

		for(i=0; i< header.length_bin + header.length_img; )
		{
			int j;
			//char buf[BUF_SIZE];
			
			memset(buf,0, BUF_SIZE);
			readn = read(fd,buf, BUF_SIZE );
			
			if( readn == -1 || readn == 0 )
				break;
			for(j=0; j<readn; j++)
			{
				i++;
				sum ^= buf[j];
			}
		}
		sum2 = sum ^ (header.length_bin + header.length_img);
		if( sum2 != header.sum )
		{
			//?? free(buf);
			fprintf(stderr,"firmware corrupt\n");
			close(fd);
			return -2;
		}
		lseek(fd, sizeof(header), SEEK_SET);
	}
	else
	{
		fprintf(stderr,"Invalid firmware size\n");
		close(fd);
		//free(buf);
		return -2;
	}
	
	*len=0;
	memset((void*)&footer,0,sizeof(footer));

	g_RunTimeState.iFlashWriteTotalSize = header.length_bin + header.length_img;
	
	if (pthread_mutex_lock(&g_ptmFlash) != 0) return -1;

	/****************************
	 * ROM IMAGE
	 ***************************/
	if( updaterom )
	{
		footer.num=0; // image number
		footer.base=0x7F000000;	// image base address
		footer.length=header.length_img+header.length_bin;	// image size
		footer.load_address=0x0000;	// image load address
		footer.exec_address=0x0000; // image exec address
		memcpy(footer.name,"rom",9);
		footer.name[9]='\0';
		footer.signature=SIGNATURE_WORD;// This is a key word to identify the image footer location
		footer.type=0;

		if(WriteFlashRom(&footer,buf,fd,read,len)!=0)
		{
			free(buf);
			close(fd);
			pthread_mutex_unlock(&g_ptmFlash);
			fprintf(stderr,"firmware error on writing rom image\n");
			return -1;
		}

		free(buf);
		close(fd);
		pthread_mutex_unlock(&g_ptmFlash);
	
		return 0;
	}
	

	/****************************
	 * IMAGE 7: linux image
	 ***************************/
	if(header.length_bin!=0 )
	{
		totallen=0;
		DelImage(7);
		footer.num=7; // image number
		if(footer.base==0)
			footer.base=0x7F020000;	// image base address
		footer.length=header.length_bin;	// image size
		footer.load_address=0x8000;	// image load address
		footer.exec_address=0x8000; // image exec address
		memcpy(footer.name,"linux.bin",9);
		footer.name[9]='\0';
		footer.signature=SIGNATURE_WORD;// This is a key word to identify the image footer location
		if(ISBINZIP(header.flag))
			footer.type=IMAGE_ACTIVE|IMAGE_COPY2RAM|IMAGE_EXEC|IMAGE_COMPRESSED; // Setting the attribution of the image
		else
			footer.type=IMAGE_ACTIVE|IMAGE_COPY2RAM|IMAGE_EXEC; // Setting the attribution of the image
		//fprintf(stderr,"start: Write image 7...\n");
		if(WriteFile2Image(&footer,buf,fd,read,len)!=0)
		{
			free(buf);
			close(fd);
			pthread_mutex_unlock(&g_ptmFlash);
			//fprintf(stderr,"firmware error on writing image 7\n");
			return -1;
		}
		//fprintf(stderr,"end:   Write image 7 ok\n");
	}
	
	////////////////////
	// IMAGE 6  : romfs
	////////////////////
	pfooter=0;
	memset((void*)&footer,0,sizeof(footer));
	if(header.length_bin!=0)
		lseek(fd,header.length_bin+sizeof(header),SEEK_SET);
	
	if(header.length_img!=0 )
	{
		totallen=0;
		DelImage(6);
		footer.num=6; // image number
		if(footer.base==0)
			footer.base=0x7F100000;
		footer.length=header.length_img;	// image size

#ifdef WINBOND_ORIG
		if(ISIMGZIP(header.flag))
		{
			footer.load_address=0x700000;	// image load address
			footer.exec_address=0x700000; // image exec address
			footer.type=IMAGE_ACTIVE|IMAGE_COPY2RAM|IMAGE_COMPRESSED; // Setting the attribution of the image
		}
		else
		{
			footer.load_address=footer.base;	// image load address
			footer.exec_address=footer.base; // image exec address
			footer.type=IMAGE_ACTIVE|IMAGE_COPY2RAM; // Setting the attribution of the image
		}
#else
		footer.type=IMAGE_ACTIVE|IMAGE_COPY2RAM; //|IMAGE_COPY2RAM;
		
		if(ISIMGZIP(header.flag))
			footer.type|=IMAGE_COMPRESSED;

		footer.exec_address=0x00D00000;
		footer.load_address=0x00D00000;
#endif

		memcpy(footer.name,"romfs.img",9);
		footer.name[9]='\0';
		footer.signature=SIGNATURE_WORD;// This is a key word to identify the image footer location
		//fprintf(stderr,"start: Write image 6...\n");
		if(WriteFile2Image(&footer,buf,fd,read,len)!=0)
		{
			free(buf);
			close(fd);
			pthread_mutex_unlock(&g_ptmFlash);
			//fprintf(stderr,"firmware error on writing image 6\n");
			return -1;
		}
		//fprintf(stderr,"end:   Write image 6 ok\n");
	}
	free(buf);
	close(fd);
	pthread_mutex_unlock(&g_ptmFlash);
	
	return 0;

}

#endif


static WR_CONFIG_PARAM_T g_ConfigParam4Flash;
static pthread_mutex_t g_ptmConfigParam4Flash = PTHREAD_MUTEX_INITIALIZER;

void SendFlashMsgWriteConfig(WR_CONFIG_PARAM_T *origConfig)
{
	MSG_T msg;

	if (pthread_mutex_lock(&g_ptmConfigParam4Flash) != 0) return;
	memcpy(&g_ConfigParam4Flash, origConfig, sizeof(WR_CONFIG_PARAM_T));
	pthread_mutex_unlock(&g_ptmConfigParam4Flash);

	msg.lMsg = MSG_FLASH_CONFIG_WRITE;
	msg.lData = 0;

	if (IsMsgExist(g_pMsgFlash, &msg)) return;
	else SendMsg(g_pMsgFlash, &msg);
}

void SendFlashMsgFirmwareUpdate()
{
	MSG_T msg;
	msg.lMsg = MSG_FLASH_FIRMWARE_UPDATE;
	msg.lData = 0;
	if (IsMsgExist(g_pMsgFlash, &msg)) return;
	else SendMsg(g_pMsgFlash, &msg);
}


int GetFlashSize()
{
#ifdef EMBED
	unsigned int iSizeBit = inpw(ROMCON);
	iSizeBit = (iSizeBit >> 16) & 0x00000007;

	return 0x00040000 << iSizeBit;
#else
	return 0;
#endif
}

void *FlashThread(void *pParam)
{
	MSG_T msg;
	int iFlashSize;
	int iOffsetConfig;

	BlockSignal();

#ifdef EMBED
	iFlashSize = GetFlashSize();
	if (iFlashSize <= 0) return NULL;
	iOffsetConfig = iFlashSize - sizeof(WR_CONFIG_PARAM_T);

#endif

	while (TRUE)
	{
		if (!GetMsg(g_pMsgFlash, &msg))
		{
			sleep(1);
			fprintf(stderr, "Msg queue error!\n");
			continue;
		}

		if (msg.lMsg == MSG_QUIT) 
			break;
		else if (msg.lMsg == MSG_FLASH_CONFIG_WRITE)
		{
			int iTry;
			WR_CONFIG_PARAM_T *pConfigParam4Write;
			WR_CONFIG_PARAM_T *pConfigParam4Check;

			if (pthread_mutex_lock(&g_ptmConfigParam4Flash) != 0) return NULL;
			
			pConfigParam4Write = malloc(sizeof(WR_CONFIG_PARAM_T));
			pConfigParam4Check = malloc(sizeof(WR_CONFIG_PARAM_T));
			if (pConfigParam4Write == NULL || pConfigParam4Check == NULL)
			{
				PRINT_MEM_OUT;
				return NULL;
			}
			
			memcpy(pConfigParam4Write, &g_ConfigParam4Flash, sizeof(WR_CONFIG_PARAM_T));
			pthread_mutex_unlock(&g_ptmConfigParam4Flash);
			
			for (iTry = 0; iTry < 20; iTry++)
			{
#ifdef EMBED			
				int iLenWrite;
				int iLenRead;

				iOffsetConfig = 0x7F1F0000; //norton
				//fprintf(stderr,"Write conf at:0x%08x...\n",iOffsetConfig);
				
				iLenWrite = WriteWinbondFlash(iOffsetConfig, sizeof(WR_CONFIG_PARAM_T), (char *)pConfigParam4Write);
				iLenRead = ReadWinbondFlash(iOffsetConfig, sizeof(WR_CONFIG_PARAM_T), (char *)pConfigParam4Check);

				//fprintf(stderr,"WriteWinbondFlash wrote: %d read:%d offset=0x%08x\n",iLenWrite,iLenRead,iOffsetConfig);
				if (iLenWrite == iLenRead && memcmp(pConfigParam4Write, pConfigParam4Check, sizeof(WR_CONFIG_PARAM_T)) == 0) 
					break;
				else fprintf(stderr,"WriteWinbondFlash write conf error, retrying\n");
#else
				FILE *fp;
				fp = fopen(FLASH_CONFIG_FILE, "wb");
				if (fp == NULL)
					fprintf(stderr, "Can not write flash!\n");
				{
					fwrite((void *)pConfigParam4Write, 1, sizeof(WR_CONFIG_PARAM_T), fp);
					fclose(fp);
				}
				break;
#endif

			}
			
			if (pConfigParam4Write != NULL)
				free(pConfigParam4Write);
			if (pConfigParam4Check != NULL)
				free(pConfigParam4Check);
		}
		else if (msg.lMsg == MSG_FLASH_FIRMWARE_UPDATE)
		{
			int iTry=0;
			int ret = 0;
			
			for( iTry = 0 ; iTry< 30; iTry++) //norton
			{
				ret = 0;

				fprintf(stderr,"v10,Write firmware %d...\n",iTry);
#ifdef EMBED
				if( (ret=WriteFirmWare(FLASH_UPDATE_TMP_FILE,&(g_RunTimeState.iFlashWriteSize)))==0 )
					break;
				g_RunTimeState.iFlashWriteSize = -1;
				if( ret == -2 )
				{
					g_RunTimeState.iFlashWriteTotalSize = -1;//-1
					g_RunTimeState.iFlashWriteSize = -1;
					break;
				}
				if( iTry>10 ) sleep(3);
			}
			
			if( ret != 0 )
			{
				//fprintf(stderr,"Write FirmWare failed\n");
				g_RunTimeState.iFlashWriteTotalSize = -1;//-1
				g_RunTimeState.iFlashWriteSize = -1;
			}

#else
			while (g_RunTimeState.iFlashWriteSize < g_RunTimeState.iFlashWriteTotalSize)
			{
				g_RunTimeState.iFlashWriteSize += 32 * 1024;
				usleep(500 * 1000);
			}
#endif			
			unlink(FLASH_UPDATE_TMP_FILE);
			g_RunTimeState.iFlashUpdateState = FIRMWARE_STOP;
		}
	}


	return NULL;
}
