#include "CommonDef.h"
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

char m_base64tab[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
#define BASE64_MAXLINE  76
#define EOL  "\r\n"

int Base64BufferSize(int iInputSize)
{
	int iOutSize = (iInputSize + 2) / 3 * 4;                    // 3:4 conversion ratio
	iOutSize += strlen(EOL) * iOutSize / BASE64_MAXLINE + 3;  // Space for newlines and NUL
	return iOutSize;
}

BOOL EncodeBase64(const char *pccInStr, int iInLen, char *pccOutStr, int iMaxOutSize, int *piOutLen)
{
	int i;
	int iInPos;
	int iOutPos;
	int iLineLen;
	int c1,c2,c3;
	char *pc;

	if (pccInStr == NULL || pccOutStr == NULL
		|| iMaxOutSize <= 0 || iInLen < 0
		|| iMaxOutSize < Base64BufferSize(iInLen))
	{
		fprintf(stderr, "Error call EncodeBase64()%d %d %d %d %d\n",
			pccInStr,
			pccOutStr,
			iMaxOutSize,
			iInLen,
			Base64BufferSize(iInLen));
		return FALSE;
	}

	//Set up the parameters prior to the main encoding loop
	iInPos = 0;
	iOutPos = 0;
	iLineLen = 0;

	// Get three characters at a time from the input buffer and encode them
	for (i=0; i<iInLen/3; ++i)
	{
		//Get the next 2 characters
		c1 = pccInStr[iInPos++] & 0xFF;
		c2 = pccInStr[iInPos++] & 0xFF;
		c3 = pccInStr[iInPos++] & 0xFF;

		//Encode into the 4 6 bit characters
		pccOutStr[iOutPos++] = m_base64tab[(c1 & 0xFC) >> 2];
		pccOutStr[iOutPos++] = m_base64tab[((c1 & 0x03) << 4) | ((c2 & 0xF0) >> 4)];
		pccOutStr[iOutPos++] = m_base64tab[((c2 & 0x0F) << 2) | ((c3 & 0xC0) >> 6)];
		pccOutStr[iOutPos++] = m_base64tab[c3 & 0x3F];
		iLineLen += 4;

		//Handle the case where we have gone over the max line boundary
		if (iLineLen >= BASE64_MAXLINE-3)
		{
			char* pc = EOL;
			pccOutStr[iOutPos++] = *pc++;
			if (*pc)
				pccOutStr[iOutPos++] = *pc;
			iLineLen = 0;
		}
	}

	// Encode the remaining one or two characters in the input buffer
	switch (iInLen % 3)
	{
    case 0:
	{
		pc = EOL;
		pccOutStr[iOutPos++] = *pc++;
		if (*pc)
			pccOutStr[iOutPos++] = *pc;
		break;
	}
    case 1:
	{
		int c1 = pccInStr[iInPos] & 0xFF;
		pccOutStr[iOutPos++] = m_base64tab[(c1 & 0xFC) >> 2];
		pccOutStr[iOutPos++] = m_base64tab[((c1 & 0x03) << 4)];
		pccOutStr[iOutPos++] = '=';
		pccOutStr[iOutPos++] = '=';
		pc = EOL;
		pccOutStr[iOutPos++] = *pc++;
		if (*pc)
			pccOutStr[iOutPos++] = *pc;
		break;
	}
    case 2:
	{
		int c1 = pccInStr[iInPos++] & 0xFF;
		int c2 = pccInStr[iInPos] & 0xFF;
		pccOutStr[iOutPos++] = m_base64tab[(c1 & 0xFC) >> 2];
		pccOutStr[iOutPos++] = m_base64tab[((c1 & 0x03) << 4) | ((c2 & 0xF0) >> 4)];
		pccOutStr[iOutPos++] = m_base64tab[((c2 & 0x0F) << 2)];
		pccOutStr[iOutPos++] = '=';
		pc = EOL;
		pccOutStr[iOutPos++] = *pc++;
		if (*pc)
			pccOutStr[iOutPos++] = *pc;
		break;
	}
    default:
		fprintf(stderr, "Unknown Error in FILE %s, LINE %d\n", __FILE__, __LINE__);
		return FALSE;
		break;
	}
	pccOutStr[iOutPos] = 0;
	*piOutLen = iOutPos;
	return TRUE;
}



/* Resolve host by gethostbyname_r.
 * 	ReturnValue.psHostent = NULL, if resolving host failed.
 */
HOST_INFO_T *ResolveHost(HOST_INFO_T *pHostInfo, const char *pcServer)
{
	int rt;
	int iHerr;
	int iHostBufLen;

	char *pcHostBuf;
	struct hostent sHost;

	if (pcServer == NULL || pHostInfo == NULL) return NULL;

	pHostInfo->pcExtraBuffer = NULL;
	pHostInfo->psHostent = NULL;
	pcHostBuf = pHostInfo->acHostSmallBuf;
	iHostBufLen = sizeof(pHostInfo->acHostSmallBuf);

	res_init(); //reload /etc/resolv.conf norton, otherwise new dns server will not take effect on PPPoE DoD mode.

	while ((rt = gethostbyname_r(pcServer, &sHost, pcHostBuf, iHostBufLen,
		&pHostInfo->psHostent, &iHerr)) == ERANGE)
	{
		iHostBufLen *= 2;
		pcHostBuf = realloc(pHostInfo->pcExtraBuffer, iHostBufLen);
		if (pcHostBuf == NULL)
		{
			PRINT_MEM_OUT;
			goto lError;
		}
		pHostInfo->pcExtraBuffer = pcHostBuf;
	}

	if (rt == 0 && pHostInfo->psHostent != NULL) goto lOk;

lError:
	fprintf(stderr, "Failed to resolve host (%s)!\n", pcServer);
	pHostInfo->psHostent = NULL;
	if (pHostInfo->pcExtraBuffer != NULL)
	{
		free(pHostInfo->pcExtraBuffer);
		pHostInfo->pcExtraBuffer = NULL;
	}
	return NULL;
lOk:
	return pHostInfo;
}

/* free the source created by ResolveHost. */
void DeleteHost(HOST_INFO_T *pHostInfo)
{
	if (pHostInfo == NULL) return;
	if (pHostInfo->pcExtraBuffer != NULL)
		free(pHostInfo->pcExtraBuffer);
}



BOOL HTTPC_Process(FD_T *pFD, BOOL bReadable, BOOL bWritable);

/*********************************************************************
** HTTPC construction & destrucion
**	do not call them directly for ever!
** var vaInit:
**	const char *pcServer,
**	const char *pcPath,
**	int iPort,
**	const char *pcHost,
**	const char *pcUserAgent,
**	const char *pcUser,
**	const char *pcPass,
**	const char *pcProxy,
**	const char *pcProxyUser,
**	const char *pcProxyPass,
**	HTTPC_RESPONSE pfunOnResponse,
**	HTTPC_REQUEST_OVER pfunOnRequestOver
*********************************************************************/
OBJHEADER *HTTPC_Init(OBJHEADER *pObj, va_list vaInit)
{
	HTTPC *pThis = (HTTPC *)pObj;

	const char *pcServer = va_arg(vaInit, const char *);
	const char *pcPath = va_arg(vaInit, const char *);
	int iPort = va_arg(vaInit, int);
	const char *pcHost = va_arg(vaInit, const char *);
	const char *pcUserAgent = va_arg(vaInit, const char *);
	const char *pcUser = va_arg(vaInit, const char *);
	const char *pcPass = va_arg(vaInit, const char *);
	const char *pcProxy = va_arg(vaInit, const char *);
	const char *pcProxyUser = va_arg(vaInit, const char *);
	const char *pcProxyPass = va_arg(vaInit, const char *);
	HTTPC_RESPONSE pfunOnResponse = va_arg(vaInit, HTTPC_RESPONSE);
	HTTPC_REQUEST_OVER pfunOnRequestOver = va_arg(vaInit, HTTPC_REQUEST_OVER);

	memset((char *)pThis + sizeof(OBJHEADER), 0, sizeof(HTTPC) - sizeof(OBJHEADER));

	pThis->tConnTimeout = 5;//timeout: 5 seconds
	pThis->pcServer = strdup(pcServer);
	if (pThis->pcServer == NULL) goto lE_strdup_pcServer;
	pThis->pcPath = strdup(pcPath);
	if (pThis->pcPath == NULL) goto lE_strdup_pcPath;
	pThis->iPort = iPort;
	if (pcHost != NULL)
	{
		pThis->pcHost = strdup(pcHost);
		if (pThis->pcHost == NULL) goto lE_strdup_pcHost;
	}
	if (pcUserAgent != NULL)
	{
		pThis->pcUserAgent = strdup(pcUserAgent);
		if (pThis->pcUserAgent == NULL) goto lE_strdup_pcUserAgent;
	}
	if (pcUser != NULL)
	{
		pThis->pcUser = strdup(pcUser);
		if (pThis->pcUser == NULL) goto lE_strdup_pcUser;
	}
	if (pcPass != NULL)
	{
		pThis->pcPass = strdup(pcPass);
		if (pThis->pcPass == NULL) goto lE_strdup_pcPass;
	}
	if (pcProxy != NULL)
	{
		pThis->pcProxy = strdup(pcProxy);
		if (pThis->pcProxy == NULL) goto lE_strdup_pcProxy;
	}
	if (pcProxyUser != NULL)
	{
		pThis->pcProxyUser = strdup(pcProxyUser);
		if (pThis->pcProxyUser == NULL) goto lE_strdup_pcProxyUser;
	}
	if (pcProxyPass != NULL)
	{
		pThis->pcProxyPass = strdup(pcProxyPass);
		if (pThis->pcProxyPass == NULL) goto lE_strdup_pcProxyPass;
	}

	pThis->pfunOnResponse = pfunOnResponse;
	pThis->pfunOnRequestOver = pfunOnRequestOver;

	if (ObjInit((OBJHEADER *)pThis, sizeof(FD_T), (OBJHEADER *)&pThis->fd, FD_Final, FD_Init,
		-1, SE_TP_HTTPC, ST_NET_INIT, 0, FALSE, FALSE, HTTPC_Process) == NULL)
		goto lE_FD_Init;
	return pObj;

lE_FD_Init:
	if (pThis->pcProxyPass != NULL) free(pThis->pcProxyPass);
lE_strdup_pcProxyPass:
	if (pThis->pcProxyUser != NULL) free(pThis->pcProxyUser);
lE_strdup_pcProxyUser:
	if (pThis->pcProxy != NULL) free(pThis->pcProxy);
lE_strdup_pcProxy:
	if (pThis->pcPass != NULL) free(pThis->pcPass);
lE_strdup_pcPass:
	if (pThis->pcUser != NULL) free(pThis->pcUser);
lE_strdup_pcUser:
	if (pThis->pcUserAgent != NULL) free(pThis->pcUserAgent);
lE_strdup_pcUserAgent:
	if (pThis->pcHost != NULL)free(pThis->pcHost);
lE_strdup_pcHost:
	if (pThis->pcPath != NULL)free(pThis->pcPath);
lE_strdup_pcPath:
	if (pThis->pcServer != NULL)free(pThis->pcServer);
lE_strdup_pcServer:
	return NULL;
}

void HTTPC_Final(OBJHEADER *pObj)
{
	HTTPC *pThis = (HTTPC *)pObj;

	ObjFinal((OBJHEADER *)&pThis->fd);

	if (pThis->pcServer != NULL)
		free(pThis->pcServer);
	if (pThis->pcPath != NULL)
		free(pThis->pcPath);
	if (pThis->pcHost != NULL)
		free(pThis->pcHost);
	if (pThis->pcUserAgent != NULL)
		free(pThis->pcUserAgent);
	if (pThis->pcUser != NULL)
		free(pThis->pcUser);
	if (pThis->pcPass != NULL)
		free(pThis->pcPass);
	if (pThis->pcProxy != NULL)
		free(pThis->pcProxy);
	if (pThis->pcProxyUser != NULL)
		free(pThis->pcProxyUser);
	if (pThis->pcProxyPass != NULL)
		free(pThis->pcProxyPass);
}


BOOL HTTPC_Process(FD_T *pFD, BOOL bReadable, BOOL bWritable)
{
	int iLen;
	HTTPC *pHttpFD;

	pHttpFD = (HTTPC *)((char *)(pFD) - (unsigned long)(&((HTTPC *)0)->fd));

	if (ST_NET_INIT == pFD->iState)
	{		
		if( ResolveHost(&pHttpFD->hostInfo, (pHttpFD->pcProxy == NULL ? pHttpFD->pcServer : pHttpFD->pcProxy)) == NULL )
		{
			log( LOG_DDNS, DDNS_RESOLV_FAIL, pHttpFD->pcServer);
		}
		
		if (pHttpFD->hostInfo.psHostent == NULL)
		{
			FD_Next(pFD, ST_NET_OVER, FALSE, FALSE, 0);
			return TRUE;
		}
		else
		{
			pHttpFD->ppcHost = pHttpFD->hostInfo.psHostent->h_addr_list;
			FD_Next(pFD, ST_NET_CONNECT, FALSE, FALSE, 0);
			return TRUE;
		}
	}
	else if (ST_NET_CONNECT == pFD->iState)
	{
		int iFl;
		struct sockaddr_in sin;

		if (pFD->iFD != -1)
		{
			close(pFD->iFD);
			pFD->iFD = -1;
		}
		if (*pHttpFD->ppcHost == NULL)
		{
			FD_Next(pFD, ST_NET_OVER, FALSE, FALSE, 0);
			return TRUE;
		}
		pFD->iFD = socket(PF_INET,SOCK_STREAM,0);
		if (pFD->iFD < 0)
		{
			fprintf(stderr, "ddns: Can not create socket!\n");
			FD_Next(pFD, ST_NET_OVER, FALSE, FALSE, 0);
			return TRUE;
		}

		/* set socket to O_NDELAY */
		iFl = fcntl(pFD->iFD, F_GETFL, 0);
		if (fcntl(pFD->iFD, F_SETFL, iFl | O_NDELAY) != 0)
		{
			fprintf(stderr, "Can not set socket fd to O_NDELAY mode.\n");
			FD_Next(pFD, ST_NET_OVER, FALSE, FALSE, 0);
			return TRUE;
		}

		/* Connect to server */
		sin.sin_family = AF_INET;
		sin.sin_port = htons(pHttpFD->iPort);	//http port number
		sin.sin_addr.s_addr = *(unsigned long*)*pHttpFD->ppcHost;
		pHttpFD->ppcHost++;
		connect(pFD->iFD,(struct sockaddr *)&sin,sizeof(sin));

		FD_Next(pFD, ST_NET_CONNECTING, TRUE, TRUE, pHttpFD->tConnTimeout * 1000);
		pHttpFD->tLastOp = time(0);
		return TRUE;
	}
	else if (ST_NET_CONNECTING == pFD->iState)
	{
		if (!(bReadable || bWritable))
		{
			time_t tThisOp = time(0);
			if (tThisOp - pHttpFD->tLastOp > pHttpFD->tConnTimeout)
			{
				FD_Next(pFD, ST_NET_CONNECT, FALSE, FALSE, 0);
			}
			return TRUE;
		}
		else
		{
			int iErrorCode;
			int iErrorCodeLen = sizeof(iErrorCode);
			if (getsockopt(pFD->iFD, SOL_SOCKET, SO_ERROR, &iErrorCode, &iErrorCodeLen) == 0)
			{
				if (iErrorCode == 0)
				{
					int iLen = 0;
					pHttpFD->tLastOp = time(0);

					FD_Next(pFD, ST_NET_WRITE, FALSE, TRUE, pHttpFD->tConnTimeout * 1000);

					if (pHttpFD->pcProxy != NULL)
						iLen = sprintf(pHttpFD->acBuffer,
								"GET HTTP://%s%s HTTP/1.1\r\n",
								pHttpFD->pcServer,
								pHttpFD->pcPath);
					else
						iLen = sprintf(pHttpFD->acBuffer,
								"GET %s HTTP/1.1\r\n",
								pHttpFD->pcPath);

					if (pHttpFD->pcHost != NULL)
						iLen += sprintf(pHttpFD->acBuffer + iLen,
								"HOST: %s\r\n",
								pHttpFD->pcHost);
					if (pHttpFD->pcUserAgent != NULL)
						iLen += sprintf(pHttpFD->acBuffer + iLen,
								"User-Agent: %s\r\n",
								pHttpFD->pcUserAgent);
					if (pHttpFD->pcUser != NULL || pHttpFD->pcPass != NULL)
					{
						char ac[256];
						int iInLen, iOutLen;

						iInLen = sprintf(pHttpFD->acBuffer + iLen, "%s%s%s",
							(pHttpFD->pcUser == NULL ? "" : pHttpFD->pcUser),
							((pHttpFD->pcUser == NULL && pHttpFD->pcPass == NULL)? "" : ":"),
							(pHttpFD->pcPass == NULL ? "" : pHttpFD->pcPass));
						if (EncodeBase64(pHttpFD->acBuffer + iLen, iInLen, ac, sizeof(ac), &iOutLen))
						{
							ac[iOutLen - 2] = '\0';//cut \r\n
							iLen += sprintf(pHttpFD->acBuffer + iLen,
								"Authorization: Basic %s\r\n",
								ac);
						}
					}
					if (pHttpFD->pcProxy != NULL)
					{
						char ac[256];
						int iInLen, iOutLen;
						if (pHttpFD->pcProxyUser == NULL && pHttpFD->pcProxyPass == NULL)
							ac[0] = '\0';
						else
						{
							iInLen = sprintf(pHttpFD->acBuffer + iLen, "%s%s%s",
								(pHttpFD->pcProxyUser == NULL ? "" : pHttpFD->pcProxyUser),
								((pHttpFD->pcProxyUser == NULL && pHttpFD->pcProxyPass == NULL)? "" : ":"),
								(pHttpFD->pcProxyPass == NULL ? "" : pHttpFD->pcProxyPass));
							if (EncodeBase64(pHttpFD->acBuffer + iLen, iInLen, ac, sizeof(ac), &iOutLen))
							{
								ac[iOutLen - 2] = '\0';//cut \r\n
							}
							else ac[0] = '\0';
						}
						iLen += sprintf(pHttpFD->acBuffer + iLen,
									"Proxy-Authorization: Basic %s\r\n",
									ac);
					}
					iLen += sprintf(pHttpFD->acBuffer + iLen,
						"ACCEPT: */*\r\n"
						"Pragma: no-cache\r\n"
						"Cache-Control: no-cache\r\n"
						"Connection: close\r\n\r\n");
					pHttpFD->iLenTotal = iLen;
					pHttpFD->iLenOK = 0;

					return TRUE;
				}
			}
			FD_Next(pFD, ST_NET_CONNECT, FALSE, FALSE, 0);
			return TRUE;
		}

	}
	else if (ST_NET_WRITE == pFD->iState)
	{
		if (bWritable)
		{
			iLen = write(pFD->iFD, pHttpFD->acBuffer + pHttpFD->iLenOK, pHttpFD->iLenTotal - pHttpFD->iLenOK);
			if (iLen > 0)
			{
				pHttpFD->tLastOp = time(0);
				pHttpFD->iLenOK += iLen;
				if (pHttpFD->iLenOK >= pHttpFD->iLenTotal)
				{
					pHttpFD->iLenOK = 0;
					pHttpFD->iLenTotal = 0;
					FD_Next(pFD, ST_NET_READ, TRUE, FALSE, pHttpFD->tConnTimeout * 1000);
				}
				return TRUE;
			}
			else if (!(errno == EWOULDBLOCK || errno == EAGAIN || errno == EPIPE))
			{
				FD_Next(pFD, ST_NET_OVER, FALSE, FALSE, 0);
				return TRUE;
			}
		}
		if (time(0) - pHttpFD->tLastOp > pHttpFD->tConnTimeout)
		{
			FD_Next(pFD, ST_NET_OVER, FALSE, FALSE, 0);
			return TRUE;
		}
	}
	else if (ST_NET_READ == pFD->iState)
	{
		time_t tThisOp = time(0);
		if (tThisOp - pHttpFD->tLastOp > pHttpFD->tConnTimeout)
		{
			if (pHttpFD->pfunOnResponse != NULL)
				(*pHttpFD->pfunOnResponse)(pHttpFD, pHttpFD->acBuffer, pHttpFD->iLenOK);
			FD_Next(pFD, ST_NET_OVER, FALSE, FALSE, 0);
			return TRUE;
		}

		if (!bReadable) return TRUE;

		iLen = read(pFD->iFD, pHttpFD->acBuffer + pHttpFD->iLenOK, sizeof(pHttpFD->acBuffer) / 2);
		if (iLen > 0)
		{
			pHttpFD->iLenOK += iLen;
			if (pHttpFD->iLenOK >= sizeof(pHttpFD->acBuffer) / 2)
			{
				if (pHttpFD->pfunOnResponse != NULL)
				{
					if (!(*pHttpFD->pfunOnResponse)(pHttpFD, pHttpFD->acBuffer, pHttpFD->iLenOK))
					{
						FD_Next(pFD, ST_NET_OVER, FALSE, FALSE, 0);
						return TRUE;
					}
				}
				pHttpFD->iLenOK = 0;
			}
			pHttpFD->tLastOp = tThisOp;
		}
		else
		{
			//fprintf(stderr, "read errno: %d\n", errno);
			if (!(errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK))
			{
				if (pHttpFD->pfunOnResponse != NULL)
					(*pHttpFD->pfunOnResponse)(pHttpFD, pHttpFD->acBuffer, pHttpFD->iLenOK);

				FD_Next(pFD, ST_NET_OVER, FALSE, FALSE, 0);
				return TRUE;
			}
		}
		return TRUE;
	}
	else if (ST_NET_OVER == pFD->iState)
	{
		if (pHttpFD->pfunOnRequestOver != NULL)
			(*pHttpFD->pfunOnRequestOver)(pHttpFD);
		if (pFD->iFD != -1) close(pFD->iFD);
		DeleteHost(&pHttpFD->hostInfo);
		return FALSE;
	}
	
	return FALSE; //norton , return TRUE ?? //
}



#ifdef WINBOND_DDNS

/*********************************************************************
** HTTPC construction for DDNS
**	do not call them directly for ever!
** var vaInit:
**	int iFD, int iType, int iState,
**	BOOL bWantRead, BOOL bWantWrite,
**	BOOL (*pfunProcessFD)(struct tagFD_T *pFD, BOOL bReadable, BOOL bWritable)
*********************************************************************/
OBJHEADER *DDNS_HTTPC_Init(OBJHEADER *pObj, va_list vaInit)
{
	OBJHEADER *pRt;


	pRt = HTTPC_Init(pObj, vaInit);
	if (pRt != NULL) ((HTTPC *)pRt)->fd.iType = SE_TP_DDNSHTTPC;
	return pRt;
}


static void DDNSRequestOver(HTTPC *pHttpFD)
{
	if (g_RunTimeState.iDDNSState != DO_DDNS_UPDATE_OK)
		g_RunTimeState.iDDNSState = DO_DDNS_UPDATE_FAIL;
}

static BOOL DDNSResponse(HTTPC *pHttpFD, char *acResponse, int iLen)
{
	char *pc = NULL;
#define DDNS_UPDATE_FAIL 3
	//write(1, acResponse, iLen);
	acResponse[iLen] = '\0';
	pc = strstr(acResponse, "\n\r\n");
	if (pc != NULL) pc += 3;
	else
	{
		pc = strstr(acResponse, "\n\n");
		if (pc != NULL) pc += 2;
	}

	if (pc != NULL)
	{
		printf("ddns state: [%s]\n", pc);
		if ( strstr(pc, "nochg" )
			|| strstr(pc, "good" ) )
		{
			g_RunTimeState.iDDNSState = DO_DDNS_UPDATE_OK;
			log(LOG_DDNS, DDNS_UPDATE_OK, NULL );
		}
		else if( strstr(pc, "nohost" ) )
			log(LOG_DDNS, DDNS_INVALID_HOST, NULL );
		else if( strstr(pc, "notfqdn") )
			log(LOG_DDNS, DDNS_NOT_FQDN, NULL );
		else if( strstr(pc, "!yours") )
			log(LOG_DDNS, DDNS_NOT_YOURS, NULL );
		else if( strstr(pc, "badauth") )
			log(LOG_DDNS, DDNS_AUTH_FAIL, NULL );
		else if( strstr(pc, "abuse") )
			log(LOG_DDNS, DDNS_ABUSE, NULL );
		else if( strstr(pc, "badagent") )
			log(LOG_DDNS, DDNS_BAD_AGENT, NULL );
		else 
			log(LOG_DDNS, DDNS_UPDATE_FAIL, NULL );

	}
	return TRUE; //norton
}


static void *AddDDNS(void *pThread, void *pParam)
{
	LISTNODE *pNode;
	LISTNODE *pNextNode;
	FD_T *pNewFD = (FD_T *)pParam;
	SELECT_THREAD_T *pThreadData = (SELECT_THREAD_T *)pThread;

	for (pNode = pThreadData->pFDList->pFirstNode;
		pNode != pThreadData->pFDList->pLastNode;
		pNode = pNextNode)
	{
		FD_T *pFD = (FD_T *)pNode->pValue;
		pNextNode = pNode->pNextNode;

		if (pFD == NULL) continue;
		if (pFD->iType == SE_TP_DDNSHTTPC)
		{
			ObjFinal((OBJHEADER *)pFD);
			DeleteNode(pNode);
		}
	}

	AddSelectFD(pNewFD);
	
	return TRUE; //norton
}

void Do_DDNS(const char *pcHost,
			unsigned long ulIP,
			const char *pcUser,
			const char *pcPass,
			const char *pcProxy,
			const char *pcProxyUser,
			const char *pcProxyPass)
{
	HTTPC *pHttpC;
	const char *pcServer = "members.dyndns.org";
	const char *pcPath = "/nic/update?system=dyndns&hostname=%s&myip=%s&wildcard=NO&backmx=NO&offline=NO&mx=";
	char acIP[24];
	char acPath[256];

	res_init(); //reload /etc/resolv.conf
	
	g_RunTimeState.iDDNSState = DO_DDNS_UPDATING;
	
	if (ulIP == 0L)
	{
		unsigned long ulPublicSubnetMask;
		unsigned long ulDefaultGateway;
		GetPubIPInfo(&ulIP, &ulPublicSubnetMask,
					&ulDefaultGateway );
	}
	if (ulIP == 0L)
	{
		log(LOG_DDNS, DDNS_NO_IP, NULL );
		g_RunTimeState.iDDNSState = DO_DDNS_UPDATE_FAIL;
		return;
	}

	if (pcHost == NULL || pcHost[0] == '\0') return;
	if (pcProxy != NULL && pcProxy[0] == '\0') pcProxy = NULL;

	IP2String(ulIP, acIP);
	if (strlen(pcPath) + strlen(acIP) + strlen(pcHost) > sizeof(acPath))
	{//not very exact compare
		fprintf(stderr, "DDNS: Too long host name!\n");
		//log(LOG_DDNS, DDNS_NO_IP, NULL );
		g_RunTimeState.iDDNSState = DO_DDNS_UPDATE_FAIL;
		return;
	}
	sprintf(acPath, pcPath, pcHost, acIP);

	if ((pHttpC = (HTTPC *)ObjInit(NULL, sizeof(HTTPC), NULL,
					HTTPC_Final, DDNS_HTTPC_Init,
					pcServer,
					acPath,
					80,
					pcServer,
					"Mozilla/4.0 (compatible; MSIE 5.00; Windows 98)",
					//"dyndnsupdate 0.6.14 - http://xzabite.org",
					pcUser,
					pcPass,
					pcProxy,
					pcProxyUser,
					pcProxyPass,
					DDNSResponse,
					DDNSRequestOver)) == NULL) return;
	InsertCustomProcess(g_pMsgSelect, &AddDDNS, (void *)&pHttpC->fd);
}
#endif

#if 0
int Config_GetDDNS(HTTPCONNECTION hConnection, LIST *pParamList, int iAction, BOOL bReturnJsType)
{
	const char *pcDDNSState;
	char ac[128];
	switch (iAction)
	{
	case CA_AUTH:
		return AUTH_ADMIN;
		break;
	case CA_CONFIG:
		printf("ddns : %d\n", g_RunTimeState.iDDNSState);
		switch (g_WebCamState.iDDNSState)
		{
		case DDNS_UPDATE_OK:
			pcDDNSState = "Updated";
			break;
		case DDNS_UPDATE_FAIL:
			pcDDNSState = "Updating Failed";
			break;
		case DDNS_UPDATTING:
			pcDDNSState = "Updating";
			break;
		default:
			pcDDNSState = "Not Update";
			break;
		}
		sprintf(ac, "Enable = %d\n", (int)g_ConfigParam.bEnableDDNS);
		AddHttpVarible(hConnection, ac, bReturnJsType);
		sprintf(ac, "User = %s\n", g_ConfigParam.acDDNSUserName);
		AddHttpVarible(hConnection, ac, bReturnJsType);
		sprintf(ac, "Pass = %s\n", g_ConfigParam.acDDNSUserPass);
		AddHttpVarible(hConnection, ac, bReturnJsType);
		sprintf(ac, "DomainName = %s\n", g_ConfigParam.acDDNSDomainName);
		AddHttpVarible(hConnection, ac, bReturnJsType);
		sprintf(ac, "Proxy = %s\n", g_ConfigParam.acProxy);
		AddHttpVarible(hConnection, ac, bReturnJsType);
		sprintf(ac, "ProxyUser = %s\n", g_ConfigParam.acProxyUser);
		AddHttpVarible(hConnection, ac, bReturnJsType);
		sprintf(ac, "ProxyPass = %s\n", g_ConfigParam.acProxyPass);
		AddHttpVarible(hConnection, ac, bReturnJsType);
		sprintf(ac, "Info = %s\n", pcDDNSState);
		AddHttpVarible(hConnection, ac, bReturnJsType);

		return 0;
		break;
	}
}

int Config_SetDDNS(HTTPCONNECTION hConnection, LIST *pParamList, int iAction, BOOL bReturnJsType)
{
	BOOL bEnableDDNS;
	const char *pcDDNSUserName;
	const char *pcDDNSUserPass;
	const char *pcDDNSDomainName;
	const char *pcProxy;
	const char *pcProxyUser;
	const char *pcProxyPass;
	unsigned long ulIP;

	switch (iAction)
	{
	case CA_AUTH:
		return AUTH_ADMIN;
		break;
	case CA_CONFIG:
		ulIP = GetIP(pParamList, "IP");//for test
		bEnableDDNS = GetBool(pParamList, "Enable");
		pcDDNSUserName = GetString(pParamList, "User");
		pcDDNSUserPass = GetString(pParamList, "Pass");
		pcDDNSDomainName = GetString(pParamList, "DomainName");
		pcProxy = GetString(pParamList, "Proxy");
		pcProxyUser = GetString(pParamList, "ProxyUser");
		pcProxyPass = GetString(pParamList, "ProxyPass");

		if (g_WebCamState.iDDNSState == DDNS_UPDATE_OK
			&& strcmp(pcDDNSUserName, g_ConfigParam.acDDNSUserName) == 0
			&& strcmp(pcDDNSUserPass, g_ConfigParam.acDDNSUserPass) == 0
			&& strcmp(pcDDNSDomainName, g_ConfigParam.acDDNSDomainName) == 0
			&& strcmp(pcProxy, g_ConfigParam.acProxy) == 0
			&& strcmp(pcProxyUser, g_ConfigParam.acProxyUser) == 0
			&& strcmp(pcProxyPass, g_ConfigParam.acProxyPass) == 0
			&& g_ConfigParam.bEnableDDNS == bEnableDDNS
			) return 0;

		pthread_mutex_lock(&g_ptmConfigParam);
		MyStrncpy(g_ConfigParam.acDDNSUserName, pcDDNSUserName, sizeof(g_ConfigParam.acDDNSUserName));
		MyStrncpy(g_ConfigParam.acDDNSUserPass, pcDDNSUserPass, sizeof(g_ConfigParam.acDDNSUserPass));
		MyStrncpy(g_ConfigParam.acDDNSDomainName, pcDDNSDomainName, sizeof(g_ConfigParam.acDDNSDomainName));
		MyStrncpy(g_ConfigParam.acProxy, pcProxy, sizeof(g_ConfigParam.acProxy));
		MyStrncpy(g_ConfigParam.acProxyUser, pcProxyUser, sizeof(g_ConfigParam.acProxyUser));
		MyStrncpy(g_ConfigParam.acProxyPass, pcProxyPass, sizeof(g_ConfigParam.acProxyPass));
		g_ConfigParam.bEnableDDNS = bEnableDDNS;
		pthread_mutex_unlock(&g_ptmConfigParam);

		if (bEnableDDNS)
			Do_DDNS(pcDDNSDomainName,
					ulIP,
					pcDDNSUserName,
					pcDDNSUserPass,
					pcProxy,
					pcProxyUser,
					pcProxyPass);

		WriteFlashMemory(&g_ConfigParam);
		return 0;
		break;
	}
}

#endif
