

#include "CommonDef.h"

int Http_Post_Init(HTTPCONNECTION hConnection, void *pParam)
{
	if (WRExternalAdmin(hConnection) == 0) return 0;

	if (GetHttpMethod(hConnection) != M_POST) return 0;

	if (pParam == (void *)Http_Config_Upload)
		SetPostDataFun(hConnection, Http_Post_Upload, pParam);
	else if (pParam == (void *)Http_Post_FMUpload)
	{
		SetPostDataFun(hConnection, Http_Post_FMUpload, NULL);
	}
	else if (pParam == (void *)Http_Post_ConfigUpload)
	{
		SetPostDataFun(hConnection, Http_Post_ConfigUpload, NULL);
	}
	else
		SetPostDataFun(hConnection, Http_Post_Data, pParam);
	return 1;
}

int Http_Post_Data(HTTPCONNECTION hConnection,
			int *piPostState,
			char **ppcPostBuf,
			int *piPostBufLen,
			int *piPostDataLen,
			char *pcFillData,
			int iFillDataLen,
			int iIsMoreData/*bool*/,
			void *pParam/*other parameter for extend use*/)
{
	if (*ppcPostBuf == NULL)
	{
		*piPostBufLen = MAX_CONFIG_POST_LENGTH;
		*ppcPostBuf = (char *)malloc(*piPostBufLen);
		*piPostDataLen = 0;
	}

	if (*ppcPostBuf == NULL || iFillDataLen + *piPostDataLen >= *piPostBufLen)
	{
		AddHttpBodyString(hConnection, "Too many post data or not enmough memory.");
		SetHttpHeader(hConnection, 200, "OK", NULL, NULL, "text/plain", TRUE);
		return 0;
	}
	
	memcpy(*ppcPostBuf + *piPostDataLen, pcFillData, iFillDataLen);
	*piPostDataLen += iFillDataLen;

	if (!iIsMoreData)
	{
		REQUEST_CALLBACK_PFUN funUseData;
		LIST *pListPost;
		int iPos;
		int iPrivilege;


		iPrivilege = GetAuthPrivilege(hConnection);
		if (iPrivilege < AUTH_USER)
		{
			free(*ppcPostBuf);
			*ppcPostBuf = NULL;
			*piPostDataLen = 0;
			*piPostBufLen = 0;
			SendAuthRequired(hConnection, AUTH_USER);

			return 1;
		}

		iPos = *piPostDataLen;

		(*ppcPostBuf)[iPos--] = '\0';
		if (iPos > 0 && (*ppcPostBuf)[iPos] == '\n')
			(*ppcPostBuf)[iPos--] = '\0';
		if (iPos > 0 && (*ppcPostBuf)[iPos] == '\r')
			(*ppcPostBuf)[iPos] = '\0';

		pListPost = ParseString(*ppcPostBuf);
		free(*ppcPostBuf);
		*ppcPostBuf = NULL;
		*piPostDataLen = 0;
		*piPostBufLen = 0;
#if DEBUG
		if (DebugShowPost(hConnection, pListPost) != 0)
#endif		
		{
			funUseData = (REQUEST_CALLBACK_PFUN)pParam;
			(*funUseData)(hConnection, pListPost);
		}
		DeleteParseString(pListPost);
		return 1;
	}
	return 1;
}

#ifdef DEBUG
int SendPostInfo(HTTPCONNECTION hConnection, LIST *pListPost)
{
	LISTNODE *pNode;
	NAMEDSTRING_T *pParseItem;
	char buf[1024];

	AddHttpBodyString(hConnection, "<font color=red>ĲΪ:</font><br>");
	for (pNode = pListPost->pFirstNode; pNode != pListPost->pLastNode; pNode = pNode->pNextNode)
	{
		if (pNode->pValue != NULL)
		{
			pParseItem = (NAMEDSTRING_T *)pNode->pValue;
			sprintf(buf, "<font color=blue>%s</font>	=	[<font color=green>%s</font>]<br>",
				(pParseItem->pcName?pParseItem->pcName:""),
				(pParseItem->pcValue?pParseItem->pcValue:"")
				);
			AddHttpBodyString(hConnection, buf);
		}
	}
	SetHttpHeader(hConnection, 200, "OK", "", "", "text/html", TRUE);

	return 0;
}

int DebugShowPost(HTTPCONNECTION hConnection, LIST *pListPost)
{
	char *pc;
	char *pcReferer;

	if (pListPost == NULL)
	{
		AddHttpBodyString(hConnection, "Error Post!");
		SetHttpHeader(hConnection, 200, "OK", "", "", "text/html", TRUE);
		return 0;
	}

	pcReferer = GetReferer(hConnection);
	if (pcReferer) pc = strchr(pcReferer, '?');
	else pc = NULL;
	if (pc)
	{
		if (strcmp(pc + 1, "Debug") == 0)
			return SendPostInfo(hConnection, pListPost);
	}
	return -1;
}
#endif


void SendReDirectHtml(HTTPCONNECTION hConnection, const char *pcRedirect, 
						const char *pcUrl,
						BOOL bShowTime,
						int iTime_InMSec,
						const char *pcInfo,
						const char *cookie )
{
	//const char *pcRedirect = "Redirect.htm";
	char acRedirect[256];
	char *pcRedirectPath = NULL;
	char *pcReferer;
	int iLen;
	char *pc1;
	char *pc2;

	pcReferer = GetReferer(hConnection);
	iLen = strlen(pcReferer);
	if (iLen <= 7)//for "http://"
		pcReferer = NULL;
	if (pcReferer != NULL)
	{
		pc1 = strchr(pcReferer + 7, '/');
		pc2 = strrchr(pcReferer + 7, '/');

		if (pc1 != NULL)
		{
			iLen = pc2 - pc1;
			if (iLen + 2 + strlen(pcRedirect) < sizeof(acRedirect))
			{
				memcpy(acRedirect, pc1, iLen);
				acRedirect[iLen] = '/';
				strcpy(acRedirect + iLen + 1, pcRedirect);
				pcRedirectPath = GetDocumentBasedPath(acRedirect);
			}
		}
	}

	SetKeepAliveMode(hConnection, FALSE);
	if (pcRedirectPath == NULL)
	{
		AddHttpBodyString(hConnection, "<script>history.back();</script>");
		SetHttpHeader(hConnection, 200, "OK", "", "Expires: 01 Jan 1970 00:00:00 GMT\r\n", "text/html", TRUE);
	}
	else
	{
		const char *pcookie = "00000000";
		if( cookie )
			pcookie = cookie;
		SendHtmlWithParam(hConnection,
			pcRedirectPath,pcUrl,bShowTime,iTime_InMSec,pcInfo,pcookie);
	}
	if( pcRedirectPath )
		free(pcRedirectPath);

}

void SendReDirectPage(HTTPCONNECTION hConnection)
{
	char pcBuf[512];
	char *pcReferer;
	pcReferer = GetReferer(hConnection);
	if (pcReferer == NULL)
		sprintf(pcBuf, "<script>top.location.reload();</script>");
	else
	{
#if 0
		sprintf(pcBuf, "Location: %s\r\nExpires: 01 Jan 1970 00:00:00 GMT\r\n", pcReferer);
		SetHttpHeader(hConnection, 302, "OK", "", pcBuf, "text/html", TRUE);
		return;
	}
#else
	sprintf(pcBuf,
		"\r\n\r\n<BODY BGCOLOR=#CEE7FF><center><p>"
		"<MARQUEE direction=right SCROLLDELAY=4 SCROLLAMOUNT=1 TRUESPEED>"
		"<font face='Comic Sans MS' size=4 >"
		"Please wait...</font>"
		"</MARQUEE>"
		"</center><p>"
		"<script>setTimeout('location=\"%s\"', 50);</script>\r\n\r\n\r\n",
		pcReferer);
	}
	AddHttpBodyString(hConnection, pcBuf);
	SetHttpHeader(hConnection, 200, "OK", "", "Expires: 01 Jan 1970 00:00:00 GMT\r\n", "text/html", TRUE);
#endif
}

void SendReDirectPage2(HTTPCONNECTION hConnection,
	char *pcUrl, BOOL bShowTime, int iTime_InMSec, char *msg )
{
	char pcBuf[512];
	
	sprintf(pcBuf,
		"<BODY BGCOLOR=#CEE7FF><center><p>"
		"<MARQUEE direction=right SCROLLDELAY=5 SCROLLAMOUNT=1 TRUESPEED>"
		"<font face='Comic Sans MS' size=4 >"
		"%s</font>"
		"</MARQUEE>"
		"</center><p>"
		"<script>setTimeout('location=\"%s\"', %d );</script>\r\n\r\n\r\n",
		msg, pcUrl, iTime_InMSec );

	AddHttpBodyString(hConnection, pcBuf);
	SetHttpHeader(hConnection, 200, "OK", "", "Expires: 01 Jan 1970 00:00:00 GMT\r\n", "text/html", TRUE);
}

void SendReDirectMain(HTTPCONNECTION hConnection)
{
	char extrahead[1024];
	char cookie[128];
	
	auth_cookie_gen(hConnection, &cookie);
	//printf("login set cookie=%s \n", cookie);
	snprintf(extrahead,sizeof(extrahead),"Location: %s\r\n%s\r\nExpires: 01 Jan 1970 00:00:00 GMT\r\n", "system_main.htm", cookie);

	AddHttpBodyString(hConnection, "Moved Temporarily" );
	SetHttpHeader(hConnection, 301, "Moved Temporarily", "", extrahead, "text/html", TRUE);
}


void SendLogoutPage(HTTPCONNECTION hConnection,  char *msg )
{
	char pcBuf[512];
	
	sprintf(pcBuf,
		"<BODY BGCOLOR=#CEE7FF><center><p>"
		"<font size=4 color=blue>"
		"%s</font>"
		"</center><p>"
		"\r\n\r\n",
		msg );

	AddHttpBodyString(hConnection, pcBuf);
	SetHttpHeader(hConnection, 200, "OK", "", "Set-Cookie: nonce=\"00000000\"\r\nExpires: 01 Jan 1970 00:00:00 GMT\r\n", "text/html", TRUE);
}
