#include "CommonDef.h"

int EMU_sem_init(EMU_sem_t *sem, unsigned int value)
{
	pthread_mutex_init(&sem->ptm, NULL);
	pthread_cond_init(&sem->ptc, NULL);
	sem->uiValue = value;
	return 0;
}

int EMU_sem_wait(EMU_sem_t *sem)
{
	if (pthread_mutex_lock(&sem->ptm) != 0) return -1;

	if (sem->uiValue == 0)
		pthread_cond_wait(&sem->ptc, &sem->ptm);

	sem->uiValue--;

	pthread_mutex_unlock(&sem->ptm);
	return 0;
}

int EMU_sem_trywait(EMU_sem_t *sem)
{
	if (pthread_mutex_lock(&sem->ptm) != 0) return -1;

	if (sem->uiValue == 0)
	{
		pthread_mutex_unlock(&sem->ptm);
		return -1;
	}
	else
	{
		sem->uiValue--;
		pthread_mutex_unlock(&sem->ptm);
		return 0;
	}
}

int EMU_sem_post(EMU_sem_t *sem)
{
	if (pthread_mutex_lock(&sem->ptm) != 0) return -1;

	sem->uiValue++;
	if (sem->uiValue == 1)
		pthread_cond_signal(&sem->ptc);

	pthread_mutex_unlock(&sem->ptm);
	return 0;
}

int EMU_sem_destroy(EMU_sem_t *sem)
{
	pthread_mutex_destroy(&sem->ptm);
	pthread_cond_destroy(&sem->ptc);
	return 0;
}



LIST *CreateMsgList()
{
	return CreateList();
}

void DeleteMsgList(LIST *plMsgList)
{
	LISTNODE *pNode;
	if (plMsgList == NULL) return;
	for (pNode = plMsgList->pFirstNode; pNode != plMsgList->pLastNode; pNode=pNode->pNextNode)
	{
		if (pNode->pValue != NULL)
			free(pNode->pValue);
	}
	DeleteList(plMsgList);
}

MSG_QUEUE *CreateMsgQueue(int iMaxMsgLimit, BOOL bFDEvent)
{
	MSG_QUEUE *pMsgQueue;

	pMsgQueue = (MSG_QUEUE *)malloc(sizeof(MSG_QUEUE));
	if (pMsgQueue == NULL)
	{
		PRINT_MEM_OUT;
		return NULL;
	}

	pMsgQueue->bFDEvent = bFDEvent;
	if (bFDEvent)
	{
		int iFL;
		if (pipe(pMsgQueue->aiFD) != 0)
		{
			free(pMsgQueue);
			return NULL;
		}
		if ((iFL = fcntl(pMsgQueue->aiFD[0], F_GETFL, 0)) != -1)
			fcntl(pMsgQueue->aiFD[0], F_SETFL, iFL | O_NDELAY);
	}

	if ((pMsgQueue->plMsgList = CreateMsgList()) == NULL)
	{
		free(pMsgQueue);
		return NULL;
	}
	pMsgQueue->iMaxMsgLimit = iMaxMsgLimit;

	if (pthread_mutex_init(&pMsgQueue->ptmMsg, NULL) != 0)
	{
		fprintf(stderr, "Init mutex for msg mutex failed!\n");
		DeleteMsgList(pMsgQueue->plMsgList);
		free(pMsgQueue);
		return NULL;
	}
	if (EMU_sem_init(&pMsgQueue->sMsgSemConsumer, 0) != 0)
	{
		fprintf(stderr, "Init semphore for msg consumer failed!\n");
		pthread_mutex_destroy(&pMsgQueue->ptmMsg);
		DeleteMsgList(pMsgQueue->plMsgList);
		free(pMsgQueue);
		return NULL;
	}
	if (iMaxMsgLimit > 0)
	{
		if (EMU_sem_init(&pMsgQueue->sMsgSemProducer, iMaxMsgLimit) != 0)
		{
			fprintf(stderr, "Init semphore for msg producer failed!\n");
			pthread_mutex_destroy(&pMsgQueue->ptmMsg);
			EMU_sem_destroy(&pMsgQueue->sMsgSemConsumer);
			DeleteMsgList(pMsgQueue->plMsgList);
			free(pMsgQueue);
			return NULL;
		}
	}

	return pMsgQueue;
}

void FreeMsgQueue(MSG_QUEUE *pMsgQueue)
{
	if (pMsgQueue == NULL) return;

	if (pMsgQueue->bFDEvent)
	{
		close(pMsgQueue->aiFD[0]);
		close(pMsgQueue->aiFD[1]);
	}

	DeleteMsgList(pMsgQueue->plMsgList);
	pthread_mutex_destroy(&pMsgQueue->ptmMsg);
	EMU_sem_destroy(&pMsgQueue->sMsgSemConsumer);
	EMU_sem_destroy(&pMsgQueue->sMsgSemProducer);
	free(pMsgQueue);
}

BOOL PopMsgNode(MSG_QUEUE *pMsgQueue, MSG_T *pMsg)
{
	MSG_T *pMsgInNode;

	if (pMsgQueue == NULL || pMsg == NULL) return FALSE;

	if (pthread_mutex_lock(&pMsgQueue->ptmMsg) != 0)
		return FALSE;

	if (pMsgQueue->plMsgList->pFirstNode == pMsgQueue->plMsgList->pLastNode)
	{
		pthread_mutex_unlock(&pMsgQueue->ptmMsg);
		return FALSE;
	}

	pMsgInNode = (MSG_T *)pMsgQueue->plMsgList->pFirstNode->pValue;
	DeleteNode(pMsgQueue->plMsgList->pFirstNode);
	if (pMsgQueue->bFDEvent)
	{
		char c;
		if (read(pMsgQueue->aiFD[0], &c, 1) != 1)
			PTE;
	}
	pthread_mutex_unlock(&pMsgQueue->ptmMsg);

	*pMsg = *pMsgInNode;
	free(pMsgInNode);
	return TRUE;
}

BOOL PushMsgNode(MSG_QUEUE *pMsgQueue, MSG_T *pMsg)
{
	LISTNODE *pNode;
	MSG_T *pMsgInNode;

	if (pMsgQueue == NULL || pMsg == NULL) return FALSE;

	pMsgInNode = (MSG_T *)malloc(sizeof(MSG_T));
	if (pMsgInNode == NULL)
	{
		PRINT_MEM_OUT;
		return FALSE;
	}
	*pMsgInNode = *pMsg;

	if (pthread_mutex_lock(&pMsgQueue->ptmMsg) != 0)
	{
		free(pMsgInNode);
		return FALSE;
	}

	pNode = AppendNode(pMsgQueue->plMsgList);
	if (pNode == NULL)
	{
		free(pMsgInNode);
		pthread_mutex_unlock(&pMsgQueue->ptmMsg);
		return FALSE;
	}

	pNode->pValue = pMsgInNode;

	if (pMsgQueue->bFDEvent)
	{
		char c = 'C';
		write(pMsgQueue->aiFD[1], &c, 1);
	}

	pthread_mutex_unlock(&pMsgQueue->ptmMsg);

	return TRUE;
}

BOOL SendMsg(MSG_QUEUE *pMsgQueue, MSG_T *pMsg)
{
	BOOL bPush;
	if (pMsgQueue->iMaxMsgLimit > 0)
		EMU_sem_wait(&pMsgQueue->sMsgSemProducer);

	bPush = PushMsgNode(pMsgQueue, pMsg);

	EMU_sem_post(&pMsgQueue->sMsgSemConsumer);
	return bPush;
}

BOOL GetMsg(MSG_QUEUE *pMsgQueue, MSG_T *pMsg)
{
	BOOL bPop;
	EMU_sem_wait(&pMsgQueue->sMsgSemConsumer);

	bPop = PopMsgNode(pMsgQueue, pMsg);

	if (pMsgQueue->iMaxMsgLimit > 0)
		EMU_sem_post(&pMsgQueue->sMsgSemProducer);

	return bPop;
}

BOOL TrySendMsg(MSG_QUEUE *pMsgQueue, MSG_T *pMsg)
{
	BOOL bPush;
	if (pMsgQueue->iMaxMsgLimit > 0)
	{
		if (EMU_sem_trywait(&pMsgQueue->sMsgSemProducer) != 0)
			return FALSE;
	}

	bPush = PushMsgNode(pMsgQueue, pMsg);

	EMU_sem_post(&pMsgQueue->sMsgSemConsumer);
	return bPush;
}

BOOL TryGetMsg(MSG_QUEUE *pMsgQueue, MSG_T *pMsg)
{
	BOOL bPop;
	if (EMU_sem_trywait(&pMsgQueue->sMsgSemConsumer) != 0)
		return FALSE;

	bPop = PopMsgNode(pMsgQueue, pMsg);

	if (pMsgQueue->iMaxMsgLimit > 0)
		EMU_sem_post(&pMsgQueue->sMsgSemProducer);

	return bPop;
}

BOOL IsMsgExist(MSG_QUEUE *pMsgQueue, MSG_T *pMsg)
{
	BOOL bFound = FALSE;
	LIST *plMsgList;
	LISTNODE *pNode;

	/* Check if pMsg is found in pMsgQueue */
	if (pthread_mutex_lock(&pMsgQueue->ptmMsg) != 0)
		return FALSE;
	plMsgList = pMsgQueue->plMsgList;
	for (pNode = plMsgList->pFirstNode; pNode != plMsgList->pLastNode; pNode = pNode->pNextNode)
	{
		if (memcmp((MSG_T *)pNode->pValue, pMsg, sizeof(MSG_T)) == 0)
		{
			bFound = TRUE;
			break;
		}
	}
	pthread_mutex_unlock(&pMsgQueue->ptmMsg);

	return bFound;
}



BOOL SELECT_MSG_Process(FD_T *pFD, BOOL bReadable, BOOL bWritable)
{
	SELECT_MSG *pSelectMsg;
	pSelectMsg = (SELECT_MSG *)((char *)(pFD) - (unsigned long)(&((SELECT_MSG *)0)->fd));
//	PTI;
	if (!bReadable) return TRUE;

//	PTI;
	if (GetMsg(pSelectMsg->pMsgQueue, pSelectMsg->pMsg))
	{
//		PTI;
		return TRUE;
	}
	else
	{
		PTE;
		pSelectMsg->pMsg->lMsg = MSG_NONE;
		return FALSE;
	}
//	PTI;
}

OBJHEADER *SELECT_MSG_Init(OBJHEADER *pObj, va_list vaInit)
{
	SELECT_MSG *pThis = (SELECT_MSG *)pObj;
	MSG_QUEUE *pMsgQueue = va_arg(vaInit, MSG_QUEUE *);
	MSG_T *pMsg = va_arg(vaInit, MSG_T *);

	if (pMsgQueue == NULL
		|| pMsgQueue->aiFD[0] < 0) goto lError;
	pThis->pMsgQueue = pMsgQueue;
	if (pMsg == NULL) goto lError;
	pThis->pMsg = pMsg;

	if (ObjInit((OBJHEADER *)pThis, sizeof(FD_T), (OBJHEADER *)&pThis->fd,
			FD_Final, FD_Init,
			pMsgQueue->aiFD[0], SE_TP_MSG, 0, -1, TRUE, FALSE, SELECT_MSG_Process) == NULL)
		goto lError;
	return pObj;
lError:
	return NULL;
}

void SELECT_MSG_Final(OBJHEADER *pObj)
{
	SELECT_MSG *pThis = (SELECT_MSG *)pObj;
	ObjFinal((OBJHEADER *)&pThis->fd);
}
