#include "CommonDef.h"

pid_t g_pidPPPoE[MAX_PPPOE_AND_UNNUM];

BOOL WRRunPPPoE( int id, 
			unsigned long loc_addr,
			unsigned long network,
			unsigned long netmask,
			unsigned long dnsaddr,
			unsigned long dnsaddr2,
			char *pcPPPoEUserName,
			char *pcPPPoEUserPass,
			char *pcPPPoEServiceName,
			char *pcPPPoEACName,
			unsigned long ulPPPoEIdle,
			BOOL bIsDialOnDemand,
			unsigned short mtu);

BOOL ppp_read_config(int id, struct ppp_info *pppd_info)
{
	FILE *fp;
	struct ppp_info info;
	char config_file[64];
	
	sprintf(config_file,"%s_%d", PPPOE_CONF_FILE, id );
	
	if (pppd_info == NULL || config_file == NULL)
	{
		fprintf(stderr, "illegal call function read_config!\n");
		return FALSE;
	}

	if ((fp=fopen(config_file, "rb")) == NULL)
	{
		return FALSE;
	}

	if (fread(&info, 1, sizeof(info), fp) == sizeof(info))
  	{
		fclose(fp);
		memcpy(pppd_info, &info, sizeof(info));
  		return TRUE;
  	}
  	else
  	{
		fprintf(stderr, "read_config, read file %s failed.\n", config_file);
		fclose(fp);
		return FALSE;
	}
}

BOOL ppp_write_config(int id, struct ppp_info *pppd_info)
{
	FILE *fp;
	char config_file[64];
	
	sprintf(config_file,"%s_%d", PPPOE_CONF_FILE, id );
	
	if (pppd_info == NULL || config_file == NULL)
	{
		fprintf(stderr, "illegal call function write_config!\n");
		return FALSE;
	}

	if ((fp=fopen(config_file,"wb")) == NULL)
	{
		fprintf(stderr, "write_config, open file %s failed.\n", config_file);
		return FALSE;
	}

	if(fwrite(pppd_info, 1, sizeof(struct ppp_info), fp) == sizeof(struct ppp_info))
  	{
		fclose(fp);
		return TRUE;
  	}
	else
	{
		fprintf(stderr, "write_config, write file %s failed.\n", config_file);
		fclose(fp);
		return FALSE;
	}
}


int WRGetPPPoEState(int id)
{
	struct ppp_info pInfo;

	if (!ppp_read_config(id, &pInfo))
		return PPPOE_NO_RUN;

	if ((pInfo.status & PPPOE_RUN_BIT) == 0L)
		return PPPOE_NO_RUN;

	if ((pInfo.status & PPPOE_DOD_BIT) == 0L)
	{
		switch (pInfo.status & PPPOE_CONN_BIT)
		{
		case CONNECTION_ON:
			return PPPOE_COMM_ON;
			break;
		case CONNECTION_TRYING:
			return PPPOE_COMM_TRYING;
			break;
		default:
			//fprintf(stderr, "common mode/offline\n");
			return PPPOE_ERROR;
		}
	}
	else
	{
		switch (pInfo.status & PPPOE_CONN_BIT)
		{
		case CONNECTION_ON:
			return PPPOE_DOD_ON;
			break;
		case CONNECTION_TRYING:
			return PPPOE_DOD_TRYING;
			break;
		case  CONNECTION_WAITING:
			return PPPOE_DOD_WAITING;
			break;
		case CONNECTION_OFF:
			return PPPOE_DOD_OFF;
			break;
		default:
			fprintf(stderr, "dial-on-demand mode/no state.\n");
			return PPPOE_ERROR;
		}
	}

	return PPPOE_ERROR;
}

#ifdef DEBUG
int Http_GetPPPoE(HTTPCONNECTION hConnection, void *pParam)
{
	char ac[2048];
	int i;
	int iCount;
	int iLen;
	char *pc;
	char acHex[4];
	struct ppp_info info;

	if (!ReadWholeFile(PPPOE_CONF_FILE, &pc, &iLen))
	{
		sprintf(ac, "File not found:\n%s\n", PPPOE_CONF_FILE);
		AddHttpBodyString(hConnection, ac);
	}
	else
	{
		iCount = 0;
		for (i=0; i<iLen; i++)
		{
			Char2Hex(pc[i], acHex);
			iCount += sprintf(ac + iCount, "%s ", acHex);
			if ((i+1)%8 == 0)
			{
				iCount += sprintf(ac + iCount, " ");
			}
			if ((i+1)%16 == 0 || i+1>=iLen)
			{
				iCount += sprintf(ac + iCount, "\n");
				AddHttpBodyString(hConnection, ac);
				iCount = 0;
			}
		}

		if (iLen == sizeof(info))
		{
			memcpy(&info, pc, sizeof(info));
			iCount = 0;
			iCount += sprintf(ac+iCount, "\nloc_addr=%s\n", info.loc_addr);
			iCount += sprintf(ac+iCount, "rem_addr=%s\n", info.rem_addr);
			for (i=0; i<3; i++)
				iCount += sprintf(ac+iCount, "dnsaddr[%d]=%s\n", i, info.dnsaddr[i]);
			iCount += sprintf(ac+iCount, "gateway=%s\n", info.gateway);
			iCount += sprintf(ac+iCount, "netmask=%s\n", info.netmask);
			iCount += sprintf(ac+iCount, "status=%08x\n", info.status);
			iCount += sprintf(ac+iCount, "live_sec=%d\n", (int)info.live_sec);
			iCount += sprintf(ac+iCount, "session_num=%d\n", (int)info.session_num);
			iCount += sprintf(ac+iCount, "bytes_in=%u\n", info.bytes_in);
			iCount += sprintf(ac+iCount, "bytes_out=%u\n", info.bytes_out);
			iCount += sprintf(ac+iCount, "StartParam.e.pcUserName=%s\n", info.StartParam.e.pcUserName);
			iCount += sprintf(ac+iCount, "StartParam.e.pcUserPass=%s\n", info.StartParam.e.pcUserPass);
			iCount += sprintf(ac+iCount, "StartParam.e.pcServiceName=%s\n", info.StartParam.e.pcServiceName);
			iCount += sprintf(ac+iCount, "StartParam.e.DisConnectTime=%d\n", (int)info.StartParam.e.DisConnectTime);
			iCount += sprintf(ac+iCount, "thttpd_pid=%d\n", (int)info.thttpd_pid);
			AddHttpBodyString(hConnection, ac);
		}
		else
			AddHttpBodyString(hConnection, "\nError in state file.\n");

		free(pc);
	}

	AddHttpBodyString(hConnection, "\nCurrent State:\n");
	switch (WRGetPPPoEState())
	{
	case PPPOE_ERROR:
		sprintf(ac, "PPPOE_ERROR");
		break;
	case PPPOE_NO_RUN:
		sprintf(ac, "PPPOE_NO_RUN");
		break;
	case PPPOE_DOD_OFF:
		sprintf(ac, "PPPOE_DOD_OFF");
		break;
	case PPPOE_DOD_TRYING:
		sprintf(ac, "PPPOE_DOD_TRYING");
		break;
	case PPPOE_DOD_ON:
		sprintf(ac, "PPPOE_DOD_ON");
		break;
	case PPPOE_COMM_TRYING:
		sprintf(ac, "PPPOE_COMM_TRYING");
		break;
	case PPPOE_DOD_WAITING:
		sprintf(ac,"PPPOE_DOD_WAITING");
	case PPPOE_COMM_ON:
		sprintf(ac, "PPPOE_COMM_ON");
		break;
	}
	sprintf(ac+strlen(ac), "\n");
	AddHttpBodyString(hConnection, ac);

	SetHttpHeader(hConnection, 200, "OK", "", "Expires: 01 Jan 1970 00:00:00 GMT\r\n", "text/plain", TRUE);
	return 0;
}
#endif


void set_pppoe_state(int id, int state)
{
	struct ppp_info pInfo;
	
	if (!ppp_read_config(id, &pInfo))
		memset(&pInfo, 0, sizeof(pInfo));
	pInfo.status =  state;
	ppp_write_config(id, &pInfo);
}

unsigned short getPidFromFile(char *filename)
{
	FILE *fp = NULL;
	char buf[24];
	int len =0;
	unsigned short pid = 0;
	
	if( (fp=fopen(filename,"rt"))==NULL )
		return 0;

	memset(buf,0,sizeof(buf));
	if( fgets(buf,sizeof(buf)-1,fp) )
	{
		len = strlen(buf);
		if( buf[len-1]=='\n' )
			buf[len-1]='\0';
		if( buf[len-2]=='\r' )
			buf[len-2]='\0';
		
		if( buf[0] )
			pid =  atoi(buf);
	}
	fclose(fp);
	
	return pid;
}


void WRKillPPPoE(int id)
{
	int i;
	char buf[255];
	
	if( id == ALL_ID )
	{
		for( i=0; i<MAX_PPPOE_AND_UNNUM; i++)
		{
			snprintf(buf,sizeof(buf)-1, "/swap/pppoe-eth1.%d.pid", i);
			g_pidPPPoE[i] = getPidFromFile(buf);
			printf("kill pid %d\n", g_pidPPPoE[i] );
			if( g_pidPPPoE[i] )
				//SafeLimitKill(g_pidPPPoE[i], SIGTERM, 5000);
				kill(g_pidPPPoE[i], SIGTERM);
			set_pppoe_state(i,CONNECTION_OFF);
		}
	}
	else
	{
		snprintf(buf,sizeof(buf)-1, "/swap/pppoe-eth1.%d.pid", id);
		g_pidPPPoE[id] = getPidFromFile(buf);
		printf("kill pid %d\n",g_pidPPPoE[ id ] );
		if( g_pidPPPoE[id] )
			//SafeLimitKill(g_pidPPPoE[id], SIGTERM, 5000);
			kill(g_pidPPPoE[id],SIGTERM);
		set_pppoe_state(id,CONNECTION_OFF);
	}
	usleep(30 * 1000);//for lesser than 20000, usleep do not really sleep.
	
}

void WRRunPPPoEClient(int id)
{
	int i;
		
	if( id == PPPOE_ID )
	{
		for( i=1; i<MAX_PPPOE_AND_UNNUM ; i++)
		{
			if( g_ConfigParam.pppoeConf[i].enable == 0 )
				continue;

		    //A.ROOT-SERVERS.NET: 198.41.0.4,  B.ROOT-SERVERS.NET: 128.9.0.107
		    // htonl(0xc2290004);
		    // htonl(0x8009006b);

			if( g_ConfigParam.pppoeConf[i].dnsaddr[0]==0 && 
				g_ConfigParam.pppoeConf[i].dnsaddr[1] == 0 )
				create_resolv( htonl(0xc2290004), htonl(0x8009006b) , i);
			else
				create_resolv( g_ConfigParam.pppoeConf[i].dnsaddr[0], g_ConfigParam.pppoeConf[i].dnsaddr[1], i );

			WRRunPPPoE(i,
				g_ConfigParam.pppoeConf[i].loc_addr,
				g_ConfigParam.pppoeConf[id].network,
				g_ConfigParam.pppoeConf[i].netmask,
				g_ConfigParam.pppoeConf[i].dnsaddr[0],
				g_ConfigParam.pppoeConf[i].dnsaddr[1],
				g_ConfigParam.pppoeConf[i].user,
				g_ConfigParam.pppoeConf[i].pass,
				g_ConfigParam.pppoeConf[i].srvname,
				g_ConfigParam.pppoeConf[i].acname,
				g_ConfigParam.pppoeConf[i].idle,
				g_ConfigParam.pppoeConf[i].demand,
				g_ConfigParam.usMtu );
			usleep(20 * 1000);//for lesser than 20000, usleep do not really sleep.
		}
	}
	else if( id == UNNUM_ID )
	{
		if( g_ConfigParam.pppoeConf[id].dnsaddr[0]==0 && 
			g_ConfigParam.pppoeConf[id].dnsaddr[1] == 0 )
			create_resolv( htonl(0xc2290004), htonl(0x8009006b) , id );
		else
			create_resolv( g_ConfigParam.pppoeConf[id].dnsaddr[0], g_ConfigParam.pppoeConf[id].dnsaddr[1], id );

		//if( g_ConfigParam.pppoeConf[i].enable )
		{
			WRRunPPPoE(id,
				g_ConfigParam.pppoeConf[id].loc_addr,
				g_ConfigParam.pppoeConf[id].network,
				g_ConfigParam.pppoeConf[id].netmask,
				g_ConfigParam.pppoeConf[id].dnsaddr[0],
				g_ConfigParam.pppoeConf[id].dnsaddr[1],
				g_ConfigParam.pppoeConf[id].user,
				g_ConfigParam.pppoeConf[id].pass,
				g_ConfigParam.pppoeConf[id].srvname,
				g_ConfigParam.pppoeConf[id].acname,
				g_ConfigParam.pppoeConf[id].idle,
				g_ConfigParam.pppoeConf[id].demand,
				g_ConfigParam.usMtu);
		}
	}
}

BOOL WRRunPPPoE( int id, 
			unsigned long loc_addr,
			unsigned long network,
			unsigned long netmask,
			unsigned long dnsaddr,
			unsigned long dnsaddr2,
			char *pcPPPoEUserName,
			char *pcPPPoEUserPass,
			char *pcPPPoEServiceName,
			char *pcPPPoEACName,
			unsigned long ulPPPoEIdle,
			BOOL bIsDialOnDemand,
			unsigned short mtu)
{
	struct ppp_info pInfo;
	char idstr[4];
	struct in_addr addr;
	
	//WRRunPPPoEClient() will call by main() after boot 
	if (!ppp_read_config(id, &pInfo))
		memset(&pInfo, 0, sizeof(pInfo));

	addr.s_addr = loc_addr;
	strlcpy(pInfo.loc_addr, inet_ntoa(addr), sizeof(pInfo.loc_addr));
	addr.s_addr = network;
	strlcpy(pInfo.network, inet_ntoa(addr), sizeof(pInfo.network));
	addr.s_addr = netmask;
	strlcpy(pInfo.netmask, inet_ntoa(addr), sizeof(pInfo.netmask));
	addr.s_addr = dnsaddr;
	strlcpy(pInfo.dnsaddr[0], inet_ntoa(addr), sizeof(pInfo.dnsaddr[0]));
	addr.s_addr = dnsaddr2;
	strlcpy(pInfo.dnsaddr[1], inet_ntoa(addr), sizeof(pInfo.dnsaddr[0]));

	strlcpy(pInfo.StartParam.e.pcUserName, pcPPPoEUserName, sizeof(pInfo.StartParam.e.pcUserName));
	strlcpy(pInfo.StartParam.e.pcUserPass, pcPPPoEUserPass, sizeof(pInfo.StartParam.e.pcUserPass));
	strlcpy(pInfo.StartParam.e.pcServiceName, pcPPPoEServiceName, sizeof(pInfo.StartParam.e.pcServiceName));
	strlcpy(pInfo.StartParam.e.pcACName, pcPPPoEACName, sizeof(pInfo.StartParam.e.pcACName));
	pInfo.StartParam.e.DisConnectTime = ulPPPoEIdle*60;
	pInfo.thttpd_pid = getpid();
	pInfo.status = CONNECTION_TRYING;
	pInfo.id = id;
	pInfo.mtu = mtu;
	
	if (bIsDialOnDemand) 
		pInfo.status |= PPPOE_DOD_BIT;

	if (!ppp_write_config(id, &pInfo)) 
		return FALSE;

	sprintf(idstr,"%d",id);
	
	//return RunProgram(&g_pidPPPoE[id], PROG_PPPOE, "pppoe", idstr,NULL);
	return RunProgram(NULL, PROG_PPPOE, "pppoe", idstr,NULL);
	
}

void SendPPPoECommand( int id, int iCommand)
{
	int iCurState;
	char buf[255];
	
	snprintf(buf,sizeof(buf)-1, "/swap/pppoe-eth1.%d.pid", id);
	g_pidPPPoE[id] = getPidFromFile(buf);

	SynProcess(&g_pidPPPoE[id]);
	iCurState = WRGetPPPoEState(id);

	if (iCurState == PPPOE_ERROR)
		fprintf(stderr, "PPPoE runs in error state!\n");
	else if (iCurState == PPPOE_NO_RUN && g_pidPPPoE[id] != 0)
	{
		fprintf(stderr, "PPPoE state error!\n");
		WRKillPPPoE(id);
	}
	else if (iCurState != PPPOE_NO_RUN && g_pidPPPoE[id] == 0)
	{
		fprintf(stderr, "PPPoE is running in unknown state\n");
		WRKillPPPoE(id); 
	}

	if (PPPOE_CMD_RESET == iCommand)
	{
		fprintf(stderr,"PPPoE: reset!\n");
		if (iCurState != PPPOE_NO_RUN)
		{
			fprintf(stderr,"pppoe pid = %d\n", g_pidPPPoE[id]);
			WRKillPPPoE(id);
		}
	}
	else if (PPPOE_CMD_DOD == iCommand|| PPPOE_CMD_COMM == iCommand)
	{
		//if (iCurState == PPPOE_COMM_TRYING || iCurState == PPPOE_COMM_ON)
			SendPPPoECommand( id, PPPOE_CMD_RESET);

		//iCurState = WRGetPPPoEState(id);
		//if (iCurState == PPPOE_NO_RUN)
		{
			usleep(2000 * 1000);//for lesser than 20000, usleep do not really sleep.
			WRRunPPPoE( id, 
					g_ConfigParam.pppoeConf[id].loc_addr,
					g_ConfigParam.pppoeConf[id].network,
					g_ConfigParam.pppoeConf[id].netmask,
					g_ConfigParam.pppoeConf[id].dnsaddr[0],
					g_ConfigParam.pppoeConf[id].dnsaddr[1],
					g_ConfigParam.pppoeConf[id].user,
					g_ConfigParam.pppoeConf[id].pass,
					g_ConfigParam.pppoeConf[id].srvname,
					g_ConfigParam.pppoeConf[id].acname,
					g_ConfigParam.pppoeConf[id].idle,
					(PPPOE_CMD_DOD == iCommand)?TRUE:FALSE,
					g_ConfigParam.usMtu);
		}
	}
#ifdef WINBOND_CMD_COMM
	else if (PPPOE_CMD_COMM == iCommand)
	{
		fprintf(stderr,"PPPoE: switch to common mode.\n");
		if (iCurState == PPPOE_DOD_OFF || iCurState == PPPOE_DOD_TRYING || iCurState == PPPOE_DOD_ON)
			SendPPPoECommand( id, PPPOE_CMD_RESET);
	}
#endif
	else if (PPPOE_CMD_CONNECT == iCommand)
	{
		fprintf(stderr,"PPPoE: connect in common mode.\n");
		SendPPPoECommand( id, PPPOE_CMD_RESET);

		usleep(2000 * 1000);//for lesser than 20000, usleep do not really sleep.
		WRRunPPPoE( id, 
				g_ConfigParam.pppoeConf[id].loc_addr,
				g_ConfigParam.pppoeConf[id].network,
				g_ConfigParam.pppoeConf[id].netmask,
				g_ConfigParam.pppoeConf[id].dnsaddr[0],
				g_ConfigParam.pppoeConf[id].dnsaddr[1],
				g_ConfigParam.pppoeConf[id].user,
				g_ConfigParam.pppoeConf[id].pass,
				g_ConfigParam.pppoeConf[id].srvname,
				g_ConfigParam.pppoeConf[id].acname,
				g_ConfigParam.pppoeConf[id].idle,
				FALSE,
				g_ConfigParam.usMtu);
	}
	else if (PPPOE_CMD_DISCONNECT == iCommand)
	{
		fprintf(stderr,"PPPoE: disconnect.\n");
		/* norton test
		if (iCurState == PPPOE_DOD_OFF
			|| iCurState == PPPOE_DOD_TRYING
			|| iCurState == PPPOE_DOD_ON
			|| iCurState == PPPOE_COMM_TRYING
			|| iCurState == PPPOE_COMM_ON)
		*/
		{
			WRKillPPPoE( id );
		}
	}

}

