#include "CommonDef.h"

extern unsigned g_model_id;
extern char g_model_name[32];

BOOL ReadFlashMemory(WR_CONFIG_PARAM_T *pwConfigParam)
{
    int i=0;
    	
    for(i=0; i<10; i++)
    {
	int iLen;

#ifdef EMBED
	int iOffsetConfig = GetFlashSize() - sizeof(WR_CONFIG_PARAM_T);
	if (iOffsetConfig < 0) return FALSE;

	iOffsetConfig = 0x7F1F0000; //norton
	iLen = ReadWinbondFlash(iOffsetConfig, sizeof(WR_CONFIG_PARAM_T), (char *)pwConfigParam);
#else
	FILE *fp;

	fp = fopen(FLASH_CONFIG_FILE, "rb");
	if (fp == NULL) return FALSE;

	iLen = fread((void *)pwConfigParam, 1, sizeof(WR_CONFIG_PARAM_T), fp);
	fclose(fp);
#endif

	if ( (iLen == sizeof(WR_CONFIG_PARAM_T)) &&
		GetConfigCheckSum(pwConfigParam) == pwConfigParam->ulCheckSum)
	{
		return TRUE;
	}
	else if( i==9)
		fprintf(stderr, "Flash config: checksum error!\n");
    }

    return FALSE;
}

BOOL WriteFlashMemory(WR_CONFIG_PARAM_T *pwConfigParam)
{
	return __WriteFlashMemory(pwConfigParam, TRUE);
}

BOOL __WriteFlashMemory(WR_CONFIG_PARAM_T *pwConfigParam, BOOL bResetCheckSum)
{
	if (bResetCheckSum)
		pwConfigParam->ulCheckSum = GetConfigCheckSum(pwConfigParam);

	SendFlashMsgWriteConfig(pwConfigParam);

	return TRUE;
}

BOOL ReadConfig(char *pcFileName )
{
	char acValue[128];

#define __ReadDefault(pcSeg, pcName, acVal) \
	{if (!ReadConfigValue(pcFileName, pcSeg, pcName, acVal, sizeof(acVal))) acVal[0] = '\0';}

	__ReadDefault("BasicSetup", "ModelName", acValue);
	MyStrncpy(g_model_name, acValue, sizeof(g_model_name));

	__ReadDefault("BasicSetup", "ModelId", acValue);
	g_model_id = strtoul(acValue,NULL,16);
	printf("model =%s , id = 0x%04x\n ",g_model_name, g_model_id );
	return 0;
}

BOOL ReadNatINI(char *pcFileName, WR_CONFIG_PARAM_T *pwConfigParam)
{
	int i;
	char acName[24];
	char acValue[128];
	LIST *pList;
	LISTNODE *pNode;


#define __ReadDefault(pcSeg, pcName, acVal) \
	{if (!ReadConfigValue(pcFileName, pcSeg, pcName, acVal, sizeof(acVal))) acVal[0] = '\0';}

	memset(pwConfigParam, 0, sizeof(WR_CONFIG_PARAM_T));

	/*  BasicSetup.htm  */
	__ReadDefault("BasicSetup", "HostName", acValue);
	MyStrncpy(pwConfigParam->acHostName, acValue, sizeof(pwConfigParam->acHostName));
	__ReadDefault("BasicSetup", "DomainName", acValue);
	MyStrncpy(pwConfigParam->acDomainName, acValue, sizeof(pwConfigParam->acDomainName));
	
	__ReadDefault("BasicSetup", "PrivateIPAddress", acValue);
	pwConfigParam->ulPrivateIPAddress = String2IP(acValue);
	__ReadDefault("BasicSetup", "PrivateSubnetMask", acValue);
	pwConfigParam->ulPrivateSubnetMask = String2IP(acValue);
	__ReadDefault("BasicSetup", "HowToGetIp", acValue);
	pwConfigParam->ulHowToGetIp = atoi(acValue);

	__ReadDefault("BasicSetup", "PublicIPAddress", acValue);
	pwConfigParam->ulPublicIpAddress = String2IP(acValue);
	__ReadDefault("BasicSetup", "PublicSubnetMask", acValue);
	pwConfigParam->ulPublicSubnetMask = String2IP(acValue);
	__ReadDefault("BasicSetup", "DefaultGateway", acValue);
	pwConfigParam->ulDefaultGateway = String2IP(acValue);
	for (i=0; i<MAX_DNS_SERVER; i++)
	{
		sprintf(acName, "DNSServer%d", i);
		__ReadDefault("BasicSetup", acName, acValue);
		pwConfigParam->aulDNSServer[i] = String2IP(acValue);
	}
	__ReadDefault("BasicSetup", "TimeZone", acValue);
	MyStrncpy(pwConfigParam->tz.va_gmt, acValue, sizeof(pwConfigParam->tz.va_gmt));
	__ReadDefault("BasicSetup", "Sntp0", acValue);
	MyStrncpy(pwConfigParam->tz.va_sntp0, acValue, sizeof(pwConfigParam->tz.va_sntp1));
	__ReadDefault("BasicSetup", "Sntp1", acValue);
	MyStrncpy(pwConfigParam->tz.va_sntp1, acValue, sizeof(pwConfigParam->tz.va_sntp1));
	__ReadDefault("BasicSetup", "Sntp2", acValue);
	MyStrncpy(pwConfigParam->tz.va_sntp2, acValue, sizeof(pwConfigParam->tz.va_sntp1));

	/*  Dhcp.htm  */
	__ReadDefault("Dhcp", "EnableDHCPServer", acValue);
	pwConfigParam->dhcp.enable = String2Bool(acValue);
	__ReadDefault("Dhcp", "DHCPLowAddress", acValue);
	pwConfigParam->dhcp.startip = String2IP(acValue);
	__ReadDefault("Dhcp", "DHCPHighAddress", acValue);
	pwConfigParam->dhcp.endip = String2IP(acValue);
	__ReadDefault("Dhcp", "DHCPLeaseTime", acValue);
	pwConfigParam->dhcp.leasetime = String2Long(acValue);

#ifdef WINBOND_URL
	/*  UrlFilter.htm  */
	__ReadDefault("UrlFilter", "EnableURLLimit", acValue);
	pwConfigParam->bEnableURLLimit = String2Bool(acValue);
	__ReadDefault("UrlFilter", "EnableWebsite", acValue);
	pwConfigParam->bEnableWebsite = String2Bool(acValue);
	for (i=0; i<MAX_URL_FILTER_ITEM; i++)
	{
		sprintf(acName, "WebSite%d", i);
		__ReadDefault("UrlFilter", acName, acValue);
		MyStrncpy(pwConfigParam->aacWebSite[i], acValue, sizeof(pwConfigParam->aacWebSite[i]));
	}
#endif

#ifdef WINBOND_IPFW
	/*  IpPortFilter.htm  */
	for (i=0; i<MAX_IP_PORT_FILTER_ITEM; i++)
	{
		LANIPFILTERITEM_T *pItem;

		sprintf(acName, "IpFilter%d", i);
		__ReadDefault("IpPortFilter", acName, acValue);

		pItem = pwConfigParam->alIpPortFilterItem + i;
		memset(pItem, 0, sizeof(LANIPFILTERITEM_T));

		pList = SplitString(acValue, ',');
		pNode = (pList==NULL?NULL:pList->pFirstNode);

		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->ulIPAddressFrom = String2IP((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->ulIPAddressTo = String2IP((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->usPortFrom = (unsigned short)String2Long((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->usPortTo = (unsigned short)String2Long((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		for (j=0; j<4; j++)
		{
			if (pNode != NULL && pNode != pList->pLastNode)
			{
				pItem->ausPortOther[j] = (unsigned short)String2Long((char *)pNode->pValue);
				pNode = pNode->pNextNode;
			}
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->cProtocolType = GetIPProtocol(pNode->pValue);
		}

		if (pList) DeleteSplitString(pList);
	}
#endif

#ifdef WINBOND_VS

	/*  VServer.htm  */
	for (i=0; i<MAX_VIRTUAL_HOST_ITEM; i++)
	{
		VIRTUALHOSTITEM_T *pItem;

		sprintf(acName, "VirtualServer%d", i);
		__ReadDefault("VServer", acName, acValue);

		pItem = pwConfigParam->avVirtualHostItem + i;
		memset(pItem, 0, sizeof(VIRTUALHOSTITEM_T));

		pList = SplitString(acValue, ',');
		pNode = (pList==NULL?NULL:pList->pFirstNode);
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->ulIPAddress = String2IP((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->usPortFrom = (unsigned short)String2Long((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->usPortTo = (unsigned short)String2Long((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->cProtocolType = GetIPProtocol((char *)pNode->pValue);
		}
		if (pList) DeleteSplitString(pList);
	}
#endif

#ifdef WINBOND_DMZ
	/*  Expose.htm  */
	__ReadDefault("Expose", "DMZHost", acValue);
	pwConfigParam->ulDMZHost = String2IP(acValue);
#endif

	/*  Admin.htm  */
	__ReadDefault("Admin", "HttpPassword", acValue);
	if(strlen(acValue) <= 2)
		pwConfigParam->acHttpPassword[0] = '\0';
	else
	{
		acValue[strlen(acValue) - 1] = '\0';
		MyStrncpy(pwConfigParam->acHttpPassword, acValue + 1, sizeof(pwConfigParam->acHttpPassword));
	}

	__ReadDefault("Admin", "LanMacReconfig", acValue);
	pwConfigParam->bLanMacReconfig = String2Bool(acValue);

	__ReadDefault("Admin", "LanMacAddress", acValue);
	pList = SplitString(acValue, ',');
	pNode = pList->pFirstNode;
	memset(pwConfigParam->acLanMacAddress, 0, sizeof(pwConfigParam->acLanMacAddress));
	for (i=0; i<6; i++)
	{
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pwConfigParam->acLanMacAddress[i] = Hex2Char((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		else break;
	}
	if (pList) DeleteSplitString(pList);

	__ReadDefault("Admin", "WanMacReconfig", acValue);
	pwConfigParam->bWanMacReconfig = String2Bool(acValue);

	__ReadDefault("Admin", "WanMacAddress", acValue);
	pList = SplitString(acValue, ',');
	pNode = pList->pFirstNode;
	memset(pwConfigParam->acWanMacAddress, 0, sizeof(pwConfigParam->acWanMacAddress));
	for (i=0; i<6; i++)
	{
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pwConfigParam->acWanMacAddress[i] = Hex2Char((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		else break;
	}
	if (pList) DeleteSplitString(pList);

	__ReadDefault("Admin", "ExternalAdmin", acValue);
	pwConfigParam->bEnableAdminWan = String2Bool(acValue);
	pwConfigParam->bEnableAdminLan = 1;

	__ReadDefault("Admin", "ExternalAdminPort", acValue);
	pwConfigParam->usExternalAdminPort = (unsigned short)String2Long(acValue);

	__ReadDefault("Admin", "UPNP", acValue);
	pwConfigParam->bEnableUPNP = String2Bool(acValue);

#ifdef WINBOND_ROUTE
	/*  DRouter.htm  */
	__ReadDefault("DRouter", "DrTX", acValue);
	pwConfigParam->cDrTX = GetRoutingProtocol(acValue);
	__ReadDefault("DRouter", "DrRX", acValue);
	pwConfigParam->cDrRX = GetRoutingProtocol(acValue);

	/*  SRouter.htm  */
	for (i=0; i<MAX_STATIC_ROUTING_ITEM; i++)
	{
		ROUTINGTABLE_T *pItem;

		sprintf(acName, "SRouting%d", i);
		__ReadDefault("SRouter", acName, acValue);

		pItem = pwConfigParam->arRoutingTableItem + i;
		memset(pItem, 0, sizeof(ROUTINGTABLE_T));

		pList = SplitString(acValue, ',');
		pNode = (pList==NULL?NULL:pList->pFirstNode);

		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->ulDestLanIPAddress = String2IP((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->ulSubnetMask = String2IP((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->ulDefaultGateway = String2IP((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->lMetric = (long)String2Long((char *)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->cNetInterface = GetNetInterface((char *)pNode->pValue);
		}
		if (pList != NULL) DeleteSplitString(pList);
	}
#endif

#ifdef WINBOND_PPPOE
	/*  PPPoE.htm  */
	//__ReadDefault("PPPoE", "EnablePPPoE", acValue);
	//pwConfigParam->bEnablePPPoE = String2Bool(acValue);
	__ReadDefault("PPPoE", "EnablePPPoEAutoConnect", acValue);
	pwConfigParam->bEnablePPPoEAutoConnect = String2Bool(acValue);
	__ReadDefault("PPPoE", "PPPoEUserName", acValue);
	MyStrncpy(pwConfigParam->acPPPoEUserName, acValue, sizeof(pwConfigParam->acPPPoEUserName));
	__ReadDefault("PPPoE", "PPPoEUserPass", acValue);
	MyStrncpy(pwConfigParam->acPPPoEUserPass, acValue, sizeof(pwConfigParam->acPPPoEUserPass));
	__ReadDefault("PPPoE", "PPPoEServiceName", acValue);
	MyStrncpy(pwConfigParam->acPPPoEServiceName, acValue, sizeof(pwConfigParam->acPPPoEServiceName));
	__ReadDefault("PPPoE", "PPPoEDisConnectTime_InSec", acValue);
	pwConfigParam->ulPPPoEDisConnectTime_InSec = String2Long(acValue);
#endif

	__ReadDefault("PPPoE", "EnableNaptUnnum", acValue);
	pwConfigParam->bEnableNaptUnnum = String2Bool(acValue);
	
#ifdef WINBOND_MODEM
	/*  Modem.htm  */
	__ReadDefault("Modem", "EnableModem", acValue);
	pwConfigParam->bEnableModem = String2Bool(acValue);
	__ReadDefault("Modem", "EnableModemAutoConnect", acValue);
	pwConfigParam->bEnableModemAutoConnect = String2Bool(acValue);
	__ReadDefault("Modem", "ModemUserName", acValue);
	MyStrncpy(pwConfigParam->acModemUserName, acValue, sizeof(pwConfigParam->acModemUserName));
	__ReadDefault("Modem", "ModemUserPass", acValue);
	MyStrncpy(pwConfigParam->acModemUserPass, acValue, sizeof(pwConfigParam->acModemUserPass));
	__ReadDefault("Modem", "ModemPhone", acValue);
	MyStrncpy(pwConfigParam->acModemPhone, acValue, sizeof(pwConfigParam->acModemPhone));
	__ReadDefault("Modem", "ModemDisConnectTime_InSec", acValue);
	pwConfigParam->ulModemDisConnectTime_InSec = String2Long(acValue);
#endif


#ifdef WINBOND_VPN
	/*  VPN.htm  */
	sprintf(acName, "VPN");
	__ReadDefault(acName, "EnableVPN", acValue);
	pwConfigParam->bEnableVPN = String2Bool(acValue);
	__ReadDefault(acName, "VPNUserName", acValue);
	MyStrncpy(pwConfigParam->acVPNUserName, acValue, sizeof(pwConfigParam->acVPNUserName));
	__ReadDefault(acName, "VPNUserPass", acValue);
	MyStrncpy(pwConfigParam->acVPNUserPass, acValue, sizeof(pwConfigParam->acVPNUserPass));
	__ReadDefault(acName, "VPNHostName", acValue);
	MyStrncpy(pwConfigParam->acVPNHostName, acValue, sizeof(pwConfigParam->acVPNHostName));
	__ReadDefault(acName, "VPNDisConnectTime_InSec", acValue);
	pwConfigParam->ulVPNDisConnectTime_InSec = String2Long(acValue);
#endif

#ifdef WINBOND_SPECIAL_APP

	/*  SpecialApp.htm  */
	for (i=0; i<MAX_SPECIAL_APP_ITEM; i++)
	{
		SPECIALAPPLICATION_T *pItem;
		pItem = pwConfigParam->asSpecialAppItem + i;
		memset(pItem, 0, sizeof(SPECIALAPPLICATION_T));

		sprintf(acName, "Name_%d", i);
		__ReadDefault("SpecialApp", acName, acValue);
		MyStrncpy(pItem->acAppName, acValue, sizeof(pItem->acAppName));

		sprintf(acName, "Value_%d", i);
		__ReadDefault("SpecialApp", acName, acValue);

		pList = SplitString(acValue, ',');
		pNode = (pList==NULL?NULL:pList->pFirstNode);
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->usInPortFrom = String2Long((char*)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->usInPortTo = String2Long((char*)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->usOutPortFrom = String2Long((char*)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->usOutPortTo = String2Long((char*)pNode->pValue);
			pNode = pNode->pNextNode;
		}
		if (pNode != NULL && pNode != pList->pLastNode)
		{
			pItem->bIsActivated = String2Bool((char*)pNode->pValue);
		}
		if (pList != NULL) DeleteSplitString(pList);
	}
#endif

#ifdef WINBOND_ADV
	/*  Advance.htm  */
	__ReadDefault("Advance", "DosEnabled", acValue);
	pwConfigParam->bEnableDosDefense = String2Bool(acValue);
	__ReadDefault("Advance", "DosRatePerHour", acValue);
	pwConfigParam->ulDosRatePerHour = String2Long(acValue);
	__ReadDefault("Advance", "EnableHidePubIP", acValue);
	pwConfigParam->bEnableHidePubIP = String2Bool(acValue);
	__ReadDefault("Advance", "EnableLanMonitor", acValue);
	pwConfigParam->bEnableLanMonitor = String2Bool(acValue);
	__ReadDefault("Advance", "EnableWanMonitor", acValue);
	pwConfigParam->bEnableWanMonitor = String2Bool(acValue);
#endif

#ifdef WINBOND_DDNS
	__ReadDefault("Advance", "EnableDDNS", acValue);
	pwConfigParam->bEnableDDNS = String2Bool(acValue);
	__ReadDefault("Advance", "DDNSDomain", acValue);
	MyStrncpy(pwConfigParam->acDDNSDomainName, acValue, sizeof(pwConfigParam->acDDNSDomainName));
	__ReadDefault("Advance", "DDNSUserName", acValue);
	MyStrncpy(pwConfigParam->acDDNSUserName, acValue, sizeof(pwConfigParam->acDDNSUserName));
	__ReadDefault("Advance", "DDNSPassword", acValue);
	MyStrncpy(pwConfigParam->acDDNSUserPass, acValue, sizeof(pwConfigParam->acDDNSUserPass));
#endif

	/* Set CheckSum */
	pwConfigParam->ulCheckSum = GetConfigCheckSum(pwConfigParam);

	return TRUE;
}

#ifdef WINBOND_ORIG

BOOL WriteNatINI(char *pcFileName, WR_CONFIG_PARAM_T *pwConfigParam)
{
	int i;
	FILE *fp;
	char ac[32];

	fp = fopen(pcFileName, "wt");
	if (fp == NULL) return FALSE;

	/* д BasicSetup.htm  */
	fprintf(fp, "[BasicSetup]\n");
	fprintf(fp, "HostName = %s\n", pwConfigParam->acHostName);
	fprintf(fp, "DomainName = %s\n", pwConfigParam->acDomainName);
	IP2String(pwConfigParam->ulPrivateIPAddress, ac);
	fprintf(fp, "PrivateIPAddress = %s\n", ac);
	IP2String(pwConfigParam->ulPrivateSubnetMask, ac);
	fprintf(fp, "PrivateSubnetMask = %s\n", ac);
	fprintf(fp, "HowToGetIp = %d\n", (int)pwConfigParam->ulHowToGetIp);
	IP2String(pwConfigParam->ulPublicIpAddress, ac);
	fprintf(fp, "PublicIPAddress = %s\n", ac);
	IP2String(pwConfigParam->ulPublicSubnetMask, ac);
	fprintf(fp, "PublicSubnetMask = %s\n", ac);
	IP2String(pwConfigParam->ulDefaultGateway, ac);
	fprintf(fp, "DefaultGateway = %s\n", ac);
	for (i=0; i<MAX_DNS_SERVER; i++)
	{
		IP2String(pwConfigParam->aulDNSServer[i], ac);
		fprintf(fp, "DNSServer%d = %s\n", i, ac);
	}
	fprintf(fp, "\n");

#ifdef WINBOND_DHCP
	/* д Dhcp.htm  */
	fprintf(fp, "[Dhcp]\n");
	fprintf(fp, "EnableDHCPServer = %d\n", (int)pwConfigParam->bEnableDHCPServer);
	IP2String(pwConfigParam->ulDHCPAddressFrom, ac);
	fprintf(fp, "DHCPLowAddress = %s\n", ac);
	IP2String(pwConfigParam->ulDHCPAddressTo, ac);
	fprintf(fp, "DHCPHighAddress = %s\n", ac);
	fprintf(fp, "\n");
#endif

#ifdef WINBOND_URL
	/* д UrlFilter.htm  */
	fprintf(fp, "[UrlFilter]\n");
	fprintf(fp, "EnableURLLimit = %d\n", (int)pwConfigParam->bEnableURLLimit);
	fprintf(fp, "EnableWebsite = %d\n", (int)pwConfigParam->bEnableWebsite);
	for (i=0; i<MAX_URL_FILTER_ITEM; i++)
		fprintf(fp, "WebSite%d = %s\n", i, pwConfigParam->aacWebSite[i]);
	fprintf(fp, "\n");
#endif

#ifdef WINBOND_IPFW
	/* д IpPortFilter.htm  */
	fprintf(fp, "[IpPortFilter]\n");
	fprintf(fp, "# IpFilter<n> = <IPFrom>, <IPTo>, <PortProm>, <PortTo>, <Port1>, <Port2>, <Port3>, <Port4>, <TCP|UDP|TCPUDP>\n");
	for (i=0; i<MAX_IP_PORT_FILTER_ITEM; i++)
	{
		LANIPFILTERITEM_T *pItem;
		pItem = pwConfigParam->alIpPortFilterItem + i;
		IP2String(pItem->ulIPAddressFrom, ac);
		fprintf(fp, "IpFilter%d = %s, ", i, ac);
		IP2String(pItem->ulIPAddressTo, ac);
		fprintf(fp, "%s, ", ac);
		fprintf(fp, "%d, %d, %d, %d, %d, %d, %s\n",
			(int)pItem->usPortFrom,
			(int)pItem->usPortTo,
			(int)pItem->ausPortOther[0],
			(int)pItem->ausPortOther[1],
			(int)pItem->ausPortOther[2],
			(int)pItem->ausPortOther[3],
			GetIPProtocolString(pItem->cProtocolType));
	}
	fprintf(fp, "\n");
#endif

#ifdef WINBOND_VS
	/* д Vserver.htm  */
	fprintf(fp, "[Vserver]\n");
	fprintf(fp, "# VirtualServer<n> = <IP>, <PortFrom>, <PortTo>, <TCP|UDP|TCPUDP>\n");
	for (i=0; i<MAX_VIRTUAL_HOST_ITEM; i++)
	{
		VIRTUALHOSTITEM_T *pItem;
		pItem = pwConfigParam->avVirtualHostItem + i;
		IP2String(pItem->ulIPAddress, ac);
		fprintf(fp, "VirtualServer%d = %s, %d, %d, %s\n", i,
			ac,
			(int)pItem->usPortFrom,
			(int)pItem->usPortTo,
			GetIPProtocolString(pItem->cProtocolType));
	}
	fprintf(fp, "\n");
#endif

#ifdef WINBOND_DMZ
	/* д Expose.htm  */
	fprintf(fp, "[Expose]\n");
	IP2String(pwConfigParam->ulDMZHost, ac);
	fprintf(fp, "DMZHost = %s\n", ac);
	fprintf(fp, "\n");
#endif

	/* д Admin.htm  */
	fprintf(fp, "[Admin]\n");
	fprintf(fp, "HttpPassword = {%s}\n", pwConfigParam->acHttpPassword);
	Char2Hex(pwConfigParam->acLanMacAddress[0], ac);
	fprintf(fp, "LanMacAddress = %s", ac);
	for (i=1; i<6; i++)
	{
		Char2Hex(pwConfigParam->acLanMacAddress[i], ac);
		fprintf(fp, ",%s", ac);
	}
	fprintf(fp, "\nLanMacReconfig = %d\n", pwConfigParam->bLanMacReconfig);
	Char2Hex(pwConfigParam->acWanMacAddress[0], ac);
	fprintf(fp, "WanMacAddress = %s", ac);
	for (i=1; i<6; i++)
	{
		Char2Hex(pwConfigParam->acWanMacAddress[i], ac);
		fprintf(fp, ",%s", ac);
	}
	fprintf(fp, "\nWanMacReconfig = %d\n", pwConfigParam->bWanMacReconfig);

	fprintf(fp, "ExternalAdmin = %d\n", (int)pwConfigParam->bEnableExternalAdmin);
	fprintf(fp, "ExternalAdminPort = %d\n", (int)pwConfigParam->usExternalAdminPort);
	fprintf(fp, "UPNP = %d\n", (int)pwConfigParam->bEnableUPNP);
	fprintf(fp, "\n");

#ifdef WINBOND_ROUTER
	/* DRouter.htm  */
	fprintf(fp, "[DRouter]\n");
	fprintf(fp, "DrTX = %s\n", GetRoutingProtocolString(pwConfigParam->cDrTX));
	fprintf(fp, "DrRX = %s\n", GetRoutingProtocolString(pwConfigParam->cDrRX));
	fprintf(fp, "\n");

	/* д SRouter.htm  */
	fprintf(fp, "[SRouter]\n");
	fprintf(fp, "# SRouting<n> = <DestIP>, <SubnetMask>, <Gateway>, <Metric>, <NetInterface>\n");
	for (i=0; i<MAX_STATIC_ROUTING_ITEM; i++)
	{
		ROUTINGTABLE_T *pItem;
		pItem = pwConfigParam->arRoutingTableItem + i;
		IP2String(pItem->ulDestLanIPAddress, ac);
		fprintf(fp, "SRouting%d = %s, ", i, ac);
		IP2String(pItem->ulSubnetMask, ac);
		fprintf(fp, "%s, ", ac);
		IP2String(pItem->ulDefaultGateway, ac);
		fprintf(fp, "%s, %d, %s\n",
			ac,
			(int)pItem->lMetric,
			GetNetInterfaceString(pItem->cNetInterface));
	}
	fprintf(fp, "\n");
#endif

#ifdef WINBOND_PPPOE
	/* PPPoE.htm  */
	fprintf(fp, "[PPPoE]\n");
	//fprintf(fp, "EnablePPPoE = %d\n", (int)pwConfigParam->bEnablePPPoE);
	fprintf(fp, "EnablePPPoEAutoConnect = %d\n", (int)pwConfigParam->bEnablePPPoEAutoConnect);
	fprintf(fp, "PPPoEUserName = %s\n", pwConfigParam->acPPPoEUserName);
	fprintf(fp, "PPPoEUserPass = %s\n", pwConfigParam->acPPPoEUserPass);
	fprintf(fp, "PPPoEServiceName = %s\n", pwConfigParam->acPPPoEServiceName);
	fprintf(fp, "PPPoEDisConnectTime_InSec = %u\n", (unsigned int)pwConfigParam->ulPPPoEDisConnectTime_InSec);
	fprintf(fp, "\n");
#endif

#ifdef WINBOND_MODEM
	/* Modem.htm  */
	fprintf(fp, "[Modem]\n");
	fprintf(fp, "EnableModem = %d\n", (int)pwConfigParam->bEnableModem);
	fprintf(fp, "EnableModemAutoConnect = %d\n", (int)pwConfigParam->bEnableModemAutoConnect);
	fprintf(fp, "ModemUserName = %s\n", pwConfigParam->acModemUserName);
	fprintf(fp, "ModemUserPass = %s\n", pwConfigParam->acModemUserPass);
	fprintf(fp, "ModemPhone = %s\n", pwConfigParam->acModemPhone);
	fprintf(fp, "ModemDisConnectTime_InSec = %u\n", (unsigned int)pwConfigParam->ulModemDisConnectTime_InSec);
	fprintf(fp, "\n");
#endif

#ifdef WINBOND_PPTP
	/* VPN.htm  */
	fprintf(fp, "[VPN]\n");
	fprintf(fp, "EnableVPN = %d\n", (int)pwConfigParam->bEnableVPN);
	fprintf(fp, "VPNUserName = %s\n", pwConfigParam->acVPNUserName);
	fprintf(fp, "VPNUserPass = %s\n", pwConfigParam->acVPNUserPass);
	fprintf(fp, "VPNHostName = %s\n", pwConfigParam->acVPNHostName);
	fprintf(fp, "VPNDisConnectTime_InSec = %u\n", (unsigned int)pwConfigParam->ulVPNDisConnectTime_InSec);
	fprintf(fp, "\n");
#endif

	/* SpecialApp.htm  */
	fprintf(fp, "[SpecialApp]\n");
	fprintf(fp, "# Name_<n> = <Application Name>\n");
	fprintf(fp, "# Value_<n> = <InPortFrom>, <InPortTo>, <OutPortFrom>, <OutPortTo>, <IsActive>\n");
	for (i=0; i<MAX_SPECIAL_APP_ITEM; i++)
	{
		SPECIALAPPLICATION_T *pItem;
		pItem = pwConfigParam->asSpecialAppItem + i;
		fprintf(fp, "Name_%d = %s\n", i, pItem->acAppName);
		fprintf(fp, "Value_%d = %d, %d, %d, %d, %d\n", i,
			(int)pItem->usInPortFrom,
			(int)pItem->usInPortTo,
			(int)pItem->usOutPortFrom,
			(int)pItem->usOutPortTo,
			(int)pItem->bIsActivated);
	}
	fprintf(fp, "\n");

#ifdef WINBOND_ADV
	/* д Advance.htm  */
	fprintf(fp, "[Advance]\n");
	fprintf(fp, "DosEnabled = %d\n", (int)pwConfigParam->bEnableDosDefense);
	fprintf(fp, "DosRatePerHour = %u\n", (unsigned int)pwConfigParam->ulDosRatePerHour);
	fprintf(fp, "EnableHidePubIP = %d\n", (int)pwConfigParam->bEnableHidePubIP);
	fprintf(fp, "EnableLanMonitor = %d\n", (int)pwConfigParam->bEnableLanMonitor);
	fprintf(fp, "EnableWanMonitor = %d\n", (int)pwConfigParam->bEnableWanMonitor);
#endif

#ifdef WINBOND_DDNS
	fprintf(fp, "EnableDDNS = %d\n", (int)pwConfigParam->bEnableDDNS);
	fprintf(fp, "DDNSDomain = %s\n", pwConfigParam->acDDNSUserName);
	fprintf(fp, "DDNSUserName = %s\n", pwConfigParam->acDDNSUserPass);
	fprintf(fp, "DDNSPassword = %s\n", pwConfigParam->acDDNSDomainName);
#endif
	fprintf(fp, "\n");

	fclose(fp);
	return TRUE;
}
#endif

BOOL ReadFactoryDefault(WR_CONFIG_PARAM_T *pwConfigParam)
{

	ReadConfig(FACTORY_DEFAULT_FILE);
	return ReadNatINI(FACTORY_DEFAULT_FILE, pwConfigParam);
}

unsigned long GetConfigCheckSum(WR_CONFIG_PARAM_T *pwConfigParam)
{
	int i;
	unsigned long ulCheckSum;
	unsigned char *puc;

	ulCheckSum=0;
	puc = (unsigned char *)pwConfigParam;
	for (i=0; i<offsetof(WR_CONFIG_PARAM_T, ulCheckSum); i++)
		ulCheckSum += (unsigned long)puc[i] + i;
	for (i=offsetof(WR_CONFIG_PARAM_T, ulCheckSum)+sizeof(pwConfigParam->ulCheckSum); i<sizeof(WR_CONFIG_PARAM_T); i++)
		ulCheckSum += (unsigned long)puc[i] + i;
	return ulCheckSum;
}

#ifdef DEBUG
int Http_GetConfig(HTTPCONNECTION hConnection, void *pParam)
{
	char *pcBuf;
	int iLen;

	int i;

	WriteNatINI(DEBUG_CONFIG_FILE, &g_ConfigParam);

	if (ReadWholeFile(DEBUG_CONFIG_FILE, &pcBuf, &iLen))
	{
		AddHttpBody(hConnection, pcBuf, iLen);
		free(pcBuf);
	}

	unlink(DEBUG_CONFIG_FILE);
	SetHttpHeader(hConnection, 200, "OK", "", "Expires: 01 Jan 1970 00:00:00 GMT\r\n", "text/plain", TRUE);
	return 0;
}
#endif
