#include "CommonDef.h"

/*********************************************************************
 * Emulate C++ objects.
*********************************************************************/
OBJHEADER *ObjInit(OBJHEADER *pParent, size_t tObjSize, OBJHEADER *pObj,
	void (*pfunFinal)(OBJHEADER *pObj),
	OBJHEADER *(*pfunInit)(OBJHEADER *pObj, va_list vaInit),
	...)
{
	int iNewCreate = 0;
	va_list vaInit;

	if (pObj == NULL && pParent != NULL)
	{
		fprintf(stderr, "No parent for new created object.\n");
		return NULL;
	}

	if (pObj == NULL)
	{
		pObj = (OBJHEADER *)malloc(tObjSize);
		iNewCreate = 1;
	}
	if (pObj == NULL) return NULL;

	//pObj->pfunInit = pfunInit;
	pObj->pfunFinal = NULL;
	if (pParent == NULL) pObj->iOffsetParent = 0;
	else pObj->iOffsetParent = (char *)pObj - (char *)pParent;
	pObj->iNewCreate = iNewCreate;

	memset((char *)pObj + sizeof(OBJHEADER), 0, tObjSize - sizeof(OBJHEADER));

	va_start(vaInit, pfunInit);
	if ((*pfunInit)(pObj, vaInit) == NULL)
	{
		if (iNewCreate) free(pObj);
		return NULL;
	}
	else
	{
		pObj->pfunFinal = pfunFinal;
		return pObj;
	}
}


void ObjFinal(OBJHEADER *pObj)
{
	if (pObj == NULL) return;

	if (pObj->iOffsetParent != 0)
	{
		//Code here to detect if a parent object exists.
		OBJHEADER *pParent = (OBJHEADER *)((char *)pObj - pObj->iOffsetParent);
		if (pParent->pfunFinal != NULL)
		{
			ObjFinal(pParent);
			return;
		}
	}

	//Code here to call final function of this object.
	if (pObj->pfunFinal != NULL)
	{
		void (*pfunFinal)(OBJHEADER *pObj) = pObj->pfunFinal;
		pObj->pfunFinal = NULL;
		(*pfunFinal)(pObj);
	}

	if (pObj->iNewCreate)
	{
		free(pObj);
	}
}





/*********************************************************************
** FD_T construction & destrucion
**	do not call them directly for ever!
** var vaInit:
**	int iFD, int iType, int iState,
**	BOOL bWantRead, BOOL bWantWrite,
**	BOOL (*pfunProcessFD)(struct tagFD_T *pFD, BOOL bReadable, BOOL bWritable)
*********************************************************************/
OBJHEADER *FD_Init(OBJHEADER *pObj, va_list vaInit)
{
	FD_T *pThis = (FD_T *)pObj;

	pThis->iFD = va_arg(vaInit, int);
	pThis->iType = va_arg(vaInit, int);
	pThis->iState = va_arg(vaInit, int);
	pThis->iConnTimeout_InMSec = va_arg(vaInit, int);
	pThis->bWantRead = va_arg(vaInit, BOOL);
	pThis->bWantWrite = va_arg(vaInit, BOOL);
	pThis->pfunProcessFD = va_arg(vaInit, FUNPROCESSFD_T);

	pThis->bReadable = 0;
	pThis->bWritable = 0;
	if (pThis->pfunProcessFD == NULL) return NULL;
	else return pObj;
}

void FD_Final(OBJHEADER *pObj)
{
	FD_T *pThis = (FD_T *)pObj;
}

inline void FD_Next(FD_T *pFD,
					int iState,
					BOOL bWantRead,
					BOOL bWantWrite,
					int iConnTimeout_InMSec)
{
	pFD->iState = iState;
	pFD->bWantRead = bWantRead;
	pFD->bWantWrite = bWantWrite;
	pFD->iConnTimeout_InMSec = iConnTimeout_InMSec;
}




/* struct for inserting a process to a running thread. */
typedef struct
{
	void *(*pfunCustomProcess)(void *pThread, void *pParam);	//the user's function
	void *pParam;	//parameters that's transferred to *pCustomProcess
} CUSTOM_PROCESS_T;

BOOL InsertCustomProcess(MSG_QUEUE *pMsgThread, void *(*pfunCustomProcess)(void *pThread, void *pParam), void *pParam)
{
	CUSTOM_PROCESS_T *pCustomProcess;
	MSG_T msg;

	if (pMsgThread == NULL || pfunCustomProcess == NULL) return FALSE;

	pCustomProcess = (CUSTOM_PROCESS_T *)malloc(sizeof(CUSTOM_PROCESS_T));
	if (pCustomProcess == NULL) return FALSE;

	pCustomProcess->pfunCustomProcess = pfunCustomProcess;
	pCustomProcess->pParam = pParam;

	msg.lMsg = MSG_CUSTOM_PROC;
	msg.lData = (long)pCustomProcess;

	if (!SendMsg(pMsgThread, &msg))
	{
		free((void *)pCustomProcess);
		return FALSE;
	}
	else return TRUE;
}


BOOL AddSelectFD(FD_T *pFD)
{
	MSG_T msg;
	if (pFD == NULL) return FALSE;
	if (pFD->pfunProcessFD == NULL) return FALSE;

	msg.lMsg = MSG_ADD_FD;
	msg.lData = (long)pFD;

	return SendMsg(g_pMsgSelect, &msg);
}


void *SelectThread(void *pParam)
{
	SELECT_THREAD_T sThreadData;	//important data in this thread.
	MSG_T *pMsg = &sThreadData.msg;	//message in the main loop of this thread
	fd_set *pfdReadSet = &sThreadData.fdReadSet;	//readfds for select
	fd_set *pfdWriteSet = &sThreadData.fdWriteSet;	//writefds for select
	LIST *pFDList;

	struct timeval tvSelect;	//timeval for select
	struct timeval *ptvSelect;
	LISTNODE *pNode;
	LISTNODE *pNextNode;
	FD_T *pFD;
	int iFD;
	int iConnTimeout_InMSec;
	BOOL bReadable;
	BOOL bWritable;
	SELECT_MSG *pSelectMsg;

	sThreadData.pFDList = CreateList();	//List that saves fds
	pFDList = sThreadData.pFDList;
	BlockSignal();
	if (pFDList == NULL)
	{
		fprintf(stderr, "Init fds list error!\n");
		return NULL;
	}

	pSelectMsg = (SELECT_MSG *)ObjInit(NULL, sizeof(SELECT_MSG), NULL,
			SELECT_MSG_Final, SELECT_MSG_Init,
			g_pMsgSelect, pMsg);
	if (pSelectMsg == NULL)
	{
		DeleteList(pFDList);
		return NULL;
	}


	/* Add msg fd */
	if ((pNode = AppendNode(pFDList)) == NULL)
	{
		ObjFinal((OBJHEADER *)pSelectMsg);
		DeleteList(pFDList);
		return NULL;
	}
	pNode->pValue = &pSelectMsg->fd;



	FD_ZERO(pfdReadSet);
	FD_ZERO(pfdWriteSet);

	while (TRUE)
	{
		if (pFDList->pFirstNode == pFDList->pLastNode)
		{//no fds, wait message
			if (!GetMsg(g_pMsgSelect, pMsg))
			{
				sleep(1);
				fprintf(stderr, "Msg queue error!\n");
				continue;
			}
		}
		else
		{//select fds.
//			fprintf(stderr, "l: %d\n", GetListLength(pFDList));PTI;
			pMsg->lMsg = MSG_NONE;

			iFD = 0;
			iConnTimeout_InMSec = -1;

			for (pNode = pFDList->pFirstNode;
				pNode != pFDList->pLastNode;
				pNode = pNode->pNextNode)
			{
				pFD = (FD_T *)pNode->pValue;
				if (pFD == NULL) continue;

				if (pFD->bWantRead || pFD->bWantWrite)
				{
					if (iFD < pFD->iFD) iFD = pFD->iFD;
				}

				if (pFD->iConnTimeout_InMSec >= 0)
				{
					if (iConnTimeout_InMSec < 0)
						iConnTimeout_InMSec = pFD->iConnTimeout_InMSec;
					else if (iConnTimeout_InMSec > pFD->iConnTimeout_InMSec)
						iConnTimeout_InMSec = pFD->iConnTimeout_InMSec;
				}

				if (pFD->bWantRead)
				{
					if (!pFD->bReadable) FD_SET(pFD->iFD, pfdReadSet);
				}
				else if (pFD->bReadable) FD_CLR(pFD->iFD, pfdReadSet);

				if (pFD->bWantWrite)
				{
					if (!pFD->bWritable) FD_SET(pFD->iFD, pfdWriteSet);
				}
				else if (pFD->bWritable) FD_CLR(pFD->iFD, pfdWriteSet);
			}

			//fprintf(stderr, "LENGTH: %d %d %d", GetListLength(pFDList), iFD, iConnTimeout_InMSec);PTI;
			/*
			for (pNode = pFDList->pFirstNode;
				pNode != pFDList->pLastNode;
				pNode = pNode->pNextNode)
			{
				pFD = (FD_T *)pNode->pValue;
				if (pFD == NULL) continue;
				fprintf(stderr, "fd: %d, crw=%d%d, wrw=%d%d, rw=%d%d\n",
						pFD->iFD,
						pFD->bReadable,
						pFD->bWritable,
						pFD->bWantRead,
						pFD->bWantWrite,
						FD_ISSET(pFD->iFD, pfdReadSet),
						FD_ISSET(pFD->iFD, pfdWriteSet));
			}
			*/

			if (iConnTimeout_InMSec < 0)
				ptvSelect = NULL;
			else
			{
				tvSelect.tv_sec = iConnTimeout_InMSec / 1000;;
				tvSelect.tv_usec = (iConnTimeout_InMSec - 1000 * tvSelect.tv_sec) * 1000;
				ptvSelect = &tvSelect;
			}
			if (select(iFD + 1, pfdReadSet, pfdWriteSet, NULL, ptvSelect) == -1)
			{
				//PTI;
				if (errno == EINTR || errno == EBADF)
				{
					continue;
				}
				else
				{
					fprintf(stderr, "Select error!\n");
					exit(-1);
				}
			}
//			PTI;
			for (pNode = pFDList->pFirstNode;
				pNode != pFDList->pLastNode;
				pNode = pNextNode)
			{
//				PTI;
				pNextNode = pNode->pNextNode;
				pFD = (FD_T *)pNode->pValue;
				if (pFD == NULL) continue;
//printf("%d \n", pFD->iFD);PTI;
				iFD = pFD->iFD;
				pFD->bReadable = bReadable = (pFD->bWantRead ? FD_ISSET(iFD, pfdReadSet) : FALSE);
				pFD->bWritable = bWritable = (pFD->bWantWrite ? FD_ISSET(iFD, pfdWriteSet) : FALSE);

//PTI;
				if (!(*pFD->pfunProcessFD)(pFD, bReadable, bWritable))
				{//return FALSE, the FD_T struct should be cleared.
//					PTI;
					if (bReadable) FD_CLR(iFD, pfdReadSet);
					if (bWritable) FD_CLR(iFD, pfdWriteSet);
					ObjFinal((OBJHEADER *)pFD);
					DeleteNode(pNode);
				}
				else if (iFD != pFD->iFD)
				{//update fd if fd is changed by *pFD->pfunProcessFD
//					PTI;
					if (bReadable)
					{
						FD_CLR(iFD, pfdReadSet);
						FD_SET(pFD->iFD, pfdReadSet);
					}
					if (bWritable)
					{
						FD_CLR(iFD, pfdWriteSet);
						FD_SET(pFD->iFD, pfdWriteSet);
					}
				}

//PTI;
			}
//			PTI;
		}

		if (pMsg->lMsg == MSG_NONE) continue;
//PTI;
		if (pMsg->lMsg == MSG_QUIT) break;
		else if (pMsg->lMsg == MSG_ADD_FD)
		{
			pFD = (FD_T *)pMsg->lData;
			if (pFD != NULL
				&& pFD->pfunProcessFD != NULL
				&& (pNode = AppendNode(pFDList)) != NULL)
			{
//				PTI;
				pFD->bReadable = FALSE;
				pFD->bWritable = FALSE;
				pNode->pValue = pFD;
			}
		}
		else if (pMsg->lMsg == MSG_CUSTOM_PROC)
		{
			CUSTOM_PROCESS_T *pCusProcess;
			pCusProcess = (CUSTOM_PROCESS_T *)pMsg->lData;
			if (pCusProcess != NULL)
			{
				(*pCusProcess->pfunCustomProcess)((void *)&sThreadData, pCusProcess->pParam);
				free(pCusProcess);
			}
		}
	}

//	PTI;
	for (pNode = pFDList->pFirstNode;
		pNode != pFDList->pLastNode;
		pNode = pNode->pNextNode)
	{
		pFD = (FD_T *)pNode->pValue;
		if (pFD == NULL) continue;
//		PTI;
		ObjFinal((OBJHEADER *)pFD);
	}

	DeleteList(pFDList);
}



