
#include "CommonDef.h"
#include "libnat.h"

void W90740Reboot(void);

/* Set a certain interface flag. */
static int set_flag(int skfd,char *ifname, short flag)
{
    struct ifreq ifr;

    strcpy(ifr.ifr_name, ifname);
    if (ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0) {
	fprintf(stderr, "%s: unknown interface: %s\n",
		ifname,	strerror(errno));
	return (-1);
    }
    strcpy(ifr.ifr_name, ifname);
    ifr.ifr_flags |= flag;
    if (ioctl(skfd, SIOCSIFFLAGS, &ifr) < 0) {
	perror("SIOCSIFFLAGS");
	return -1;
    }
    return (0);
}


/* Clear a certain interface flag. */
static int clr_flag(int skfd,char *ifname, short flag)
{
    struct ifreq ifr;

    strcpy(ifr.ifr_name, ifname);
    if (ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0) {
	fprintf(stderr, "%s: unknown interface: %s\n",
		ifname, strerror(errno));
	return -1;
    }
    strcpy(ifr.ifr_name, ifname);
    ifr.ifr_flags &= ~flag;
    if (ioctl(skfd, SIOCSIFFLAGS, &ifr) < 0) {
	perror("SIOCSIFFLAGS");
	return -1;
    }
    return (0);
}


int SetWanDuplex(unsigned long mode)
{
	unsigned long duplexFlag=NATA_WAN_AUTO_NEG;
	
	switch (mode )
	{
	case 1:
		duplexFlag = NATA_WAN_100_FULL_DUPLEX;
		break;
	case 2:
		duplexFlag = NATA_WAN_100_HALF_DUPLEX;
		break;
	case 3:
		duplexFlag = NATA_WAN_10_FULL_DUPLEX;
		break;
	case 4:
		duplexFlag = NATA_WAN_10_HALF_DUPLEX;
		break;
	case 0: /* AUTO */
	default:
		duplexFlag = NATA_WAN_AUTO_NEG;
		break;
	}

	NATA_SetDuplex( duplexFlag );
	return 0;
}

BOOL SetMacAddress(const char *pcInterface, const char *pcMacAddress)
{
	struct ifreq ifr;
	int fd;
	int rt;

	if (pcMacAddress == NULL || pcInterface == NULL)
	{
		fprintf(stderr, "illegal call function GetMacAddress!\n");
		return FALSE;
	}

	if ((fd = socket(AF_INET,SOCK_DGRAM,0)) < 0) return FALSE;
	strcpy(ifr.ifr_name, pcInterface);

	clr_flag(fd, ifr.ifr_name, IFF_UP);

	ifr.ifr_hwaddr.sa_family = ARPHRD_ETHER;
	memcpy(ifr.ifr_hwaddr.sa_data, pcMacAddress, 6);

	rt = ioctl(fd,SIOCSIFHWADDR,&ifr);

	set_flag(fd, ifr.ifr_name, (IFF_UP | IFF_RUNNING));

	close(fd);
	return (rt >= 0 ? TRUE : FALSE);
}

char *GetMacAddress(const char *pcInterface, char *pcMacAddress)
{
	struct ifreq ifr;
	int fd;

	if (pcMacAddress == NULL || pcInterface == NULL)
	{
		fprintf(stderr, "illegal call function GetMacAddress!\n");
		return NULL;
	}

	if ((fd = socket(AF_INET,SOCK_DGRAM,0)) < 0) return NULL;
	strcpy(ifr.ifr_name, pcInterface);
	if (ioctl(fd,SIOCGIFHWADDR,&ifr) < 0)
	{
		close(fd);
		memset(pcMacAddress, 0, 6);
		return NULL;
	}
	close(fd);

	memcpy(pcMacAddress, ifr.ifr_hwaddr.sa_data, 6);
	return pcMacAddress;
}


BOOL InitNetInterface(const char *pcInterface)
{
	struct ifreq ifr;
	int fd;
	printf("Initializing %s...\n", pcInterface);

	if ((fd = socket(AF_INET,SOCK_DGRAM,0)) < 0)
	{
		fprintf(stderr, "errno = %d\n", errno);
		return FALSE;
	}
	strcpy(ifr.ifr_name, pcInterface);

	clr_flag(fd, ifr.ifr_name, IFF_UP);
	set_flag(fd, ifr.ifr_name, (IFF_UP | IFF_RUNNING));
	close(fd);
	printf("Initialize %s... over \n", pcInterface);
	return TRUE;
}

unsigned long GetGeneralIP(const char *pcInterface, int iRequest)
{
	struct ifreq ifr;
	int fd;
	unsigned long ulRt;

	if (pcInterface == NULL)
	{
		fprintf(stderr, "illegal call function GetGeneralIP!\n");
		return 0L;
	}

	if ((fd = socket(AF_INET,SOCK_DGRAM,0)) < 0) return 0L;
	strcpy(ifr.ifr_name, pcInterface);
	if (ioctl(fd, iRequest, &ifr) < 0)
		ulRt = 0;
	else
		ulRt = (*(struct sockaddr_in *)&(ifr.ifr_addr)).sin_addr.s_addr;
	close(fd);

	return ulRt;
}


static int set_host_info(char *filename, unsigned int local_ip, char * local_name)
{
	FILE *fp;
	char num[32];
	fp = fopen(filename, "w+");

	if (fp == NULL)
	{
    		fprintf(stderr, "open   file %s:  failed\n", filename);
		return -1;
	}

	IP2String(local_ip, num);
	strcat(num, "\twinbond");
	if(fputs(num,fp)<0)
	{
		fclose(fp);
		return -1;
	}
	fclose(fp);
	return 0;
}



BOOL SetGeneralIP(const char *pcInterface, unsigned long ulIP, unsigned long ulNetmask)
{
	struct ifreq ifr;
	int fd;
	struct sockaddr_in *pAddr;

	if (pcInterface == NULL)
	{
		fprintf(stderr, "illegal call function SetGeneralIP!\n");
		return FALSE;
	}

	if (strncmp(pcInterface, "ppp", 3) == 0)
	{
		fprintf(stderr, "call SetGeneralIP on pppN!\n");
		return FALSE;
	}

	if ((fd = socket(AF_INET,SOCK_DGRAM,0)) < 0) return FALSE;
	strcpy(ifr.ifr_name, pcInterface);

	pAddr = (struct sockaddr_in *)&(ifr.ifr_addr);
	bzero(pAddr, sizeof(struct sockaddr_in));
	pAddr->sin_addr.s_addr = ulIP;
	pAddr->sin_family = AF_INET;
	if (ioctl(fd, SIOCSIFADDR, &ifr) < 0)
	{
		close(fd);
		return FALSE;
	}

	pAddr = (struct sockaddr_in *)&(ifr.ifr_addr);
	bzero(pAddr, sizeof(struct sockaddr_in));
	pAddr->sin_addr.s_addr = ulNetmask;
	pAddr->sin_family = AF_INET;
	if (ioctl(fd, SIOCSIFNETMASK, &ifr) < 0)
	{
		close(fd);
		return FALSE;
	}

	close(fd);

	if (strcmp(pcInterface, G_PC_ETH0) == 0)
		NotifyLocalIP2Nat();

	if (strcmp(pcInterface, G_PC_ETH1) == 0)
	{
		set_host_info(HOSTS_FILE, ulIP, "winbond");
		NotifyPubIP2Nat(G_PC_ETH1);
	}
	return TRUE;
}

unsigned long GetIPAddress(const char *pcInterface)
{
	return GetGeneralIP(pcInterface, SIOCGIFADDR);
}

#ifdef EMBED
static inline _syscall1(void, set_local_ip, unsigned long *,args);
static inline _syscall1(void, set_out_ip, unsigned long *,args);
#else
void set_local_ip(unsigned long *args)
{
	char ac[16];
	IP2String((unsigned long)args[0], ac);
	fprintf(stderr, "http server - nat: local ip %s!\n", ac);
}

void set_out_ip(unsigned long *args)
{
	char ac[16];
	IP2String((unsigned long)args[0], ac);
	fprintf(stderr, "http server - nat: out ip %s!\n", ac);
}
#endif

void NotifyLocalIP2Nat()
{
	unsigned long ulCurIPAddress;
	unsigned long args[1];
	ulCurIPAddress = GetIPAddress(G_PC_ETH0);
	args[0]=(unsigned long)ulCurIPAddress;
	set_local_ip(args);
}


void NotifyPubIP2Nat(const char *pcInterface)
{
	unsigned long ulCurIPAddress;
	unsigned long args[1];
	ulCurIPAddress = GetIPAddress(pcInterface);
	args[0]=(unsigned long)ulCurIPAddress;
	set_out_ip(args);
}

unsigned long GetNetMask(const char *pcInterface)
{
	return GetGeneralIP(pcInterface, SIOCGIFNETMASK);
}

#define PROC_ROUTE_FILE_LINE_LEN 128
#define PROC_ROUTE_FILE "/proc/net/route"
static BOOL WRReadRouterTable(PROC_ROUTE_ITEM_T **ppItem, int *piCount)
{
	char *pcFileBuf;
	int iFileLength;
	int iCount;
	PROC_ROUTE_ITEM_T *pItem;
	int i;

	if (!(ReadWholeFile(PROC_ROUTE_FILE, &pcFileBuf, &iFileLength)))
		return FALSE;

	iCount = iFileLength / PROC_ROUTE_FILE_LINE_LEN - 1;
	if (iCount<=0)
	{
		free(pcFileBuf);
		return FALSE;
	}
	if( iCount > 20 ) //maximum 20 entries
	{
		printf("route entries = %d , reset to %d\n", iCount, 20);
		iCount = 20;
	}
	if ((pItem = malloc(sizeof(PROC_ROUTE_ITEM_T)*iCount)) == NULL)
	{
		free(pcFileBuf);
		PRINT_MEM_OUT;
		return FALSE;
	}
	memset(pItem, 0, sizeof(PROC_ROUTE_ITEM_T)*iCount);

	for (i=0; i<iCount; i++)
	{
		char *pc;
		char *pc1;
		pc = pcFileBuf+(i+1)*PROC_ROUTE_FILE_LINE_LEN;
		pc1 = strchr(pc, '\t');
		if (pc1 == NULL || pc1-pc >= PROC_ROUTE_FILE_LINE_LEN)
		{
			fprintf(stderr, "Error in file: %s\n", PROC_ROUTE_FILE);
			free(pcFileBuf);
			return FALSE;
		}

		memcpy(pItem[i].acIface, pc, (pc1-pc>7?7:pc1-pc));
		sscanf(pc1, "\t%lx\t%lx\t%lx\t%d\t%d\t%d\t%lx\t%d\t%d\t%d",
			&pItem[i].ulDestination,
			&pItem[i].ulGateway,
			&pItem[i].ulFlags,
			&pItem[i].iRefCnt,
			&pItem[i].iUse,
			&pItem[i].iMetric,
			&pItem[i].ulMask,
			&pItem[i].iMTU,
			&pItem[i].iWindow,
			&pItem[i].iIRTT);
	}

	free(pcFileBuf);

	*ppItem = pItem;
	*piCount = iCount;
	return TRUE;
}

/* norton : skip pcInterface parameter, because of multi-pppoe have multiple default route */
unsigned long GetGateway(const char *pcInterface)
{
	PROC_ROUTE_ITEM_T *pItem;
	int iItemCount;
	int i, lastMetric = 32767;
	unsigned long ulRt = 0;

	if (!WRReadRouterTable(&pItem, &iItemCount))
	{
		pItem = NULL;
		iItemCount = 0;
		return 0;
	}
	for (i = 0; i < iItemCount; i++)
	{
		if((pItem[i].ulFlags&RTF_GATEWAY)&&(pItem[i].ulFlags&RTF_UP)
			&& (pItem[i].ulDestination==0) && (pItem[i].ulMask==0) )
		{
			//test if (strcmp(pItem[i].acIface, pcInterface) == 0)
			{
				if( pItem[i].iMetric < lastMetric )
				{
					lastMetric = pItem[i].iMetric;
					ulRt = pItem[i].ulGateway;
				}
				//test break;
			}
		}
	}
	free(pItem);
	return ulRt;
}

static void set_sockaddr(struct sockaddr_in *sin, unsigned long addr, unsigned short port)
{
	sin->sin_family = AF_INET;
	sin->sin_addr.s_addr = addr;
	sin->sin_port = port;
}

static BOOL ChangeRoute(unsigned long ulIP,
			unsigned long ulGateway,
			unsigned long ulNetmask,
			short sMetric,
			const char *pcInterface,
			int iAction)
{
	struct rtentry rt;
	int fd;

	//fprintf(stderr,"\nChangeRoute: %x %x %x %d %s ",
	//	ulIP,ulGateway,ulNetmask,sMetric,pcInterface );
		
	if ((fd = socket(AF_INET,SOCK_DGRAM,0)) < 0) return FALSE;

	memset(&rt, 0, sizeof(rt));
	set_sockaddr((struct sockaddr_in *) &rt.rt_dst, ulIP, 0);
	set_sockaddr((struct sockaddr_in *) &rt.rt_gateway, ulGateway, 0);
	set_sockaddr((struct sockaddr_in *) &rt.rt_genmask, ulNetmask, 0);
	rt.rt_metric = sMetric + 1;

	if( pcInterface && pcInterface[0] )
		rt.rt_dev = (char *)pcInterface;

	rt.rt_flags = RTF_UP;
	if (ulGateway != 0L)
    		rt.rt_flags |= RTF_GATEWAY;
	if ((ulIP & 0xFF000000) != 0L && ulNetmask == 0xFFFFFFFF)
		rt.rt_flags |= RTF_HOST;

	//useless code: rt.rt_flags |= RTF_DYNAMIC|RTF_MODIFIEDRTF_USR1;	
	rt.rt_window = 16384;/* set this as a mark to save into flash */
	rt.rt_flags |= RTF_WINDOW;
	
	if (ioctl(fd, iAction, &rt) == -1)
	{
		//fprintf(stderr, "change route ioctl errno = %d\n", errno);
		close(fd);
		return FALSE;
	}
	close(fd);
	return TRUE;
}

BOOL AddRoute(unsigned long ulIP,
				unsigned long ulGateway,
				unsigned long ulNetmask,
				short sMetric,
				const char *pcInterface)
{
	return ChangeRoute(ulIP, ulGateway, ulNetmask, sMetric, pcInterface, SIOCADDRT);
}

BOOL DelRoute(unsigned long ulIP,
				unsigned long ulGateway,
				unsigned long ulNetmask,
				short sMetric,
				const char *pcInterface)
{
	return ChangeRoute(ulIP, ulGateway, ulNetmask, sMetric, pcInterface, SIOCDELRT);
}

BOOL SetGateway(const char *pcInterface, unsigned long ulIP)
{
	unsigned ulOldGateway;
	int rt = TRUE;

	ulOldGateway = GetGateway(pcInterface);
	if (ulOldGateway != 0L)
	{
		if (!DelRoute(0, ulOldGateway, 0, 0, pcInterface))
		{
			fprintf(stderr, "Cannot del default route (%d).\n", errno);
			rt = FALSE;
		}
	}

	if (ulIP != 0)
	{
		if (!AddRoute(0, ulIP, 0, 0, pcInterface))
		{
			fprintf(stderr, "Cannot add default route (%d).\n", errno);
			rt = FALSE;
		}
	}
	return rt;
}


#define DNS_SERVER_FILE "/swap/resolv.conf"
BOOL SetDNSServer(unsigned long *pulDNS)
{
	FILE *fp;
	int i;
	char ac[128];
	int iLen;

	iLen = 0;
	for (i = 0; i < 3; i++)
	{
		if (pulDNS[i] != 0L)
		{
			iLen += sprintf(ac + iLen, "nameserver ");
			//iLen += sprintf(ac + iLen, G_PC_ETH1 );
			//iLen += sprintf(ac + iLen, " ");
			iLen += IP2String(pulDNS[i], ac + iLen);
			iLen += sprintf(ac + iLen, "\n");
		}
	}
	//iLen += sprintf(ac + iLen, "search localdomain\n");

	if ((fp = fopen(DNS_SERVER_FILE, "wb")) == NULL)
		return FALSE;
	iLen = fwrite(ac, iLen, 1, fp);
	fclose(fp);
	return (iLen == 1);
}

/*
 * create_resolv - create the replacement resolv.conf file
 */
void
create_resolv(unsigned long peerdns1, unsigned long peerdns2, int session)
{
    FILE *f;
    char buf[32];
    
    snprintf(buf,sizeof(buf),"/swap/resolv.conf.%d", session );

    f = fopen(buf, "w+");
    if (f == NULL) {
	printf("Failed to create %s", buf );
	return;
    }
	
    if (peerdns1)
	fprintf(f, "nameserver ppp%d %s\n", session, int_ntoa(peerdns1));

    if (peerdns2)
	fprintf(f, "nameserver ppp%d %s\n", session, int_ntoa(peerdns2));

    if (ferror(f))
	printf("Write failed to %s", buf );

    fclose(f);

    //create /etc/resolv.conf for DDNS use 
    if( session == 1 || session == 0 )
    {
	strcpy(buf,"/etc/resolv.conf");

	f = fopen(buf, "w+");
	if (f == NULL) {
		printf("Failed to create %s", buf);
		return;
	}

	if (peerdns1)
		fprintf(f, "#by httpd\nnameserver %s\n", int_ntoa(peerdns1));

	if (peerdns2)
		fprintf(f, "nameserver %s\n", int_ntoa(peerdns2));

	if (ferror(f))
		printf("Write failed to %s", buf );

	fclose(f);
    }
}

BOOL LimitKill(pid_t *pg_pid, int iSignal, unsigned long ulLimitTime_InMiniSec)
{
	unsigned long i;
	if (*pg_pid == 0) return TRUE;

	if (kill(*pg_pid, iSignal) != 0)
	{
		fprintf(stderr, "Can not kill process %d.\n", *pg_pid);
		return FALSE;
	}

	ulLimitTime_InMiniSec /= 50;
	for (i = 0; i < ulLimitTime_InMiniSec && *pg_pid != 0; i++)
		usleep(50000);//for lesser than 20000, usleep do not really sleep.

	if (*pg_pid != 0)
	{
		fprintf(stderr, "Can not kill process %u in %lu ms.\n", *pg_pid, ulLimitTime_InMiniSec);
		return FALSE;
	}
	else return TRUE;
}


BOOL SafeLimitKill(pid_t pid, int iSignal, unsigned long ulLimitTime_InMiniSec)
{
	int i;
	if ((pid_t)0 == pid) return TRUE;

	if (kill(pid, iSignal) != 0)
	{
		if (ESRCH == errno) return TRUE;
		else return FALSE;
	}

	for (i = 0; i < ulLimitTime_InMiniSec / 50; i++)
	{
		char cState = GetProcessState(pid);

		if ('E' == cState)//Process not exists.
			break;
		else if ('Z' == cState)//Defunct ("zombie") process
		{
			waitpid(pid, 0, 0);
			break;
		}
		else usleep(50 * 1000);//for lesser than 20000, usleep do not really sleep.
	}

	if (i < ulLimitTime_InMiniSec) return TRUE;
	else
	{
		fprintf(stderr, "Can not kill process %d in %lu ms.\n", pid, ulLimitTime_InMiniSec);
		return FALSE;
	}
}

void WRSigChildDie(int iSignal)
{
	pid_t pidChild;
	int iReturnStatus;

	if (iSignal != SIGCHLD) return;

	pidChild = waitpid(-1, &iReturnStatus, 0);

	if (pidChild <=0)
	{//something goes error!
		return;
	}

	printf("Child %d exit!\n", pidChild);
	return;
}

BOOL SetMtu(const char *pcInterface, long lMtu)
{
	struct ifreq ifr;
	int fd;

	if (pcInterface == NULL)
	{
		fprintf(stderr, "illegal call function GetIPAddress!\n");
		return FALSE;
	}

	if ((fd = socket(AF_INET,SOCK_DGRAM,0)) < 0) return FALSE;
	strcpy(ifr.ifr_name, pcInterface);
	ifr.ifr_mtu = lMtu;

	if (ioctl(fd, SIOCSIFMTU, &ifr) < 0)
	{
		close(fd);
		return FALSE;
	}

	close(fd);
	return TRUE;
}


long GetMtu(char *pcInterface)
{
	struct ifreq ifr;
	int fd;

	if (pcInterface == NULL)
	{
		fprintf(stderr, "illegal call function GetIPAddress!\n");
		return 0;
	}

	if ((fd = socket(AF_INET,SOCK_DGRAM,0)) < 0) return 0;
	strcpy(ifr.ifr_name, pcInterface);

	if (ioctl(fd, SIOCGIFMTU, &ifr) < 0)
	{
		close(fd);
		return 0;
	}

	close(fd);
	return ifr.ifr_mtu;
}

void fwstate_check(int enable)
{
	if( enable )
	{
		iptables("ipt -D STATE -p ip -j ACCEPT");
		iptables("ipt -t mangle -D FORWARD -p ip -j ACCEPT");
	}
	else
	{
		iptables("ipt -I STATE -p ip -j ACCEPT");
		iptables("ipt -t mangle -I FORWARD -p ip -j ACCEPT");
	}
}

void enable_nat()
{
	int i=0;
	char cmd[64];

	iptables("ipt -t nat -A POSTROUTING -o eth1 -j MASQUERADE");
	for(i=1; i<MAX_PPPOE_AND_UNNUM; i++)
	{
		sprintf(cmd,"ipt -t nat -A POSTROUTING -o ppp%d -j MASQUERADE",i);
		iptables(cmd);
	}
	
	if( g_ConfigParam.bEnableNaptUnnum )
	{
		sprintf(cmd,"ipt -t nat -A POSTROUTING -o ppp0 -j MASQUERADE");
		iptables(cmd);
	}
}

void disable_nat()
{
	int i=0;
	char cmd[64];
	
	iptables("ipt -t nat -D POSTROUTING -o eth1 -j MASQUERADE");
	for(i=1; i<MAX_PPPOE_AND_UNNUM; i++)
	{
		sprintf(cmd,"ipt -t nat -D POSTROUTING -o ppp%d -j MASQUERADE",i);
		iptables(cmd);
	}
	
	if( g_ConfigParam.bEnableNaptUnnum == 0 )
	{
		sprintf(cmd,"ipt -t nat -D POSTROUTING -o ppp0 -j MASQUERADE");
		iptables(cmd);
	}
	
}

void ResetNat()
{
/*
	struct Enable_nat eNat;
	struct MTCPMSS mMTC;

	strcpy(eNat.interface, pcInterface);
	strcpy(mMTC.interface, pcInterface);
	
	disable_nat();
	enable_nat(&eNat);

	Del_TCPMSS(&mMTC);
	Add_TCPMSS(&mMTC);
*/
	disable_nat();
	enable_nat();
}

static char acOuterNetworkInterface[8];

char *WRSetOuterNetworkInterface(const char *pcInterface)
{
	if (pcInterface == NULL) return NULL;
	MyStrncpy(acOuterNetworkInterface, pcInterface, sizeof(acOuterNetworkInterface));
	return acOuterNetworkInterface;	
}

const char *WRGetOuterNetworkInterface()
{
	if (g_ConfigParam.ulHowToGetIp == GETIP_PPP || g_ConfigParam.ulHowToGetIp == GETIP_UNNUM )
		return acOuterNetworkInterface;
	else return G_PC_ETH1;
}

unsigned long GetDefaultWanIP()
{
	char buf[18];
	
	if (g_ConfigParam.ulHowToGetIp == GETIP_PPP )
		strcpy(buf,G_PC_PPP1);
	else if (g_ConfigParam.ulHowToGetIp == GETIP_UNNUM )
		strcpy(buf,G_PC_PPP0);
	else
		strcpy(buf, G_PC_ETH1 );
	
	return GetIPAddress(buf);
	
}


void nat_install( int enable )
{
	if( enable == 0 )
		disable_nat();
	else
		ResetNat();//WRGetOuterNetworkInterface());
}


void WRSigSetRuleConnected(int iSignal)
{
	HTTPC *pHttpC;
	char acPath[64];
	sprintf(acPath, "/__SetRulesInHttpdThread__.cgi?Signal=%d", iSignal);
	
	fprintf(stderr, "Signal recevied: %d\n", iSignal);
	
	if ((pHttpC = (HTTPC *)ObjInit(NULL, sizeof(HTTPC), NULL,
			HTTPC_Final, HTTPC_Init,
			"127.0.0.1",
			acPath,
			g_RunTimeState.iPort[1],
			NULL,
			NULL,
			NULL,
			NULL,
			NULL,
			NULL,
			NULL,
			NULL,
			NULL)) == NULL) return;
	AddSelectFD(&pHttpC->fd);
}

int lan_service(unsigned long old_lan, unsigned long new_lan, unsigned long netmask, int updateDhcp )
{
	
	NotifyLocalIP2Nat();
	//WRConfigVserver(g_ConfigParam.avVirtualHostItem, TRUE);
	//WRSetExternalPort((g_ConfigParam.bEnableExternalAdmin ? g_ConfigParam.usExternalAdminPort : 0), TRUE);
	dnsProxyInstall();
	//service_install(1,"dns","-i", "eth0",NULL);
	WRConfigUPNP(FALSE,1);
	WRConfigUPNP(g_ConfigParam.bEnableUPNP,1);

	if( updateDhcp )
	{
		dhcpd_update(old_lan, new_lan, netmask );
		dhcpd_install();
	}
	
	//g_ConfigParam.dhcp.subnet = g_ConfigParam.ulPrivateSubnetMask ;
	//WRSetDHCPAddressRange(g_ConfigParam.ulDHCPAddressFrom, g_ConfigParam.ulDHCPAddressTo);
	return 0;
}

int wan_service(int sess_no)
{
	unsigned long def_wan_ip = 0;

	def_wan_ip = GetDefaultWanIP();
	
	//ResetNat();

	//WRConfigVserver(g_ConfigParam.avVirtualHostItem, TRUE);
	//WRSetExternalPort((g_ConfigParam.bEnableExternalAdmin ? g_ConfigParam.usExternalAdminPort : 0), TRUE);
	//WRConfigDMZ(g_ConfigParam.ulDMZHost, TRUE);
	
	route_install();
	dnsProxyInstall();
	//service_install(1,"dns","-i","eth0",NULL);

#ifdef WINBOND_DDNS

	if( 
		//def_wan_ip && 
		g_ConfigParam.bEnableDDNS  && (sess_no == 0 ) )
	{
		pid_t* pidList;
		pid_t pidRt;
		int i=0;
		
		pidList = FindPidByName( PROG_DDNS );

		if (!pidList)
		{
			ddns_writeconf(  DDNS_CONF );
			ddns_install();
		}
		else
		{
			pidRt = *pidList;
			for( i=0; *(pidList+i); i++)
			{
				kill((pid_t)*(pidList+i), SIGHUP);
			}
			free(pidList);
		}		
		/*
		Do_DDNS(g_ConfigParam.acDDNSDomainName,
			def_wan_ip,
			g_ConfigParam.acDDNSUserName,
			g_ConfigParam.acDDNSUserPass,
			g_ConfigParam.acProxy,
			g_ConfigParam.acProxyUser,
			g_ConfigParam.acProxyPass);
		*/
	}

#endif

	return 0;
}


void WRSigSetRuleConnected_ByCGI(int iSignal)
{
	//int iCurState;
	//const char *g_pcInterface = G_PC_ETH1 ;
	//struct ppp_info pInfo;
	int id = 0;
	
	//printf("WRSigSetRuleConnected_ByCGI got signal %u \n", iSignal );
	
	if (iSignal != SIGPWR && iSignal != SIGWINCH && iSignal != SIGIO) return;
	if (iSignal == SIGPWR && g_ConfigParam.ulHowToGetIp != GETIP_DHCP) return;
	
	//do not check this according VPN do not set to GETIP_PPP
	if (iSignal == SIGWINCH && g_ConfigParam.ulHowToGetIp != GETIP_PPP &&
		g_ConfigParam.ulHowToGetIp != GETIP_UNNUM ) return;

	id = g_ConfigParam.ulHowToGetIp == GETIP_UNNUM ? UNNUM_ID : PPPOE_ID;

	if (iSignal == SIGWINCH) //pppoe send signal
	{
		WRSetOuterNetworkInterface(g_ConfigParam.ulHowToGetIp == GETIP_UNNUM ? "ppp0" : "ppp1" );
		printf("pppoe signal\n");
	}

	else if (iSignal == SIGPWR) //dhcp send signal
	{
		unsigned long ulIP = GetIPAddress(G_PC_ETH1);
		printf("Set dhcp client IP %x.\n", ntohl(ulIP));
		set_host_info(HOSTS_FILE, ulIP, "winbond");
		//g_pcInterface = G_PC_ETH1;
		WRSetOuterNetworkInterface( G_PC_ETH1 );
	}
	else
		WRSetOuterNetworkInterface( G_PC_ETH1 );
	
#ifdef WINBOND_PPTP
	if (iSignal == SIGIO)
	{
		if (ppp_read_config(VPN_CONF_FILE, &pInfo))
		{
			g_pcInterface = pInfo.ppp_if;
		}
		else
		{
			fprintf(stderr, "Can not read VPN configuration.\n");
			return;
		}
		
		iCurState = WRGetVPNState();
		if (iCurState != VPN_DOD_ON && iCurState != VPN_COMM_ON)
		{
			fprintf(stderr, "Receive signal SIGIO in wrong state.\n");
			return;
		}		
	}
#endif

	if (iSignal == SIGWINCH) //pppoe send signal
	{
		char pppoe_state[16];
		unsigned long pppoe_life=0, pppoe_sid=0;

		memset(pppoe_state,0,sizeof(pppoe_state));

		get_pppoe_info(1 /* first pppoe session */, 
				&pppoe_state[0],
				&pppoe_sid,
				&pppoe_life);

		if( strcmp(pppoe_state,"ON") == 0 )
		{
			printf("wan server( 0 )\n");
			wan_service(0);
		}
		else
		{
			printf("wan server( 1 )\n");
			wan_service(1);
		}
	}
	else
		wan_service(0);
	
}


int Config_SetRules(HTTPCONNECTION hConnection, LIST *pParamList, int iAction, XML *pReturnXML)
{
	int iSignal;

	switch (iAction)
	{
	case CA_AUTH:
		return AUTH_ANY;
		break;
	case CA_CONFIG:
		iSignal = GetLong(pParamList, "Signal");
		//printf("Config_SetRules: signal=%u\n", iSignal);
		WRSigSetRuleConnected_ByCGI(iSignal);
		break;
	default:
		break;
	}
	return 0;
}

void WRSigPPPChanged(int iSignal)
{
	fprintf(stderr, "Add code for WRSigPPPoEChanged here.\n");
}

void WRSIGTERM(int iSignal)
{
	MSG_T msg;
	msg.lMsg = MSG_QUIT;

	if (g_pMsgFlash != NULL) SendMsg(g_pMsgFlash, &msg);
	if (g_pMsgSelect != NULL) SendMsg(g_pMsgSelect, &msg);

	if (g_ptdFlash != 0)
	{
		pthread_join(g_ptdFlash, NULL);
		g_ptdFlash = 0;
	}
	if (g_pMsgFlash != NULL)
	{
		FreeMsgQueue(g_pMsgFlash);
		g_pMsgFlash = NULL;
		//printf("Delete msg queue (flash), Ok!\n");
	}

	if (g_ptdSelect != 0)
	{
		pthread_join(g_ptdSelect, NULL);
		g_ptdSelect = 0;
	}
	if (g_pMsgSelect != NULL)
	{
		FreeMsgQueue(g_pMsgSelect);
		g_pMsgSelect = NULL;
		//printf("Delete msg queue (select), Ok!\n");
	}
}

void BlockSignal()
{
	sigset_t newmask;
	sigemptyset(&newmask);
	sigaddset(&newmask, SIGTERM);
	sigaddset(&newmask, SIGINT);
	sigaddset(&newmask, SIGPIPE);
	pthread_sigmask(SIG_SETMASK, &newmask, NULL);
}

static BOOL IsIntDigit(char *pcStr)
{
	char *pc;
	for (pc=pcStr; *pc!='\0'; pc++)
	{
		if (*pc > '\40' && (*pc < '0' || *pc > '9')) return FALSE;
	}
	return TRUE;
}

/* This function comes from kill.c */
/* find_pid_by_name()
 *
 *  This finds the pid of the specified process.
 *  Currently, it's implemented by rummaging through
 *  the proc filesystem.
 *
 *  Returns a list of all matching PIDs
 */
#define READ_BUF_SIZE	50
pid_t* FindPidByName(char* pcPidName)
{
	DIR *dir;
	struct dirent *next;
	pid_t* pidList=NULL;
	int i=0;

	dir = opendir("/proc");
	if (!dir)
	{
		fprintf(stderr, "Cannot open /proc");//CXH_MODIFY
		return NULL;
	}

	while ((next = readdir(dir)) != NULL)
	{
		FILE *status;
		char filename[READ_BUF_SIZE];
		char buffer[READ_BUF_SIZE];
		char name[READ_BUF_SIZE];

		/* If it isn't a number, we don't want it */
		//if (!isdigit(*next->d_name)) continue;

		if (!IsIntDigit(next->d_name)) continue;//CXH_MODIFY

		sprintf(filename, "/proc/%s/status", next->d_name);
		if (! (status = fopen(filename, "r")) ) {
			continue;
		}
		if (fgets(buffer, READ_BUF_SIZE-1, status) == NULL) {
			fclose(status);
			continue;
		}
		fclose(status);

		/* Buffer should contain a string like "Name:   binary_name" */
		sscanf(buffer, "%*s %s", name);
		if (strcmp(name, pcPidName) == 0) {
			pidList=realloc( pidList, sizeof(pid_t) * (i+2));
			pidList[i++]=strtol(next->d_name, NULL, 0);
		}
	}

	closedir(dir);//CXH_MODIFY

	if (pidList)
		pidList[i]=0;
	return pidList;
}

/*
       E   Exit, no such process. (xhchen added)

       D   uninterruptible sleep (usually IO)
       R   runnable (on run queue)
       S   sleeping
       T   traced or stopped
       Z   a defunct ("zombie") process

       For BSD formats and when the "stat" keyword is used, additional letters
       may be displayed:

       W   has no resident pages
       <   high-priority process
       N   low-priority task
       L   has pages locked into memory (for real-time and custom IO)
*/
char GetProcessState(pid_t pid)
{
	char acFileName[READ_BUF_SIZE];
	char acBuffer[READ_BUF_SIZE];
	FILE *fp;
	char cReturn = 'E';

	sprintf(acFileName, "/proc/%d/status", (int)pid);

	//sprintf(acFileName, "Gnu_License");
	if ((fp = fopen(acFileName, "r")) != NULL)
	{
		while (1)
		{
			if (fgets(acBuffer, READ_BUF_SIZE - 1, fp) == NULL) break;

			if (sscanf(acBuffer, "State%*s %c", &cReturn) == 1)
				break;

			while (acBuffer[strlen(acBuffer) - 1] != '\n'
				&& fgets(acBuffer, READ_BUF_SIZE - 1, fp) != NULL);
		}
		fclose(fp);
	}
	return cReturn;
}

void SynProcess(pid_t *pid)
{
	pid_t pidThis = *pid;
	switch (GetProcessState(pidThis))
	{
	case 'E':
		*pid = 0;
		break;
	case 'Z':
		if (waitpid(pidThis, 0, 0) == pidThis)
			*pid = 0;
		break;
	default:;
	}
}

#ifdef DEBUG
int Http_Shell_Ps(HTTPCONNECTION hConnection, void *pParam)
{
	DIR *dir;
	struct dirent *next;

	dir = opendir("/proc");
	if (!dir)
		fprintf(stderr, "Cannot open /proc");//CXH_MODIFY
	{
		while ((next = readdir(dir)) != NULL)
		{
			FILE *status;
			char filename[READ_BUF_SIZE];
			char buffer[READ_BUF_SIZE];
			char name[READ_BUF_SIZE];

			/* If it isn't a number, we don't want it */
			//if (!isdigit(*next->d_name)) continue;

			if (!IsIntDigit(next->d_name)) continue;//CXH_MODIFY

			sprintf(filename, "/proc/%s/status", next->d_name);
			if (! (status = fopen(filename, "r")) ) {
				continue;
			}
			if (fgets(buffer, READ_BUF_SIZE-1, status) == NULL) {
				fclose(status);
				continue;
			}
			fclose(status);

			/* Buffer should contain a string like "Name:   binary_name" */
			sscanf(buffer, "%*s %s", name);

			{
				int iLen;
				char *pc;
				if (ReadWholeFile(filename, &pc, &iLen))
				{
					AddHttpBody(hConnection, pc, iLen);
					AddHttpBodyString(hConnection, "\n\n");
					free(pc);
				}
			}
		}

		closedir(dir);//CXH_MODIFY
	}
	SetHttpHeader(hConnection, 200, "OK", "", "Expires: 01 Jan 1970 00:00:00 GMT\r\n", "text/plain", TRUE);
	return 0;
}

#endif

int sysinfo(HTTPCONNECTION hConnection, void *pParam)
{
	char buf[64];
	
	snprintf(buf,sizeof(buf),"product %s version %s\r\n", WRGetProductName(), WRGetProductVersion());

	AddHttpBody(hConnection, buf, strlen(buf) );
	SetHttpHeader(hConnection, 200, "OK", "", "Expires: 01 Jan 1970 00:00:00 GMT\r\n", "text/plain", TRUE);
	return 0;
}

int postinfo(HTTPCONNECTION hConnection, void *pParam)
{
	LIST *pList = (LIST *)pParam;
	const char *mac0=NULL, *mac1=NULL;
	int reboot = 0;
	unsigned int macseg[6]={0,0,0,0,0,0};
	char buf[6];
	int i;
	char body[64];
	int iOffsetConfig = 0x7F010000;
	char aryTmp2[6];
	int fail=0;
	
	/* Read page parameters. */
	mac0 = GetString(pList, "mac0");
	mac1 = GetString(pList, "mac1");
	reboot = GetLong(pList, "reboot");
	
	printf("mac0=%s, mac1=%s, reboot=%d\n", mac0, mac1, reboot );

	// read MAC0
	iOffsetConfig = iOffsetConfig + 8 ;
	ReadWinbondFlash(iOffsetConfig, sizeof(buf), buf);
	printf("\norig MAC0:");
	for( i=0; i<6; i++ )
		printf("%X:", buf[i]);

	// write MAC0
	if( mac0 && sscanf(mac0,"%X-%X-%X-%X-%X-%X", 
		&macseg[0],&macseg[1],&macseg[2],&macseg[3],&macseg[4],&macseg[5]) == 6 )
	{
		printf("\nmac->%02x:%02x:%02x:%02x:%02x:%02x\n",
			macseg[0],macseg[1],macseg[2],macseg[3],macseg[4],macseg[5] );
		for(i=0; i<6; i++)
			buf[i] = (char *)macseg[i];
		WriteWinbondFlash(iOffsetConfig, sizeof(buf), buf);

		//check MAC0
		ReadWinbondFlash(iOffsetConfig, sizeof(aryTmp2), aryTmp2);
		for( i=0; i<6; i++ )
		{
			if( aryTmp2[i] != buf[i] )
				fail++;
		}
	}
	
	// read MAC1
	iOffsetConfig = iOffsetConfig + 12;
	ReadWinbondFlash(iOffsetConfig, sizeof(buf), buf);
	printf("\norig MAC1:");
	for( i=0; i<6; i++ )
		printf("%X:", buf[i]);

	// write MAC1
	if( mac1 && sscanf(mac1,"%X-%X-%X-%X-%X-%X", 
		&macseg[0],&macseg[1],&macseg[2],&macseg[3],&macseg[4],&macseg[5]) == 6 )
	{
		printf("\nmac->%02x:%02x:%02x:%02x:%02x:%02x\n",
			macseg[0],macseg[1],macseg[2],macseg[3],macseg[4],macseg[5] );
		for(i=0; i<6; i++)
			buf[i] = (char *)macseg[i];
		WriteWinbondFlash(iOffsetConfig, sizeof(buf), buf);

		//check MAC0
		ReadWinbondFlash(iOffsetConfig, sizeof(aryTmp2), aryTmp2);
		for( i=0; i<6; i++ )
		{
			if( aryTmp2[i] != buf[i] )
				fail++;
		}
	}

	SetKeepAliveMode(hConnection, FALSE);
	snprintf(body,sizeof(body),"postinfo %s\r\n", fail ? "fail" : "ok");

	AddHttpBody(hConnection, buf, strlen(buf) );
	SetHttpHeader(hConnection, 200, "OK", "", "Expires: 01 Jan 1970 00:00:00 GMT\r\n", "text/plain", TRUE);
	sleep(1);
	
	if ( reboot )
		W90740Reboot();
		
	return 0;
}

BOOL KillAllProc(char *pcProcName)
{
	pid_t myPid;
	pid_t* pidList;
	pid_t* pidThis;

	myPid = getpid();

	pidList = FindPidByName(pcProcName);

	if (!pidList)
		return FALSE;

	for(pidThis=pidList; *pidThis!=0; pidThis++)
	{
		if (*pidThis == myPid) continue;
		if (kill(*pidThis, SIGTERM) != 0)
			fprintf(stderr, "Could not kill pid '%d'\n", *pidThis);
	}
	free(pidList);
	
	return TRUE;
}

int WRExternalAdmin(HTTPCONNECTION hConnection)
{

	if (g_ConfigParam.bEnableAdminWan) return -1;

#if 0
	unsigned long ulPrivateNetmask;
	unsigned long ulPrivateIP;
	unsigned long ulClientIP;

	ulPrivateIP = GetIPAddress(G_PC_ETH0);
	ulPrivateNetmask = GetNetMask(G_PC_ETH0);
	ulClientIP = String2IP((char *)inet_ntoa(GetClientAddr(hConnection)));

	if ((ulPrivateIP & ulPrivateNetmask) != (ulClientIP & ulPrivateNetmask))
	{
		SetHttpHeader(hConnection, 404, "Not Found", "", "Expires: 01 Jan 1970 00:00:00 GMT\r\n", "text/html", TRUE);
		return 0;
	}
#endif

	return -1;


}

void WRSetFactoryDefault()
{
	g_ConfigParam.ulCheckSum = GetConfigCheckSum(&g_ConfigParam) + 1;
	__WriteFlashMemory(&g_ConfigParam, FALSE);
}


void W90740Reboot(void)
{
	volatile unsigned int iValue;
	
	KillAllProc(PROG_PPPOE); //let it send out PADT to server
	RunProgram(NULL,"/bin/reboot",NULL);
	
	WRSIGTERM(0);
	exit(-1);
	
	iValue = *((volatile unsigned int *)0xFFF0000C);
	iValue |= 1;
	*((volatile unsigned int *)0xFFF0000C) = iValue;
	
}


void WRReSetRouter()
{
	WRSIGTERM(0);
#ifdef EMBED
	W90740Reboot();
#endif
}

void W90740ResetWatchDogTimer(void)
{
	volatile unsigned int iValue;
	
#if 0
	iValue = *((volatile unsigned int *)0xFFF81000);
	printf("\nWatchDogTimer (0xfff81000) = %x ", iValue);
	iValue = 0x0005;
	*((volatile unsigned int *)0xFFF81000) = iValue;
	iValue = *((volatile unsigned int *)0xFFF81000);
	printf("\nWatchDogTimer (0xfff81000) = %x ", iValue);	
#endif
	
	iValue = *((volatile unsigned int *)0xFFF83000);
	printf("\nWatchDogTimer (GPIO12) (0xfff83000) = %x ", iValue);
	//iValue &= ~0x00c0;
	//iValue &= ~0xc0;
	iValue &= ~0x1000;
	*((volatile unsigned int *)0xFFF83000) = iValue;
	iValue = *((volatile unsigned int *)0xFFF83000);
	printf("\nWatchDogTimer (GPIO12)(0xfff83000) = %x \n", iValue);	
}


BOOL RunProgram(pid_t *ppidFork, char *pcPath, ...)
{
	char *pcArg[20];
	int i;
	va_list arg;
	pid_t pidTmp;
	volatile BOOL bExecvSuccess = TRUE;

	if (ppidFork == NULL)
	{
		//fprintf(stderr, "Call RunProgram(), no pid set.\n");
		ppidFork = &pidTmp;
	}

	pcArg[0] = pcPath;
	va_start(arg, pcPath);
	for (i = 1; i < sizeof(pcArg) / sizeof(char *) - 1; i++)
	{
		pcArg[i] = va_arg(arg, char *);
		if (pcArg[i] == NULL) break;
		if ( *pcArg[i] == '\0' ) break;
	}
	pcArg[i + 1] = NULL;
	va_end(arg);

	if (vfork() == 0)
	{
#ifdef XX
		int fd = getdtablesize();
		
		fprintf(stderr,"Running %s.\n", pcArg[0]);
		while( fd >= 0 )
		{
			close(fd);
			fd--;
		}
#endif
		*ppidFork = getpid();
		execvp(pcArg[0], pcArg);
		printf("execv %s error!\n", pcArg[0]);
		*ppidFork = 0;
		bExecvSuccess = FALSE;
		_exit(0);
	}

	return bExecvSuccess;
}
