#include "CommonDef.h"


XML *AppendXMLArray(XML *pXML, const char *pcName)
{
	LISTNODE *pNode;
	XML *pNodeXML;
	XML *pAppend;
	if (pXML == NULL
		|| pXML->plAttrib == NULL
		|| pcName == NULL) return NULL;
	
	pAppend = CreateXML(pcName);
	if (pAppend == NULL) return NULL;
	
	for (pNode = pXML->plSubXML->pFirstNode;
		pNode != pXML->plSubXML->pLastNode;
		pNode = pNode->pNextNode)
	{
		pNodeXML = (XML *)pNode->pValue;
		if (pNodeXML == NULL) continue;
		
		if (strcmp(GetString(pNodeXML->plAttrib, "Array"), pcName) == 0)
		{
			if (!AppendXML(pNodeXML, pAppend))
			{
				DeleteXML(pAppend);
				return NULL;
			}
			else return pAppend;
		}

		if (strcmp(pNodeXML->pcName, pcName) == 0)
		{
			XML *pArrayXML = CreateXML("Array");
			if (pArrayXML != NULL
				&& AppendXML(pXML, pArrayXML))
			{
				SetString(pArrayXML->plAttrib, "Array", pcName);
				if (DetachXML(pXML, pNodeXML))
				{
					if (!AppendXML(pArrayXML, pNodeXML))
						DeleteXML(pNodeXML);
				}
				if (!AppendXML(pArrayXML, pAppend))
				{
					DeleteXML(pAppend);
					return NULL;
				}
				else return pAppend;
			}
			else if (pArrayXML != NULL)
				DeleteXML(pAppend);

			return NULL;
		}
	}
	
	if (!AppendXML(pXML, pAppend))
	{
		DeleteXML(pAppend);
		return NULL;
	}
	else return pAppend;
}

	

XML *ReturnTypeBegin(HTTPCONNECTION hConnection, LIST *pQueryList)
{
	XML *pRt;
	int i;
	const char *pcName;
	const char *apcType[] = {"JsVar", "JsObj", "XML"};

	if (pQueryList == NULL) return NULL;

	
	for (i = 0; i < sizeof(apcType) / sizeof(const char *); i++)
	{
		pcName = GetString(pQueryList, apcType[i]);
		if (pcName != NULL && pcName[0] != 0) break;
	} while (0);

	if (!(i < sizeof(apcType) / sizeof(const char *)))
		return CreateXML("Txt");

	pRt = CreateXML(apcType[i]);
	if (pRt != NULL)
	{
		if (!SetString(pRt->plAttrib, "Name", pcName))
			DeleteXML(pRt);
		else return pRt;
	}
	return NULL;
}


void Dump_HttpAdd(const char *s, void *pParam)
{
	HTTPCONNECTION *pConn = (HTTPCONNECTION *)pParam;
	if (pConn != NULL)
		AddHttpBodyString(*pConn, s);
}


static void inline HTN(HTTPCONNECTION hConnection, int iCount)
{
	int i;
	for (i = 0; i < iCount; i++)
		AddHttpBodyString(hConnection, "\t");
}


void DumpJsVar(XML *pXML, void *pParam, BOOL bJS)
{
	LISTNODE *pNode;
	char *pc;
	HTTPCONNECTION *pConn = (HTTPCONNECTION *)pParam;
	if (pXML == NULL) return;

	if (pXML->plSubXML != NULL)
	{
		for (pNode = pXML->plSubXML->pFirstNode;
			pNode != pXML->plSubXML->pLastNode;
			pNode = pNode->pNextNode)
		DumpJsVar(pNode->pValue, pParam, bJS);
	}
	else
	{
		XML *pParent = GetParentXML(pXML);
		if (pParent != NULL)
		{
			if (bJS)
			{
				if ((pc = GetWebInCString(pParent->pcName)) != NULL)
				{
					AddHttpBodyString(*pConn, pc);
					free(pc);
				}
				AddHttpBodyString(*pConn, " = ");
				if ((pc = GetWebInCString(pXML->pcName)) != NULL)
				{
					AddHttpBodyString(*pConn, pc);
					free(pc);
				}
				AddHttpBodyString(*pConn, "\\n");
			}
			else
			{
				AddHttpBodyString(*pConn, pParent->pcName);
				AddHttpBodyString(*pConn, " = ");
				AddHttpBodyString(*pConn, pXML->pcName);
				AddHttpBodyString(*pConn, "\n");
			}
		}
	}
}


char *GetCleanVariableName(const char *pcVaribleName)
{
	char *pcBrk;
	char *pc = strdup(pcVaribleName);
	if (pc == NULL) return NULL;
	if (pc[0] == '\0')
	{
		free(pc);
		return NULL;
	}

	if (!((pc[0] >= 'a' && pc[0] <= 'z') || (pc[0] >= 'A' && pc[0] <= 'Z') || pc[0] == '_'))
		pc[0] = 'V';
	for (pcBrk = pc; ((*pcBrk >= '0' && *pcBrk <= '9') || (*pcBrk >= 'a' && *pcBrk <= 'z') || (*pcBrk >= 'A' && *pcBrk <= 'Z') || *pcBrk == '_'); pcBrk++)
		;
	*pcBrk = '\0';
	return pc;
}

#define IS_ONLY_ONE_NODE(pList) ((pList)->pFirstNode->pNextNode == (pList)->pLastNode)
#define IS_TO_LAST_NODE(pNode) ((pNode)->pNextNode == (pNode)->pList->pLastNode)
void DumpJsObj(XML *pXML, void *pParam, int iTabs, BOOL bInArray)
{
	LISTNODE *pNode;
	char *pc;
	HTTPCONNECTION *pConn = (HTTPCONNECTION *)pParam;

	if (pXML == NULL) return;
	
	
	if (pXML->plAttrib == NULL)
	{//pure text node
		HTN(*pConn, iTabs);
		AddHttpBodyString(*pConn, "\"");
		
		pc = GetWebInCString(pXML->pcName);
		if (pc != NULL)
		{
			AddHttpBodyString(*pConn, pc);
			free(pc);
		}
		AddHttpBodyString(*pConn, "\"");
	}
	else if (pXML->plAttrib->pFirstNode != pXML->plAttrib->pLastNode)
	{//array
		HTN(*pConn, iTabs);
		pc = GetCleanVariableName(GetString(pXML->plAttrib, "Array"));
		if (pc != NULL)
		{
			AddHttpBodyString(*pConn, pc);
			free(pc);
		}
		else AddHttpBodyString(*pConn, "VName");
		AddHttpBodyString(*pConn, " :\n");
		HTN(*pConn, iTabs);
		AddHttpBodyString(*pConn, "[\n");
		for (pNode = pXML->plSubXML->pFirstNode;
			pNode != pXML->plSubXML->pLastNode;
			pNode = pNode->pNextNode)
		{
			XML *pSub = (XML *)pNode->pValue;
			if (pSub == NULL) continue;
			DumpJsObj(pSub, pParam, iTabs + 1, TRUE);
		}
		HTN(*pConn, iTabs);
		if (IS_TO_LAST_NODE(pXML->pParXMLNode))
			AddHttpBodyString(*pConn, "]\n");
		else AddHttpBodyString(*pConn, "],\n");
	}
	else
	{//struct
		BOOL bStruct;
		for (bStruct = FALSE, pNode = pXML->plSubXML->pFirstNode;
			pNode != pXML->plSubXML->pLastNode;
			pNode = pNode->pNextNode)
		{
			XML *pNodeXML = (XML *)pNode->pValue;
			if (pNodeXML != NULL && pNodeXML->plSubXML != NULL)
			{
				bStruct = TRUE;
				break;
			}
		}
		
		if (!bInArray)
		{
			HTN(*pConn, iTabs);
			pc = GetCleanVariableName(pXML->pcName);
			if (pc != NULL)
			{
				AddHttpBodyString(*pConn, pc);
				free(pc);
			}
			else AddHttpBodyString(*pConn, pXML->pcName);
			AddHttpBodyString(*pConn, ":\n");
		}
		if (bStruct)
		{
			HTN(*pConn, iTabs);
			AddHttpBodyString(*pConn, "{\n");
		}

		for (pNode = pXML->plSubXML->pFirstNode;
			pNode != pXML->plSubXML->pLastNode;
			pNode = pNode->pNextNode)
		{
			XML *pNodeXML = (XML *)pNode->pValue;
			if (pNodeXML != NULL)
				DumpJsObj(pNodeXML, pParam, iTabs + 1, FALSE);
		}
		if (bStruct)
		{
			HTN(*pConn, iTabs);
			if (IS_TO_LAST_NODE(pXML->pParXMLNode))
				AddHttpBodyString(*pConn, "}\n");
			else AddHttpBodyString(*pConn, "},\n");
		}
		else
		{
			if (!IS_TO_LAST_NODE(pXML->pParXMLNode))
				AddHttpBodyString(*pConn, ",\n");
			else AddHttpBodyString(*pConn, "\n");
		}
	}
#if 0
	if (pXML->plSubXML != NULL)
	{
		BOOL bStruct;
		BOOL bOnlyOneNode;

		bOnlyOneNode = IS_ONLY_ONE_NODE(pXML->plSubXML);
		
		for (bStruct = FALSE, pNode = pXML->plSubXML->pFirstNode;
			pNode != pXML->plSubXML->pLastNode;
			pNode = pNode->pNextNode)
		{
			XML *pNodeXML = (XML *)pNode->pValue;
			if (pNodeXML != NULL && pNodeXML->plSubXML != NULL)
			{
				bStruct = TRUE;
				break;
			}
		}
		
		HTN(*pConn, iTabs);
		pc = strdup(pXML->pcName);
		if (pc != NULL)
		{
			if (pc[0] == '\0') AddHttpBodyString(*pConn, "VName");
			else
			{
				char *pcBrk;
				if (!((pc[0] >= 'a' && pc[0] <= 'z') || (pc[0] >= 'A' && pc[0] <= 'Z') || pc[0] == '_'))
					pc[0] = 'V';
				for (pcBrk = pc; ((*pcBrk >= '0' && *pcBrk <= '9') || (*pcBrk >= 'a' && *pcBrk <= 'z') || (*pcBrk >= 'A' && *pcBrk <= 'Z') || *pcBrk == '_'); pcBrk++)
					;
				
				*pcBrk = '\0';
				AddHttpBodyString(*pConn, pc);
			}
			free(pc);
		}
		else AddHttpBodyString(*pConn, pXML->pcName);
		
		if (!bStruct && bOnlyOneNode)
			AddHttpBodyString(*pConn, ": ");
		else
		{
			AddHttpBodyString(*pConn, ":\n");
			HTN(*pConn, iTabs);
			if (bStruct)
				AddHttpBodyString(*pConn, "{\n");
			else if (!bOnlyOneNode)
				AddHttpBodyString(*pConn, "[\n");
		}
				
		if (!bStruct && bOnlyOneNode)
			DumpJsObj(pXML->plSubXML->pFirstNode->pValue, pParam, 0);
		else
			for (pNode = pXML->plSubXML->pFirstNode;
				pNode != pXML->plSubXML->pLastNode;
				pNode = pNode->pNextNode)
			DumpJsObj(pNode->pValue, pParam, iTabs + 1);
		
		if (bStruct)
		{
			HTN(*pConn, iTabs);
			AddHttpBodyString(*pConn, "}");
		}
		else if (!bOnlyOneNode)
		{
			AddHttpBodyString(*pConn, "\n");
			HTN(*pConn, iTabs);
			AddHttpBodyString(*pConn, "]");
		}
		
		if (IS_TO_LAST_NODE(pXML->pParXMLNode))
			AddHttpBodyString(*pConn, "\n");
		else
			AddHttpBodyString(*pConn, ",\n");
	}
	else
	{
		HTN(*pConn, iTabs);
		AddHttpBodyString(*pConn, "\"");
		
		pc = GetWebInCString(pXML->pcName);
		if (pc != NULL)
		{
			AddHttpBodyString(*pConn, pc);
			free(pc);
		}
		
		if (IS_TO_LAST_NODE(pXML->pParXMLNode))
			AddHttpBodyString(*pConn, "\"");
		else
			AddHttpBodyString(*pConn, "\",\n");			
	}
#endif	
}

void ReturnTypeEnd(HTTPCONNECTION hConnection, LIST *pList, XML *pReturnXML)
{
	const char *pcName;
	LISTNODE *pNode;
	const char *pcMimeType = "text/plain";
	
	if (pReturnXML != NULL)
	{
		pcName = GetString(pReturnXML->plAttrib, "Name");
		if (strcmp(pReturnXML->pcName, "Txt") == 0)
		{
			DumpJsVar(pReturnXML, &hConnection, FALSE);
		}
		else if (strcmp(pReturnXML->pcName, "JsVar") == 0)
		{
			const char *pcFun;
			AddHttpBodyString(hConnection, pcName);
			AddHttpBodyString(hConnection, "=\"");
			for (pNode = pReturnXML->plSubXML->pFirstNode;
				pNode != pReturnXML->plSubXML->pLastNode;
				pNode = pNode->pNextNode)
			{
				DumpJsVar(pNode->pValue, &hConnection, TRUE);
			}
			AddHttpBodyString(hConnection, "\";\n");			
			pcFun = GetString(pList, "OnJs");
			if (pcFun != NULL && pcFun[0] != '\0')
			{
				AddHttpBodyString(hConnection, pcFun);
				AddHttpBodyString(hConnection, "(this);\n");
			}			
		}
		else if (strcmp(pReturnXML->pcName, "JsObj") == 0)
		{
			const char *pcFun;
			AddHttpBodyString(hConnection, pcName);
			AddHttpBodyString(hConnection, "=\n{\n");
			for (pNode = pReturnXML->plSubXML->pFirstNode;
				pNode != pReturnXML->plSubXML->pLastNode;
				pNode = pNode->pNextNode)
			{
				DumpJsObj(pNode->pValue, &hConnection, 1, 0);
			}
			AddHttpBodyString(hConnection, "};\n");
			pcFun = GetString(pList, "OnJs");
			if (pcFun != NULL && pcFun[0] != '\0')
			{
				AddHttpBodyString(hConnection, pcFun);
				AddHttpBodyString(hConnection, "(this);\n");
			}			
		}
		else if (strcmp(pReturnXML->pcName, "XML") == 0)
		{
			DumpXML(pReturnXML, Dump_HttpAdd, (void *)&hConnection);
			pcMimeType = "text/xml";
		}
		DeleteXML(pReturnXML);
	}
	SetHttpHeader(hConnection, 200, "OK", "", G_PC_NO_CACHE_HEADER, pcMimeType, TRUE);
}


void AddHttpValue(XML *pReturnXML, const char *pcString, const char *pcValue)
{
	XML *pAppend = NULL;
	XML *pText;
	if (pReturnXML == NULL || pcString == NULL) return;

#if 0	
	/* Check if pcString already exists.*/
	for (pNode = pReturnXML->plSubXML->pFirstNode;
		pNode != pReturnXML->plSubXML->pLastNode;
		pNode = pNode->pNextNode)
	{
		pAppend = (XML *)pNode->pValue;
		if (pAppend == NULL) continue;
		if (strcmp(pAppend->pcName, pcString) == 0) break;
	}
	
	if (pNode == pReturnXML->plSubXML->pLastNode)
	{
		pAppend = CreateXML(pcString);
		if (pAppend != NULL) pAppend = AppendXML(pReturnXML, pAppend);
	}
#else
	pAppend = AppendXMLArray(pReturnXML, pcString);
#endif	
	if (pAppend != NULL)
	{
		pText = CreateXMLText(pcValue);
		if (pText != NULL) AppendXML(pAppend, pText);
	}
}


int Http_SendRedirectRequest(HTTPCONNECTION hConnection, LIST *pQueryList)
{
	char acExtraHeader[256];
	const char *pcUrl = GetString(pQueryList, "RedirectUrl");
	BOOL bShowTime = GetBool(pQueryList, G_PC_REDIRECT_TIME_SHOW);
	int iTime_InMSec = GetLong(pQueryList, G_PC_REDIRECT_TIME_DEFER);
	const char *pcInfo = GetString(pQueryList, G_PC_REDIRECT_INFO);

	if (pcUrl == NULL || pcUrl[0] == '\0') return 1;

	ClearHttpSendData(hConnection);

	if (iTime_InMSec <= 0)
	{
		sprintf(acExtraHeader, "Location: %s\r\n", pcUrl);
		SetHttpHeader(hConnection, 302, "OK", "", acExtraHeader, "text/html", TRUE);
	}
	else
	{
		const char *pcRedirect = "Redirect.htm";
		char acRedirect[256];
		char *pcRedirectPath = NULL;
		char *pcReferer;
		int iLen;
		char *pc1;
		char *pc2;

		pcReferer = GetReferer(hConnection);

		if (pcReferer == NULL) iLen = 0;
		else iLen = strlen(pcReferer);
		if (iLen <= 7)//for "http://"
			pcReferer = NULL;

		if (pcReferer != NULL)
		{
			pc1 = strchr(pcReferer + 7, '/');
			pc2 = strrchr(pcReferer + 7, '/');

			if (pc1 != NULL)
			{
				iLen = pc2 - pc1;
				if (iLen + 2 + strlen(pcRedirect) < sizeof(acRedirect))
				{
					memcpy(acRedirect, pc1, iLen);
					acRedirect[iLen] = '/';
					strcpy(acRedirect + iLen + 1, pcRedirect);
					pcRedirectPath = GetDocumentBasedPath(acRedirect);
				}
			}
		}

		SetKeepAliveMode(hConnection, FALSE);
		if (pcRedirectPath == NULL)
		{
			char *pcHtml =
"<!--$B<script language=JavaScript>"
"sUrl = '%s';"
"bShowTime = %d;"
"iTime_InMSec = %d;"
"sInfo = '%s';"
"</script>-->"
"<script>"
"document.write(sInfo);"
"setTimeout('location=\"'+sUrl+'\";', iTime_InMSec);"
"document.write('<form name=CF>The page is to be redirected to:<br> <a href=\"' + sUrl + '\">' + sUrl + '</a> in ' + (iTime_InMSec/1000) + ' seconds.</form>');"
"</script>";
			SendBufferWithParam(hConnection,
				pcHtml,
				strlen(pcHtml),
				pcUrl,
				bShowTime,
				iTime_InMSec,
				pcInfo);
		}
		else
		{
			SendHtmlWithParam(hConnection,
				pcRedirectPath,
				pcUrl,
				bShowTime,
				iTime_InMSec,
				pcInfo);
			free(pcRedirectPath);
		}
	}
	return 0;
}


static int Config_SubNotFound(HTTPCONNECTION hConnection, LIST *pParamList, int iAction, XML *pReturnXML)
{
	switch (iAction)
	{
	case CA_AUTH:
		return AUTH_ANY;
		break;
	case CA_CONFIG:
		return 0;
		break;
	}
	return 0;
}


typedef struct
{
	LIST *pParamList;
	SUBCMD_PFUN funSub;
} SUBCMD_RUN_STRUCT;

SUBCMD_RUN_STRUCT *CreateSubCmdRunStruct(const char *pcParamString)
{
	REQUEST_CALLBACK_PFUN pfunRequestCallBack;
	SUBCMD_RUN_STRUCT *pRt;
	const char *pcSubCmd;
	void *pFun;

	pRt = (SUBCMD_RUN_STRUCT *)malloc(sizeof(SUBCMD_RUN_STRUCT));
	if (pRt == NULL)
	{
		PRINT_MEM_OUT;
		return NULL;
	}
	pRt->pParamList = ParseString(pcParamString);
	pcSubCmd = GetString(pRt->pParamList, "Cmd");
	if (IsRegisterEmbedFunEx(pcSubCmd, &pfunRequestCallBack, NULL, &pFun)
		&& pfunRequestCallBack == Http_SubCmdInit
		&& pFun != NULL)
	{
		pRt->funSub = (SUBCMD_PFUN)pFun;
		return pRt;
	}
	else
	{
		//DeleteParseString(pRt->pParamList);
		//free(pRt);
		//return NULL;
		pRt->funSub = Config_SubNotFound;
		fprintf(stderr, "Sub cmd not found: %s\n", pcSubCmd);
		return pRt;
	}
}

void DeleteSubCmdRunStruct(SUBCMD_RUN_STRUCT *pStruct)
{
	if (pStruct == NULL) return;
	if (pStruct->pParamList != NULL)
		DeleteParseString(pStruct->pParamList);
	free(pStruct);
}

int Http_CommonCmdRun(HTTPCONNECTION hConnection, char *pcCmdParam, int iCmdParamLen, SUBCMD_PFUN funConfig)
{
	LIST *pList;
	LIST *pListSeg;
	LIST *pListSub = NULL;
	LISTNODE *pNode;
	int iPrivilege;
	int iAuthRequired;
	XML *pReturnXML;
	SUBCMD_RUN_STRUCT *pSubCmdStruct;

	pList = ParseString(pcCmdParam?pcCmdParam:"");

	pListSeg = SplitBuffer(pcCmdParam, iCmdParamLen, -1, NULL, "&Cmd=");
	for (pNode = pListSeg->pFirstNode; pNode != pListSeg->pLastNode; pNode = pNode->pNextNode)
	{
		if (pNode->pValue)
		{
			SPLIT_ITEM_T *pSplitItem;
			char *pcSubStr;
			pSplitItem = (SPLIT_ITEM_T *)pNode->pValue;
			pcSubStr = (char *)malloc(5 + pSplitItem->pcSegEnd - pSplitItem->pcSegStart);
			if (pcSubStr != NULL)
			{
				int iOffSet;

				if (pNode != pListSeg->pFirstNode)
				{
					memcpy(pcSubStr, "Cmd=", 4);
					iOffSet = 4;
				}
				else iOffSet = 0;
				memcpy(pcSubStr + iOffSet, pSplitItem->pcSegStart, pSplitItem->pcSegEnd - pSplitItem->pcSegStart);
				pcSubStr[iOffSet + pSplitItem->pcSegEnd - pSplitItem->pcSegStart] = '\0';

				pSubCmdStruct = CreateSubCmdRunStruct(pcSubStr);
				if (pSubCmdStruct != NULL)
				{
					if (pListSub == NULL) pListSub = CreateList();
					if (pListSub != NULL)
					{
						LISTNODE *pAppNode = AppendNode(pListSub);
						if (pAppNode != NULL)
							pAppNode->pValue = pSubCmdStruct;
					}
				}
				free(pcSubStr);
			}
			else
				PRINT_MEM_OUT;
		}
	}
	DeleteSplitBuffer(pListSeg);

	iAuthRequired = AUTH_ANY;

	if (pListSub != NULL)
	{
		/* Get the highest privilege of all sub-commands. */
		for (pNode = pListSub->pFirstNode; pNode != pListSub->pLastNode; pNode = pNode->pNextNode)
		{
			int iAuthRequiredNew;
			pSubCmdStruct = (SUBCMD_RUN_STRUCT *)pNode->pValue;
			if (pSubCmdStruct == NULL) continue;

			iAuthRequiredNew = (*pSubCmdStruct->funSub)(hConnection, pSubCmdStruct->pParamList, CA_AUTH, NULL);
			if (iAuthRequiredNew > iAuthRequired)
				iAuthRequired = iAuthRequiredNew;
		}
		iPrivilege = GetAuthPrivilege(hConnection);

		if (iPrivilege < iAuthRequired)
		{
			SendAuthRequired(hConnection, iAuthRequired);
		}
		else
		{
			pReturnXML = ReturnTypeBegin(hConnection, pList);
			for (pNode = pListSub->pFirstNode; pNode != pListSub->pLastNode; pNode = pNode->pNextNode)
			{
				XML *pSubXML;
				
				pSubCmdStruct = (SUBCMD_RUN_STRUCT *)pNode->pValue;
				if (pSubCmdStruct == NULL
					|| pSubCmdStruct->funSub == Config_SubNotFound) continue;

				
				pSubXML = AppendXMLArray(pReturnXML, GetString(pSubCmdStruct->pParamList, "Cmd"));
				if (pSubXML == NULL) continue;
				
				(*pSubCmdStruct->funSub)(hConnection, pSubCmdStruct->pParamList, CA_CONFIG, pSubXML);
			}
			ReturnTypeEnd(hConnection, pList, pReturnXML);

			for (pNode = pListSub->pFirstNode; pNode != pListSub->pLastNode; pNode = pNode->pNextNode)
			{
				pSubCmdStruct = (SUBCMD_RUN_STRUCT *)pNode->pValue;
				if (pSubCmdStruct == NULL) continue;

				if (Http_SendRedirectRequest(hConnection, pSubCmdStruct->pParamList) == 0)
					break;
			}
		}

		for (pNode = pListSub->pFirstNode; pNode != pListSub->pLastNode; pNode = pNode->pNextNode)
		{
			pSubCmdStruct = (SUBCMD_RUN_STRUCT *)pNode->pValue;
			DeleteSubCmdRunStruct(pSubCmdStruct);
		}
		DeleteList(pListSub);
	}
	else
	{
		if (Http_SendRedirectRequest(hConnection, pList) != 0)
		{
			pReturnXML = ReturnTypeBegin(hConnection, pList);
			ReturnTypeEnd(hConnection, pList, pReturnXML);
		}
	}

	DeleteParseString(pList);
	return 0;
}


int Http_SubCmdRun(HTTPCONNECTION hConnection, char *pcCmdParam, int iCmdParamLen, SUBCMD_PFUN funConfig)
{
	LIST *pList;
	int rt;
	int iPrivilege;
	int iAuthRequired;
	XML *pReturnXML;

	pList = ParseString(pcCmdParam?pcCmdParam:"");

	iAuthRequired = (*funConfig)(hConnection, pList, CA_AUTH, FALSE);
	iPrivilege = GetAuthPrivilege(hConnection);

	if (iPrivilege < iAuthRequired)
	{
		SendAuthRequired(hConnection, iAuthRequired);
		rt = 0;
	}
	else
	{
		pReturnXML = ReturnTypeBegin(hConnection, pList);
		rt = (*funConfig)(hConnection, pList, CA_CONFIG, pReturnXML);
		ReturnTypeEnd(hConnection, pList, pReturnXML);
		Http_SendRedirectRequest(hConnection, pList);
	}

	DeleteParseString(pList);
	return rt;
}

typedef struct
{
	void *pParam;
	SUBCMD_CALLER_PFUN funCmdCaller;
} CO_PARAM_T;

int Http_PostDataGet(HTTPCONNECTION hConnection,
								int *piPostState,
								char **ppcPostBuf,
								int *piPostBufLen,
								int *piPostDataLen,
								char *pcFillData,
								int iFillDataLen,
								int iIsMoreData/*bool*/,
								void *pParam/*other parameter for extend use*/)
{
	if (*piPostBufLen == 0)
	{//Initial state, no post data received.
		*piPostBufLen = GetContentLength(hConnection) + 8;
		if (*piPostBufLen < 8) *piPostBufLen = 8;
		if (*piPostBufLen < MAX_POST_LENGTH)
		{
			if ((*ppcPostBuf = (char *)malloc(*piPostBufLen)) == NULL)
			{
				PRINT_MEM_OUT;
			}
			*piPostDataLen = 0;
		}
	}

	if (*ppcPostBuf == NULL)
	{//Not in initial state, malloc failed or post data too longer.
		if (!iIsMoreData)
		{
			AddHttpBodyString(hConnection, "Too many post data or not enmough memory.");
			SetHttpHeader(hConnection, 200, "OK", NULL, NULL, "text/plain", TRUE);
			return 1;
		}
	}
	else
	{//Not in initial state, malloc successfully.
		if (*piPostDataLen + iFillDataLen < *piPostBufLen)
		{
			memcpy(*ppcPostBuf + *piPostDataLen, pcFillData, iFillDataLen);
			*piPostDataLen += iFillDataLen;
		}

		if (!iIsMoreData)
		{
			(*ppcPostBuf)[*piPostDataLen] = '\0';
			if (*piPostDataLen >= 1 && (*ppcPostBuf)[*piPostDataLen - 1] == '\n')
			{
				(*ppcPostBuf)[--*piPostDataLen] = '\0';
				if (*piPostDataLen >= 1 && (*ppcPostBuf)[*piPostDataLen - 1] == '\r')
					(*ppcPostBuf)[--*piPostDataLen] = '\0';
			}

			if (pParam != NULL)
			{
				CO_PARAM_T *pCoParam = (CO_PARAM_T *)pParam;
				(*pCoParam->funCmdCaller)(hConnection, *ppcPostBuf, *piPostDataLen, (SUBCMD_PFUN)(pCoParam->pParam));
				free(pCoParam);
			}
			free(*ppcPostBuf);
			*ppcPostBuf = NULL;
			*piPostDataLen = *piPostBufLen = 0;
			return 1;
		}
	}

	return 1;
}

int Http_ConnInit(HTTPCONNECTION hConnection, void *pParam, SUBCMD_CALLER_PFUN funCaller)
{
	char *pcQuery;
	CO_PARAM_T *pCoParam;

	switch (GetHttpMethod(hConnection))
	{
	case M_POST:
		if ((pCoParam = (CO_PARAM_T *)malloc(sizeof(CO_PARAM_T))) != NULL)
		{
			pCoParam->pParam = pParam;
			pCoParam->funCmdCaller = funCaller;
			SetPostDataFun(hConnection, Http_PostDataGet, (void *)pCoParam);
		}
		else
			PRINT_MEM_OUT;
		return 1;
	case M_GET:
		pcQuery = GetQueryString(hConnection);
		if (pcQuery == NULL) pcQuery = "";
		return (*funCaller)(hConnection, pcQuery, strlen(pcQuery), pParam);
	}
	
	return 1; //norton
}

int Http_CommonCmdInit(HTTPCONNECTION hConnection, void *pParam)
{
	return Http_ConnInit(hConnection, NULL, Http_CommonCmdRun);
}

int Http_SubCmdInit(HTTPCONNECTION hConnection, void *pParam)
{
	return Http_ConnInit(hConnection, pParam, Http_SubCmdRun);
}
