#include "CommonDef.h"
#include "aos_fw.h"

#ifdef EMBED
static inline _syscall2(void,nat_ip_filter_reset,unsigned long *,args,int,count);
#endif //#else

char *protolist[]={"ip","tcp","udp","icmp","igmp","ah","esp","unknown"};
char *chainlist[]={"INBOUND","OUTBOUND","SELF","UNKNOWN"};
char *actionlist[]={"ACCEPT","DROP","REJECT","UNKNOWN"};

char *domainfile[]={"/swap/pdin.list","/swap/pdout.list","/swap/pdme.list"};
#define DOMAINFILE "/swap/pdomain.list"

#define FW_INIT_RULE	"/etc/fwinit.rule"
#define RULE_TRY	1
#define RULE_FORCE	2

char *strlwr(char *string)
{
    unsigned int c;
    unsigned char *p;

    p = string;
    while ((c = *p) != 0) {
	*p++ = tolower(c);
    }

    return (string);
}	

char *cdr2ip(unsigned long msk)
{
	static char buf[18];
	struct in_addr addr;
	
	addr.s_addr = htonl((0xffffffff << (32-msk) ));
	strcpy(buf,(char *)inet_ntoa(addr));

	return &buf[0];
}


int get_acl()
{
	int i=0;
	
	for(i=0; i<MAX_IP_PORT_FILTER_ITEM; i++)
	{
		if( g_ConfigParam.fw_acl_pool[i].id == 0 )
			return i;
	}
	return -1;
}

void put_acl(int i)
{
	if( i >= 0 && i<MAX_IP_PORT_FILTER_ITEM)
		memset( &g_ConfigParam.fw_acl_pool[i],0,sizeof(fw_acl));
}

int get_rule_count(int chain)
{
	int i=0,count=0;
	int *fw_acl_list=NULL;
	
	if( chain == CHAIN_IN )
	{
		fw_acl_list = &g_ConfigParam.fw_acl_in[0];
	}
	else if( chain == CHAIN_OUT )
	{
		fw_acl_list = &g_ConfigParam.fw_acl_out[0]; 
	}
	else
	{
		fw_acl_list = &g_ConfigParam.fw_acl_me[0]; 
	}
	
	for(i=0; i<MAX_IP_PORT_FILTER_ITEM; i++)
	{
		if( fw_acl_list[i] )
			count++;
	}

	return count;
}

int fw_move_backward( int *fw_acl_list, int index )
{
	int i=0;
	
	if( index < 0 )  return 0;
	
	//fprintf(stderr,"\nbackward from %d ",index);
	/* do this after del */
	for( i=index; i<MAX_IP_PORT_FILTER_ITEM-1; i++)
	{
		fw_acl_list[i] = fw_acl_list[i+1];
	}
	fw_acl_list[i] = 0;

	return 0;
}

int fw_move_forward( int *fw_acl_list, int index )
{
	int i=0, last=0;
	
	if( index < 0 )  return 0;
	
	/* do this before ins */
	for( i=index; i<MAX_IP_PORT_FILTER_ITEM; i++)
	{
		if( fw_acl_list[i] == 0 ) 
		{
			last = i;
			break;
		}
	}
	for( i=last; i>index; i--)
	{
		fw_acl_list[i] = fw_acl_list[i-1];
	}
	
	return 0;
}

int fw_find_acl( int chain, fw_acl *fw_acl_cur )
{
	int cur_id = -1;
	int i;
	int *fw_acl_list=NULL;
	
	if( chain == CHAIN_IN )
	{
		fw_acl_list = &g_ConfigParam.fw_acl_in[0];
	}
	else if( chain == CHAIN_OUT )
	{
		fw_acl_list = &g_ConfigParam.fw_acl_out[0]; 
	}
	else
	{
		fw_acl_list = &g_ConfigParam.fw_acl_me[0]; 
	}
	for(i=0; i<MAX_IP_PORT_FILTER_ITEM; i++)
	{
		if( fw_acl_list[i] != 0 )
		{
			int id = fw_acl_list[i]-1;
			
			if( id >= 0 && 
				g_ConfigParam.fw_acl_pool[ id ].chain == fw_acl_cur->chain &&
				g_ConfigParam.fw_acl_pool[ id ].srctype == fw_acl_cur->srctype &&
				g_ConfigParam.fw_acl_pool[ id ].src == fw_acl_cur->src &&
				g_ConfigParam.fw_acl_pool[ id ].smsk == fw_acl_cur->smsk &&
				g_ConfigParam.fw_acl_pool[ id ].dsttype == fw_acl_cur->dsttype &&
				g_ConfigParam.fw_acl_pool[ id ].dst == fw_acl_cur->dst &&
				g_ConfigParam.fw_acl_pool[ id ].dmsk == fw_acl_cur->dmsk &&
				g_ConfigParam.fw_acl_pool[ id ].sporttype == fw_acl_cur->sporttype &&
				g_ConfigParam.fw_acl_pool[ id ].sport[0] == fw_acl_cur->sport[0] &&
				g_ConfigParam.fw_acl_pool[ id ].sport[1] == fw_acl_cur->sport[1] &&
				g_ConfigParam.fw_acl_pool[ id ].dporttype == fw_acl_cur->dporttype &&
				g_ConfigParam.fw_acl_pool[ id ].dport[0] == fw_acl_cur->dport[0] &&
				g_ConfigParam.fw_acl_pool[ id ].dport[1] == fw_acl_cur->dport[1] &&
				g_ConfigParam.fw_acl_pool[ id ].proto == fw_acl_cur->proto &&
				g_ConfigParam.fw_acl_pool[ id ].action == fw_acl_cur->action &&
				g_ConfigParam.fw_acl_pool[ id ].icmptype == fw_acl_cur->icmptype &&
				strncmp(g_ConfigParam.fw_acl_pool[ id ].route, fw_acl_cur->route, sizeof(fw_acl_cur->route) )==0 &&
				strncmp(g_ConfigParam.fw_acl_pool[ id ].dstdomain, fw_acl_cur->dstdomain, sizeof(fw_acl_cur->dstdomain) )==0 
			)
			{
				cur_id = i;
				break;
			}
		}
	}
	

	return cur_id;
}


void fw_del( int chain, int cur_id)
{
	int *fw_acl_list=NULL;
	
	if( chain == CHAIN_IN )
	{
		fw_acl_list = &g_ConfigParam.fw_acl_in[0];
	}
	else if( chain == CHAIN_OUT )
	{
		fw_acl_list = &g_ConfigParam.fw_acl_out[0]; 
	}
	else
	{
		fw_acl_list = &g_ConfigParam.fw_acl_me[0]; 
	}
	
	if( cur_id < MAX_IP_PORT_FILTER_ITEM && cur_id >= 0 )
	{
		if( fw_acl_list[cur_id] != 0 )
			put_acl( fw_acl_list[cur_id]-1 );
		fw_acl_list[cur_id] = 0;
	}
	fw_move_backward(fw_acl_list,cur_id);
}

void fw_deledit( int chain, int noused_id, fw_acl *fw_acl_cur, int op, int opt )
{
	int *fw_acl_list=NULL;
	int cur_id = -1;
	
	if( chain == CHAIN_IN )
	{
		fw_acl_list = &g_ConfigParam.fw_acl_in[0];
	}
	else if( chain == CHAIN_OUT )
	{
		fw_acl_list = &g_ConfigParam.fw_acl_out[0]; 
	}
	else
	{
		fw_acl_list = &g_ConfigParam.fw_acl_me[0]; 
	}
	
	cur_id = fw_find_acl( chain, fw_acl_cur );
	
	if( cur_id < MAX_IP_PORT_FILTER_ITEM && cur_id >= 0 )
	{
		if( fw_acl_list[cur_id] != 0 )
			put_acl( fw_acl_list[cur_id]-1 );
		fw_acl_list[cur_id] = 0;

		fw_move_backward(fw_acl_list,cur_id);
	}
}

int fw_addedit( int chain, int cur_id, fw_acl *fw_acl_cur, int op, int opt )
{
	int *fw_acl_list=NULL;
	int candidate = 0;
	
	if( chain == CHAIN_IN )
	{
		fw_acl_list = &g_ConfigParam.fw_acl_in[0];
	}
	else if( chain == CHAIN_OUT )
	{
		fw_acl_list = &g_ConfigParam.fw_acl_out[0]; 
	}
	else
	{
		fw_acl_list = &g_ConfigParam.fw_acl_me[0]; 
	}
	
	if( cur_id >= MAX_IP_PORT_FILTER_ITEM )
		return -1;
			
	if( opt == RULE_TRY && op == AOS_FW_RULE_ADD && get_rule_count(chain) >= MAX_IP_PORT_FILTER_ITEM-1 )
		return -1;
	
	if( op == AOS_FW_RULE_ADD )
	{
		candidate = get_acl();
		if( candidate == -1 )
			return -1;
	}
	else
	{
		if( cur_id >= get_rule_count(chain) )
		{
			fprintf(stderr,"\nExceed acl pool boundary");
			return -1;
		}
		candidate = fw_acl_list[cur_id]-1;
		if( candidate < 0 )
		{
			fprintf(stderr,"\nInvalid index to acl pool");
			return -1;
		}
	}
	//fprintf(stderr,"\n2:dporttype=%d ",g_ConfigParam.fw_acl_cur.dporttype );
	
	g_ConfigParam.fw_acl_pool[candidate].id = candidate+1;
	g_ConfigParam.fw_acl_pool[candidate].chain = fw_acl_cur->chain;
	g_ConfigParam.fw_acl_pool[candidate].srctype = fw_acl_cur->srctype;
	g_ConfigParam.fw_acl_pool[candidate].dsttype = fw_acl_cur->dsttype;
	g_ConfigParam.fw_acl_pool[candidate].sporttype = fw_acl_cur->sporttype;
	g_ConfigParam.fw_acl_pool[candidate].dporttype = fw_acl_cur->dporttype;
	g_ConfigParam.fw_acl_pool[candidate].src = fw_acl_cur->src;
	g_ConfigParam.fw_acl_pool[candidate].dst = fw_acl_cur->dst;
	g_ConfigParam.fw_acl_pool[candidate].smsk = fw_acl_cur->smsk;
	g_ConfigParam.fw_acl_pool[candidate].dmsk = fw_acl_cur->dmsk;
	g_ConfigParam.fw_acl_pool[candidate].action = fw_acl_cur->action;
	g_ConfigParam.fw_acl_pool[candidate].log = fw_acl_cur->log;
	g_ConfigParam.fw_acl_pool[candidate].proto = fw_acl_cur->proto;
	
	strlcpy(g_ConfigParam.fw_acl_pool[candidate].route, fw_acl_cur->route, sizeof(g_ConfigParam.fw_acl_pool[candidate].route));
	strlcpy(g_ConfigParam.fw_acl_pool[candidate].dstdomain, fw_acl_cur->dstdomain, sizeof(g_ConfigParam.fw_acl_pool[candidate].dstdomain));

	if( g_ConfigParam.fw_acl_pool[candidate].proto == PROTO_TCP ||
		g_ConfigParam.fw_acl_pool[candidate].proto == PROTO_UDP )
	{
		g_ConfigParam.fw_acl_pool[candidate].sport[0] = fw_acl_cur->sport[0];
		g_ConfigParam.fw_acl_pool[candidate].sport[1] = fw_acl_cur->sport[1];
		g_ConfigParam.fw_acl_pool[candidate].dport[0] = fw_acl_cur->dport[0];
		g_ConfigParam.fw_acl_pool[candidate].dport[1] = fw_acl_cur->dport[1];
	}
	else
	{
		g_ConfigParam.fw_acl_pool[candidate].sport[0] = 0;
		g_ConfigParam.fw_acl_pool[candidate].sport[1] = 0;
		g_ConfigParam.fw_acl_pool[candidate].dport[0] = 0;
		g_ConfigParam.fw_acl_pool[candidate].dport[1] = 0;
	}

	if( g_ConfigParam.fw_acl_pool[candidate].proto == PROTO_ICMP )
	{
		g_ConfigParam.fw_acl_pool[candidate].dport[0] = fw_acl_cur->icmptype;
		g_ConfigParam.fw_acl_pool[candidate].dport[1] = fw_acl_cur->icmptype;
	}
	
	if( op == AOS_FW_RULE_ADD )
	{
		fw_move_forward( fw_acl_list, cur_id );
		fw_acl_list[cur_id] = candidate+1;
	}
	return 0;

}


static int set_fw_rule(fw_acl *fw_acl_list, struct aos_fw_acl *aos_fw_acl)
{
	struct in_addr addr;
	
	if( fw_acl_list->id > MAX_IP_PORT_FILTER_ITEM )
		return -1;

	if( fw_acl_list->chain >= CHAIN_LAST )
		return -2;

	strcpy(aos_fw_acl->chain, chainlist[fw_acl_list->chain]);
	aos_fw_acl->srctype = fw_acl_list->srctype;
	addr.s_addr = fw_acl_list->src;
	strcpy(aos_fw_acl->src1,(char *)inet_ntoa(addr));
	addr.s_addr = fw_acl_list->smsk;
	strcpy(aos_fw_acl->src2,(char *)inet_ntoa(addr));

	if( aos_fw_acl->srctype == AOS_FW_NET_SELF )
	{
		//strcpy(aos_fw_acl->src1,"0.255.0.255");
		//strcpy(aos_fw_acl->src2,"0.255.0.255");
	}
	aos_fw_acl->dsttype = fw_acl_list->dsttype;
	addr.s_addr = fw_acl_list->dst;
	strcpy(aos_fw_acl->dst1,(char *)inet_ntoa(addr));
	addr.s_addr = fw_acl_list->dmsk;
	strcpy(aos_fw_acl->dst2,(char *)inet_ntoa(addr));
	
	if( aos_fw_acl->dsttype == AOS_FW_NET_SELF )
	{
		//strcpy(aos_fw_acl->dst1,"0.255.0.255");
		//strcpy(aos_fw_acl->dst2,"0.255.0.255");
	}
	else if( aos_fw_acl->dsttype == AOS_FW_NET_DOMAIN )
	{
		strcpy(aos_fw_acl->dstdomain, fw_acl_list->dstdomain);
	}
	if( fw_acl_list->proto >= PROTO_LAST )
		return -3;

	strcpy(aos_fw_acl->proto,protolist[fw_acl_list->proto]);
	aos_fw_acl->icmptype = fw_acl_list->dport[0]; /* icmptype value is stored in dport !! */
	aos_fw_acl->sport1 = fw_acl_list->sport[0];
	aos_fw_acl->sport2 = fw_acl_list->sport[1];
	aos_fw_acl->dport1 = fw_acl_list->dport[0];
	aos_fw_acl->dport2 = fw_acl_list->dport[1];
	aos_fw_acl->action = fw_acl_list->action;
	aos_fw_acl->log    = fw_acl_list->log;
	
	if( fw_acl_list->chain == CHAIN_OUT )
	{
		if( strcmp(fw_acl_list->route,"AUTO")!=0 )
		{
			//TEST ONLY pppX interface ??
			strlcpy(aos_fw_acl->route, fw_acl_list->route,sizeof(aos_fw_acl->route));
		}
	}
	aos_fw_acl->op = AOS_FW_RULE_ADD;
	aos_fw_acl->objtype = AOS_FW_FILTER;

	return 0;
}

int policy_install()
{
	struct aos_fw_acl acl;
	
	/* flush rule first */
	acl.op = AOS_FW_RULE_POLICY;
	acl.objtype = AOS_FW_FILTER;
	acl.action = g_ConfigParam.fw_rule ? AOS_FW_DROP : AOS_FW_ACCEPT;
	strcpy(acl.chain,"FORWARD");	/* we use mangle table now, so it become useless now!!, 
					   please check /etc/fwinit.rule */
	aos_fw_rule( &acl );

	/* SELF ACCESS */
	strcpy(acl.chain,"INPUT");
	aos_fw_rule( &acl );

	/* OUTPUT chain should be allow by default, otherwise dhcp can't work */
	/* INPUT chain should allow port 80 */
	
	return 0;
}

void acl_rule_install(int chain)
{
	struct aos_fw_acl acl;
	int i, len;
	FILE *fp;
	int *fw_acl_list = NULL;
	char buf[128];
	
	if( chain >= CHAIN_LAST )
	{
		printf("invalid chain %d\n",chain);
		return;
	}
	/* flush filter rule first */
	acl.op = AOS_FW_RULE_FLUSH;
	acl.objtype = AOS_FW_FILTER;
	strcpy(acl.chain,chainlist[chain]); /* INBOUND/OUTBOUND/ME */
	aos_fw_rule( &acl );

	/* flush mangle rule first only if chain==OUTBOUND */
	if( chain == CHAIN_OUT )
	{
		acl.op = AOS_FW_RULE_FLUSH;
		acl.objtype = AOS_FW_MANGLE;
		strcpy(acl.chain,"OUTBOUND"); /* NOTE: policy route rules are stored in mangle table FORWARD chain */
		aos_fw_rule( &acl );
	
		iptables("ipt -A OUTBOUND -p ip -j ACCEPT");
	}
	
	if( chain == CHAIN_IN )
		fw_acl_list = &g_ConfigParam.fw_acl_in[0];
	else if( chain == CHAIN_OUT )
		fw_acl_list = &g_ConfigParam.fw_acl_out[0];
	else fw_acl_list = &g_ConfigParam.fw_acl_me[0];

	for (i=0; i<get_rule_count(chain); i++)
	{
		memset(&acl,0,sizeof(acl));
		if( set_fw_rule(&g_ConfigParam.fw_acl_pool[fw_acl_list[i]-1],&acl) == 0 )
		{
			acl.objtype = AOS_FW_FILTER;
			if( chain != CHAIN_OUT )
				aos_fw_rule( &acl );
			
			//set policy route only if /* action == ACCEPT &&*/ chain == OUT
			if( chain == CHAIN_OUT /* && acl.action == AOS_FW_ACCEPT  */
				/* && g_ConfigParam.fw_acl_pool[fw_acl_list[i]-1].route[0] 
				  because we move outbound rules to mangle table */ 
			)
			{
				acl.objtype = AOS_FW_MANGLE;
				strcpy(acl.chain, "OUTBOUND" ); 
				aos_fw_rule( &acl );
			}
		}
	}
	
	//policydomain('M',""); this will be called in aos_fw_rule();
	policydomain('C',""); //clean mark domain entries

	fp = fopen(DOMAINFILE,"wt");
	if( fp )
	{
	   int count=0;
	   
	   for( chain=0; chain< CHAIN_LAST; chain++)
	   {
		if( chain == CHAIN_IN )
			fw_acl_list = &g_ConfigParam.fw_acl_in[0];
		else if( chain == CHAIN_OUT )
			fw_acl_list = &g_ConfigParam.fw_acl_out[0];
		else fw_acl_list = &g_ConfigParam.fw_acl_me[0];

	   	count = get_rule_count(chain);
	   	for (i=0; i<count; i++)
		{
			//printf("\nk=%d i=%d dm=%s",k,i,g_ConfigParam.fw_acl_pool[fw_acl_list[i]-1].dstdomain[0]);
			if( g_ConfigParam.fw_acl_pool[fw_acl_list[i]-1].dstdomain[0] )
			{
				if( g_ConfigParam.fw_acl_pool[fw_acl_list[i]-1].route[0] )
					len = snprintf(buf,sizeof(buf),"%s\t%s\n", 
						g_ConfigParam.fw_acl_pool[fw_acl_list[i]-1].route,
						g_ConfigParam.fw_acl_pool[fw_acl_list[i]-1].dstdomain );
				else
				{
					len = snprintf(buf,sizeof(buf),"%s\t%s\n", 
						"AUTO",
						g_ConfigParam.fw_acl_pool[fw_acl_list[i]-1].dstdomain );
				}
				fwrite( buf, len, 1, fp );
			}
		}
	   }
	   fclose(fp);
	}
}

int WRWriteAclRuleArgument(HTTPCONNECTION hConnection, char *buf)
{
	int i=0,j=0,num=0;
	int *fw_acl_list = &g_ConfigParam.fw_acl_out[0];
	
	//WARNING: buf pointer may not large enough! overflow !
	if( g_ConfigParam.fw_acl_cur.chain == CHAIN_IN )
		fw_acl_list = &g_ConfigParam.fw_acl_in[0];
	else if( g_ConfigParam.fw_acl_cur.chain == CHAIN_OUT )
		fw_acl_list = &g_ConfigParam.fw_acl_out[0]; 
	else fw_acl_list = &g_ConfigParam.fw_acl_me[0]; 

	i+=sprintf(buf+i,"\nvar va_fw_enable='%d';", g_ConfigParam.fw_rule);
	i+=sprintf(buf+i,"\nvar va_max_dm='%d';", MAX_POLICY_DOMAIN);
	i+=sprintf(buf+i,"\nvar va_max_acl='%d';", MAX_IP_PORT_FILTER_ITEM);
	num = get_acl();
	i+=sprintf(buf+i,"\nvar va_cur_acl='%d';", num==-1 ? MAX_IP_PORT_FILTER_ITEM: num );

	/* chain should be placed before memset(..) if not OP_EDIT  */
	i+=sprintf(buf+i,"\nvar va_chain='%d';",
		g_ConfigParam.fw_acl_cur.chain);
	g_ConfigParam.fw_acl_cur.id = get_rule_count(g_ConfigParam.fw_acl_cur.chain)+1;
	i+=sprintf(buf+i,"\nvar va_id = '%d';",
		g_ConfigParam.fw_acl_cur.id);

	memset(&g_ConfigParam.fw_acl_cur,0,sizeof(g_ConfigParam.fw_acl_cur));

	i+=sprintf(buf+i,"\nvar va_op = '%d';",
		g_ConfigParam.fw_acl_cur.op);

	//////////////////
	// fw_acl_list 
	//////////////////
	
	//src
	i+=sprintf(buf+i,"\nvar va_siptypelist = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n'%d',",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].srctype);
	}
	i+=sprintf(buf+i,"\n0);");

	i+=sprintf(buf+i,"\nvar va_siplistbeg = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		struct in_addr addr;
		
		addr.s_addr = g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].src;
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n'%s',",inet_ntoa(addr));
	}
	i+=sprintf(buf+i,"\n'0.0.0.0');");

	i+=sprintf(buf+i,"\nvar va_siplistend = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		struct in_addr addr;
		
		addr.s_addr = g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].smsk;
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n'%s',",inet_ntoa(addr));
	}
	i+=sprintf(buf+i,"\n'0.0.0.0');");

	//dst	
	i+=sprintf(buf+i,"\nvar va_diptypelist = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n'%d',",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].dsttype);
	}
	i+=sprintf(buf+i,"\n0);");
	
	i+=sprintf(buf+i,"\nvar va_diplistbeg = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		struct in_addr addr;
		
		if( g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].dsttype == AOS_FW_NET_DOMAIN )
			i+=sprintf(buf+i,"\n'%s',",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].dstdomain);
		else
		{
			addr.s_addr = g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].dst;
			if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
				i+=sprintf(buf+i,"\n'%s',",inet_ntoa(addr));
		}
	}
	i+=sprintf(buf+i,"\n'0.0.0.0');");

	i+=sprintf(buf+i,"\nvar va_diplistend = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		struct in_addr addr;
		
		addr.s_addr = g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].dmsk;
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n'%s',",inet_ntoa(addr));
	}
	i+=sprintf(buf+i,"\n'0.0.0.0');");

	//sport
	i+=sprintf(buf+i,"\nvar va_sporttypelist = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n%d,",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].sporttype);
	}
	i+=sprintf(buf+i,"\n'0');");
	
	i+=sprintf(buf+i,"\nvar va_sportlistbeg = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n%d,",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].sport[0]);
	}
	i+=sprintf(buf+i,"\n0);");

	i+=sprintf(buf+i,"\nvar va_sportlistend = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n%d,",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].sport[1]);
	}
	i+=sprintf(buf+i,"\n0);");
	
	//dport
	i+=sprintf(buf+i,"\nvar va_dporttypelist = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n%d,",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].dporttype);
	}
	i+=sprintf(buf+i,"\n0);");

	i+=sprintf(buf+i,"\nvar va_dportlistbeg = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n%d,",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].dport[0]);
	}
	i+=sprintf(buf+i,"\n0);");

	i+=sprintf(buf+i,"\nvar va_dportlistend = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n%d,",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].dport[1]);
	}
	i+=sprintf(buf+i,"\n0);");

	i+=sprintf(buf+i,"\nvar va_actionlist = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n%d,",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].action);
	}
	i+=sprintf(buf+i,"\n0);");

	i+=sprintf(buf+i,"\nvar va_routelist = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
		{
			if( g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].route[0] )
				i+=sprintf(buf+i,"\n'%s',", g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].route );
			else i+=sprintf(buf+i,"\n'%s',", "AUTO" );
		}
	}
	i+=sprintf(buf+i,"\n'');");

	i+=sprintf(buf+i,"\nvar va_protolist = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n%d,",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].proto);
	}
	i+=sprintf(buf+i,"\n0);");

	i+=sprintf(buf+i,"\nvar va_loglist = new Array (");
	for(j=0; j<MAX_IP_PORT_FILTER_ITEM ; j++)
	{
		if( fw_acl_list[j] && g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].id )
			i+=sprintf(buf+i,"\n%d,",g_ConfigParam.fw_acl_pool[fw_acl_list[j]-1].log);
	}
	i+=sprintf(buf+i,"\n0);");
	
	return i;
}


int get_acl_in(HTTPCONNECTION hConnection, void *pParam)
{
	char *pcFile;

	if (WRExternalAdmin(hConnection) == 0) return 0;
	if ((pcFile = GetRequestFilePath(hConnection)) == NULL) return 0;

	g_ConfigParam.fw_acl_cur.chain = 0;
	SendHtmlWithParam(hConnection, pcFile, WRWriteAclRuleArgument);

	memset(&g_ConfigParam.fw_acl_cur,0,sizeof(g_ConfigParam.fw_acl_cur));

	return 0;
}
int get_acl_out(HTTPCONNECTION hConnection, void *pParam)
{
	char *pcFile;

	if (WRExternalAdmin(hConnection) == 0) return 0;
	if ((pcFile = GetRequestFilePath(hConnection)) == NULL) return 0;

	g_ConfigParam.fw_acl_cur.chain = 1;
	SendHtmlWithParam(hConnection, pcFile, WRWriteAclRuleArgument);

	memset(&g_ConfigParam.fw_acl_cur,0,sizeof(g_ConfigParam.fw_acl_cur));

	return 0;
}
int get_acl_me(HTTPCONNECTION hConnection, void *pParam)
{
	char *pcFile;

	if (WRExternalAdmin(hConnection) == 0) return 0;
	if ((pcFile = GetRequestFilePath(hConnection)) == NULL) return 0;

	g_ConfigParam.fw_acl_cur.chain = 2;
	SendHtmlWithParam(hConnection, pcFile, WRWriteAclRuleArgument);

	memset(&g_ConfigParam.fw_acl_cur,0,sizeof(g_ConfigParam.fw_acl_cur));

	return 0;
}

int get_acl_rule(HTTPCONNECTION hConnection, void *pParam)
{
	char *pcFile;
	char *pcPath;
	
	if (WRExternalAdmin(hConnection) == 0) return 0;
	if ((pcFile = GetRequestFilePath(hConnection)) == NULL) return 0;

	if ((pcPath = GetRequestPath(hConnection)) == NULL ) return 0;

	if( strstr(pcPath,"acl_in")==0 )
		g_ConfigParam.fw_acl_cur.chain = 0;
	else if( strstr(pcPath,"acl_out")==0 )
		g_ConfigParam.fw_acl_cur.chain = 1;
	else g_ConfigParam.fw_acl_cur.chain = 2; 
	
	SendHtmlWithParam(hConnection, pcFile, WRWriteAclRuleArgument);

	memset(&g_ConfigParam.fw_acl_cur,0,sizeof(g_ConfigParam.fw_acl_cur));

	return 0;
}

/* op == AOS_FW_RULE_ADD or AOS_FW_RULE_DEL */
void fw_rule(int op, int chain,  int action, int proto, 
	int stype, unsigned long src, unsigned long srcmask, 
	int sporttype, unsigned short sport0, unsigned short sport1,
	int dtype, unsigned long dst, unsigned long dstmask,
	int dporttype, unsigned short dport0, unsigned short dport1  )
{
	int cur_id = get_rule_count(chain);

	memset(&g_ConfigParam.fw_acl_cur, 0, sizeof(g_ConfigParam.fw_acl_cur));
	
	g_ConfigParam.fw_acl_cur.chain = chain;
	g_ConfigParam.fw_acl_cur.op = (op == AOS_FW_RULE_ADD) ? OP_ADD : OP_DEL;
	g_ConfigParam.fw_acl_cur.id = cur_id;
	g_ConfigParam.fw_acl_cur.log = 0;

	//src
	g_ConfigParam.fw_acl_cur.srctype = stype;
	g_ConfigParam.fw_acl_cur.src = src;
	g_ConfigParam.fw_acl_cur.smsk = srcmask;

	//dst
	g_ConfigParam.fw_acl_cur.dsttype = dtype;
	g_ConfigParam.fw_acl_cur.dst = dst;
	g_ConfigParam.fw_acl_cur.dmsk = dstmask;

	//sport
	g_ConfigParam.fw_acl_cur.sporttype = sporttype;
	g_ConfigParam.fw_acl_cur.sport[0] = sport0;
	g_ConfigParam.fw_acl_cur.sport[1] = sport1;

	//dport
	g_ConfigParam.fw_acl_cur.dporttype = dporttype;
	g_ConfigParam.fw_acl_cur.dport[0] = dport0;
	g_ConfigParam.fw_acl_cur.dport[1] = dport1;

	g_ConfigParam.fw_acl_cur.action = action;

	g_ConfigParam.fw_acl_cur.proto = proto;
	g_ConfigParam.fw_acl_cur.icmptype = 0;
	
	if( op == AOS_FW_RULE_ADD )
	{
		if( fw_find_acl( g_ConfigParam.fw_acl_cur.chain, &g_ConfigParam.fw_acl_cur ) == -1 )
			fw_addedit( g_ConfigParam.fw_acl_cur.chain, cur_id , &g_ConfigParam.fw_acl_cur, AOS_FW_RULE_ADD, RULE_TRY );
	}
	else if( op == AOS_FW_RULE_DEL )
	{
		fw_deledit( g_ConfigParam.fw_acl_cur.chain, cur_id , &g_ConfigParam.fw_acl_cur, AOS_FW_RULE_DEL, RULE_TRY );
	}
	WriteFlashMemory(&g_ConfigParam);
	//acl_rule_install(g_ConfigParam.fw_acl_cur.chain);
	
	memset(&g_ConfigParam.fw_acl_cur,0,sizeof(g_ConfigParam.fw_acl_cur));

}

int post_acl_rule(HTTPCONNECTION hConnection, void *pParam)
{
	LIST *pList;
	int cur_id = 0;
	int move_id = 0;
	const char *route = NULL, *domain = NULL;
	pList = (LIST *)pParam;

	/* Read page parameters. */
	g_ConfigParam.fw_acl_cur.chain = GetLong(pList,"va_chain");
	g_ConfigParam.fw_acl_cur.op = GetLong(pList,"va_op");
	g_ConfigParam.fw_acl_cur.id = GetLong(pList,"In_AclID");
	g_ConfigParam.fw_acl_cur.log = GetLong(pList,"In_Log");
	move_id = GetLong(pList,"In_MoveID") -1 ;
	cur_id = g_ConfigParam.fw_acl_cur.id -1 ;

	//src
	g_ConfigParam.fw_acl_cur.srctype = GetLong(pList,"In_SrcIpType");

	if( g_ConfigParam.fw_acl_cur.srctype == 0 ) //any
	{
		g_ConfigParam.fw_acl_cur.src = 0;
		g_ConfigParam.fw_acl_cur.smsk = 0;
	}
	else if( g_ConfigParam.fw_acl_cur.srctype == 3 ) //self
	{
		//g_ConfigParam.fw_acl_cur.src = 0xff00ff00;
		//g_ConfigParam.fw_acl_cur.smsk = 0xffffffff;	
	}
	else
	{
		g_ConfigParam.fw_acl_cur.src = GetIP(pList,"In_SrcIpBeg");
		g_ConfigParam.fw_acl_cur.smsk = GetIP(pList,"In_SrcIpEnd");	
	}
	//dst
	g_ConfigParam.fw_acl_cur.dsttype = GetLong(pList,"In_DestIpType");

	if( g_ConfigParam.fw_acl_cur.dsttype == 0 ) //any
	{
		g_ConfigParam.fw_acl_cur.dst = 0;
		g_ConfigParam.fw_acl_cur.dmsk = 0;
	}
	else if( g_ConfigParam.fw_acl_cur.dsttype == 3 ) //self
	{
		//g_ConfigParam.fw_acl_cur.dst = 0xff00ff00;
		//g_ConfigParam.fw_acl_cur.dmsk = 0xffffffff;	
	}
	else if( g_ConfigParam.fw_acl_cur.dsttype == 4 ) //domain
	{
		g_ConfigParam.fw_acl_cur.dst = 0x0;
		g_ConfigParam.fw_acl_cur.dmsk = 0x0;
		domain = GetString(pList, "In_DestIpBeg");
		if( domain == NULL || domain[0] == '\0' )
			g_ConfigParam.fw_acl_cur.dstdomain[0] = '*';
		else
			strncpy(g_ConfigParam.fw_acl_cur.dstdomain, domain, sizeof(g_ConfigParam.fw_acl_cur.dstdomain)-1);
		strlwr(g_ConfigParam.fw_acl_cur.dstdomain);
	}
	else
	{
		g_ConfigParam.fw_acl_cur.dst = GetIP(pList,"In_DestIpBeg");
		g_ConfigParam.fw_acl_cur.dmsk = GetIP(pList,"In_DestIpEnd");	
	}

	
	//sport
	g_ConfigParam.fw_acl_cur.sporttype = GetLong(pList,"In_SrcPortType");

	if( g_ConfigParam.fw_acl_cur.sporttype == 0 ) //any
	{
		g_ConfigParam.fw_acl_cur.sport[0] = 0;
		g_ConfigParam.fw_acl_cur.sport[1] = 0;
	}
	else
	{
		g_ConfigParam.fw_acl_cur.sport[0] = GetLong(pList,"In_SrcPortBegin");
		 if( g_ConfigParam.fw_acl_cur.sporttype == 1 ) //single
			g_ConfigParam.fw_acl_cur.sport[1] = g_ConfigParam.fw_acl_cur.sport[0];
		else
			g_ConfigParam.fw_acl_cur.sport[1] = GetLong(pList,"In_SrcPortEnd");
	}

	//dport
	g_ConfigParam.fw_acl_cur.dporttype = GetLong(pList,"In_DestPortType");
	//fprintf(stderr,"\ndporttype=%d ",g_ConfigParam.fw_acl_cur.dporttype );
	
	if( g_ConfigParam.fw_acl_cur.dporttype == 0 ) //any
	{
		g_ConfigParam.fw_acl_cur.dport[0] = 0;
		g_ConfigParam.fw_acl_cur.dport[1] = 0;
	}
	else
	{
		g_ConfigParam.fw_acl_cur.dport[0] = GetLong(pList,"In_DestPortBegin");
		 if( g_ConfigParam.fw_acl_cur.dporttype == 1 ) //single
			g_ConfigParam.fw_acl_cur.dport[1] = g_ConfigParam.fw_acl_cur.dport[0];
		else
			g_ConfigParam.fw_acl_cur.dport[1] = GetLong(pList,"In_DestPortEnd");
	}

	g_ConfigParam.fw_acl_cur.action = GetLong(pList, "In_Action");
	
	route = GetString(pList, "In_Route");
	if( route )
		strlcpy( g_ConfigParam.fw_acl_cur.route, route, sizeof(g_ConfigParam.fw_acl_cur.route));

	g_ConfigParam.fw_acl_cur.proto = GetLong(pList, "In_AclProtocol");
	g_ConfigParam.fw_acl_cur.icmptype = GetLong(pList, "In_IcmpType");
	
	//submit: edit,del,add,insert
	{
		if( g_ConfigParam.fw_acl_cur.op == OP_ADD )
		{
			{
				// possible exceed max, add fail but del ok!!
				fw_addedit( g_ConfigParam.fw_acl_cur.chain, move_id , &g_ConfigParam.fw_acl_cur, AOS_FW_RULE_ADD, RULE_TRY );
			
			}

#ifdef XX
			fw_addedit( g_ConfigParam.fw_acl_cur.chain, cur_id , &g_ConfigParam.fw_acl_cur, AOS_FW_RULE_ADD, RULE_TRY );

				// #jason_start
				fw_addedit( g_ConfigParam.fw_acl_cur.chain, move_id , &g_ConfigParam.fw_acl_cur, AOS_FW_RULE_ADD, RULE_FORCE );
			
				if( move_id < cur_id )
					cur_id++;
				fw_del( g_ConfigParam.fw_acl_cur.chain, cur_id );
				// #jason_end 
#endif
		}
		else if( g_ConfigParam.fw_acl_cur.op == OP_EDIT )
		{
			//fprintf(stderr,"\ncurid=%d moveid=%d dporttype=%d", cur_id, move_id ,g_ConfigParam.fw_acl_cur.dporttype);
			if( cur_id == move_id )
			{
				fprintf(stderr,"\nmodified rule");
				fw_addedit( g_ConfigParam.fw_acl_cur.chain, cur_id , &g_ConfigParam.fw_acl_cur, AOS_FW_RULE_EDIT, RULE_TRY );
			}
			else 
			{
				fprintf(stderr,"\nmodified rule & moved ");
				// possible exceed max, add fail but del ok!!
				fw_addedit( g_ConfigParam.fw_acl_cur.chain, move_id , &g_ConfigParam.fw_acl_cur, AOS_FW_RULE_ADD, RULE_FORCE );
			
				if( move_id < cur_id )
					cur_id++;
				fw_del( g_ConfigParam.fw_acl_cur.chain, cur_id );
			}
		}
		else if( g_ConfigParam.fw_acl_cur.op == OP_DEL )
		{
			fw_del( g_ConfigParam.fw_acl_cur.chain, cur_id );
		}
		
		WriteFlashMemory(&g_ConfigParam);
		acl_rule_install(g_ConfigParam.fw_acl_cur.chain);
		//WebLog(CL_PRIVILEGE_ADMIN, CL_SET_IPFILTER, NULL, hConnection);
		nat_ip_filter_reset( NULL,0 );
		
		memset(&g_ConfigParam.fw_acl_cur,0,sizeof(g_ConfigParam.fw_acl_cur));
	}
	//request: edit,add,insert
	
	SetKeepAliveMode(hConnection, FALSE);
	SendReDirectPage(hConnection);

	return 0;
}


void acl_init()
{
	FILE *fp;
	char buf[128];
	
	if( (fp=fopen(FW_INIT_RULE,"rt")) == NULL)
		return;
	
	memset(buf,0,sizeof(buf));
	while( fgets(buf,sizeof(buf),fp) )
	{
		if( strncmp(buf,"ipt ",4) != 0 )
			continue;
		iptables(buf); //init
		memset(buf,0,sizeof(buf));
	}
	//initialized from dos_install()
	//fwstate_check(g_ConfigParam.fw_rule);
	fclose(fp);

}

void acl_install()
{
	acl_rule_install(CHAIN_IN);
	acl_rule_install(CHAIN_OUT);
	acl_rule_install(CHAIN_ME);
}

void fwrule_default()
{

	/* ALLOW IP FROM ANY TO ANY OUTBOUND */
	fw_rule(AOS_FW_RULE_ADD, CHAIN_OUT, AOS_FW_ACCEPT, PROTO_IP,
		AOS_FW_NET_ANY, 0, 0,
		AOS_FW_PORT_ANY, 0, 0,
		AOS_FW_NET_ANY, 0, 0,
		AOS_FW_PORT_ANY, 0, 0 );
#ifdef XX
		
	/* ALLOW UDP PORT 53 FROM LAN TO ME */
	fw_rule(AOS_FW_RULE_ADD, CHAIN_ME, AOS_FW_ACCEPT, PROTO_UDP,
		AOS_FW_NET_ANY, 0, 0,
		AOS_FW_PORT_ANY, 0, 0,
		AOS_FW_NET_SELF, 0, 0,
		AOS_FW_PORT_SINGLE, 53, 53 );

	/* ALLOW UDP PORT 67 FROM LAN TO ME */
	fw_rule(AOS_FW_RULE_ADD, CHAIN_ME, AOS_FW_ACCEPT, PROTO_UDP,
		AOS_FW_NET_ANY, 0, 0,
		AOS_FW_PORT_ANY, 0, 0,
		AOS_FW_NET_SELF, 0, 0,
		AOS_FW_PORT_SINGLE, 67, 67 );

	/* ALLOW TCP PORT 80 FROM LAN TO ME */
	fw_rule(AOS_FW_RULE_ADD, CHAIN_ME, AOS_FW_ACCEPT, PROTO_TCP,
		AOS_FW_NET_ANY, 0, 0,
		AOS_FW_PORT_ANY, 0, 0,
		AOS_FW_NET_SELF, 0, 0,
		AOS_FW_PORT_SINGLE, 80, 80 );
#endif

	// should call acl_rule_install() later !!!! 
	g_ConfigParam.bEnableDnsProxy = 1;
	g_ConfigParam.bEnablePingLan = 1;
	g_ConfigParam.bEnablePingWan = 0;
	g_ConfigParam.bEnableAdminLan = 1;
	g_ConfigParam.bEnableAdminWan = 0;
}


int acl_service( int enable, unsigned port1 )
{
	char cmd[132];
	
	snprintf(cmd,sizeof(cmd),"ipt -D SERVICE -p tcp -m tcp -d 0.255.0.255 --dport %u -j DROP", port1 );
	iptables(cmd);

	if( enable )
	{
		snprintf(cmd,sizeof(cmd),"ipt -A SERVICE -p tcp -m tcp -d 0.255.0.255 --dport %u -j DROP", port1 );
		iptables(cmd);
	}
	return TRUE;
}
