#include "CommonDef.h"

int last_dhcp_enable = -1;
extern unsigned long old_lan_ip;

BOOL WRRenewDHCPClient()
{
	SynProcess(&g_pidDhcpClient);
	if (g_pidDhcpClient != 0)
	{
		kill((pid_t)g_pidDhcpClient, SIGHUP);
		return TRUE;
	}
	else
		return RunProgram(&g_pidDhcpClient, "/bin/dhcpc", "-h", getHostname(), NULL);
}

BOOL WRRunDHCPClient()
{
	SynProcess(&g_pidDhcpClient);
	if (g_pidDhcpClient != 0) return TRUE;
	return RunProgram(&g_pidDhcpClient, "/bin/dhcpc", "-h", getHostname(), NULL);
}

BOOL WRStopDHCPClient()
{
	SynProcess(&g_pidDhcpClient);
	if (g_pidDhcpClient == 0) return TRUE;
	return SafeLimitKill(g_pidDhcpClient, SIGTERM, 5000);
}

static pid_t WRGetDhcpPID(void)
{
	pid_t* pidList;
	pid_t pidRt;

	pidList = FindPidByName("dhcpd");

	if (!pidList)
	{
		fprintf(stderr, "No process named \"dhcpd\"!\n");
		return 0;
	}

	pidRt = *pidList;
	if (*(pidList+1) != 0)
	{/* to many processes named "dhcpd", can not know who is the correct one. */
		free(pidList);
		fprintf(stderr, "Too many processes named \"dhcpd\"!\n");
		return 0;
	}
	free(pidList);

	return pidRt;
}

#ifdef WINBOND_DHCP

static BOOL WRReadConfig_Dhcp(struct d_config *init_conf)
{
	FILE *in;
	char buffer[80],*token,*line,*val,*opt;

	if (!(in = fopen(DHCP_CONFIG_FILE, "r")))
	{
		if (!(in = fopen(DHCP_INIT_CONFIG_FILE, "r")))
			return FALSE;
	}

	memset(buffer,0,80);
	while (fgets(buffer, 80, in))
	{
		if (strchr(buffer, '\n')) *(strchr(buffer, '\n')) = '\0';
		if (strchr(buffer, '#')) *(strchr(buffer, '#')) = '\0';

		token = buffer + strspn(buffer, " \t");
		if (*token == '\0') continue;

		line = token + strcspn(token, " \t=");
		if (*line == '\0') continue;

		*line = '\0';
		line++;
		line = line + strspn(line, " \t=");
		if (*line == '\0') continue;

		if(!strcasecmp(token,"start"))
			strncpy(init_conf->start, line, 48);
		if(!strcasecmp(token,"end"))
			strncpy(init_conf->end, line, 48);
		if((!strcasecmp(token,"opt"))||(!strcasecmp(token,"option")))
		{
			if(!(opt=strtok(line, " \t=")))
			{
				fclose(in);
				return FALSE;
			}

			val=strtok(NULL, ", \t");
			if(val)
			{
				if(!strcasecmp(opt,"subnet"))
					strncpy(init_conf->subnet, val, 48);
				if(!strcasecmp(opt,"dns"))
					strncpy(init_conf->dns, val, 48);
				if(!strcasecmp(opt,"router"))
					strncpy(init_conf->router, val, 48);
				if(!strcasecmp(opt,"wins"))
					strncpy(init_conf->wins, val, 48);
				if(!strcasecmp(opt,"lease"))
					strncpy(init_conf->lease, val, 48);
			}
		}
		memset(buffer,0,80);
	}

	fclose(in);
	return TRUE;
}
#endif


BOOL WRIsEnableDHCPServer( void )
{
	FILE *fp;

	fp = fopen(DHCP_STATUS_FILE, "r");
	if (fp == NULL)
	{
		fprintf(stderr, "Can not open DHCP_STATUS_FILE:%s.\n", DHCP_STATUS_FILE);
		return FALSE;
	}
	else
	{
		int iStatus;
		fscanf(fp, "%d", &iStatus);
		fclose(fp);
		return (iStatus != 0);
	}
}

BOOL ToggleDhcpLan( int enable )
{
	char cmd[132];
	char action='I';
	
	action = enable ? 'I':'D';

	if( last_dhcp_enable != g_ConfigParam.dhcp.enable )
	{
		snprintf(cmd,sizeof(cmd),"ipt -%c SERVICE -i eth0 -p udp -m udp --dport %u -j ACCEPT", action, 67 );
		iptables(cmd);
	}
	
	/* init last_dhcp_enable in main.c() */
	last_dhcp_enable = g_ConfigParam.dhcp.enable;
	
	return TRUE;
}

BOOL WREnableDHCPServer(BOOL bIsEnable)
{
	pid_t pDhcpID;

	pDhcpID = WRGetDhcpPID();
	if( pDhcpID )
		kill((pid_t)pDhcpID, SIGINT);

	if ( /* (pDhcpID == (pid_t)0) && */ bIsEnable )
	{
		struct stat st;
		int i;
		unlink(DHCP_STATUS_FILE);
		RunProgram(NULL, "/bin/dhcpd", NULL);

		/* ͬdhcpd */
		for (i = 0; i < 50; i++)
		{
			usleep(100 * 1000);
			if (stat(DHCP_STATUS_FILE, &st) == 0) break;
		}
		if (!(i < 50))
			fprintf(stderr, "dhcpd can not run to normal state in 5 sec!\n");

	}
	ToggleDhcpLan( bIsEnable );
	
	return TRUE;
}

#ifdef WINBOND_DHCP

BOOL WRGetDHCPInfo(unsigned long *pulDHCPAddressFrom, unsigned long *pulDHCPAddressTo, 
	unsigned long *pulLeasedTime_InSec, unsigned long *pulRemainTime_InSec,
	unsigned long *pulDHCPDns1, unsigned long *pulDHCPDns2 )
{
	struct d_config dDhcpParam;
	FILE *fp;

	memset(&dDhcpParam, 0, sizeof(dDhcpParam));
	if (WRReadConfig_Dhcp(&dDhcpParam))
	{
		if (pulDHCPAddressFrom != NULL)
			*pulDHCPAddressFrom = String2IP(dDhcpParam.start);
		if (pulDHCPAddressTo != NULL)
			*pulDHCPAddressTo = String2IP(dDhcpParam.end);
		if (pulLeasedTime_InSec != NULL)
			*pulLeasedTime_InSec = String2Long(dDhcpParam.lease);
		if (pulRemainTime_InSec != NULL)
			fprintf(stderr, "Dhcp RemainTime not implemented.\n");
		if (pulDHCPDns1 != NULL)
			*pulDHCPDns1 = String2IP(dDhcpParam.dns);
		if (pulDHCPDns2 != NULL)
			*pulDHCPDns2 = String2IP(dDhcpParam.dns);
		if (pulDHCPWins != NULL)
			*pulDHCPDns2 = String2IP(dDhcpParam.dns);
		return TRUE;
	}
	else
	{
		fprintf(stderr, "Cannot read dhcp config before!\n");
		return FALSE;
	}
}
#endif

int  dhcpd_conf_check()
{
	if( g_ConfigParam.dhcp.startip == 0 )
		g_ConfigParam.dhcp.startip = g_ConfigParam.ulPrivateIPAddress+1;

	if(  g_ConfigParam.dhcp.endip == 0 )
		g_ConfigParam.dhcp.endip = g_ConfigParam.ulPrivateIPAddress | 0xfe;
		
	if( g_ConfigParam.dhcp.dns == 0 )
		g_ConfigParam.dhcp.dns = g_ConfigParam.ulPrivateIPAddress;

	/*
	if( g_ConfigParam.dhcp.wins == 0 )
		g_ConfigParam.dhcp.wins = g_ConfigParam.ulPrivateIPAddress;
	*/
	
	if( g_ConfigParam.dhcp.defgw == 0 )
		g_ConfigParam.dhcp.defgw = g_ConfigParam.ulPrivateIPAddress;
		
	if(  g_ConfigParam.dhcp.subnet == 0 )
		 g_ConfigParam.dhcp.subnet = 0x00ffffff;

	if(  g_ConfigParam.dhcp.leasetime < MIN_LEASETIME )
		g_ConfigParam.dhcp.leasetime = MIN_LEASETIME ;

	if(  g_ConfigParam.dhcp.leasetime > MAX_LEASETIME )
		g_ConfigParam.dhcp.leasetime = MAX_LEASETIME ;
		
	return 0;
}

BOOL WriteDHCPConf()
{
	FILE *fp;
	struct in_addr addr;
	
	if (!(fp = fopen(DHCP_CONFIG_FILE, "wt")))
	{
		fprintf(stderr, "Cannot open dhcp config file: %s\n", DHCP_CONFIG_FILE);
		return FALSE;
	}
	
	addr.s_addr = g_ConfigParam.dhcp.startip;
	fprintf(fp, "start	%s\n", inet_ntoa(addr));
	addr.s_addr = g_ConfigParam.dhcp.endip;
	fprintf(fp, "end	%s\n", inet_ntoa(addr));
	addr.s_addr = g_ConfigParam.dhcp.dns;
	fprintf(fp, "opt	dns	%s ", inet_ntoa(addr));
	if( g_ConfigParam.dhcp.dns1 )
	{
		addr.s_addr = g_ConfigParam.dhcp.dns1;
		fprintf(fp, " %s\n", inet_ntoa(addr));
	}
	else fprintf(fp,"\n");
	
	addr.s_addr = g_ConfigParam.dhcp.subnet;
	fprintf(fp, "opt	subnet	%s\n", inet_ntoa(addr));
	addr.s_addr = g_ConfigParam.dhcp.defgw;
	fprintf(fp, "opt	router	%s\n", inet_ntoa(addr));
	if( g_ConfigParam.dhcp.wins )
	{
		addr.s_addr = g_ConfigParam.dhcp.wins;
		fprintf(fp, "opt	wins	%s", inet_ntoa(addr));
	}
	if( g_ConfigParam.dhcp.wins1 )
	{
		addr.s_addr = g_ConfigParam.dhcp.wins1;
		fprintf(fp, " %s\n",inet_ntoa(addr));
	}
	else fprintf(fp,"\n");
	
	fprintf(fp, "opt	lease	%lu\n", g_ConfigParam.dhcp.leasetime);
	fclose(fp);

	return FALSE;
}

/* use 255.255.255.0 for largest netmask */

void dhcpd_update(unsigned long old_lan, unsigned long new_lan, unsigned long new_netmask)
{
	unsigned long startip, endip, netmask = new_netmask;
	
	if( ntohl(new_netmask) < 0xffffff00 )
		netmask = 0x00ffffff;

	startip = g_ConfigParam.dhcp.startip & ~(netmask);
	endip = g_ConfigParam.dhcp.endip & ~(netmask);
	
	g_ConfigParam.dhcp.startip = new_lan & netmask;
	g_ConfigParam.dhcp.startip = g_ConfigParam.dhcp.startip | startip;
	
	g_ConfigParam.dhcp.endip = new_lan & netmask;
	g_ConfigParam.dhcp.endip = g_ConfigParam.dhcp.endip | endip;
	
	if( g_ConfigParam.dhcp.dns == old_lan )
		g_ConfigParam.dhcp.dns = new_lan;

	if( g_ConfigParam.dhcp.dns1 == old_lan )
		g_ConfigParam.dhcp.dns1 = new_lan;
	
	if( g_ConfigParam.dhcp.wins == old_lan )
		g_ConfigParam.dhcp.wins = new_lan;

	if( g_ConfigParam.dhcp.wins1 == old_lan )
		g_ConfigParam.dhcp.wins1 = new_lan;
		
	if( g_ConfigParam.dhcp.defgw == old_lan )
		g_ConfigParam.dhcp.defgw = new_lan;
		
	g_ConfigParam.dhcp.subnet = netmask;

}

void dhcpd_install()
{
	g_ConfigParam.dhcp.subnet = GetNetMask(G_PC_ETH0);
	dhcpd_conf_check();
	WriteDHCPConf();
	WREnableDHCPServer( g_ConfigParam.dhcp.enable );
}
	

int get_dhcp(HTTPCONNECTION hConnection, void *pParam)
{
	char *pcFile;

	if (WRExternalAdmin(hConnection) == 0) return 0;
	if ((pcFile = GetRequestFilePath(hConnection)) == NULL) return 0;

	dhcpd_conf_check();

	SendHtmlWithParam(hConnection, pcFile,
		g_ConfigParam.ulPrivateSubnetMask,
		g_ConfigParam.dhcp.enable,
		g_ConfigParam.dhcp.startip,
		g_ConfigParam.dhcp.endip,
		g_ConfigParam.dhcp.leasetime,
		g_ConfigParam.dhcp.dns,
		g_ConfigParam.dhcp.dns1,
		g_ConfigParam.dhcp.wins,
		g_ConfigParam.dhcp.wins1,
		g_ConfigParam.dhcp.defgw );

	return 0;
}

int post_dhcp(HTTPCONNECTION hConnection, void *pParam)
{
	LIST *pList;
	pList = (LIST *)pParam;
	
	/* Read page parameters. */
	g_ConfigParam.dhcp.enable = GetLong(pList,"va_enable");
	g_ConfigParam.dhcp.startip = GetIP(pList, "va_startip");
	g_ConfigParam.dhcp.endip  = GetIP(pList, "va_endip");
	g_ConfigParam.dhcp.defgw = GetIP(pList, "va_defgw");
	g_ConfigParam.dhcp.dns = GetIP(pList, "va_dns");
	g_ConfigParam.dhcp.dns1 = GetIP(pList, "va_dns1");
	g_ConfigParam.dhcp.wins = GetIP(pList, "va_wins");
	g_ConfigParam.dhcp.wins1 = GetIP(pList, "va_wins1");
	g_ConfigParam.dhcp.leasetime = GetLong(pList, "va_leasetime");
	

	SetKeepAliveMode(hConnection, FALSE);
	
	if( g_ConfigParam.dhcp.defgw != g_ConfigParam.ulPrivateIPAddress )
	{
		FOR_CHANGE_IP_T *pIP;
		
		pIP = (FOR_CHANGE_IP_T *)malloc(sizeof(FOR_CHANGE_IP_T));
		if (pIP != NULL)
		{
			old_lan_ip = g_ConfigParam.ulPrivateIPAddress;
			
			pIP->ulIP = g_ConfigParam.dhcp.defgw;
			pIP->ulNetmask = g_ConfigParam.ulPrivateSubnetMask;
			pIP->updateDhcp = 1;
		
			SetSendDataOverFun(hConnection, WRToBeChangeLANIP, pIP);
			g_ConfigParam.ulPrivateIPAddress = g_ConfigParam.dhcp.defgw;
			
			WRConfigVserver(g_ConfigParam.avVirtualHostItem, TRUE); //reinstall virtual server rule
		}

		char acUrl[64];
		strcpy(acUrl, "http://");
		
		IP2String(pIP->ulIP, acUrl + strlen(acUrl));
		SendReDirectHtml(hConnection, "message.htm", acUrl, FALSE, 0, "MSG_RECONNECT_PROMPT", NULL );		
		//SendReDirectHtml(hConnection, "Redirect.htm", acUrl, TRUE, 5000, "MSG_CHANGE_IP_ADDR", NULL );
	}
	else
	{
		dhcpd_install();
		SendReDirectPage(hConnection);
	}
	
	WriteFlashMemory(&g_ConfigParam);

	
	return 0;
}
