
#include "CommonDef.h"

char *con_mode_str[4]={ "Static", "DHCP", "PPPoE", "PPPoE Unnumbered" };
int sysctlRead(  char *mib, unsigned long *value);

#define SYSCTL_IP_SESS_TCP		"/proc/sys/net/ipv4/ip_conntrack_count_tcp"
#define SYSCTL_IP_SESS_UDP		"/proc/sys/net/ipv4/ip_conntrack_count_udp"
#define SYSCTL_IP_SESS_ICMP	"/proc/sys/net/ipv4/ip_conntrack_count_icmp"

int sysctlReadString( const char *mib,  char * value, int len)
{
	FILE* fp;

	fp = fopen(mib,"r");
	
	if(fp==NULL)
	{
		fprintf(stderr,"\nsysctl open error %s", mib );
		return -1;
	}
	
	fgets(value, len, fp);
	len = strlen(value);
	if( value[len] == '\r' || value[len]=='\n' )
		value[len] = 0;
	fclose(fp);
	
	return 0;
}

unsigned long getUptime(void)
{
	char value[16];
	unsigned long sec=0, msec=0;
	
	memset(value,0,sizeof(value));
	if( sysctlReadString( "/proc/uptime", value, sizeof(value) ) == 0 )
	{
		sscanf( value, "%lu.%lu", &sec, &msec);
	}
	return sec;
}

static int WRWriteInterfaceArgument(HTTPCONNECTION hConnection, char *pcBuf)
{
	int i,iCount=0;
	char ac[6];
	char aacLanMac[6][4];
	char aacWanMac[6][4];
	time_t t;
	char con_mode[24];
	char buf[48];
	struct ppp_info pppd_info;

	char uptime [64];
	unsigned long totalsec = getUptime();
	unsigned int days,hr,min,sec;	
	days = totalsec / (86400);
	hr = (totalsec / 3600)%24;
	min = (totalsec / 60)%60;
	sec = totalsec % 60;

    	sprintf(uptime, "%02u days %02u:%02u:%02u" ,days, hr, min, sec);

	if (GetMacAddress(G_PC_ETH0, ac))
		for (i=0; i<6; i++)
			Char2Hex(ac[i], aacLanMac[i]);
	else
		for (i=0; i<6; i++)
			strcpy(aacLanMac[i], "??");

	if (GetMacAddress(G_PC_ETH1, ac))
		for (i=0; i<6; i++)
			Char2Hex(ac[i], aacWanMac[i]);
	else
		for (i=0; i<6; i++)
			strcpy(aacWanMac[i], "??");

	time(&t);
	strcpy(con_mode, con_mode_str[g_ConfigParam.ulHowToGetIp] );

	g_ConfigParam.ulPrivateIPAddress = GetIPAddress(G_PC_ETH0);
	g_ConfigParam.ulPrivateSubnetMask = GetNetMask(G_PC_ETH0);
	GetPubIPInfo(&g_ConfigParam.ulPublicIpAddress,
		&g_ConfigParam.ulPublicSubnetMask,
		&g_ConfigParam.ulDefaultGateway );
	
	snprintf(buf,sizeof(buf),"%s %s",WRGetProductName(),WRGetProductVersion());

	iCount += sprintf(pcBuf+iCount, "\nvar va_firm = '%s';", buf);
	sprintf(buf,"%s",ctime(&t));
	buf[strlen(buf)-1] = 0;
	iCount += sprintf(pcBuf+iCount, "\nvar va_datetime = '%s';",buf);
	iCount += sprintf(pcBuf+iCount, "\nvar va_hostname = '%s';", g_ConfigParam.acHostName);
	iCount += sprintf(pcBuf+iCount, "\nvar va_domain = '%s';", g_ConfigParam.acHostName);
	iCount += sprintf(pcBuf+iCount, "\nvar va_lanip = '%s';", int_ntoa(g_ConfigParam.ulPrivateIPAddress));
	iCount += sprintf(pcBuf+iCount, "\nvar va_lanmask = '%s';",int_ntoa(g_ConfigParam.ulPrivateSubnetMask));
	iCount += sprintf(pcBuf+iCount, "\nvar va_conmode = '%s';", con_mode_str[g_ConfigParam.ulHowToGetIp]);
	iCount += sprintf(pcBuf+iCount, "\nvar va_wanip = '%s';", int_ntoa(g_ConfigParam.ulPublicIpAddress));
	iCount += sprintf(pcBuf+iCount, "\nvar va_wanmask = '%s';", int_ntoa(g_ConfigParam.ulPublicSubnetMask));
	iCount += sprintf(pcBuf+iCount, "\nvar va_uptime = '%s';", uptime);
	iCount += sprintf(pcBuf+iCount, "\nvar va_defgw = '%s';",int_ntoa(g_ConfigParam.ulDefaultGateway));
	iCount += sprintf(pcBuf+iCount, "\nvar va_sess_tcp = '%d';", sysctlRead( SYSCTL_IP_SESS_TCP, NULL ) );
	iCount += sprintf(pcBuf+iCount, "\nvar va_sess_udp = '%d';", sysctlRead( SYSCTL_IP_SESS_UDP , NULL) );
	iCount += sprintf(pcBuf+iCount, "\nvar va_sess_icmp = '%d';", sysctlRead( SYSCTL_IP_SESS_ICMP, NULL ) );
	
	sprintf(buf,"%s-%s-%s-%s-%s-%s", aacLanMac[0], aacLanMac[1], aacLanMac[2], aacLanMac[3], aacLanMac[4], aacLanMac[5] );
	iCount += sprintf(pcBuf+iCount, "\nvar va_lanmac = '%s';",buf);
	sprintf(buf,"%s-%s-%s-%s-%s-%s", aacWanMac[0], aacWanMac[1], aacWanMac[2], aacWanMac[3], aacWanMac[4], aacWanMac[5] );
	iCount += sprintf(pcBuf+iCount, "\nvar va_wanmac = '%s';",buf);

	iCount += sprintf(pcBuf+iCount, "\nvar va_dns = new Array(\n");
	for( i=0; i<3; i++)
		iCount += sprintf(pcBuf+iCount,"\t'%s',\n", int_ntoa(g_ConfigParam.aulDNSServer[i]));
	iCount += sprintf(pcBuf+iCount,"\t'');");

	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoeip = new Array(\n");
	for( i=0; i<2; i++)
	{
		memset( &pppd_info,0,sizeof(struct ppp_info));
		if( !ppp_read_config(i+1, &pppd_info) || (pppd_info.status & PPPOE_CONN_BIT)!=CONNECTION_ON )
			memset( &pppd_info, 0,sizeof(struct ppp_info));
		
		iCount += sprintf(pcBuf+iCount,"\t'%s',\n", g_ConfigParam.pppoeConf[i+1].enable ? pppd_info.got_loc_addr : "Disabled" );
	}
	iCount += sprintf(pcBuf+iCount,"\t'');");

	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoemask = new Array(\n");
	for( i=0; i<2; i++)
	{
		memset( &pppd_info,0,sizeof(struct ppp_info));
		if( !ppp_read_config(i+1, &pppd_info) || (pppd_info.status & PPPOE_CONN_BIT)!=CONNECTION_ON )   
			memset( &pppd_info, 0,sizeof(struct ppp_info));
		iCount += sprintf(pcBuf+iCount,"\t'%s',\n", g_ConfigParam.pppoeConf[i+1].enable ? pppd_info.got_netmask : "Disabled" );
	}
	iCount += sprintf(pcBuf+iCount,"\t'');");
	
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoegw = new Array(\n");
	for( i=0; i<2; i++)
	{
		memset( &pppd_info,0,sizeof(struct ppp_info));
		if( !ppp_read_config(i+1, &pppd_info) || (pppd_info.status & PPPOE_CONN_BIT)!=CONNECTION_ON )   
			memset( &pppd_info, 0,sizeof(struct ppp_info));
		iCount += sprintf(pcBuf+iCount,"\t'%s',\n", g_ConfigParam.pppoeConf[i+1].enable ? pppd_info.got_gateway : "Disabled" );
	}
	iCount += sprintf(pcBuf+iCount,"\t'');");
	
	
	return iCount;
}

int get_system_info(HTTPCONNECTION hConnection, void *pParam)
{
	char *pcFile;
	
	if (WRExternalAdmin(hConnection) == 0) return 0;
	if ((pcFile = GetRequestFilePath(hConnection)) == NULL) return 0;

	
	SendHtmlWithParam(hConnection, pcFile, WRWriteInterfaceArgument);
	return 0;
}

