
#include "CommonDef.h"
#include "libnat.h"

enum wan_action { ACTION_SET=0, ACTION_CONNECT, ACTION_DISCONNECT, ACTION_RELOAD };
int g_iAction = 0;
int g_tmpConnMode = 0;
int g_curPPPoEId = 0;
unsigned long old_lan_ip = 0;

/*
void strlcpy( char *dst, const char *src, unsigned long len)
{
	int i;
	
	for(i=0; i<len; i++)
		*(dst+i) = *(src+i);
	*(dst+i)=0;
}
*/

const char *getHostname(void)
{
	if( g_ConfigParam.acHostName[0] == '\0' )
		return "HOSTNAME";
	return g_ConfigParam.acHostName;
}

unsigned short getMinMaxMtu(int HowToGetIp , int type )
{
	if(  HowToGetIp == GETIP_DHCP || HowToGetIp == GETIP_MAN )
	{
		if( type == MTU_MAX )
			return MAX_STATIC_MTU;
		else
			return MIN_STATIC_MTU;
	}
	else
	{
		if( type == MTU_MAX )
			return MAX_PPPOE_MTU;
		else
			return MIN_PPPOE_MTU;
	}
}

unsigned short getSysMtu(int HowToGetIp, unsigned short mtu)
{
	int max_mtu = MAX_STATIC_MTU;
	int min_mtu = MIN_STATIC_MTU;
	int def_mtu = MAX_STATIC_MTU;
	
	if(  HowToGetIp == GETIP_DHCP || HowToGetIp == GETIP_MAN )
	{
		max_mtu = MAX_STATIC_MTU;
		def_mtu = MAX_STATIC_MTU;
		min_mtu = MIN_STATIC_MTU;
	}
	else
	{
		max_mtu = MAX_PPPOE_MTU;
		def_mtu = DEF_PPPOE_MTU;
		min_mtu = MIN_PPPOE_MTU;
	}
	
	if( mtu == 0 )
		mtu = def_mtu;
	else if( mtu > max_mtu )
		mtu = max_mtu;
	else if( mtu < min_mtu )
		mtu = min_mtu;
	
	return mtu;
	
}

int wipeout(char *name)
{
	FILE *fp;
	
	if( (fp=fopen(name,"wt"))==NULL )
		return -1;
	fclose(fp);
	return 0;
}

/* host order */
const char *int_ntoa(unsigned long ipaddr)
{
	static char ip[18];
	
	sprintf(&ip[0],"%lu.%lu.%lu.%lu", (ipaddr&0xff),
			(ipaddr>>8)&0xff,(ipaddr>>16)&0xff,ipaddr>>24);
	//ipaddr>>24, (ipaddr>>16)&0xff, (ipaddr>>8)&0xff, (ipaddr&0xff));
	
	return ip;
}

int get_pppoe_info( int id, char *pcPPPoEConnectionStatus,
		long *lPPPoESessionID,
		long *lPPPoEConnectingTime_InSec )
{
	struct ppp_info pInfo;
	char *pcPPPoEIp;

	memset(&pInfo,0,sizeof(pInfo));
	
	if (!ppp_read_config( id, &pInfo)
		|| !(pInfo.status & PPPOE_RUN_BIT))
	{
		strcpy(pcPPPoEConnectionStatus ,"OFF" );
		*lPPPoESessionID = 0;
		*lPPPoEConnectingTime_InSec = 0;
		pcPPPoEIp = "";
	}
	else
	{
		*lPPPoESessionID = 0;
		*lPPPoEConnectingTime_InSec = 0;
		pcPPPoEIp = "";
		switch (pInfo.status & PPPOE_CONN_BIT)
		{
		case CONNECTION_OFF:
			strcpy(pcPPPoEConnectionStatus, "OFF");
			break;
		case CONNECTION_ON:
			strcpy(pcPPPoEConnectionStatus, "ON");
			*lPPPoESessionID = pInfo.session_num;
			*lPPPoEConnectingTime_InSec = pInfo.live_sec;
			pcPPPoEIp = pInfo.loc_addr;
			break;
		case CONNECTION_TRYING:
			strcpy(pcPPPoEConnectionStatus, "CONNECTING");
			break;
		case CONNECTION_WAITING:
			strcpy(pcPPPoEConnectionStatus, "WAITING");
			break;
		default:
			pcPPPoEConnectionStatus = "";
			break;
		}
	}
	
	return 0;
}

/* return 1: change
          0: no change
 */
 
int post_pppoe_info(void *pParam)
{
	LIST 	*pList;
	BOOL 	bIsChange = FALSE;
	const char *pcPPPoEUserName=NULL;
	const char *pcPPPoEUserPass=NULL;
	const char *pcPPPoEServiceName=NULL;
	const char *pcPPPoEACName=NULL;
	unsigned long ulPPPoEIdle=0;
	unsigned long ulPPPoEDemand = 0;
	unsigned long ulPPPoEEnable = 0;
	unsigned long ulPPPoEIPAddr = 0;
	unsigned long ulPPPoENetwork = 0;
	unsigned long ulPPPoENetmask = 0;
	unsigned long ulPPPoEDNS[2] = {0,0};
	unsigned short usMtu = 0;
	
	int bEnableNaptUnnum = 0;
	int	id=0;
	int 	iAction=0;
	int	ulHowToGetIp=0;
	int	demand=0;
	char	buf[64];
	
	/* Last params or status */
	pList = (LIST *)pParam;

	/* Read page parameters. */
	iAction = g_iAction;
	id = g_curPPPoEId;
	ulHowToGetIp = g_tmpConnMode;
	
	ulPPPoEEnable = GetLong(pList, "va_pppoe_enable");
	pcPPPoEUserName = GetString(pList, "Txt_PPPoEUser");
	pcPPPoEUserPass = GetString(pList, "Txt_PPPoEPass");
	pcPPPoEServiceName = GetString(pList, "Txt_PPPoESeviceName");
	pcPPPoEACName = GetString(pList, "Txt_PPPoEACName");
	ulPPPoEIdle = GetLong(pList, "Num_PPPoEIdle");
	ulPPPoEDemand = GetBool(pList, "Num_PPPoEDemand");
	bEnableNaptUnnum  = GetBool(pList, "va_pppoe_naptunnum");
	usMtu = GetLong(pList, "Num_pppoe_mtu");
	
	ulPPPoEIPAddr = GetIP(pList, "va_pppoe_ipaddr");
	ulPPPoENetwork = GetIP(pList, "va_pppoe_network");
	ulPPPoENetmask = GetIP(pList, "va_pppoe_netmask");
	ulPPPoEDNS[0] = GetIP(pList, "va_pppoe_dns1");
	ulPPPoEDNS[1] = GetIP(pList, "va_pppoe_dns2");

	
	if( ulHowToGetIp == GETIP_UNNUM )
	{
		ulPPPoEEnable = 1;
		g_curPPPoEId = UNNUM_ID;
		id = UNNUM_ID;
		if( bEnableNaptUnnum != g_ConfigParam.bEnableNaptUnnum )
		{
			char cmd[64];
			
			sprintf(cmd,"ipt -t nat -%c POSTROUTING -o ppp0 -j MASQUERADE", bEnableNaptUnnum ? 'A' : 'D');
			iptables(cmd);
			g_ConfigParam.bEnableNaptUnnum = bEnableNaptUnnum;
			bIsChange = TRUE;
		}
	}
	fprintf(stderr,"\nPPPoE action=%d ",iAction);

	if (iAction == ACTION_DISCONNECT )
	{
		WRKillPPPoE( id );
		//SendPPPoECommand( id, PPPOE_CMD_DISCONNECT);
	}
	else if (iAction == ACTION_RELOAD )
	{
		g_curPPPoEId = id;
	}
	else //apply, connect buttion
	{
		if( usMtu != g_ConfigParam.usMtu )
		{
			g_ConfigParam.usMtu  = usMtu;
			bIsChange = TRUE;
		}
		if ( ulPPPoEEnable != g_ConfigParam.pppoeConf[id].enable )
		{
			g_ConfigParam.pppoeConf[id].enable = ulPPPoEEnable;
			bIsChange = TRUE;
		}
		if( ulPPPoEIPAddr != g_ConfigParam.pppoeConf[id].loc_addr )
		{
			g_ConfigParam.pppoeConf[id].loc_addr = ulPPPoEIPAddr;
			bIsChange = TRUE;
		}
		if( ulPPPoENetmask != g_ConfigParam.pppoeConf[id].netmask )
		{
			g_ConfigParam.pppoeConf[id].netmask = ulPPPoENetmask;
			bIsChange = TRUE;
		}
		if( ulPPPoENetwork != g_ConfigParam.pppoeConf[id].network )
		{
			g_ConfigParam.pppoeConf[id].network = ulPPPoENetwork;
			bIsChange = TRUE;
		}
		if( ulPPPoEDNS[0] != g_ConfigParam.pppoeConf[id].dnsaddr[0] )
		{
			g_ConfigParam.pppoeConf[id].dnsaddr[0] = ulPPPoEDNS[0];
			bIsChange = TRUE;
		}
		if( ulPPPoEDNS[1] != g_ConfigParam.pppoeConf[id].dnsaddr[1] )
		{
			g_ConfigParam.pppoeConf[id].dnsaddr[1] = ulPPPoEDNS[1];
			bIsChange = TRUE;
		}	
		if (ulPPPoEIdle != g_ConfigParam.pppoeConf[id].idle)
		{
			g_ConfigParam.pppoeConf[id].idle = ulPPPoEIdle;
			bIsChange = TRUE;
		}
		if ( ulPPPoEDemand != g_ConfigParam.pppoeConf[id].demand )
		{
			g_ConfigParam.pppoeConf[id].demand = ulPPPoEDemand;
			bIsChange = TRUE;
			//demand = ulPPPoEDemand; 
		}
		demand = ulPPPoEDemand; //should always set, because disconnect button will clear
		//demand value 
		
		if ( g_ConfigParam.pppoeConf[id].demand == 0 )
			g_ConfigParam.pppoeConf[id].idle = 0 ;
		
		if (strcmp(pcPPPoEUserName, g_ConfigParam.pppoeConf[id].user) != 0)
		{
			memset(g_ConfigParam.pppoeConf[id].user,0,sizeof(g_ConfigParam.pppoeConf[id].user));
			strlcpy(&g_ConfigParam.pppoeConf[id].user[0], pcPPPoEUserName, sizeof(g_ConfigParam.pppoeConf[0].user));
			bIsChange = TRUE;
		}
		if (strcmp(pcPPPoEUserPass, g_ConfigParam.pppoeConf[id].pass) != 0)
		{
			memset(g_ConfigParam.pppoeConf[id].pass,0,sizeof(g_ConfigParam.pppoeConf[id].pass));
			strlcpy(&g_ConfigParam.pppoeConf[id].pass[0], pcPPPoEUserPass, sizeof(g_ConfigParam.pppoeConf[0].pass));
			bIsChange = TRUE;
		}
		if (strcmp(pcPPPoEServiceName, g_ConfigParam.pppoeConf[id].srvname) != 0)
		{
			memset(g_ConfigParam.pppoeConf[id].srvname,0,sizeof(g_ConfigParam.pppoeConf[id].srvname));
			strlcpy(&g_ConfigParam.pppoeConf[id].srvname[0], pcPPPoEServiceName, sizeof(g_ConfigParam.pppoeConf[0].srvname));
			bIsChange = TRUE;
		}
		if (strcmp(pcPPPoEACName, g_ConfigParam.pppoeConf[id].acname) != 0)
		{
			memset(g_ConfigParam.pppoeConf[id].acname,0,sizeof(g_ConfigParam.pppoeConf[id].acname));
			strlcpy(&g_ConfigParam.pppoeConf[id].acname[0], pcPPPoEACName, sizeof(g_ConfigParam.pppoeConf[0].acname));
			bIsChange = TRUE;
		}

		snprintf(buf, sizeof(buf), "/swap/resolv.conf.%d", id );
		wipeout(buf);
		
		if (iAction == ACTION_SET)
		{
			if( ulPPPoEEnable == 0 )
				WRKillPPPoE( id );//SendPPPoECommand(id, PPPOE_CMD_DISCONNECT );
			else
			{
				if (demand) 
				{
					fprintf(stderr,"\nDOD..");
					SendPPPoECommand(id, PPPOE_CMD_DOD);
				}
				else
				{
					fprintf(stderr,"\nCOMM...");
					SendPPPoECommand(id, PPPOE_CMD_COMM);
				}
			}
		}
		else if (iAction == ACTION_CONNECT)
		{
			fprintf(stderr,"\nCONNECT...");
			SendPPPoECommand( id, PPPOE_CMD_CONNECT);
		}
	}

	dnsProxyInstall();

	return bIsChange;
}

static int WRWriteInterfaceArgument(HTTPCONNECTION hConnection, char *pcBuf)
{
	int id = 0;
	int iCount = 0;
	char pppoe_state[16];
	unsigned long pppoe_life=0, pppoe_sid=0;
	int ulHowToGetIp = 0;
	
	if( g_iAction == ACTION_RELOAD )
		id = g_curPPPoEId;
	else if( g_ConfigParam.ulHowToGetIp == GETIP_UNNUM )
		id = UNNUM_ID; /* set to UNNUM_ID if not RELOAD and HowToGetIp == UNNUM, -> reboot case */
	else id = 1; /* first pppoe id */

	ulHowToGetIp = g_iAction==ACTION_RELOAD ? g_tmpConnMode : (int)g_ConfigParam.ulHowToGetIp;
	
	memset(pppoe_state,0,sizeof(pppoe_state));

	get_pppoe_info(id, 
			&pppoe_state[0],
			&pppoe_sid,
			&pppoe_life);
	
	iCount += sprintf(pcBuf+iCount, "\nvar va_hostname = '%s';", g_ConfigParam.acHostName);
	iCount += sprintf(pcBuf+iCount, "\nvar va_domain = '%s';", g_ConfigParam.acHostName);
	iCount += sprintf(pcBuf+iCount, "\nvar va_lan_ipaddr = '%s';", int_ntoa(GetIPAddress(G_PC_ETH0)));
	iCount += sprintf(pcBuf+iCount, "\nvar va_lan_netmask = '%s';",int_ntoa(GetNetMask(G_PC_ETH0)));
	iCount += sprintf(pcBuf+iCount, "\nvar va_dyn_stat = '%d';", ulHowToGetIp );
	iCount += sprintf(pcBuf+iCount, "\nvar va_startup_stat = '%lu';", g_ConfigParam.ulHowToGetIp );
	iCount += sprintf(pcBuf+iCount, "\nvar va_ipaddr_static = '%s';", int_ntoa(g_ConfigParam.ulPublicIpAddress));
	iCount += sprintf(pcBuf+iCount, "\nvar va_mask_static = '%s';", int_ntoa(g_ConfigParam.ulPublicSubnetMask));
	iCount += sprintf(pcBuf+iCount, "\nvar va_defgw_static = '%s';",int_ntoa(g_ConfigParam.ulDefaultGateway));
	iCount += sprintf(pcBuf+iCount, "\nvar va_dns_prim_static = '%s';", int_ntoa(g_ConfigParam.aulDNSServer[0]));
	iCount += sprintf(pcBuf+iCount, "\nvar va_dns_sec_static = '%s';", int_ntoa(g_ConfigParam.aulDNSServer[1]));
	iCount += sprintf(pcBuf+iCount, "\nvar va_dns_third_static = '%s';", int_ntoa(g_ConfigParam.aulDNSServer[2]));

	iCount += sprintf(pcBuf+iCount, "\nvar va_mtu = '%d';", g_ConfigParam.usMtu );

	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_id = %d;", id );

	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_ipaddr = '%s';", int_ntoa(g_ConfigParam.pppoeConf[id].loc_addr));
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_network = '%s';", int_ntoa(g_ConfigParam.pppoeConf[id].network));
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_netmask = '%s';",int_ntoa(g_ConfigParam.pppoeConf[id].netmask));
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_dns1 = '%s';", int_ntoa(g_ConfigParam.pppoeConf[id].dnsaddr[0]));
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_dns2 = '%s';", int_ntoa(g_ConfigParam.pppoeConf[id].dnsaddr[1]));
	
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_enable = %d;",g_ConfigParam.pppoeConf[id].enable);
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_user = '%s';", g_ConfigParam.pppoeConf[id].user);
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_pass = '%s';", g_ConfigParam.pppoeConf[id].pass);
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_srvname = '%s';", g_ConfigParam.pppoeConf[id].srvname);
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_acname = '%s';", g_ConfigParam.pppoeConf[id].acname);
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_demand = %d;", g_ConfigParam.pppoeConf[id].demand);
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_idle = %lu;", g_ConfigParam.pppoeConf[id].idle);
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_state = '%s';", pppoe_state);
	iCount += sprintf(pcBuf+iCount, "\nvar va_pppoe_naptunnum = %u;", g_ConfigParam.bEnableNaptUnnum);

	g_iAction = ACTION_SET;
	
	return iCount;
}

int get_system_main(HTTPCONNECTION hConnection, void *pParam)
{
	char *pcFile;
	
	if (WRExternalAdmin(hConnection) == 0) return 0;
	if ((pcFile = GetRequestFilePath(hConnection)) == NULL) return 0;

	//SendHtmlWithParam(hConnection, pcFile, url);
	SendReDirectHtml(hConnection, "message.htm", "index.htm", TRUE, 0, "MSG_LOGIN", NULL );
	
	return 0;
}


int get_system_setup(HTTPCONNECTION hConnection, void *pParam)
{
	char *pcFile;
	if (WRExternalAdmin(hConnection) == 0) return 0;
	if ((pcFile = GetRequestFilePath(hConnection)) == NULL) return 0;

	SetKeepAliveMode(hConnection, FALSE);
	SendHtmlWithParam(hConnection, pcFile, WRWriteInterfaceArgument);

	return 0;
}

int WRToBeChangeLANIP(HTTPCONNECTION hConnection, time_t *tLastFill, void *pParam )
{
	FOR_CHANGE_IP_T *pIP = (FOR_CHANGE_IP_T *)pParam;
	if (pIP != NULL)
	{
		if (SetGeneralIP(G_PC_ETH0, pIP->ulIP, pIP->ulNetmask))
		{
			g_ConfigParam.ulPrivateIPAddress = pIP->ulIP;
			g_ConfigParam.ulPrivateSubnetMask = pIP->ulNetmask;
			
			lan_service(old_lan_ip, pIP->ulIP, pIP->ulNetmask, pIP->updateDhcp );	//reinstall LAN service
			
			WriteFlashMemory(&g_ConfigParam);
		}

		free(pIP);
	}

	SetSendDataOverFun(hConnection, NULL, NULL);
	return 0;
}


int post_system_setup(HTTPCONNECTION hConnection, void *pParam)
{
	LIST *pList;
	BOOL bIsChange, bIsChangeLan ;
	int i;
	BOOL bPublicIPChanged = FALSE;
	unsigned ulPrivateIPToBe = 0L;

	/* Config params */
	const char *pcHostName;
	const char *pcDomainName;
	
	unsigned long ulPrivateIPAddress;
	unsigned long ulPrivateSubnetMask;
	unsigned long ulHowToGetIp;
	unsigned long ulPublicIpAddress;
	unsigned long ulPublicSubnetMask;
	unsigned long ulDefaultGateway;
	unsigned long aulDNSServer[MAX_DNS_SERVER];
	unsigned short usMtu = 0;

	pList = (LIST *)pParam;

	/* Read page parameters. */
	pcHostName = GetString(pList, "va_hostname");
	pcDomainName = GetString(pList, "va_domainname");
	ulPrivateIPAddress = GetIP(pList, "va_lan_ipaddr");
	ulPrivateSubnetMask = GetIP(pList, "va_lan_netmask");
	ulHowToGetIp = (unsigned long)GetLong(pList, "va_dyn_stat");
	ulPublicIpAddress = GetIP(pList, "va_ipaddr_static");
	ulPublicSubnetMask = GetIP(pList, "va_mask_static");
	ulDefaultGateway = GetIP(pList, "va_defgw_static");
	aulDNSServer[0] = GetIP(pList, "va_dns_prim_static");
	aulDNSServer[1] = GetIP(pList, "va_dns_sec_static");
	aulDNSServer[2] = GetIP(pList, "va_dns_third_static");

	g_iAction = GetLong(pList,"action");
	g_curPPPoEId = GetLong(pList,"va_pppoe_id");
	g_tmpConnMode = ulHowToGetIp;
	if( ulHowToGetIp == GETIP_UNNUM )	
		g_curPPPoEId = UNNUM_ID;
		
	/* Config */
	bIsChange = FALSE;
	bIsChangeLan = FALSE;

	SetKeepAliveMode(hConnection, FALSE);

	if( g_iAction == ACTION_RELOAD )
	{
		SendReDirectPage(hConnection);
		return 0;
	}

	
	/*----------------- LAN PART ------------------- */
	if (strcmp(pcHostName, g_ConfigParam.acHostName) != 0)
	{
		MyStrncpy(g_ConfigParam.acHostName, pcHostName, sizeof(g_ConfigParam.acHostName));
		bIsChangeLan = TRUE;
	}

	if (strcmp(pcDomainName, g_ConfigParam.acDomainName) != 0)
	{
		MyStrncpy(g_ConfigParam.acDomainName, pcDomainName, sizeof(g_ConfigParam.acDomainName));
		bIsChangeLan = TRUE;
	}

	if (ulPrivateIPAddress != GetIPAddress(G_PC_ETH0)
		|| ulPrivateSubnetMask != GetNetMask(G_PC_ETH0))
	{
		FOR_CHANGE_IP_T *pIP;
		pIP = (FOR_CHANGE_IP_T *)malloc(sizeof(FOR_CHANGE_IP_T));
		if (pIP != NULL)
		{
			old_lan_ip = g_ConfigParam.ulPrivateIPAddress;
			
			pIP->ulIP = ulPrivateIPAddress;
			pIP->ulNetmask = ulPrivateSubnetMask;
			pIP->updateDhcp = 1;
			
			SetSendDataOverFun(hConnection, WRToBeChangeLANIP, pIP );
			ulPrivateIPToBe = ulPrivateIPAddress;
			g_ConfigParam.ulPrivateIPAddress = ulPrivateIPAddress;
			
			WRConfigVserver(g_ConfigParam.avVirtualHostItem, TRUE); //reinstall virtual server rule
			bIsChangeLan = TRUE;
		}
	}

	/*----------------- WAN PART------------------- */
	
	if (ulHowToGetIp != g_ConfigParam.ulHowToGetIp)
	{
		bIsChange = TRUE;
	}

	usMtu = getSysMtu( ulHowToGetIp, usMtu );

	SetGeneralIP("eth0:1", 0L, 0L );

	if (ulHowToGetIp == GETIP_MAN)
	{
		int dnsChanged = FALSE;
		
		NATA_StopPPPoE(0);
		
		if (g_ConfigParam.ulHowToGetIp == GETIP_DHCP)
			WRStopDHCPClient();

		else if (g_ConfigParam.ulHowToGetIp == GETIP_PPP||g_ConfigParam.ulHowToGetIp == GETIP_UNNUM)
			WRKillPPPoE(ALL_ID); //SendPPPoECommand(ALL_ID, PPPOE_CMD_DISCONNECT);

		if( SetDNSServer(aulDNSServer) )
		{
			for (i = 0; i < MAX_DNS_SERVER; i++)
			{
				g_ConfigParam.aulDNSServer[i] = aulDNSServer[i];
				bIsChange = TRUE;
			}
		}

		for( i=0; i<3; i++)
		{
			if( aulDNSServer[i] != g_ConfigParam.aulDNSServer[i] )
			{
				dnsChanged = TRUE;
				break;
			}
		
		}
		if (  (ulHowToGetIp != g_ConfigParam.ulHowToGetIp) || (ulPublicIpAddress != g_ConfigParam.ulPublicIpAddress) 
			|| (ulPublicSubnetMask != g_ConfigParam.ulPublicSubnetMask) )
		{
			if (SetGeneralIP(G_PC_ETH1,
				ulPublicIpAddress,
				ulPublicSubnetMask))
			{
				g_ConfigParam.ulPublicIpAddress = ulPublicIpAddress;
				g_ConfigParam.ulPublicSubnetMask = ulPublicSubnetMask;
			
				bIsChange = TRUE;
				bPublicIPChanged = TRUE;
				
			}
		}

		usMtu = GetLong(pList,"Num_static_mtu");
		if( usMtu != g_ConfigParam.usMtu )
		{
			if( SetMtu( G_PC_ETH1, usMtu ) == TRUE )
			{
				g_ConfigParam.usMtu = usMtu;
				bIsChange = TRUE;
			}
		}
		
		if (SetGateway(G_PC_ETH1, ulDefaultGateway))
		{
			g_ConfigParam.ulDefaultGateway = ulDefaultGateway;
		}
		
 		wan_service(0);

	}
	else if (ulHowToGetIp == GETIP_DHCP)
	{
		NATA_StopPPPoE(0);
		
		if (g_ConfigParam.ulHowToGetIp == GETIP_PPP||g_ConfigParam.ulHowToGetIp == GETIP_UNNUM)
			//SendPPPoECommand(ALL_ID, PPPOE_CMD_DISCONNECT);
			WRKillPPPoE(ALL_ID);

		usMtu = GetLong(pList,"Num_dyn_mtu");
		if( usMtu != g_ConfigParam.usMtu )
		{
			if( SetMtu( G_PC_ETH1, usMtu ) == TRUE )
			{
				g_ConfigParam.usMtu = usMtu;
				bIsChange = TRUE;
			}
		}
			
		if (ulHowToGetIp != g_ConfigParam.ulHowToGetIp ||(g_iAction ==  ACTION_SET &&  g_ConfigParam.ulPublicIpAddress == 0L) )
		{
			WRRunDHCPClient();
			bIsChange = TRUE;
		}
		else if ( g_iAction == ACTION_CONNECT )
		{
			WRRenewDHCPClient();
		}
		else if (  g_iAction ==  ACTION_DISCONNECT ) 
		{
			
			WRStopDHCPClient();
			
			SetGeneralIP(G_PC_ETH1, 0L, 0L);
			g_ConfigParam.ulPublicIpAddress = 0L;
			g_ConfigParam.ulPublicSubnetMask = 0L;

			SetGateway(G_PC_ETH1, 0L);
			g_ConfigParam.ulDefaultGateway = 0L;
		}
		else // ACTION_SET
		{
			GetPubIPInfo(&ulPublicIpAddress,&ulPublicSubnetMask,&ulDefaultGateway );
			if (ulPublicIpAddress == 0L)
				WRRunDHCPClient();
		}
	}

/* changing netmask/ip will not take effect immediately because of using SendFuncOver.... 
	else if (ulHowToGetIp == GETIP_PPP || ulHowToGetIp == GETIP_UNNUM )
	{
		wipeout("/swap/resolv.conf");
		
		if (g_ConfigParam.ulHowToGetIp == GETIP_DHCP)
			WRStopDHCPClient();
		
		SetGeneralIP("eth1", 0L, 0L);

		if( ulHowToGetIp != g_ConfigParam.ulHowToGetIp)
		{
			KillAllProc(PROG_PPPOE);
			//WRKillPPPoE(ALL_ID);	//multiple  pppoe (>3) running 
		}
		
		bIsChange |= post_pppoe_info( pParam);
	}
*/	

	if (ulPrivateIPToBe != 0L)
	{
		char acUrl[64];
		strcpy(acUrl, "http://");
		
		if ( GetLong(pList, "va_RedirectOrNot") ) //MSG_CHANGE_IP_ADDR => Changing IP or netmask ...
		{
			IP2String(ulPrivateIPToBe, acUrl + strlen(acUrl));
			SendReDirectHtml(hConnection, "Redirect.htm", acUrl, TRUE, 5000, "MSG_CHANGE_IP_ADDR", NULL );
		}
		else
		{
			SendReDirectHtml(hConnection, "message.htm", acUrl, FALSE, 0, "MSG_RECONNECT_PROMPT", NULL );		
		}
	}
	else
	{
		SendReDirectPage(hConnection);
		
	}
	

	if (ulHowToGetIp == GETIP_PPP || ulHowToGetIp == GETIP_UNNUM )
	{
		/* -- don't wipeout :consider the case:
			a. manually connect pppoe 0
			b. manually connect pppoe 1 -> wipeout /etc/resolv.conf!!! 
		wipeout("/swap/resolv.conf");
		*/

		if( ulHowToGetIp == GETIP_UNNUM )
		{
			wipeout("/swap/resolv.conf");
			wipeout("/swap/resolv.conf.1");
			wipeout("/swap/resolv.conf.2");
		}

		if (g_ConfigParam.ulHowToGetIp == GETIP_DHCP)
			WRStopDHCPClient();
		
		SetGeneralIP("eth1", 0L, 0L);

		/* kill all ppp process */
		if( ulHowToGetIp != g_ConfigParam.ulHowToGetIp)
		{
			KillAllProc(PROG_PPPOE);
			//WRKillPPPoE(ALL_ID);	//multiple  pppoe (>3) running 
		}
		
		g_ConfigParam.ulHowToGetIp = ulHowToGetIp;
		if( bIsChangeLan )
			usleep(1000 * 1000);//for lesser than 20000, usleep do not really sleep.

		/* run ppp process */
		bIsChange |= post_pppoe_info( pParam); /* HowToGet change */
	}

	if (bIsChange|bIsChangeLan)
	{
		g_ConfigParam.ulHowToGetIp = ulHowToGetIp;
		WriteFlashMemory(&g_ConfigParam);
	}


	return 0;
}


