
#include <getopt.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
//#include <dlfcn.h>
#include <ctype.h>
#include <stdarg.h>
#include <limits.h>
#include <unistd.h>
#include <iptables.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <net/if.h>
#include <arpa/inet.h>
              
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_multiport.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ipt_TCPMSS.h>
#include <linux/netfilter_ipv4/ipt_limit.h>
#include <linux/netfilter_ipv4/ip_tables.h>

#include "aos_fw.h"
#include "aos_fw_defs.c"

int policydomain( char action, char *cmdstr );

unsigned long GetLocalIP(char *interface)
{
	int fd = -1;
	struct ifreq ifr;
	struct sockaddr_in *sin;	

	fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	ifr.ifr_addr.sa_family = AF_INET;
	strcpy(ifr.ifr_name, interface);
	
	if (ioctl(fd, SIOCGIFADDR, &ifr) == 0)
	{
		sin = (struct sockaddr_in *) &ifr.ifr_addr;		
	//	printf("*sin->sin_addr.s_addr=%0x\n",sin->sin_addr.s_addr);
		close(fd);
		return sin->sin_addr.s_addr;
	}	
	else
	{
		close(fd);
		return 0;
	}	
}

#define FREE_E	{\
			if(e)\
			free(e);\
			e=NULL;\
		}
//#define printf iprintf
extern struct ipt_entry *e;
extern struct in_addr *saddrs, *daddrs;

#define FREE_ADDR {\
			if(daddrs)\
				free(daddrs);\
				daddrs=NULL;\
			if(saddrs)\
				free(saddrs);\
				saddrs=NULL;\
		}

extern int sockfd;
extern int socket_count;


int Add_Port(struct Alt_Port *wport)
{
	int ret;
	iptc_handle_t handle = NULL;
		
	ret = do_command("PREROUTING",Add_Alt_Port,wport,28,CMD_INSERT,&handle);
	if (ret)
		ret = iptc_commit(&handle);
	if (!ret)
	{
		if(handle)
		{
			close(sockfd);
			free(handle);
			FREE_E
			FREE_ADDR
		}
		return 0;
	}
	else
	{
		FREE_E
		FREE_ADDR
	}
	
	return 1;
}

int Del_Port(struct Alt_Port *wport)
{
	int ret;
	iptc_handle_t handle = NULL;

	ret = do_command("PREROUTING",Del_Alt_Port,wport,28,CMD_DELETE,&handle);
	if (ret)
		ret = iptc_commit(&handle);
	if (!ret)
	{
		if(handle)
		{
			close(sockfd);
			free(handle);
			FREE_E
			FREE_ADDR
		}
		//printf("\n Del_Port iptc_commit failed!\n");
		return 0;
	}
	else
	{
		FREE_E
		FREE_ADDR
		//printf("\n Del_Port iptc_commit ok!\n");
	}
	
	return 1;
}


int Add_Upnp_Port(struct Upnp_Port *uport)
{
	int ret;
	iptc_handle_t handle = NULL;


	ret = do_command("PREROUTING",Add_UPnP_Port,uport,28,CMD_INSERT,&handle);
	if (ret)
		ret = iptc_commit(&handle);
	if (!ret)
	{
		if(handle)
		{
			close(sockfd);
			free(handle);
			FREE_E
			FREE_ADDR
		}
		//printf("\n INSERT_UPnP_Port iptc_commit failed!\n");
		return 0;
	}
	else
	{
		FREE_E
		FREE_ADDR
		//printf("\n INSERT_UPnP_Port iptc_commit ok!\n");
	}
	
	return 1;
}
int Del_Upnp_Port(struct Upnp_Port *uport)
{
	int ret;
	iptc_handle_t handle = NULL;

	ret = do_command("PREROUTING",Del_UPnP_Port,uport,28,CMD_DELETE,&handle);
	if (ret)
		ret = iptc_commit(&handle);
	if (!ret)
	{
		if(handle)
		{
			close(sockfd);
			free(handle);
			FREE_E
			FREE_ADDR
		}
		//printf("\n Del_UPnP_Port iptc_commit failed!\n");
		return 0;
	}
	else
	{
		FREE_E
		FREE_ADDR
		//printf("\n Del_UPnP_Port iptc_commit ok!\n");
	}
	
	return 1;
}

#ifdef XX

int Add_Upnp_Port(struct Upnp_Port *param)
{
	char buf[256];
	
	snprintf(buf,sizeof(buf),"iptables -t nat -I PREROUTING %s -d %s --dport %d -j DNAT --to-destination %s:%d",
		((struct Upnp_Port*)param)->Proto,
		((struct Upnp_Port*)param)->ExtIP, ((struct Upnp_Port*)param)->ExtPort,
		((struct Upnp_Port*)param)->IntIP, ((struct Upnp_Port*)param)->IntPort );
	printf("\nAdd_Upnp_Port:%s", buf); fflush(stdout);
	iptables(buf);

	return 1;
}
int Del_Upnp_Port(struct Upnp_Port *param)
{
	char buf[256];
	
	snprintf(buf,sizeof(buf),"iptables -t nat -D PREROUTING %s -d %s --dport %d -j DNAT --to-destination %s:%d",
		((struct Upnp_Port*)param)->Proto,
		((struct Upnp_Port*)param)->ExtIP, ((struct Upnp_Port*)param)->ExtPort,
		((struct Upnp_Port*)param)->IntIP, ((struct Upnp_Port*)param)->IntPort );
	printf("\nDel_Upnp_Port:%s", buf); fflush(stdout);
	iptables(buf);

	return 1;
}

#endif

struct ipt_entry *
Generate_entry(int type,void *param,struct ipt_entry *fw)
{
	struct ipt_entry *e;
	struct ipt_entry_target *mytarget;
	int len,tsize,bsize;
  	unsigned long theIP __attribute__ ((aligned (4)));
		
	struct ipt_tcp mytcp;
	struct ip_nat_multi_range mymulti_range;
	int spt;//,ept;
	
	switch(type)
	{

		case Add_Alt_Port:
		case Del_Alt_Port:
			sscanf(((struct Alt_Port*)param)->mport,"%d",&spt);//ExtPort
			//tip=inet_addr(((struct Virtual_Server *)param)->ip);
			len=sizeof(struct ipt_entry)+(sizeof(struct ipt_entry_target)+sizeof(struct ipt_tcp))+(sizeof(struct ipt_entry_target)+sizeof(struct ip_nat_multi_range));
					//112				32				12								20
			e=(void*)malloc(len);	
			memset(e,0,len);	
			*e=*fw;
			e->target_offset=sizeof(struct ipt_entry)+(sizeof(struct ipt_entry_target)+sizeof(struct ipt_tcp));
			e->next_offset=len;	
			
			tsize=sizeof(struct ipt_entry_target)+sizeof(struct ipt_tcp);
			mytarget=(void*)malloc(tsize);	
			memset(mytarget,0,tsize);	
			mytarget->u.target_size=tsize;	
		//	printf("((struct Virtual_Server *)param)->proto=%s\n",((struct Virtual_Server *)param)->proto);
			strcpy(mytarget->u.user.name,"tcp");
				
			memset(&mytcp,0,sizeof(struct ipt_tcp));
			mytcp.spts[0]=0x0000;
			mytcp.spts[1]=0xffff;	
			mytcp.dpts[0]=spt;
			mytcp.dpts[1]=spt;	
			memcpy(mytarget->data,&mytcp,sizeof(struct ipt_tcp));
			memcpy(e->elems,mytarget,mytarget->u.target_size);
			free(mytarget);
			
			bsize=tsize;
			tsize=sizeof(struct ipt_entry_target)+sizeof(struct ip_nat_multi_range);
			mytarget=(void*)malloc(tsize);	
			memset(mytarget,0,tsize);	
			mytarget->u.target_size=tsize;	
			strcpy(mytarget->u.user.name,"DNAT");
			
			theIP=GetLocalIP("eth0");//IntIP
			memset(&mymulti_range,0,sizeof(struct ip_nat_multi_range));
			mymulti_range.rangesize=1;
			mymulti_range.range[0].flags=3;
			mymulti_range.range[0].min_ip=theIP;
			mymulti_range.range[0].max_ip=theIP;
			//mymulti_range.range[0].min.tcp.port=0x5000;
			mymulti_range.range[0].min.tcp.port=htons(55530);
			//mymulti_range.range[0].max.tcp.port=0x5000;		
			mymulti_range.range[0].max.tcp.port=htons(55530);
			memcpy(mytarget->data,&mymulti_range,sizeof(struct ip_nat_multi_range));	
			memcpy(e->elems+bsize,mytarget,mytarget->u.target_size);
			free(mytarget);
			
			break;
		
		case Add_UPnP_Port:
		case Del_UPnP_Port:
			//sscanf(((struct Alt_Port*)param)->mport,"%d",&spt);//ExtPort
			spt=((struct Upnp_Port*)param)->ExtPort;
			//printf("spt===%d\n",spt);
			//tip=inet_addr(((struct Virtual_Server *)param)->ip);
			
			len=sizeof(struct ipt_entry)+(sizeof(struct ipt_entry_target)+sizeof(struct ipt_tcp))+(sizeof(struct ipt_entry_target)+sizeof(struct ip_nat_multi_range));
				//112				32				12								20
			e=(void*)malloc(len);	
			memset(e,0,len);	
			*e=*fw;
			e->target_offset=sizeof(struct ipt_entry)+(sizeof(struct ipt_entry_target)+sizeof(struct ipt_tcp));
			e->next_offset=len;	
			tsize=sizeof(struct ipt_entry_target)+sizeof(struct ipt_tcp);
			mytarget=(void*)malloc(tsize);	
			memset(mytarget,0,tsize);	
			mytarget->u.target_size=tsize;	
			//	printf("((struct Virtual_Server *)param)->proto=%s\n",((struct Virtual_Server *)param)->proto);
				
			strcpy(mytarget->u.user.name,((struct Upnp_Port*)param)->Proto);
			memset(&mytcp,0,sizeof(struct ipt_tcp));
			mytcp.spts[0]=0x0000;
			mytcp.spts[1]=0xffff;	
			mytcp.dpts[0]=spt;
			mytcp.dpts[1]=spt;	
			memcpy(mytarget->data,&mytcp,sizeof(struct ipt_tcp));
			memcpy(e->elems,mytarget,mytarget->u.target_size);
			free(mytarget);
		
			bsize=tsize;
			tsize=sizeof(struct ipt_entry_target)+sizeof(struct ip_nat_multi_range);
			mytarget=(void*)malloc(tsize);	
			memset(mytarget,0,tsize);	
			mytarget->u.target_size=tsize;	
			strcpy(mytarget->u.user.name,"DNAT");
			
			//theIP=GetLocalIP("eth0");//IntIP
			theIP=inet_addr(((struct Upnp_Port*)param)->IntIP);
			memset(&mymulti_range,0,sizeof(struct ip_nat_multi_range));
			mymulti_range.rangesize=1;
			mymulti_range.range[0].flags=3;
			mymulti_range.range[0].min_ip=theIP;
			mymulti_range.range[0].max_ip=theIP;
			//mymulti_range.range[0].min.tcp.port=0x5000;
			mymulti_range.range[0].min.tcp.port=htons(((struct Upnp_Port*)param)->IntPort);
			//mymulti_range.range[0].max.tcp.port=0x5000;		
			mymulti_range.range[0].max.tcp.port=htons(((struct Upnp_Port*)param)->IntPort);
			memcpy(mytarget->data,&mymulti_range,sizeof(struct ip_nat_multi_range));	
			memcpy(e->elems+bsize,mytarget,mytarget->u.target_size);
			free(mytarget);
			
			break;
			
		default:
//			printf("The rule can't been support!\n");
			e=0;
			break;
	}
	
	return e;
	
}

int
aos_fw_net_parse( struct aos_fw_acl *aos_fw_acl, char *rulematch )
{
	char rulesrc[AOS_FW_RULE_SRC_LEN];
	char ruledst[AOS_FW_RULE_SRC_LEN];
	char rulesport[AOS_FW_RULE_SRC_LEN];
	char ruledport[AOS_FW_RULE_SRC_LEN];
	char ruledomain[AOS_FW_RULE_DOMAIN_LEN];
	
	int i=0,j=0;
	
	memset(rulesrc,0,sizeof(rulesrc));
	memset(ruledst,0,sizeof(ruledst));
	memset(rulesport,0,sizeof(rulesport));
	memset(ruledport,0,sizeof(ruledport));
	memset(ruledomain,0,sizeof(ruledomain));
	
	switch ( aos_fw_acl->srctype )
	{
	case AOS_FW_NET_ANY:
		break;

	case AOS_FW_NET_IP:
		if( aos_fw_acl->src1[0] )
			snprintf(rulesrc,sizeof(rulesrc),"-s %s ",aos_fw_acl->src1);
		break;
		
	case AOS_FW_NET_SUBNET:
		if( aos_fw_acl->src1[0] && aos_fw_acl->src2[0] )
			snprintf(rulesrc,sizeof(rulesrc),"-s %s/%s ",aos_fw_acl->src1,
				aos_fw_acl->src2);
		break;
	
	case AOS_FW_NET_RANGE:
		if( aos_fw_acl->src1[0] && aos_fw_acl->src2[0] )
		{
			snprintf(rulesrc,sizeof(rulesrc)-i,"-s %s-%s ",aos_fw_acl->src1,aos_fw_acl->src2);
		}
		break;
	case AOS_FW_NET_LIST:
		break;
	case AOS_FW_NET_SELF:
		snprintf(rulesrc,sizeof(rulesrc)-i,"-s %s ","0.255.0.255");//special ip address
		break;
	}
	
	//printf("objtype=%d dsttype=%d\n", aos_fw_acl->objtype , aos_fw_acl->dsttype );
	switch ( aos_fw_acl->dsttype )
	{
	case AOS_FW_NET_ANY:
		break;
	case AOS_FW_NET_IP:
		if( aos_fw_acl->objtype == AOS_FW_VIRTUAL_SERVER||
			aos_fw_acl->objtype == AOS_FW_NAT_DNAT)
			snprintf(ruledst,sizeof(ruledst),"-m self --dstif wan " );
		else if( aos_fw_acl->dst1[0] )
			snprintf(ruledst,sizeof(ruledst),"-d %s ",
				aos_fw_acl->dst1);
		break;
		
	case AOS_FW_NET_SUBNET:
		if( aos_fw_acl->dst1[0] && aos_fw_acl->dst2[0] )
			snprintf(ruledst,sizeof(ruledst),"-d %s/%s ",
				aos_fw_acl->dst1,
				aos_fw_acl->dst2);
		break;
	case AOS_FW_NET_SELF:
		snprintf(ruledst,sizeof(ruledst)-i,"-d %s ","0.255.0.255");//special ip address
		break;

	case AOS_FW_NET_DOMAIN:
		snprintf(ruledomain,sizeof(ruledomain)-1,"-m policydomain --dst-domain %s ",aos_fw_acl->dstdomain);
		break;
	
	case AOS_FW_NET_RANGE:
		if( aos_fw_acl->dst1[0] && aos_fw_acl->dst2[0] )
		{
			snprintf(ruledst,sizeof(ruledst),"-d %s-%s ",
				aos_fw_acl->dst1,
				aos_fw_acl->dst2);
		}
		break;
	case AOS_FW_NET_LIST:
		break;
	}

	if( strcmp(aos_fw_acl->proto,"tcp")==0 ||
		strcmp(aos_fw_acl->proto,"udp")==0 )
	{
		i = 0;
		j = 0;
		
		/* write '-m tcp/udp/icmp' only if port != 0 */
		if( (aos_fw_acl->sport1|aos_fw_acl->sport2|aos_fw_acl->dport1|aos_fw_acl->dport2|
			aos_fw_acl->tport1|aos_fw_acl->tport2) )
			i+=snprintf(rulesport+i,sizeof(rulesport)-i,"-m %s ",aos_fw_acl->proto );
		
		//sport		
		if( (aos_fw_acl->sport1 != 0 || aos_fw_acl->sport2 != 0) )
		{
			i+=snprintf(rulesport+i,sizeof(rulesport)-i,"--sport %d",aos_fw_acl->sport1 );
			if( aos_fw_acl->sport2 != 0 )
				i+=snprintf(rulesport+i,sizeof(rulesport)-i,":%d",aos_fw_acl->sport2 );
		}
		i = 0;
		//dport
		if( (aos_fw_acl->dport1 != 0 || aos_fw_acl->dport2 != 0) )
		{
			j+=snprintf(ruledport+j,sizeof(ruledport)-j,"--dport %d",aos_fw_acl->dport1 );
			if( aos_fw_acl->dport2 != 0 )
				j+=snprintf(ruledport+j,sizeof(ruledport)-j,":%d",aos_fw_acl->dport2 );
		}
	}
	else if( strcmp(aos_fw_acl->proto,"icmp")==0 )
	{
		i = 0;
		if( aos_fw_acl->icmptype != 0xff &&
			aos_fw_acl->icmptype != 19 /* all type */ &&
			aos_fw_acl->icmptype != 20 /* all type */  )
			j+=snprintf(ruledport+j,sizeof(ruledport)-j," -m icmp --icmp-type %d",aos_fw_acl->icmptype );
	}
	else
	{
		aos_fw_acl->tport1 = 0;
		aos_fw_acl->tport2 = 0;
	}
	
	//fprintf(stderr,"\nrulesrc=%s dst=%s sport=%s dport=%s ",
	//	rulesrc,ruledst,rulesport,ruledport);
	return sprintf(rulematch,"%s %s %s %s %s ",
				rulesrc,
				ruledst,
				rulesport,
				ruledport,
				ruledomain );
}

int 
aos_fw_rule( struct aos_fw_acl *aos_fw_acl )
{
	//char rulematch[AOS_FW_RULE_SRC_LEN*2+AOS_FW_RULE_PORT_LEN*2+AOS_FW_RULE_MATCH_EXT];
	//char ruleaction[AOS_FW_RULE_SRC_LEN*2+AOS_FW_RULE_PORT_LEN*2+AOS_FW_RULE_MATCH_EXT+AOS_FW_RULE_ACTION_EXT];
	char rulecmd[AOS_FW_RULE_SRC_LEN*2+AOS_FW_RULE_PORT_LEN*2+AOS_FW_RULE_MATCH_EXT+AOS_FW_RULE_ACTION_EXT+AOS_FW_RULE_CMD_EXT+AOS_FW_RULE_DOMAIN_LEN];
	int ret = 0;
	int i=0;

	memset(rulecmd,0,sizeof(rulecmd));
	if( aos_fw_acl->op == AOS_FW_RULE_FLUSH )
	{
		if( aos_fw_acl->objtype == AOS_FW_FILTER )
			i+=snprintf(rulecmd, sizeof(rulecmd), "ipt -t filter -F %s",aos_fw_acl->chain);
		else if( aos_fw_acl->objtype == AOS_FW_NAT_DNAT||aos_fw_acl->objtype == AOS_FW_NAT_SNAT||
			aos_fw_acl->objtype == AOS_FW_NAT_BINAT )
			i+=snprintf(rulecmd, sizeof(rulecmd), "ipt -t nat -F %s",aos_fw_acl->chain);
		else if( aos_fw_acl->objtype == AOS_FW_MANGLE )
		{
			/* OUTBOUND flush -> flush policydomain, mangle table:FORWARD */
			
			policydomain('M',"");
			i+=snprintf(rulecmd, sizeof(rulecmd), "ipt -t mangle -F %s",aos_fw_acl->chain);
		}
		else
		{
			printf("\nincorrect rule chain : %s", aos_fw_acl->chain);
			return -1;
		}

		iptables(rulecmd);
		return 0;
	}
	if( aos_fw_acl->op == AOS_FW_RULE_POLICY )
	{
		if( aos_fw_acl->objtype == AOS_FW_FILTER )
			i+=snprintf(rulecmd, sizeof(rulecmd), "ipt -t filter -P %s %s",aos_fw_acl->chain, aos_fw_acl->action==AOS_FW_ACCEPT? "ACCEPT" : "DROP" );
		else
			i+=snprintf(rulecmd, sizeof(rulecmd), "ipt -t nat -P %s %s",aos_fw_acl->chain, aos_fw_acl->action==AOS_FW_ACCEPT? "ACCEPT" : "DROP" );

		//printf("\naos_fw_rule=%s ",rulecmd);
		iptables(rulecmd);
		return 0;
	}
	
	if( aos_fw_acl->op != AOS_FW_RULE_ADD &&
		aos_fw_acl->op != AOS_FW_RULE_DEL &&
		aos_fw_acl->op != AOS_FW_RULE_INS )
		return -1;
	//
	// the rest are for  aos_fw_acl->op == AOS_FW_RULE_ADD, AOS_FW_RULE_DEL (virtual server, admin implied rules, ...) //
	//
	switch( aos_fw_acl->objtype )
	{
	// case 0:
	case AOS_FW_FILTER:
		i=0;
		i+=snprintf(rulecmd+i,sizeof(rulecmd)-i,"ipt -%c %s -p %s ",
				aos_fw_ruleop[aos_fw_acl->op],
				aos_fw_acl->chain,
				aos_fw_acl->proto);
		i+=aos_fw_net_parse( aos_fw_acl, &rulecmd[0]+i);
#ifdef FILTER_ROUTE
		if( aos_fw_acl->route[0] )
		{
			/* if no 'continue' , packet will not be NAT */
			i+=snprintf(rulecmd+i,sizeof(rulecmd)-i," %s -j ROUTE --continue --oif %s",
				aos_fw_acl->log?"-l" : "", aos_fw_acl->route );
		}
		else
#endif
		{
			i+=snprintf(rulecmd+i,sizeof(rulecmd)-i," -j %s %s", aos_fw_ruleaction[aos_fw_acl->action], 
				aos_fw_acl->log?"-l" : "" );
		}
		break;

	case AOS_FW_VIRTUAL_SERVER:
	case AOS_FW_NAT_DNAT:
		i=0;
		if( aos_fw_acl->taddr1[0] == 0 )
			return -1;

		/* note: '-q' implies do filter */
		/* removed '-q' option since v1.40 2005/07/04 */

		/* '-a' set state ACCEPT,  FORWARD chain should accept state ACCPET traffic' */
		//i+=snprintf(rulecmd+i,sizeof(rulecmd)-i,"ipt -t nat -%c %s -i ! eth0 -p %s %s",
		i+=snprintf(rulecmd+i,sizeof(rulecmd)-i,"ipt -t nat -%c %s -a -p %s %s",
				aos_fw_ruleop[aos_fw_acl->op],
				"VSERVER",	/*PREROUTING */
				aos_fw_acl->proto,
				aos_fw_acl->log?"-l" : "" );
		i+=aos_fw_net_parse( aos_fw_acl, &rulecmd[0]+i );
		i+=snprintf(rulecmd+i,sizeof(rulecmd)-i," -j DNAT --to-destination %s-%s",
				aos_fw_acl->taddr1,
				aos_fw_acl->taddr2 );
		if( aos_fw_acl->tport1 && aos_fw_acl->tport2 )
			i+=snprintf(rulecmd+i,sizeof(rulecmd)-i,":%d-%d",
				aos_fw_acl->tport1,
				aos_fw_acl->tport2);
		break;

	case AOS_FW_MANGLE:
		i=0;
		i+=snprintf(rulecmd+i,sizeof(rulecmd)-i,"ipt -t mangle -%c %s -p %s ",
				aos_fw_ruleop[aos_fw_acl->op],
				aos_fw_acl->chain,
				aos_fw_acl->proto);
		i+=aos_fw_net_parse( aos_fw_acl, &rulecmd[0]+i);
		if( aos_fw_acl->route[0] && strncmp(&aos_fw_acl->route[0],"AUTO",4) )	/* if no '--continue', packet will not be NAT */
			i+=snprintf(rulecmd+i,sizeof(rulecmd)-i," %s -j ROUTE --continue --oif %s", 
				aos_fw_acl->log?"-l" : " ",aos_fw_acl->route );
//#ifdef MANGLE_FILTER
		//else	return -1;
		else
		{
			i+=snprintf(rulecmd+i,sizeof(rulecmd)-i," %s -j %s ", 
				aos_fw_acl->log?"-l" : " ", aos_fw_ruleaction[aos_fw_acl->action] );
		}
//#endif
			
		break;

#ifdef NORTON_NAT
	case AOS_FW_NAT_SNAT:
		i=0;
		if( aos_fw_acl->taddr1[0] == 0 )
			return -1;
		i+=snprintf(rulecmd+i,sizeof(rulecmd)-i,"ipt -t nat -%c %s -p %s ",
				aos_fw_ruleop[aos_fw_acl->op],
				"POSTROUTING",
				aos_fw_acl->proto);
		i+=aos_fw_net_parse( aos_fw_acl, &rulecmd[0]+i);
		i+=snprintf(rulecmd+i,sizeof(rulecmd)-i," -j SNAT --to-source %s-%s:%d-%d",
				aos_fw_acl->taddr1,
				aos_fw_acl->taddr2,
				aos_fw_acl->tport1,
				aos_fw_acl->tport2);
		break;

	case AOS_FW_NAT_BINAT:
		i=0;
		if( aos_fw_acl->taddr1[0] == 0 )
			return -1;
		i+=snprintf(rulecmd+i,sizeof(rulecmd)-i,"ipt -t nat -%c %s -p %s -m %s ",
				aos_fw_ruleop[aos_fw_acl->op],
				"PREROUTING",
				aos_fw_acl->proto,
				aos_fw_acl->proto );
		i+=aos_fw_net_parse( aos_fw_acl, &rulecmd[0]+i);
		i+=snprintf(rulecmd+i,sizeof(rulecmd)-i," -j DNAT --to-destination %s-%s:%d-%d",
				aos_fw_acl->taddr1,
				aos_fw_acl->taddr2,
				aos_fw_acl->tport1,
				aos_fw_acl->tport2);
		ret = iptables(rulecmd);

		i=0;
		memset(rulecmd,0,sizeof(rulecmd));
		i+=snprintf(rulecmd+i,sizeof(rulecmd)-i,"ipt -t nat -%c %s -p %s -m %s ",
				aos_fw_ruleop[aos_fw_acl->op],
				"POSTROUTING",
				aos_fw_acl->proto,
				aos_fw_acl->proto );
		i+=aos_fw_net_parse( aos_fw_acl, &rulecmd[0]+i);
		i+=snprintf(rulecmd,sizeof(rulecmd)-i," -j SNAT --to-source %s-%s:%d-%d",
				aos_fw_acl->taddr1,
				aos_fw_acl->taddr2,
				aos_fw_acl->tport1,
				aos_fw_acl->tport2);
		break;
#endif
	default:
		fprintf(stderr,"\nnot correct objtype for rule");
		break;
	}

	//fprintf(stderr,"\naos_fw_rule2=%s ",rulecmd);
	if( ret == 0 )
		ret = iptables(rulecmd);
	
	if( aos_fw_acl->dsttype == AOS_FW_NET_DOMAIN )
	{
		//snprintf(rulecmd,sizeof(rulecmd)-1,"pd -N %s",aos_fw_acl->dstdomain);
		//fprintf(stderr,"\npolicydomain=%s ",rulecmd);
		//policydomain(rulecmd);
		policydomain('N',aos_fw_acl->dstdomain);
	}
	
	return ret;
	
}
