#ifndef _NAT_H
#define _NAT_H

#include "aos_defs.h"

enum AOS_FW_RULE_OP {AOS_FW_RULE_ADD=0,AOS_FW_RULE_INS,AOS_FW_RULE_DEL,AOS_FW_RULE_EDIT,AOS_FW_RULE_FLUSH,AOS_FW_RULE_POLICY};
enum AOS_FW_OBJ_TYPE {AOS_FW_FILTER,AOS_FW_NAT_DNAT,AOS_FW_NAT_SNAT,AOS_FW_NAT_BINAT,AOS_FW_VIRTUAL_SERVER,AOS_FW_MANGLE};
enum AOS_FW_RULE_ACTION {AOS_FW_ACCEPT=0,AOS_FW_DROP,AOS_FW_REJECT,AOS_FW_LIMIT};
enum AOS_FW_NET_TYPE {AOS_FW_NET_ANY=0,AOS_FW_NET_IP,AOS_FW_NET_SUBNET,AOS_FW_NET_SELF,AOS_FW_NET_DOMAIN,AOS_FW_NET_RANGE,AOS_FW_NET_LIST};
enum AOS_FW_PORT_TYPE {AOS_FW_PORT_ANY=0,AOS_FW_PORT_SINGLE,AOS_FW_PORT_RANGE,AOS_FW_PORT_LIST};
enum { OP_ADD=0,OP_INS=1,OP_DEL=2,OP_EDIT=3, OP_LAST };
enum { CHAIN_IN=0, CHAIN_OUT=1, CHAIN_ME=2, CHAIN_LAST };
enum { PROTO_IP=0,PROTO_TCP,PROTO_UDP,PROTO_ICMP,PROTO_IGMP,PROTO_AH,PROTO_ESP,PROTO_LAST };

#define AOS_FW_PROTO_LEN	8
#define AOS_FW_IP_LEN		16
#define AOS_FW_RULE_SRC_LEN     48
#define AOS_FW_RULE_PORT_LEN    16
#define AOS_FW_RULE_ACTION_EXT  24
#define AOS_FW_RULE_MATCH_EXT   24
#define AOS_FW_RULE_CMD_EXT     24
#define AOS_FW_CHAIN_LEN	16
#define AOS_FW_RULE_DOMAIN_LEN	100

#define OPT_NONE	0x00000U
#define OPT_NUMERIC	0x00001U
#define OPT_SOURCE	0x00002U
#define OPT_DESTINATION	0x00004U
#define OPT_PROTOCOL	0x00008U
#define OPT_JUMP	0x00010U
#define OPT_VERBOSE	0x00020U
#define OPT_EXPANDED	0x00040U
#define OPT_VIANAMEIN	0x00080U
#define OPT_VIANAMEOUT	0x00100U
#define OPT_FRAGMENT    0x00200U
#define OPT_LINENUMBERS 0x00400U
#define OPT_COUNTERS	0x00800U
#define NUMBER_OF_OPT	12

#define Enable_NAT 1
#define Add_MDMZ 2
#define Del_MDMZ 3
#define Add_MTCPMSS 4
#define Del_MTCPMSS 5
#define Add_Virtrual_Server 6
#define Del_Virtrual_Server 7
#define HTTP_Server 8

#define Add_MDefense 9
#define Del_MDefense 10
#define Add_MHide 11
#define Del_MHide 12

#define Add_Alt_Port 13
#define Del_Alt_Port 14

#define Add_FireWall 15
#define Del_FireWall 16

#define Add_UPnP_Port 17
#define Del_UPnP_Port 18

#define CMD_NONE		0x0000U
#define CMD_INSERT		0x0001U
#define CMD_DELETE		0x0002U
#define CMD_DELETE_NUM		0x0004U
#define CMD_REPLACE		0x0008U
#define CMD_APPEND		0x0010U
#define CMD_LIST		0x0020U
#define CMD_FLUSH		0x0040U
#define CMD_ZERO		0x0080U
#define CMD_NEW_CHAIN		0x0100U
#define CMD_DELETE_CHAIN	0x0200U
#define CMD_SET_POLICY		0x0400U
#define CMD_CHECK		0x0800U
#define CMD_RENAME_CHAIN	0x1000U
#define NUMBER_OF_CMD	13

#define NFC_BYPASS		0x10000

struct Enable_nat
{
	char interface[8];//"ppp0"
};

struct MTCPMSS
{
	char interface[8];//"ppp0"
};

struct winDMZ
{
	//char interface[8];//"eth0"
	//char out_ip[16];
	char ip[16];
}__attribute__ ((aligned (4)));

struct Virtual_Server
{
	//char interface[8];//"eth0"
	char proto[8];//"tcp","udp","both"
	char sport[8];//start port
	char eport[8];//end port
	char ip[16]; //target ip
}__attribute__ ((aligned (4)));

struct Defense
{
	char mrate[48];//  xxx/minute xxx/second  xxx/hour xxx/day
}__attribute__ ((aligned (4)));

struct Hide
{
	char mproto[8];
}__attribute__ ((aligned (4)));

struct Alt_Port
{
	char mport[8];
}__attribute__ ((aligned (4)));

struct Upnp_Port
{
	char Proto[4];	
	char ExtIP[16];
	int ExtPort;	
	char IntIP[16];
	int IntPort;
}__attribute__ ((aligned (4)));

#ifdef WINBOND_ORIG
struct aos_fw_rule
{
	uint32 src,dst;
	uint32 smsk,dmsk;
	uint16 number;		/* rule number */
	u_char proto;		/* ip protocol */
	u_char sporttype;	/* for tcp/udp */
	u_char dporttype;	/* port type: single,range,list */
	uint16 sports[2];
	uint16 dports[2];
	u_char icmptypes[IP_FW_ICMPTYPES_DIM]; /* icmp types bitmap */
	u_char comment[16];
	u_char action;

}__attribute__ ((aligned (4)));
#else

struct aos_fw_acl 
{
	unsigned int objtype;	//AOS_FW_FILTER,AOS_FW_NAT_DNAT,AOS_FW_NAT_SNAT,AOS_FW_NAT_BINAT
	char src1[AOS_FW_IP_LEN];
	char src2[AOS_FW_IP_LEN];
	char dst1[AOS_FW_IP_LEN];
	char dst2[AOS_FW_IP_LEN];
	char taddr1[AOS_FW_IP_LEN];
	char taddr2[AOS_FW_IP_LEN];
	char chain[AOS_FW_CHAIN_LEN];
	unsigned char op;	//AOS_FW_RULE_INS,AOS_FW_RULE_DEL,AOS_FW_RULE_APPEND,AOS_FW_RULE_FLUSH
	unsigned char srctype;
	unsigned char dsttype;
	unsigned char taddrtype;
	unsigned int sport1;
	unsigned int sport2;
	unsigned int dport1;
	unsigned int dport2;
	unsigned int tport1;
	unsigned int tport2;
	char proto[AOS_FW_PROTO_LEN];
	unsigned int icmptype;
	unsigned char action;
	unsigned char log;
	u_char route[IFNAMSIZ+1];
	u_char dstdomain[64];

}__attribute__ ((aligned (4)));
#endif


int Add_Upnp_Port(struct Upnp_Port *uport);
int Del_Upnp_Port(struct Upnp_Port *uport);

int Add_DMZ(struct winDMZ *dmz);
int Del_DMZ(struct winDMZ *dmz);

int Add_virtual_server(struct Virtual_Server *vs);
int Del_virtual_server(struct Virtual_Server *vs);

int Add_TCPMSS();
int Del_TCPMSS();

int Add_Defense(struct Defense *mdf);
int Del_Defense(struct Defense *mdf);

int Add_Hide();
int Del_Hide();

int Reject_HTTP_Server();

int Add_Port(struct Alt_Port *wport);
int Del_Port(struct Alt_Port *wport);

unsigned long GetLocalIP(char *interface);

enum aos_fw_chain_type { PREROUTING=0, INPUT, FORWARD, OUTPUT, POSTROUTING, AOS_FW_CHAIN_MAX };

int aos_fw_add_rule();
int aos_fw_del_rule();
int aos_fw_update_rule();
int aos_fw_rule( struct aos_fw_acl *);
int iptables(char *);

#endif
