#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <ctype.h>
#include <stdlib.h>
#include <getopt.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <linux/netfilter_ipv4/policy_domain.h>

#ifdef XX
static ip_pool_t high_nr(void)
{
	policydomain_request req;
	int reqlen = sizeof(req);

	req.op = IP_POOL_HIGH_NR;
	if (0 > getsockopt(fd, SOL_IP, SO_IP_POOL, &req, &reqlen)) {
		fprintf(stderr,
			"IP_POOL_HIGH_NR failed: %s\n",
			errno2msg(IP_POOL_HIGH_NR, errno));
		exit(1);
	}
	return req.index;
}

void do_policydomain_list(ip_pool_t pool)
{
	policydomain_request req;
	int reqlen = sizeof(req);
	u_int32_t first_ip;
	u_int32_t last_ip;

	req.op = IP_POOL_LOOKUP;
	req.index = pool;
	req.addr = req.addr2 = 0;
	reqlen = sizeof(req);
	if (0 == getsockopt(fd, SOL_IP, SO_IP_POOL, &req, &reqlen)) {
		struct in_addr ina;
		ina.s_addr = req.addr;
		printf("%s %s", get_name(req.index), inet_ntoa(ina));
		ina.s_addr = req.addr2;
		printf(" %s", inet_ntoa(ina));
		first_ip = ntohl(req.addr);
		last_ip = ntohl(req.addr2);
		if (flag_u) {
			req.op = IP_POOL_USAGE;
			req.index = pool;
			reqlen = sizeof(req);
			if (0 == getsockopt(fd, SOL_IP, SO_IP_POOL,
						&req, &reqlen)) {
				printf(" %d %d", req.addr, req.addr2);
			} else {
				printf(" - -");
			}
		}
		printf("\n");

	}
}

void do_policydomain_list_all(void)
{
	ip_pool_t i, highest = high_nr();

	for (i=0; i<=highest; i++)
		do_policydomain_list(i);
}

int do_policydomain_statistic(void)
{
	policydomain_request req;
	int reqlen = sizeof(req);

	req.op = POLICYDOMAIN_HIGH_NR;
	if (0 > getsockopt(fd, SOL_IP, SO_POLICYDOMAIN, &req, &reqlen)) {
		fprintf(stderr,	"statistic failed\n");
	}
	return req.ans;
}

int do_policydomain_flush(void)
{
	policydomain_request req;
	int reqlen = sizeof(req);

	req.op = POLICYDOMAIN_FLUSH_DOMAIN;
	if (0 > getsockopt(fd, SOL_IP, SO_POLICYDOMAIN, &req, &reqlen)) {
		fprintf(stderr,	"flush failed\n");
	}
	return req.ans;
}

int do_policydomain_flush(void)
{
	policydomain_request req;
	int reqlen = sizeof(req);

	req.op = POLICYDOMAIN_FLUSH_DOMAIN;
	if (0 > getsockopt(fd, SOL_IP, SO_POLICYDOMAIN, &req, &reqlen)) {
		fprintf(stderr,	"flush failed\n");
	}
	return req.ans;
}

int do_policydomain_list_all(void)
{
	policydomain_request req;
	int reqlen = sizeof(req);

	req.op = POLICYDOMAIN_LIST_ALL;
	if (0 > getsockopt(fd, SOL_IP, SO_POLICYDOMAIN, &req, &reqlen)) {
		fprintf(stderr,	"list failed\n");
	}
	return req.ans;
}

int do_policydomain_del(char *domain)
{
	policydomain_request req;
	int reqlen = sizeof(req);

	req.op = POLICYDOMAIN_DEL_DOMAIN;
	strncpy(req.domain, domain, sizeof(req.domain));
	if (0 > getsockopt(fd, SOL_IP, SO_POLICYDOMAIN, &req, &reqlen)) {
		fprintf(stderr,"del failed\n");
	}
	return req.ans;
}

int do_policydomain_new(char *domain)
{
	policydomain_request req;
	int reqlen = sizeof(req);

	req.op = POLICYDOMAIN_ADD_DOMAIN;
	strncpy(req.domain, domain, sizeof(req.domain));
	if (0 > getsockopt(fd, SOL_IP, SO_POLICYDOMAIN, &req, &reqlen)) {
		fprintf(stderr,"new failed\n");
	}
	return req.ans;
}
#endif

int
policydomain( char action, char *domain )
{
	int fd=-1;
	
	policydomain_request req;
	int reqlen = sizeof(req);
        
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0) {
		fprintf(stderr, "cannot get DGRAM socket: %s\n",
			strerror(errno));
		return -1;
	}

	switch(action)
	{
	case 'N':
	case 'I':
	case 'A':
		req.op = POLICYDOMAIN_ADD_DOMAIN;
		strncpy(req.domain, domain, sizeof(req.domain));
		break;
	case 'D':
		req.op = POLICYDOMAIN_DEL_DOMAIN;
		strncpy(req.domain, domain, sizeof(req.domain));
		break;
	case 'F':
		req.op = POLICYDOMAIN_FLUSH_DOMAIN;
		break;
	case 'C':
		req.op = POLICYDOMAIN_CLEAN_DOMAIN;
		break;
	case 'M':
		req.op = POLICYDOMAIN_MARK_DOMAIN;
		break;
	case 'L':
		req.op = POLICYDOMAIN_LIST_ALL;
		break;
	case 'S':
		req.op = POLICYDOMAIN_HIGH_NR;
		break;
	default:
		fprintf(stderr,"policydomain: no command specified\n");
		close(fd);
		return -1;
	}

	if (0 > getsockopt(fd, SOL_IP, SO_POLICYDOMAIN, &req, &reqlen)) {
		fprintf(stderr,	" failed\n");
	}

	close(fd);
	
	return req.ans;
}

#ifndef EMBED

int do_policydomain(int argc, char **argv);

static void policy_usage(char *msg)
{
	if (msg) fprintf(stderr, "ERROR: %s\n", msg);
	fprintf(stderr,
"Usage: policydomain -N domain\n"
"\t\tadd a new domain\n"
"\tpolicydomain -D domain\n"
"\t\tdel a domain\n"
"\tpolicydomain -F domain\n"
"\t\tflush all domain\n"
"\tpolicydomain -M\n"
"\t\tmark all domain to be purged\n"
"\tpolicydomain -C\n"
"\t\tpurge all marked domain\n" );
	return;
}

int main(int argc, char **argv)
{
	int fd=-1;
	int ret = -1;
	int opt;

	while (EOF != (opt=getopt( argc, argv, "N:D:MCFLS")))
	switch(opt) {
		case 'F':
			ret = policydomain( opt, "");
			break;
		case 'M':
			ret = policydomain( opt, "");
			break;
		case 'C':
			ret = policydomain( opt, "");
			break;
		case 'N':
			ret = policydomain( opt, optarg );
			break;
		case 'D':
			ret = policydomain( opt, optarg );
			break;
		case 'L':
			ret = policydomain( opt, "");
			break;
		case 'S':
			ret = policydomain( opt, "");
			break;
		case 'h':
		default:
			policy_usage("policydomain: no operation specified");
	}
	
	return ret;
}
#endif

