/****************************************************************************
* Copyright (c) 2002 - 2003 Windond Electronics Corp.
* All rights reserved.
*
* FILE NAME: natlib.c
*
* DESCRIPTION: NAT Functions of Nucleus Ethernet Library for W90N740 NAT.
*
* AUTHOR: Jen-Hao Tsai, PC30
****************************************************************************/
//#include "nucleus.h"
//#include "externs.h"
//#include "tcp_errs.h"
#define  IP_ADDR_SIZE 4
#define  MAC_ADDR_SIZE 6
#define   UINT16  unsigned short
#define   UINT32  unsigned long
#define   UINT8   unsigned char

#include "740_defs.h"
//#include "740_nat.h"



/* Enable NAT Function */
void NAT_Enable()
{
 	NATCMD = NATEN;
}


/* Disable NAT Function */
void NAT_Disable()
{
 	NATCMD = 0;
}


/* Clear NAT entry Hit Counter */
void NAT_HitCount_Clear(int entry)
{
 	NATCCLR(entry / 16) = 1 << ((entry % 16) + 16);
}


/* Clear All NAT Hit Counters */
void NAT_HitCount_ClearAll()
{
 	int i = 0;
 
 	for (i = 0; i < 4; i++)	
    	NATCCLR(i) = 0xffff0000;
}


/* Clear NAT entry Processed Hit Counter */
void NAT_ProcCount_Clear(int entry)
{
 	NATCCLR(entry / 16) = 1 << (entry % 16);
}


/* Clear All NAT Processed Hit Counters */
void NAT_ProcCount_ClearAll()
{
 	int i = 0;
 
 	for (i = 0; i < 4; i++)	
    	NATCCLR(i) = 0x0000ffff;
}


/* Read NAT entry Hit Counter */
int  NAT_HitCount_Read(int entry)
{
 	int count;

 	count = (int)(NATCFG(entry) >> 16);

 	return count;
}


/* Get the entry with Minimum NAT Hit Counter */
/*
int  NAT_HitCount_MinEntry()
{
 	int count, min, entry, i;

 	entry = 0;
 	min = (int)(NATCFG(entry) >> 16);

 	for (i = 1; i < NATTbl_Entry_Num; i++)
    {
     	count = (int)(NATCFG(i) >> 16);
     	if (count < min)
       	{
        	min = count;
        	entry = i;
       	}
    }

 	return entry;
}
*/

/* Read NAT entry Processed Hit Counter */
int  NAT_ProcCount_Read(int entry)
{
 	int count;

 	count = (int)((NATCFG(entry) & 0x00000700) >> 8);

 	return count;
}


/* Get the entry with Minimum NAT Processed Hit Counter */
/*
int  NAT_ProcCount_MinEntry()
{
 	int count, min, entry, i;

 	entry = 0;
 	min = (int)((NATCFG(entry) & 0x00000700) >> 8);

 	for (i = 1; i < NATTbl_Entry_Num; i++)
    {
     	count = (int)((NATCFG(i) & 0x00000700) >> 8);
     	if (count<min)
       	{
       		min = count;
        	entry = i;
       	}
    }

 return entry;
}
*/

/* Enable entry x of NAT Table */
void NAT_Entry_Config(int entry, UINT32 value)
{
 	NATCFG(entry) = value;
}


/* Enable entry x of NAT Table */
void NAT_Entry_Enable(int entry)
{
 	NATCFG(entry) |= NAT_ExEN;
}


/* Disable entry x of NAT Table */
void NAT_Entry_Disable(int entry)
{
 	NATCFG(entry) &= ~NAT_ExEN;
}


/* IP Address Comparison Enable at entry x */
void NAT_AddrCompare_Enable(int entry)
{
 	NATCFG(entry) |= NAT_AxCE;
}


/* IP Address Comparison Disable at entry x */
void NAT_AddrCompare_Disable(int entry)
{
 	NATCFG(entry) &= ~NAT_AxCE;
}


/* Port Number Comparison Enable at entry x */
void NAT_PortCompare_Enable(int entry)
{
 	NATCFG(entry) |= NAT_PxCE;
}


/* Port Number Comparison Disable at entry x */
void NAT_PortCompare_Disable(int entry)
{
 	NATCFG(entry) &= ~NAT_PxCE;
}


/* IP Address Replacement Enable at entry x */
void NAT_AddrReplace_Enable(int entry)
{
 	NATCFG(entry) |= NAT_AxRE;
}


/* IP Address Replacement Disable at entry x */
void NAT_AddrReplace_Disable(int entry)
{
 	NATCFG(entry) &= ~NAT_AxRE;
}


/* Port Number Replacement Enable at entry x */
void NAT_PortReplace_Enable(int entry)
{
 	NATCFG(entry) |= NAT_PxRE;
}


/* Port Number Replacement Disable at entry x */
void NAT_PortReplace_Disable(int entry)
{
 	NATCFG(entry) &= ~NAT_PxRE;
}


/* Inverse Comparison and  Replacement at entry x */
void NAT_Inverse_Enable(int entry)
{
 	NATCFG(entry) |= NAT_Inverse;
}


/* Disable Inverse Comparison and  Replacement at entry x */
void NAT_Inverse_Disable(int entry)
{
 	NATCFG(entry) &= ~NAT_Inverse;
}


/* Packet Discard at entry x */
void NAT_Discard_Enable(int entry)
{
 	NATCFG(entry) |= NAT_Discard;
}


/* Disable Packet Discard at entry x */
void NAT_Discard_Disable(int entry)
{
 	NATCFG(entry) &= ~NAT_Discard;
}


/* Packet Discard at entry x */
void NAT_NOP_Enable(int entry)
{
 	NATCFG(entry) |= NAT_NOP;
}


/* Disable Packet Discard at entry x */
void NAT_NOP_Disable(int entry)
{
 	NATCFG(entry) &= ~NAT_NOP;
}


/* External MAC Address Configuration */
void NAT_EXMAC_Config(UINT8 *addr)
{
 	UINT32 	addrtemp;
 	int 	i;

 	addrtemp = 0;
 	for (i = 0; i < (int)MAC_ADDR_SIZE - 2; i++)
    addrtemp = (addrtemp << 8) | addr[i] ;
 	EXMACM = addrtemp;

 	addrtemp = 0;
 	for (i = (int)(MAC_ADDR_SIZE - 2); i < (int)MAC_ADDR_SIZE; i++)
    	addrtemp = (addrtemp << 8) | addr[i] ;
 	addrtemp = (addrtemp << 16) ;
 	EXMACL = addrtemp;
}


/* Internal MAC Address Configuration */
void NAT_INMAC_Config(UINT8 *addr)
{
 	UINT32 addrtemp;
 	int i;

 	addrtemp = 0;
 	for (i = 0; i < (int)MAC_ADDR_SIZE - 2; i++)
    	addrtemp = (addrtemp << 8) | addr[i] ;
 	INMACM = addrtemp;

 	addrtemp = 0;
 	for (i = (int)(MAC_ADDR_SIZE - 2); i < (int)MAC_ADDR_SIZE; i++)
    	addrtemp = (addrtemp << 8) | addr[i] ;
 	addrtemp = (addrtemp << 16) ;
 	INMACL = addrtemp;
}


/* Local Station MAC Address Configuration */
void NAT_LSMAC_Config(int entry, UINT8 *addr)
{
 	UINT32 addrtemp;
 	int i;

 	addrtemp = 0;
 	//printk("MAC2\n");
 	for (i = 0; i < (int)MAC_ADDR_SIZE - 2; i++)
    	addrtemp = (addrtemp << 8) | addr[i] ;
 	//printk("MAC3\n");
 	LSMACM(entry) = addrtemp;
	//printk("NAT_LSMAC %x\n",addrtemp);
 	addrtemp = 0;
 	for (i = (int)(MAC_ADDR_SIZE - 2); i<(int)MAC_ADDR_SIZE; i++)
    	addrtemp = (addrtemp << 8) | addr[i] ;
 	addrtemp = (addrtemp << 16) ;
 	//printk("MAC2\n");
 	LSMACL(entry) = addrtemp;
 	//printk("MAC1\n");
}


/* Remote Station MAC Address Configuration */
void NAT_RSMAC_Config(int entry, UINT8 *addr)
{
 	UINT32 addrtemp;
 	int i;

 	addrtemp = 0;
 	for (i = 0; i< (int)MAC_ADDR_SIZE - 2; i++)
    	addrtemp = (addrtemp << 8) | addr[i] ;
 	RSMACM(entry) = addrtemp;

 	addrtemp = 0;
 	for (i = (int)(MAC_ADDR_SIZE - 2); i < (int)MAC_ADDR_SIZE; i++)
    	addrtemp = (addrtemp << 8) | addr[i] ;
 	addrtemp = (addrtemp << 16) ;
 	RSMACL(entry) = addrtemp;
}


/* Masquerading IP Address Configuration at entry x */
void NAT_MASAD_Config(int entry, UINT8 *addr)
{
 	UINT32 addrtemp;
 	int i;

 	addrtemp = 0;
 	for (i = 0; i < (int)IP_ADDR_SIZE; i++)
    	addrtemp = (addrtemp << 8) | addr[i] ;
 	MASAD(entry) = addrtemp;
}


/* Masquerading Port Number Configuration at entry x */
void NAT_MASPN_Config(int entry, UINT16 *port)
{
 	MASPN(entry) = *port;
}


/* Local Station IP Address Configuration at entry x */
void NAT_LSAD_Config(int entry, UINT8 *addr)
{
 	UINT32 	addrtemp;
 	int 	i;

 	addrtemp = 0;
 	for (i = 0; i < (int)IP_ADDR_SIZE; i++)
    {
    		//printk("%x ",addr[i]);
    		addrtemp = (addrtemp << 8) | addr[i] ;
    }		
 	
 	LSAD(entry) = addrtemp;
}


/* Local Station Port Number Configuration at entry x */
void NAT_LSPN_Config(int entry, UINT16 *port)
{
 	LSPN(entry) = *port;
}

void NAT_Entry_ClearAll()
{
  unsigned  long i,	* p=0xFFF06800;
  for(i=0;i<64*8;i++)
    *(p+i)=0;
}	
