#ifndef _W740NAT_H_
#define _W740NAT_H_

#ifdef __KERNEL__
#include <linux/netfilter_ipv4/ip_conntrack.h>
#endif

//#include <linux/netfilter_ipv4/ip_conntrack_core.h>

#define _NATA_BASE	0xfff06000

//NATA Control and status Register
#define NATCMD		(_NATA_BASE+0)  //NAT Command Register
#define NATCCLR		(_NATA_BASE+4)  //NAT Counter clear Register
#define NATCFG0		(_NATA_BASE+40) //NAT Entry 0 Configuration Register
//Address lookup and Replacement Registers
#define EXMACM		(_NATA_BASE+100) 
#define EXMACL		(_NATA_BASE+104)
#define INMACM		(_NATA_BASE+108)
#define INMACL		(_NATA_BASE+10c)
#define NATA_ENTRY_START (_NATA_BASE+200)
#define NATA_ENTRY_SIZE	 64
//NAT command REgister 
#define NATCMD_NATEN     1

//NAT Entry x Configuration Registers
#define NATCFG_ExEn     1  //Entry x Comparison Enble bit
#define NATCFG_AxCE	0x2 //	(1<<1) //ip Address Comparion  Enable at entry x
#define NATCFG_PxCE	0x4 //	(1<<2)
#define NATCFG_AxRE	0x8 //	(1<<3)
#define NATCFG_PxRE	0x10 //	(1<<4) //Port Number Repacement Enalbe at Entry x
#define NATCFG_Inverse	0x20 //(1<<5) //Inverse Comparion adn Replacement bit
#define NATCFG_Discard	0x40 //(1<<6)
#define NATCFG_NOP	0x80 //	(1<<7)
#define NATCFG_CNTx	0x100	//(1<<8)
#define NATCFG_EHCNTx	0x10000	//(1<<16)


#define NATEN       0x00000001  // NAT Enable bit
// NAT Entry Configuration Register(NATCFG)
#define NAT_NOP     0x00000080  // No Operation bit
#define NAT_Discard 0x00000040  // Packet Discard bit
#define NAT_Inverse 0x00000020  // Inverse Comparison and Replacement bit
#define NAT_PxRE    0x00000010  // Port Number Replacement Enable
#define NAT_AxRE    0x00000008  // IP Address Replacement Enable
#define NAT_PxCE    0x00000004  // Port Number Comparison Enable
#define NAT_AxCE    0x00000002  // IP Address Comparison Enable
#define NAT_ExEN    0x00000001  // Entry Enable bit


#define NATA_UDP  2
#define NATA_TCP  1

//////////////
#define SECS *HZ
#define MINS * 60 SECS
#define HOURS * 60 MINS
#define DAYS * 24 HOURS
#define ONEDAY 86400 SECS

/* PPPoE Header Struct */
struct NATA_PPPOE_PPP_HEADER_REQ
{
	 
	unsigned char  ver : 4;
	unsigned char  type : 4;
	unsigned char  code;
	unsigned short sid;
	unsigned short length;
	
	unsigned short protocl;
	unsigned  long PPPoE_rmad;
	char  remote_mac[8];
	unsigned long  start_pppoe;

#ifdef ACCELERATE_SESSION1_ONLY

#else
	unsigned char session;
#endif

};

struct NATA_PPPOE_PPP_HEADER
{
	 
	unsigned char  ver : 4;
	unsigned char  type : 4;
	unsigned char  code;
	unsigned char  sid;
	unsigned short length;
	unsigned short protocl;
	//unsigned  long PPPoE_rmad;
};

/*
*  NATA Entry struct 
*/
#define NATA_ENTRY_FLAG_EMPTY 0
#define NATA_ENTRY_FLAG_HEAD  1
#define NATA_ENTRY_FLAG_RESERVE 2
#define NATA_ENTRY_FLAG_NONE	3
typedef struct _NATA_ENTRY
{
	unsigned long flag;
	unsigned long index;
	unsigned long time;
	
	unsigned long rsad;
	unsigned long rspn;
	
	unsigned long msad;
	unsigned long mspn ; //NAT Masquerading Port Number Entry 	
	
	unsigned long lsad  ;	
	unsigned long lspn  ;//NAT  local Port Number Entry
	
	//masq ip;
		
	unsigned  char	LocMAC[8];	
	unsigned  char	RsMAC[8];
	struct _NATA_ENTRY  * next;
	struct _NATA_EMTRY  * prev;
	struct ip_conntrack  *ct;
	struct sk_buff *skb;
	int  tcp_times;	
	int  proto;

#ifdef ACCELERATE_SESSION1_ONLY

#else

#define	PPPOE_MAX_SESSION	5
	unsigned char	session; /* session index */
#endif

}NATA_ENTRY;

#ifdef __KERNEL__
struct   _NATA_Table
{
	int  size;  
	unsigned long num_use;     
	NATA_ENTRY  * head;
	
	NATA_ENTRY  * empty_list;
	NATA_ENTRY  * reserve_list;
	
	unsigned long valid;   //ã
	unsigned long LanIp;
	unsigned long WanIp;
	unsigned char Lanmac[ETH_ALEN]; //ڲMAC 
	unsigned char Wanmac[ETH_ALEN]; //ⲿMAC
	
#ifdef ACCELERATE_SESSION1_ONLY
	unsigned short PPPoE_SessionID;
	unsigned short PPPoE_PPP_Protocol;
	
	unsigned char PPPoE_rmmac[ETH_ALEN]; //ⲿMAC
#else
	unsigned short PPPoE_SessionID[PPPOE_MAX_SESSION];
	unsigned short PPPoE_PPP_Protocol;

	unsigned char PPPoE_rmmac[PPPOE_MAX_SESSION][ETH_ALEN]; //ⲿMAC
#endif

	unsigned   short PPPoE_rmad;
	struct timer_list timer;	
	unsigned long  start_pppoe;
	NATA_ENTRY    table[64];        //Entry data
};
extern  struct _NATA_Table  NATA_Table; 
#endif

struct nata_req
{
    int cmd;   
    union{
    	struct NATA_PPPOE_PPP_HEADER_REQ pppoe_header;
    	NATA_ENTRY  entry;
    	unsigned long wanip;
    	unsigned long lanip;
	unsigned long duplex;
    	char buf[0];
    	char mac[6];
    }type1;
};
/* NATA command type */
#define NATA_ADD_ENTRY 0x01
#define NATA_DEL_ENTRY 0x02
#define NATA_SET_PPPOE_HEADER 0x03
#define NATA_GET_STATUS	0x04
#define NATA_START		0x05
#define NATA_STOP		0x06
#define NATA_RESET		0x07

#define NATA_SET_MAC0		0x10
#define NATA_SET_MAC1		0x11

/*-- norton */
#define NATA_SET_WAN_DUPLEX			0x15
#define NATA_WAN_100_FULL_DUPLEX		0x16
#define NATA_WAN_100_HALF_DUPLEX		0x17
#define NATA_WAN_10_FULL_DUPLEX		0x18
#define NATA_WAN_10_HALF_DUPLEX		0x19
#define NATA_WAN_AUTO_NEG			0x20

/* norton-- */


#define MAC0_FLASH_ADDR  ((char *) 0x7F010008)
#define MAC1_FLASH_ADDR  ((char *) 0x7F010014)

#define LAN 0
#define WAN 1
#define MAC_LEN 6
#define PPPOE_HEADER_LEN 8


#define TCP_TIMEOUT		10
#define UDP_TIMEOUT	3		/* special workaround hack for msn conference call */
#define IPT_TCP_TIMEOUT	3600
#define IPT_UDP_TIMEOUT	600
#define IPT_DNS_TIMEOUT	15

#define IPT_TCP_CLOSE_TIMEOUT	30	/* close */	/* norton */
#define IPT_TCP_RST_TIMEOUT	10	/* close */

#define DNS_PORT	53

extern void disable_entry(int i);
extern int NATA_Stop (void);
extern int NATA_Start (void);
extern int NATA_Add_Entry(NATA_ENTRY *entry);
extern int NATA_Del_Entry(NATA_ENTRY *entry);
extern int NATA_SetMAC(int , const char *);

#endif
