#include "usb_prism_wireless.h"

/*
 * Version Information
 */
#define DRIVER_VERSION "v1.00 for Linux 2.4"
#define EMAIL "ypxin@winbond.com.tw"
#define DRIVER_AUTHOR "ypxin@winbond.com.tw"
#define DRIVER_DESC "Prism USB Wireless Networking Driver"

/* table of devices that work with this driver */
static struct usb_device_id uprism_table [] = {
	{ USB_DEVICE(PRISM_VENDOR, PRISM_PRODUCT) },
	{ }					/* Terminating entry */
};

static struct usb_driver uprism_driver = {
	name:		"uprism",
	id_table:	uprism_table,
	probe:		uprism_probe,
	disconnect:	uprism_disconnect
};



/****************************************************************************
 *
 *  USB routines
 *
 ***************************************************************************/
static void *
uprism_probe(struct usb_device *dev, unsigned int ifnum,
	    const struct usb_device_id *id)
{
	struct usb_interface_descriptor *interface;
	struct usb_endpoint_descriptor *endpoint;
	struct usb_W99683 *w683;
	int i;
	int registered = 0;

	KDEBUG("probing for device..., %d\n", ifnum);
#ifdef UPDEBUG
	usb_show_device(dev);
#endif
#if 0	
	/* We don't handle multi-config cameras */
	if (dev->descriptor.bNumConfigurations != 1)
		return NULL;

	interface = &dev->actconfig->interface[ifnum].altsetting[0];
	endpoint = interface[ifnum].endpoint;
	
	/*printk("endponts: %d\n", interface->bNumEndpoints);
	for(i = 0; i < interface->bNumEndpoints; i++)
	{
		if(IS_EP_BULK_IN(endpoint[i]))
			printk("get bulk in: %d\n", i);
	}*/
	/* Checking vendor/product should be enough, but what the hell */
/*	if (interface->bInterfaceClass != 0xFF)
		return NULL;
	if (interface->bInterfaceSubClass != 0x00)
		return NULL;*/

	/* Since code below may sleep, we use this as a lock */
	MOD_INC_USE_COUNT;

	if ((w683 = kmalloc(sizeof(*w683), GFP_KERNEL)) == NULL) {
		err("couldn't kmalloc ov struct");
		goto error_out;
	}

	w683_debug = w683;
	memset(w683, 0, sizeof(*w683));
	
	//printk("probe: user: %d\n", w683->user);
	w683->dev = dev;
	w683->iface = interface->bInterfaceNumber;
	w683->user = 0;
	//printk("probe: user: %d\n", w683->user);
	if ((dev->descriptor.idVendor != W99683_VENDOR) ||
	    (dev->descriptor.idProduct != W99683i_PRODUCT)) {
		err("Unknown product ID 0x%x or Vendor ID 0x%x", dev->descriptor.idProduct, dev->descriptor.idVendor);
		goto error_dealloc;
	}

	init_waitqueue_head(&w683->wq);

	init_MUTEX(&w683->param_lock);
	init_MUTEX(&w683->lock);

	memcpy(&w683->vdev, &vdev_template, sizeof(vdev_template));
	w683->vdev.priv = w683;

	/* Use the next available one */
	if (video_register_device(&w683->vdev, VFL_TYPE_GRABBER, video_nr) < 0) {
		err("video_register_device failed");
		goto error;
	}

	info("Device registered on minor %d", w683->vdev.minor);
	
	w683->rawbuf = kmalloc(W683BUF_SIZE, GFP_KERNEL);
	if (!w683->rawbuf)
		goto error;
#ifdef CONFIG_BOARD_W90N740	
	w683->rawbuf = (char *)((unsigned long)w683->rawbuf|0x80000000);
#endif		
	ioctlbuf = kmalloc(64*sizeof(unsigned char), GFP_KERNEL);
	if (!ioctlbuf)
		goto error1;
#ifdef CONFIG_BOARD_W90N740	
	ioctlbuf = (unsigned char *)((unsigned long)ioctlbuf|0x80000000);	
#endif	
	MOD_DEC_USE_COUNT;
     	return w683;
error1:
	kfree(w683->rawbuf);
error:
	err("Camera initialization failed");

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
	/* Safe to call even if entry doesn't exist */
#endif

	usb_driver_release_interface(&W99683_driver,
		&dev->actconfig->interface[w683->iface]);

error_dealloc:
	if (w683) {
		kfree(w683);
		w683 = NULL;
	}

error_out:
	MOD_DEC_USE_COUNT;
#endif
	return NULL;
}

static void
uprism_disconnect(struct usb_device *dev, void *ptr)
{
}


/******************************************************************************
 *
 *  Module Routines
 *
 ******************************************************************************/
static int __init 
usb_prism_init(void)
{
#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
        proc_prism_create();
#endif
	printk("init start user: %x\n", &uprism_driver);
	if(&uprism_driver == NULL)
		return -1;

	if (usb_register(&uprism_driver) < 0) {
		printk("Can't register prism usb driver\n");
		return -1;
	}
	
	info(DRIVER_VERSION " : " DRIVER_DESC);
	return 0;
}

static void __exit 
usb_prism_exit(void)
{
	usb_deregister(&uprism_driver);
	info("driver deregistered");

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
        proc_prism_destroy();
#endif
}

module_init(usb_prism_init);
module_exit(usb_prism_exit);
