#ifndef _IP_CONNTRACK_MSNMS_H
#define _IP_CONNTRACK_MSNMS_H
/* MSNMS tracking. */

#ifdef __KERNEL__

#include <linux/netfilter_ipv4/lockhelp.h>

/* Protects msnms part of conntracks */
DECLARE_LOCK_EXTERN(ip_msnms_lock);

#define MSNMS_PORT	1863

#endif /* __KERNEL__ */

enum MSN_INVITE { MSN_INVITE_NONE, MSN_WHITEBOARD_INVITE, MSN_WINMSG_FTP_INVITE, MSN_VIDEO_INVITE };

#define MIN_CONTEXT_LEN	20
#define MAX_PORTS 	8
#define WINMSG_FTP_PORT	6891
#define T120_PORT	1503
#define T120_TTL	60
#define MSN_SIP_TIMEOUT	3600*HZ
#define MSN_SIP_CLOSING	3*HZ

#define MAX_INVITE_NUM	30

enum ip_ct_msnms_type
{
	IP_CT_MSNMS_NONE,
	IP_CT_MSNMS_ACCEPT,
	IP_CT_MSNMS_REPLYACCEPT,
	IP_CT_MSNMS_CONTEXT,
	IP_CT_SIP_REQUEST,
	IP_CT_SIP_STATUS,
	IP_CT_MSNMS_REPLYCONTEXT,
	IP_CT_MSNMS_GUID,
	IP_CT_MSNMS_PORT,
	IP_CT_MSNMS_INVITECOOKIE,
	IP_CT_MSNMS_INVITE,
	IP_CT_MSNMS_CANCEL,
	IP_CT_MSNMS_REPLYCANCEL,
	IP_CT_MSNMS_CALLID
};

/* This structure is per expected connection */
struct ip_ct_msnms_expect
{
	/* We record seq number and length of msnms ip/port text here: all in
	 * host order. */

 	/* sequence number of IP address in packet is in ip_conntrack_expect */
	u_int32_t len;			/* length of IP address */
	u_int32_t ctx_len;		/* content-length */
	enum ip_ct_msnms_type msnmstype;/* PORT or PASV ? */

	char pattern[4][32];		/* pattern going to be replace */
	char newpattern[4][32];		/* new pattern */
	u_int16_t plen[4];	
	u_int16_t newplen[4];	

	u_int32_t cookie;		/* identifier for each invite */

};

/* This structure exists only once per master */
struct ip_ct_msnms_master {

	unsigned char is_msnp;		/* msn control protocol ? */
	unsigned char invite_type;	/* invite type */
	unsigned char inviter;		/* Am i invite? */
	unsigned char bye;		/* disconnect it */
	unsigned char proto;		/* protocol of this session */	
	u_int32_t cookie;		/* identifier for each invite */
	u_int32_t priv_addr;		/* internal ip address */
	u_int16_t priv_port; 		/* TCP port that was to be used */
	u_int32_t pub_addr;		/* public ip address */
	u_int16_t pub_port; 		/* public TCP port that was to be used */

	u_int32_t peer_addr;		/* real peer ip address */
	u_int16_t peer_port;
	
	u_int32_t callid;		/* sip call id : use for locate new/old ssession info */
	u_int16_t media_port_caller[2];	/* media port : m=application m=audio m=video */
	u_int16_t media_port_callee[2];	/* media port : m=application m=audio m=video */
	unsigned long ttl;		/* live time */
	u_int16_t index;		/* index of this record */
};

struct ip_ct_msnms_master *session_find(u_int32_t cookie);
struct ip_ct_msnms_master *session_add(struct ip_ct_msnms_master *rec);
int session_del(u_int32_t cookie);
void session_clean(struct ip_ct_msnms_master *rec);
extern struct ip_ct_msnms_master msnms_session[MAX_INVITE_NUM];

#endif /* _IP_CONNTRACK_MSNMS_H */
