#ifndef _IP_CONNTRACK_TUPLE_H
#define _IP_CONNTRACK_TUPLE_H

/* USE_RX3042_PPTP_PATCH  defined in config.h */
#ifdef CONFIG_IP_NF_PPTP
#ifndef USE_RX3042_PPTP_PATCH
#define USE_RX3042_PPTP_PATCH 1
#endif
#endif
//#define USE_MANIP_PROTO_PPTP_PATCH	1	//if enabled, virtual server port range will fail!!!!

/* A `tuple' is a structure containing the information to uniquely
  identify a connection.  ie. if two packets have the same tuple, they
  are in the same connection; if not, they are not.

  We divide the structure along "manipulatable" and
  "non-manipulatable" lines, for the benefit of the NAT code.
*/

/* The protocol-specific manipulable parts of the tuple: always in
   network order! */
union ip_conntrack_manip_proto
{
	/* Add other protocols here. */
#if defined(CONFIG_IP_NF_PPTP)
	u_int32_t all;
#else
	u_int16_t all; //test by norton u_int16_t all;
#endif
	struct {
#if defined(CONFIG_IP_NF_PPTP)  && defined(USE_MANIP_PROTO_PPTP_PATCH)
		u_int16_t dummy;
#endif
		u_int16_t port;
	} tcp;
	struct {
#if defined(CONFIG_IP_NF_PPTP)  && defined(USE_MANIP_PROTO_PPTP_PATCH)
		u_int16_t dummy;
#endif
		u_int16_t port;
	} udp;
	struct {
#if defined(CONFIG_IP_NF_PPTP)  && defined(USE_MANIP_PROTO_PPTP_PATCH)
		u_int16_t dummy;
#endif
		u_int16_t id;
	} icmp;
#if defined(CONFIG_IP_NF_PPTP)
	struct {
		u_int32_t key;
	} gre;
#endif

};

/* The manipulable part of the tuple. */
struct ip_conntrack_manip
{
	u_int32_t ip;
	union ip_conntrack_manip_proto u;
};

/* This contains the information to distinguish a connection. */
struct ip_conntrack_tuple
{
	struct ip_conntrack_manip src;

	/* These are the parts of the tuple which are fixed. */
	struct {
		u_int32_t ip;
		union {
			/* Add other protocols here. */
#if defined(CONFIG_IP_NF_PPTP)  && defined(USE_RX3042_PPTP_PATCH)
			u_int32_t	 all; //test by norton u_int64_t all;	//u_int32_t all;
			u_int16_t dummy16;
			u_int8_t dummy8;
#elif defined(CONFIG_IP_NF_PPTP) 
			u_int32_t all;
#else
			u_int16_t all; //test u_int16_t all;
#endif
			struct {
#if defined(CONFIG_IP_NF_PPTP)  && defined(USE_RX3042_PPTP_PATCH)
				//u_int32_t dummy32;
				u_int16_t dummy16;
#endif
				u_int16_t port;
			} tcp;
			struct {
#if defined(CONFIG_IP_NF_PPTP)  && defined(USE_RX3042_PPTP_PATCH)
				//u_int32_t dummy32;
				u_int16_t dummy16;
#endif
				u_int16_t port;
			} udp;
			struct {
#if defined(CONFIG_IP_NF_PPTP)  && defined(USE_RX3042_PPTP_PATCH)
				//u_int32_t dummy32;
				u_int16_t dummy16;
#endif
				u_int8_t type, code;
			} icmp;

#ifdef CONFIG_IP_NF_PPTP
			struct {
				u_int32_t key; //test u_int32_t key
				u_int16_t protocol;
				u_int8_t version;
			} gre;
#endif
		} u;

		/* The protocol. */
		u_int16_t protonum;
	} dst;
}__attribute__ ((packed));

enum ip_conntrack_dir
{
	IP_CT_DIR_ORIGINAL,
	IP_CT_DIR_REPLY,
	IP_CT_DIR_MAX
};

#ifdef __KERNEL__

#ifdef CONFIG_IP_NF_PPTP

#define DUMP_TUPLE(tp)						\
DEBUGP("tuple %p: %u %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u\n",	\
       (tp), (tp)->dst.protonum,				\
       NIPQUAD((tp)->src.ip), ntohl((tp)->src.u.all),		\
       NIPQUAD((tp)->dst.ip), ntohl((tp)->dst.u.all))

#define DUMP_TUPLE_RAW(x) 						\
	DEBUGP("tuple %p: %u %u.%u.%u.%u:0x%08x -> %u.%u.%u.%u:0x%08x\n",\
	(x), (x)->dst.protonum,						\
	NIPQUAD((x)->src.ip), ntohl((x)->src.u.all), 			\
	NIPQUAD((x)->dst.ip), ntohl((x)->dst.u.all))
#else

#define DUMP_TUPLE(tp)						\
DEBUGP("tuple %p: %u %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u\n",	\
       (tp), (tp)->dst.protonum,				\
       NIPQUAD((tp)->src.ip), ntohs((tp)->src.u.all),		\
       NIPQUAD((tp)->dst.ip), ntohs((tp)->dst.u.all))

#define DUMP_TUPLE_RAW(x) 						\
	DEBUGP("tuple %p: %u %u.%u.%u.%u:0x%08x -> %u.%u.%u.%u:0x%08x\n",\
	(x), (x)->dst.protonum,						\
	NIPQUAD((x)->src.ip), ntohs((x)->src.u.all), 			\
	NIPQUAD((x)->dst.ip), ntohs((x)->dst.u.all))


#endif

#define CTINFO2DIR(ctinfo) ((ctinfo) >= IP_CT_IS_REPLY ? IP_CT_DIR_REPLY : IP_CT_DIR_ORIGINAL)

/* If we're the first tuple, it's the original dir. */
#define DIRECTION(h) ((enum ip_conntrack_dir)(&(h)->ctrack->tuplehash[1] == (h)))

/* Connections have two entries in the hash table: one for each way */
struct ip_conntrack_tuple_hash
{
	struct list_head list;

	struct ip_conntrack_tuple tuple;

	/* this == &ctrack->tuplehash[DIRECTION(this)]. */
	struct ip_conntrack *ctrack;
};

#endif /* __KERNEL__ */

static inline int ip_ct_tuple_src_equal(const struct ip_conntrack_tuple *t1,
				        const struct ip_conntrack_tuple *t2)
{
	return t1->src.ip == t2->src.ip
		&& t1->src.u.all == t2->src.u.all;
}

static inline int ip_ct_tuple_dst_equal(const struct ip_conntrack_tuple *t1,
				        const struct ip_conntrack_tuple *t2)
{
	return t1->dst.ip == t2->dst.ip
		&& t1->dst.u.all == t2->dst.u.all
		&& t1->dst.protonum == t2->dst.protonum;
}

static inline int ip_ct_tuple_equal(const struct ip_conntrack_tuple *t1,
				    const struct ip_conntrack_tuple *t2)
{
	return ip_ct_tuple_src_equal(t1, t2) && ip_ct_tuple_dst_equal(t1, t2);
}

static inline int ip_ct_tuple_mask_cmp(const struct ip_conntrack_tuple *t,
				       const struct ip_conntrack_tuple *tuple,
				       const struct ip_conntrack_tuple *mask)
{
	return !(((t->src.ip ^ tuple->src.ip) & mask->src.ip)
		 || ((t->dst.ip ^ tuple->dst.ip) & mask->dst.ip)
		 || ((t->src.u.all ^ tuple->src.u.all) & mask->src.u.all)
		 || ((t->dst.u.all ^ tuple->dst.u.all) & mask->dst.u.all)
		 || ((t->dst.protonum ^ tuple->dst.protonum)
		     & mask->dst.protonum));
}

#endif /* _IP_CONNTRACK_TUPLE_H */
