
/* DNS extension for IP connection tracking. */
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/ctype.h>
#include <net/checksum.h>
#include <net/udp.h>

#include <linux/netfilter_ipv4/lockhelp.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/policy_domain.h>

#if 1
#define DP(args...) if(fw_debug==1) printk(args)
#else
#define DP(format, args...)
#endif
extern int fw_debug;

//#define PACKETSZ	512		/* maximum packet size */
#define MAXDNAME	1025		/* maximum domain name */
#define INT16SZ		2		/* for systems without 16-bit ints */
#define FORMERR		1		/* format error */
#define DNS_PORT	53		
#define DNS_DATA_MIN	12
#define DNS_DATA_OFFSET	8
#define DNS_HDR_DOMAIN_OFFSET	12
#define DNS_ANS_TYPE_A_HDR_SIZE	16

#define DNS_NAME_PTR		0xc0		/* 0xc0 = 192 indicate a pointer */

#define DNS_CLASS		0x01		/* inet class */
#define DNS_CLASS_LEN		2
#define DNS_TYPE_LEN		2

#define DNS_TYPE_A		0x01		/* host address */
#define DNS_TYPE_NS		0x02		/* name server */
#define DNS_TYPE_CNAME		0x05		/* cname */
#define DNS_TYPE_MX		0x0f		/* Mail exchange type */

#define DNS_QUERY		0x0		/* standard query */
#define DNS_NO_ERROR		0x0		/* no error */
#define DNS_RESPONSE		0x1		/* response */
#define DNS_NO_SUCH_NAME	0x11		/* no such name */

#define T_A		1		/* host address */
#define T_PTR		12		/* domain name pointer */
#define T_MX		15		/* mail routing information */

/*
 * Values for class field
 */

#define C_IN		1		/* the arpa internet */

#define C_ANY		255		/* wildcard match */

#define GET_SHORT(cp) ((unsigned char )(*cp)<<8 | (unsigned char )*(cp+1))
#define GET_LONG(cp) ((unsigned char )(*cp)<<24 | (unsigned char )*(cp+1)<<16 | (unsigned char )*(cp+2)<<8 | (unsigned char )*(cp+3))


#define htons(A)  ((((A) & 0xff00) >> 8) |(((A) & 0x00ff) << 8))

typedef struct {
	unsigned	id :16;		/* query identification number */

	unsigned	rd :1;		/* recursion desired */
	unsigned	tc :1;		/* truncated message */
	unsigned	aa :1;		/* authoritative answer */
	unsigned	opcode :4;	/* purpose of message */
	unsigned	qr :1;		/* response flag */
			/* fields in fourth byte */
	unsigned	rcode :4;	/* response code */
	unsigned	cd: 1;		/* checking disabled by resolver */
	unsigned	ad: 1;		/* authentic data from named */
	unsigned	unused :1;	/* unused bits (MBZ as of 4.9.3a3) */
	unsigned	ra :1;		/* recursion available */

			/* remaining bytes */
	unsigned	qdcount :16;	/* number of question entries */
	unsigned	ancount :16;	/* number of answer entries */
	unsigned	nscount :16;	/* number of authority entries */
	unsigned	arcount :16;	/* number of resource entries */
} dns_hdr;

typedef struct {
	unsigned		name0 :8;	/* name of query */
	unsigned		name1 :8;	/* name of query */
	unsigned		type0 :8;	/* type of answer */
	unsigned		type1 :8;	/* type of answer */
	unsigned		class :16;	/* class of answer */
	unsigned		ttl0 :8;	/* ttl of answer */
	unsigned	 	ttl1 :8;	/* ttl of answer */
	unsigned	 	ttl2 :8;	/* ttl of answer */
	unsigned	 	ttl3 :8;	/* ttl of answer */
	unsigned	 	len0  :8;	/* length of following answer */
	unsigned		len1  :8;
	unsigned 		addr0 :8;	/* ip addr of type a answer of query */
	unsigned		addr1 :8;	/* ip addr of type a answer of query */
	unsigned 		addr2 :8;	/* ip addr of type a answer of query */
	unsigned		addr3 :8;	/* ip addr of type a answer of query */
} dns_ans_hdr;

#define DNS_IP_ADDR	(ans->addr0<<24|ans->addr1<<16|ans->addr2<<8|ans->addr3)
#define DNS_IP_TTL	(ans->ttl0<<24 |ans->ttl1<<16 |ans->ttl2<<8 |ans->ttl3 )

asmlinkage void sys_nat_url_filter_reset(unsigned long *args,char count,char flag)
{

}

static int get_objectname(unsigned char *dnshdr, unsigned char **rr,
			  char *string, int k)
{
    unsigned int len;
    int	i;

    while ((len = (**rr)) != 0) {
	*rr += 1;
	if (len & 0xC0) {
	    unsigned offset;
	    unsigned char *p;

	    offset = ((len & ~0xc0) << 8) + **rr;
	    p = &dnshdr[offset];
	    k = get_objectname(dnshdr, &p, string, k);

	    break;
	}
	else if (len < 64) {
	    for (i=0; i < len; i++) {
		string[k++] = **rr;
		*rr += 1;
	    }

	    string[k++] = '.';
	}
    }

    *rr += 1;
    string[k] = 0;

    return (k);
}
DECLARE_LOCK(ip_dns_lock);
struct module *ip_conntrack_dns = THIS_MODULE;

#define MAX_PORTS 1
static int ports[MAX_PORTS];
static int ports_c;

/*

+----+    +-------------+    +-------------+
|hash|--->|id1,domain 1 |--->|id1,domain 2 |---> ...
| id1|    +-------------+    +-------------+
+----+          |              |
|hash|          V              V
| id2|      +------------+
+----+      |dm1,ip1,ttl1|
|hash|      +------------+
| id3|             |
+----+             V
|hash|      +------------+
| id4|      |dm1,ip2,ttl2|
+----+      +------------+
                   |
                   ... 256 entries

hash_id = hash_dm(domain_name);

test dns: 
	xyz.twbbs.org
	www.google.com
	www.microsoft.com
	www.nctu.edu.tw
	msia.twbbs.org
	www.nchu.edu.tw
	norton.dyndns.org	
*/

#ifdef XX

asmlinkage void sys_nat_dns_reset(unsigned long *args,char count,char flag)
{
	LOCK_BH(&ip_dns_lock);
	
	clean_domain();
	
	UNLOCK_BH(&ip_dns_lock);
	
	return;
}


static int process_request(dns_hdr *header,char* name)
{
  unsigned char *ansp, *namep, *p, *p1 ;
  int qtype, qclass;
  
  char *cp;
  unsigned int j,l,q;
  int qdcount = ntohs(header->qdcount);
  int opcode = header->opcode;
  
  if (!qdcount)
    {
    //  header->rcode = htons(FORMERR);
    //  header->qr = 1;
    //  header->ra = 1;
     // return sizeof(dns_hdr);
      return 0;
    }
  
  /* determine end of question section (we put answers there) */
  p = ansp = (char *)(header+1);
  for (q=0; q<qdcount; q++)
    {
      while (1)
	{
	  if (((*ansp) & 0xc0) == 0xc0) /* pointer for name compression */
	    {
	      ansp += 2;
	      break;
	    }
	  else if (*ansp) /* another segment */
	    ansp += (*ansp) + 1;
	  else            /* end */
	    {
	      ansp++;
	      break;
	    }
	}
      ansp += 4; /* class and type */
    }

  /* now process each question, answers go in RRs after the question */
  for (q=0; q<qdcount; q++)
    {
      /* save pointer to name for copying into answers */
      namep = p;

      /* now extract name as .-concatenated string into name */
      cp = name;
      p1 = NULL;
      while ((l = *p++))
	{
	  if ((l & 0xc0) == 0xc0) /* pointer */
	    { 
	      /* get offset */
	      l = (l&0x3f) << 8;
	      l |= *p++;
	      if (!p1) /* first jump, save location to go back to */
		p1 = p;
	      p = l + (unsigned char *)header;
	    }
	  else
	    {
	      for(j=0; j<l; j++) {
		*cp++ = tolower(*p);
		p++;
	      }
	      *cp++ = '.';
	    }
	}
      *--cp = 0; /* overwrite last period */
	  
      if (p1) /* we jumped via compression */
	p = p1;

      GETSHORT(qtype, p); 
      GETSHORT(qclass, p);
  
#if 0
     // if ((opcode == QUERY) && (qclass == C_IN) &&((qtype == T_PTR)||(qtype == T_A)|| (qtype == T_MX)))
      if ((opcode == QUERY) && (qclass == C_IN) &&(qtype == T_A))
      {
      		printk("In QUERY T_A \n");
		return 1;
	}
	else
	{
		printk("In other \n");
		return 0;
	}
	    
#endif

	return 1;
    
    }
 
}
#endif


void strlwr(char *source_str)
{
	while(*source_str)
	{
		if((*source_str>='A')&&(*source_str<='Z'))
			*source_str=*source_str+32;
		source_str++;
	}

	return;
}

static unsigned short parse_name( char *buf, char *name, unsigned int len)
{
	unsigned short i,temp_len=0;
	char *ptr;
	
	if( buf==NULL ) return 0;
	ptr = buf;
	
	temp_len = (unsigned short)*ptr;
	i = (unsigned short)*ptr;
	ptr++;
		
	while( i-- )	
	{
		ptr++;
		if( i==0 )
		{
			temp_len += (unsigned short)(*ptr+1);
			i = (unsigned short)*ptr;
			ptr++;
		}
		else
		{
			/* warning: name may contain '\0' character */
			if( *ptr == '\0' )
				return -1;
		}
	}
	strncpy(name,buf+1,len-1); /* just copy it because '\0' at the end */
	for( i=0; name[i]; i++)
	{
		if( name[i]>=0x01 && name[i]<=0x20 )
			name[i] = '.';
	}

	return temp_len;
}

static int update_domain_ip(const char *data, unsigned int datalen )
{
	int i, p_len, cname_rr=0;
	char domain[64];
	unsigned char *p = NULL, *v;
	dns_hdr *dnshdr = (dns_hdr *)data;
	dns_ans_hdr *ans = NULL;
	unsigned short query_type; 
	unsigned short query_class;
	unsigned long query_ttl;
	unsigned short query_name_len=0;
	unsigned short query_addr_len;
	unsigned long query_addr;
	unsigned char *here;
	unsigned long offset = 12;
	char alias_name[64];
	char query_name[64];
	
		/* query www.microsoft.com will result rcode=1 */
	if( dnshdr->opcode != DNS_QUERY || dnshdr->qr != DNS_RESPONSE ||
		dnshdr->rcode == DNS_NO_SUCH_NAME || 
		ntohs(dnshdr->ancount) <= 0 )
	{
		DP("\n skip response: opcode=%d qr=%d rcode=%d ancount=%d",
			dnshdr->opcode, dnshdr->qr, dnshdr->rcode, ntohs(dnshdr->ancount) );
		return 0;
	}
	
	strncpy( &domain[0], ((char*)dnshdr)+DNS_HDR_DOMAIN_OFFSET+1, sizeof(domain)); /* domain name start with length of x of x.y.z , so offset+1 */
	domain[63] = '\0';
	
	/* check malicious domain name to against Dns DoS */
	/* ...*/
	
	p_len=strlen(domain);
	if(p_len==0) return 0;
	strlwr(domain);
	
	for(i=0; i<p_len; i++)
	{
		if( domain[i]>=0x01 && domain[i]<=0x20)
			domain[i] = '.'; /* replace length of sub-part of domain, to '.' */
	}
	DP("\n\nDNS query %s\n",domain);

#ifdef XX
	pdomain = find_match_domain(domain);
	if ( pdomain == NULL )
		return 0;	/* not interest policy route domain */
#endif
	
	/*********************************************************************************
	 * parse dns response to get ip(s), ttl(s) 
	 * get ip,ttl from Answer RRs if query type == A (1)
	 * get ip,ttl from Additional RRs if query type = CNAME (0x05),MX (0x0f), NS (0x02)
	 *********************************************************************************/
	if( ntohs(dnshdr->qdcount) == 1  )
	{
		ans  = (dns_ans_hdr *) ( ((char*)dnshdr)+DNS_HDR_DOMAIN_OFFSET+1+p_len+1+DNS_TYPE_LEN+DNS_CLASS_LEN);
			 /* hostname is ended with '\0', so +1 */
		p = (char *)ans;
	}
	else
	{
		DP(" domain = %s, query number = %d \n", domain, ntohs(dnshdr->qdcount) );
		return 0;
	}
	
	/* weak to against DNS DoS, DNS poision */
	{
		here = p;
		offset = get_objectname( data, &here , alias_name, 0 );
	}
	
	DP("ans:%d question=%d\n", ntohs(dnshdr->ancount), ntohs(dnshdr->qdcount) );
	
	/* parse answer record */
	for( i=0; ans && i< ntohs(dnshdr->ancount) ; i++)
	{
		p = (char *)ans;
		
		DP("\nans:%d &ans=%lu dnshdr=%lu &data=%lu name=%0x|%0x|%0x|%0x|%0x", i, ans,dnshdr, data, *(p-1),p[0], p[1], p[2], p[3] );
		if( (((char*)ans) + DNS_ANS_TYPE_A_HDR_SIZE ) > (data+datalen) )
		{
			DP("\nans record endoff =%lu exceed datalen=%u \n",((char*)ans)+DNS_ANS_TYPE_A_HDR_SIZE,(data+datalen));
			return -1;
		}
		{
			here = p;
			offset = get_objectname( data , &here , domain, 0 );
		}
		if( p[0] < DNS_NAME_PTR )
		{
			DP("\nDNS_PTR raw: name=%02x type=%02x class=%02x len=%04x ttl=%08x",
				(ans->name0<<8|ans->name1), (ans->type0<<8|ans->type1), ntohs(ans->class), (ans->len0<<8|ans->len1), DNS_IP_TTL);
			query_name_len = offset; /* parse_name( p, &query_name[0], sizeof(query_name) ); */
			v = p + query_name_len + 1 /* start of string char, end of string char already counted in parse_name()  */ + 2 /*type len*/ + 2 /*class len*/ + 4 /*ttl len*/;
			query_addr_len = GET_SHORT( v );
			v = (p + query_name_len + 1);
			query_type = GET_SHORT( v );
			v = p + query_name_len + 1 + 2;
			query_class = GET_SHORT( v );
			v = p + query_name_len + 1 + 2 + 2;
			query_ttl =  GET_LONG( v );
			if( query_type == htons(DNS_TYPE_A) )
			{
				v = p + query_name_len + 1 + 2 + 2 + 4 + 2;
				query_addr = GET_LONG( v );
				add_record( domain, htonl(query_addr) , query_ttl);
				if( cname_rr )
					add_record( alias_name, htonl(DNS_IP_ADDR), DNS_IP_TTL );
			}
			
			DP("\nDNS_PTR new: name=%s type=%02x class=%02x len=%04x ttl=%08x", 
				domain, query_type, query_class, query_addr_len, query_ttl );

			/* next ptr */
			ans = (dns_ans_hdr*)( p + query_name_len + 1 + 2 + 2 + 4 + 2 /* size of len */ + (query_addr_len) );
		}
		else
		{
			DP("\nMAYBE TYPE A name=%02x type=%02x class=%02x len=%04x ttl=%08x", 
				(ans->name0<<8|ans->name1), (ans->type0<<8|ans->type1), ntohs(ans->class), (ans->len0<<8|ans->len1), DNS_IP_TTL );
			if( (ans->type0<<8|ans->type1) == (DNS_TYPE_A) )
			{
				add_record( domain, htonl(DNS_IP_ADDR), DNS_IP_TTL );
				if( cname_rr )
					add_record( alias_name, htonl(DNS_IP_ADDR), DNS_IP_TTL );
			}
			else if( (ans->type0<<8|ans->type1) == (DNS_TYPE_CNAME) )
				cname_rr = 1;
			/* next ptr */
			ans = (dns_ans_hdr*)( p + DNS_ANS_TYPE_A_HDR_SIZE - 4 + (ans->len0<<8|ans->len1)); /* compute in wrong way ? */
		}
	}
	/* parse authorize record */
	for( i=0; ans && i< ntohs(dnshdr->nscount) ; i++)
	{
		p = (char *)ans;
		 
		DP("\nauthorize:%d &ans=%lu dnshdr=%lu &data=%lu ", ntohs(dnshdr->nscount),ans,dnshdr, data);
		if( (((char*)ans) + DNS_ANS_TYPE_A_HDR_SIZE ) > (data+datalen) )
		{
			DP("ans record endoff =%lu exceed datalen=%u \n",((char*)ans)+DNS_ANS_TYPE_A_HDR_SIZE,(data+datalen));
			return -1;
		}
		
		{
			here = p;
			offset = get_objectname( data, &here , domain, 0 );
		}
		if( p[0] < DNS_NAME_PTR )
		{
			DP("ans->name=%02x:%02x \n", ans->name0, ans->name1 );
			query_name_len = parse_name( (char*)ans, &query_name[0], sizeof(query_name) );
			v = ( p +  query_name_len + 1 /* start of string char, end of string char already counted in parse_name()  */ + 2 /*type len*/ + 2 /*class len*/ + 4 /*ttl len*/);
			query_addr_len = GET_SHORT ( v );
			v = ( p + query_name_len + 1 );
			query_type = GET_SHORT ( v );
			v = ( p + query_name_len + 1 + 2 );
			query_class = GET_SHORT ( v );
			v = ( p + query_name_len + 1 + 2 + 2 );
			query_ttl =  GET_LONG ( v );
			/* next ptr */
			ans = (dns_ans_hdr*)( ((char*)ans) + query_name_len + 1 + 2 + 2 + 4 + 2 /* size of len */ + (query_addr_len) );
			DP("\nparsing name=%02x type=%02x class=%02x len=%04x ttl=%08x", 
				query_name, (query_type), (query_class), (query_addr_len), (query_ttl) );
		}
		else
			ans = (dns_ans_hdr*)( ((char*)ans) + DNS_ANS_TYPE_A_HDR_SIZE - 4 + (ans->len0<<8|ans->len1)); /* compute in wrong way ? */
			
	}
	/* parse additional record */
	for( i=0; ans && i< ntohs(dnshdr->arcount) ; i++)
	{
		p = (char *)ans;
		
		DP("\nadditional:%d &ans=%lu dnshdr=%lu &data=%lu ", ntohs(dnshdr->arcount), ans,dnshdr, data);
		if( (((char*)ans) + DNS_ANS_TYPE_A_HDR_SIZE ) > (data+datalen) )
		{
			DP("ans record endoff =%lu exceed datalen=%u \n",((char*)ans)+DNS_ANS_TYPE_A_HDR_SIZE,(data+datalen));
			return -1;
		}
		{
			here = p;
			offset = get_objectname( data, &here , domain, 0 );
		}
		DP("\nparsing name=%02x type=%02x class=%02x len=%04x ttl=%08x", 
			(ans->name0<<8|ans->name1), (ans->type0<<8|ans->type1), ntohs(ans->class), (ans->len0<<8|ans->len1), DNS_IP_TTL );
		if( (ans->type0<<8|ans->type1) == (DNS_TYPE_A) )
		{
			add_record( domain, htonl(DNS_IP_ADDR), DNS_IP_TTL );
		}
		ans = (dns_ans_hdr*)( ((char*)ans) + DNS_ANS_TYPE_A_HDR_SIZE - 4 + (ans->len0<<8|ans->len1)); /* compute in wrong way ? */
	}
	
	return 1;
}


/* FIXME: This should be in userspace.  Later. */
static int filter_help(const struct iphdr *iph, size_t len,
		struct ip_conntrack *ct,
		enum ip_conntrack_info ctinfo)
{
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	const char *data = (const char *)udph + 8;
	unsigned int udplen = len - iph->ihl * 4;
	unsigned int datalen = udplen - 8;

	LOCK_BH(&ip_dns_lock);
     
#ifdef XX
	int dir = CTINFO2DIR(ctinfo);
	if (IP_CT_DIR_ORIGINAL == dir) {  /* ??? */
		UNLOCK_BH(&ip_dns_lock);
		return NF_ACCEPT;
	}
#endif
	
	/* Not whole UDP header */
	if (udplen < sizeof(struct udphdr)) {
		DP("Not whole UDP headeru\n");
		UNLOCK_BH(&ip_dns_lock);
		return NF_ACCEPT;
	}
	
	if( datalen > DNS_DATA_MIN )
	{
		update_domain_ip( data, datalen ); /* dns query header+data */
	}
	UNLOCK_BH(&ip_dns_lock);
	return NF_ACCEPT;
}

static struct ip_conntrack_helper dns;

/* Not __exit: called from init() */
static void fini(void)
{
	int i;
	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {
		ip_conntrack_helper_unregister(&dns);
	}
}

static int __init init(void)
{
	int i, ret;
	
	printk("Policy Domain DNS v0.6 conntrack\n");

	if (ports[0] == 0)
		ports[0] = DNS_PORT;

	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {
		memset(&dns, 0, sizeof(struct ip_conntrack_helper));
		
		dns.tuple.src.u.udp.port = htons(ports[i]);
		dns.tuple.dst.protonum = IPPROTO_UDP;
		dns.mask.src.u.udp.port = 0xFFFF;
		dns.mask.dst.protonum = 0xFFFF;
		dns.max_expected = 0;//lsshi add
		dns.timeout = 0;
		dns.flags = IP_CT_HELPER_F_REUSE_EXPECT;
		dns.me = ip_conntrack_dns;
		dns.help = filter_help;
		
		ret = ip_conntrack_helper_register(&dns);

		if (ret) {
			fini();
			return ret;
		}
		ports_c++;
	}
	
	return 0;
}

module_init(init);
module_exit(fini);
