/* CHANGES:
 *
 *	20040831: initial for new INVITATION
 *
 *	20040926: use try_ipport for "MSNMS_ACCEPT" compatible to win messenger 4.7 video, msn6.2 whiteboard
 *   
 *	video: windows 4.7 (priv:192.168.1.100) -> windows 4.7 (pub:10.10.31.8)
 * 	1. ->	Invitation-Command: INVITE
 *	2. <-   Invitation-Command: ACCEPT
 *		Context-Data: Requested:SIP_A,SIP_V,;
 *		Request-Data: IP-Address:
 *		IP-Address: 10.10.31.8
 *	3. ->	Invitation-Command: ACCEPT
 *		IP-Address: 192.168.1.100:11519
 *	4. <-	192.168.1.100:11519 <- SIP-REQUEST
 *	
 *
 *	20041003
 *		use try_content_type() to determine whether to parse whole packet
 *		fixed sizeof(MSN_HEADER_END) problem, should be sizeof(MSN_HEADER_END)-1 or strlen(...)
 *
 *	ftp: msn62 (priv:192.168.1.100) -> UPNP -> msn62 (pub:10.10.31.8)
 *	1. ->	Content-Type: x/...p2p
 *	2. <-	IPv4Internal-Address: ..
 *	3. ->	TCP SYN (direct connect ok, note A.)
 *	4. ->	Data transfer...
 *
 *	ftp: msn62 (priv:192.168.1.100) <- UPNP <- msn62 (pub:10.10.31.8)
 *	1. <-	Content-Type: x/...p2p
 *	2. ->	IPv4Internal-Address: 192.168.1.100 ...
 *	3. <-	TCP SYN (direct connect failed, note B.)
 *	4. <-	Data transfer via MSN server ...
 *
 *
 *	Note:
 *		a. data connection may not close after transfered, can be reused by peer to send back data.
 *		b. if data connection is exist, no new connection will be created.
 */
 
/* MSNMS extension for IP connection tracking. */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <linux/ctype.h>
#include <net/checksum.h>
#include <net/tcp.h>
#include <net/udp.h>

#include <linux/netfilter_ipv4/lockhelp.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>
#include <linux/netfilter_ipv4/ip_conntrack_msnms.h>

DECLARE_LOCK(ip_msnms_lock);
DECLARE_RWLOCK_EXTERN(ip_conntrack_lock);

struct module *ip_conntrack_msnms = THIS_MODULE;

#define MSN_HEADER_END		"\r\n\r\n"
#define CONTENT_TYPE_OPT	"Content-Type: "

enum ctx_type { CTX_TYPE_NONE=0, CTX_TYPE_MSMSGSINVITE=1, CTX_TYPE_PLAINTEXT, CTX_TYPE_UNKNOWN=0xff };

#define MSN_WHITEBOARD_GUID	"{1DF57D09-637A-4ca5-91B9-2C3EDAAF62FE}"
#define MSN_VIDEO_GUID		"{02D3C01F-BF30-4825-A83A-DE7AF41648AA}"
#define MSN_WINMSG_FTP_GUID	"{5D3E02AB-6190-11d3-BBBB-00C04F795683}"


static int ports[MAX_PORTS];
static int ports_c = 0;
#ifdef MODULE_PARM
MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_PORTS) "i");
#endif

int msn_alg_enable = 1;
int msn_whiteboard = 0;
int msn_winftp = 1;
int msn_video = 0;
int msn_audio = 0;
int msn_debug = 0;

static int loose = 0;
MODULE_PARM(loose, "i");

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

extern int rdr_clean(struct ip_conntrack* ct);
extern unsigned int ROUTER_LOCAL_IP;
extern unsigned int ROUTER_LOCAL_MASK;

#ifdef LAN_NET

int lan_net(unsigned long ip)
{
	return ((ip&ROUTER_LOCAL_MASK)==(ROUTER_LOCAL_IP&ROUTER_LOCAL_MASK));
}
#endif

extern int from_lan(struct ip_conntrack *ct);		/* define in ip_conntrack_spec.c */

/*------------ BEGIN : MSN SESSION STRUCTURE -------------------- */

struct ip_ct_msnms_master msnms_session[MAX_INVITE_NUM];

struct ip_ct_msnms_master *session_find(u_int32_t cookie)
{
	int i;
	
	if( cookie == 0 )
		return NULL;
	
	for(i=0;i<MAX_INVITE_NUM;i++)
	{
		if( msnms_session[i].cookie == cookie)
			return &msnms_session[i];
	}
	return NULL;
}

/* CALL by sip_request_help(),sip_status_help() ..*/
struct ip_ct_msnms_master *session_find_by_ip(int req_type, unsigned char protocol, int fromlan, u_int32_t sip, u_int32_t dip,
	u_int16_t sport, u_int16_t dport , u_int32_t callid)
{
	int i;
	
	for(i=0;i<MAX_INVITE_NUM;i++)
	{
		if( msnms_session[i].proto && msnms_session[i].proto != protocol )
			continue;
		if( msnms_session[i].ttl < jiffies )
			continue;

		/* use sport for distingushing old/new session info */
		if( msnms_session[i].callid && msnms_session[i].callid != callid)
			continue;
		
		/* consider the case:
			a. invite but reject/idle -> 
						peer_addr = 0, peer_port =0, 
						pub_ddr=0, pub_port =0, 
						priv_addr=0, priv_port = 0
		*/
		if( fromlan )
		{

			if( /* msnms_session[i].priv_addr && (impossible be zero)*/ msnms_session[i].priv_addr != sip)
				continue;
			if( /* msnms_session[i].peer_addr && (impossible be zero)*/ msnms_session[i].peer_addr != dip)
				continue;
			
			if( /* msnms_session[i].peer_port && (impossible be zero)*/ msnms_session[i].peer_port != dport)
				continue;
			break;
		}
		else
		{
			if( /* msnms_session[i].pub_addr && (impossible be zero)*/ msnms_session[i].pub_addr != dip)
				continue;
			/* windows messenger ftp protocol: peer_addr may be zero */
			if( msnms_session[i].peer_addr && msnms_session[i].peer_addr != sip)
				continue;
			if( /* msnms_session[i].pub_port && (impossible be zero)*/ msnms_session[i].pub_port != dport)
				continue;
			break;			
		}
	}
	if( i<MAX_INVITE_NUM )
		return &msnms_session[i];
	else
	{
		for(i=0;i<MAX_INVITE_NUM;i++)
		{
			if( msnms_session[i].ttl==0 ) continue;
			
			if(msn_debug)
			{
				printk("i=%d ttl=%lu now=%lu callid=%u  peer_addr=%u.%u.%u.%u port=%u\n",	
					i, msnms_session[i].ttl, jiffies, msnms_session[i].callid, 
					NIPQUAD(msnms_session[i].peer_addr), msnms_session[i].peer_port);
				printk("\t\tpub_addr=%u.%u.%u.%u port=%u priv_addr=%u.%u.%u.%u port=%u\n",
					NIPQUAD(msnms_session[i].pub_addr), msnms_session[i].pub_port,
					NIPQUAD(msnms_session[i].priv_addr), msnms_session[i].priv_port );
			}
		}
		
	}
	return NULL;
}


struct ip_ct_msnms_master *session_add(struct ip_ct_msnms_master *rec)
{
	int i;
	int cand=MAX_INVITE_NUM;
	
	if( rec == NULL )
		return NULL;
	
	for(i=0;i<MAX_INVITE_NUM;i++)
	{
		if( msnms_session[i].cookie == rec->cookie)
			break;
		if( cand == (MAX_INVITE_NUM) && msnms_session[i].ttl < jiffies )
			cand = i;
	}
	if( i<MAX_INVITE_NUM)
		cand = i;
	if( cand < MAX_INVITE_NUM && cand >= 0 )
	{
		memcpy( &msnms_session[cand], rec, sizeof(struct ip_ct_msnms_master));
		msnms_session[cand].index = cand;
		msnms_session[cand].ttl = jiffies + MSN_SIP_TIMEOUT*HZ;
		return &msnms_session[cand];
	}
	return NULL;	
}

int session_refresh( struct ip_ct_msnms_master *rec, unsigned long ttl)
{
	if( rec == NULL ) return -1;

	rec->ttl = jiffies + ttl;
	return 0;
}

int session_del(u_int32_t cookie)
{
	int i;
	
	if( cookie == 0 )
		return -1;
	
	for(i=0;i<MAX_INVITE_NUM;i++)
	{
		if( msnms_session[i].cookie == cookie)
		{
			memset(&msnms_session[i],0,sizeof(struct ip_ct_msnms_master));
			return i;
		}
	}
	return -1;
}

void session_clean(struct ip_ct_msnms_master *rec)
{
	if( rec == NULL )
		return ;
	
	memset(rec,0,sizeof(struct ip_ct_msnms_master));
}
/*------------END: MSN SESSION STRUCTURE -------------------- */



static int try_guid(const char *data, size_t dlen, u_int32_t array[6], char term);
static int try_ipaddr(const char *data, size_t dlen, u_int32_t array[6], char term);
static int try_ipport(const char *data, size_t dlen, u_int32_t array[6], char term);
static int try_sipipport(const char *data, size_t dlen, u_int32_t array[6], char term);
static int try_length(const char *data, size_t dlen, u_int32_t array[6], char term);
static int try_cmd(const char *data, size_t dlen, u_int32_t array[6], char term);
static int try_callid(const char *data, size_t dlen, u_int32_t array[6], char term);
static int sip_request_help(const struct iphdr *iph, size_t len,
		     struct ip_conntrack *ct,
		     enum ip_conntrack_info ctinfo);

static struct msnms_search {
	enum ip_conntrack_dir dir;
	const char *pattern;
	size_t plen;
	char skip;
	char term;
	enum ip_ct_msnms_type msnmstype;
	int (*getnum)(const char *, size_t, u_int32_t[], char);
} search[] = {
	{
		IP_CT_DIR_ORIGINAL,
		/* "Invitation-Command: INVITE",	sizeof("Invitation-Command: INVITE") - 1, ' ', '\r', */
		"Application-GUID: ", sizeof("Application-GUID: ") - 1, ' ', '\r',
		IP_CT_MSNMS_GUID,
		try_guid,
	},
	{
		IP_CT_DIR_REPLY,
		/* "Invitation-Command: INVITE",	sizeof("Invitation-Command: INVITE") - 1, ' ', '\r', */
		"Application-GUID: ", sizeof("Application-GUID: ") - 1, ' ', '\r',
		IP_CT_MSNMS_GUID,
		try_guid,
	},
	{
		IP_CT_DIR_ORIGINAL,
		/* "Invitation-Command: ACCEPT",	sizeof("Invitation-Command: ACCEPT") - 1, ' ', '\r', */
		"IP-Address: ", sizeof("IP-Address: ") - 1, ' ', '\r',
		IP_CT_MSNMS_ACCEPT,
		try_ipport,//try_ipaddr,
	},
	{
		/* msn 6.2 */
		IP_CT_DIR_ORIGINAL,
		/* "Invitation-Command: Context", sizeof("Invitation-Command: Context") - 1, ' ', '\r', */
		"Context-Data: ", sizeof("Context-Data: ") - 1, ' ', '\r',
		IP_CT_MSNMS_CONTEXT,
		try_ipport,
	},
	{
		IP_CT_DIR_REPLY,
		/* IP-Address: a.b.c.d\r\n */
		"IP-Address: ", sizeof("IP-Address: ") - 1, ' ', '\r',
		IP_CT_MSNMS_REPLYACCEPT,
		try_ipaddr,
	},
	{
		/* msn 6.2 */
		IP_CT_DIR_REPLY,
		/* "Invitation-Command: Context", sizeof("Invitation-Command: Context") - 1, ' ', '\r', */
		"Context-Data: ", sizeof("Context-Data: ") - 1, ' ', '\r',
		IP_CT_MSNMS_REPLYCONTEXT,
		try_ipport,
	},
	{
		IP_CT_DIR_ORIGINAL,
		/* "Invitation-Command: CANCEL", sizeof("Invitation-Command: CANCEL") - 1, ' ', '\r', */
		"Invitation-Command: CANCEL", sizeof("Invitation-Command: CANCEL") - 1, ' ', '\r',
		IP_CT_MSNMS_CANCEL,
		NULL,
	},
	{
		IP_CT_DIR_REPLY,
		/* "Invitation-Command: CANCEL", sizeof("Invitation-Command: CANCEL") - 1, ' ', '\r', */
		"Invitation-Command: CANCEL", sizeof("Invitation-Command: CANCEL") - 1, ' ', '\r',
		IP_CT_MSNMS_REPLYCANCEL,
		NULL,
	},

};

static struct msnms_search sip_request[] = {
	{
		IP_CT_DIR_ORIGINAL,
		/* Contact: <sip:10.10.31.8:10166>\r\n */
		"Contact: <sip:", sizeof("Contact: <sip:") - 1, ' ', '>',
		IP_CT_SIP_REQUEST,
		try_sipipport,
	},
};


static struct msnms_search sip_status[] = {
	{
		IP_CT_DIR_ORIGINAL,
		/* Contact: <sip:10.10.31.8:10166>\r\n */
		"Contact: <sip:", sizeof("Contact: <sip:") - 1, ' ', '>',
		IP_CT_SIP_STATUS,
		try_sipipport,
	},
};


static struct msnms_search sip_statuslength[] = {
	{
		IP_CT_DIR_ORIGINAL,
		/* Context-Length: YY\r\n */
		"Content-Length: ", sizeof("Content-Length: ") - 1, ' ', '\r',
		IP_CT_SIP_STATUS,
		try_length,
	},

};

static struct msnms_search sip_requestlength[] = {
	{
		IP_CT_DIR_ORIGINAL,
		/* Context-Length: YY\r\n */
		"Content-Length: ", sizeof("Content-Length: ") - 1, ' ', '\r',
		IP_CT_SIP_REQUEST,
		try_length,
	},

};
static struct msnms_search winmsg_port[] = {
	{
		IP_CT_DIR_ORIGINAL,
		"Port: ", sizeof("Port: ") - 1, ' ', '\r',
		IP_CT_MSNMS_PORT,
		try_length,
	},
};

static struct msnms_search sip_callid[] = {
	{
		IP_CT_DIR_ORIGINAL,
		"Call-ID: ", sizeof("Call-ID: ") - 1, ' ', '\r',
		IP_CT_MSNMS_CALLID,
		try_callid,
	},
};



static struct msnms_search invite_cookie[] = {
	{
		IP_CT_DIR_ORIGINAL,
		"Invitation-Cookie: ", sizeof("Invitation-Cookie: ") - 1, ' ', '\r',
		IP_CT_MSNMS_INVITECOOKIE,
		try_length,
	},
	{
		IP_CT_DIR_REPLY,
		"Invitation-Cookie: ", sizeof("Invitation-Cookie: ") - 1, ' ', '\r',
		IP_CT_MSNMS_INVITECOOKIE,
		try_length,
	},
};


static unsigned long
strtou16( char *pbuf)
{
	int n=0;
	unsigned long pval = 0;
	
	while( pbuf[n] >= '0' && pbuf[n] <= '9' )
	{
		pval = (pval *10) + (pbuf[n]-'0');
		n++;
	}
	return pval;
}

char *strnstr(const char *haystack, const char *needle, size_t haystacklen)
{
    char *p;
    ssize_t plen;
    ssize_t len = strlen(needle);

    if (*needle == '\0')	/* everything matches empty string */
	return (char *) haystack;

    plen = haystacklen;
    for (p = (char *) haystack; p != NULL;
	 p = memchr(p + 1, *needle, plen - 1)) {
	plen = haystacklen - (p - haystack);

	if (plen < len)
	    return NULL;

	if (strnicmp(p, needle, len) == 0)
	    return (p);
    }
    return NULL;
}


static int try_number(const char *data, size_t dlen, u_int32_t array[],
		      int array_size, char sep, char term)
{
	u_int32_t i, len;

	memset(array, 0, sizeof(array[0])*array_size);

	/* Keep data pointing at next char. */
	for (i = 0, len = 0; len < dlen && i < array_size; len++, data++) {
		if (*data >= '0' && *data <= '9') {
			array[i] = array[i]*10 + *data - '0';
		}
		else if (*data == sep)
			i++;
		else {
			/* Unexpected character; true if it's the
			   terminator and we're finished. */
			//if (*data == term && i == array_size - 1)
				return len;

			/*
			DEBUGP("Char %u (got %u nums) `%u' unexpected\n",
			       len, i, *data);
			return 0;
			*/
		}
	}
	DEBUGP("Failed to fill %u numbers separated by %c\n", array_size, sep);

	return 0;
}

/* Grab port: number up to delimiter */
static int get_port(const char *data, int start, size_t dlen, u_int32_t array[2])
{
	u_int16_t port = 0;
	int i;

	for (i = start; i < dlen; i++) {
		if (data[i] >= '0' && data[i] <= '9')
			port = port*10 + data[i] - '0';
		else /* Some other crap */
		{
			array[0] = port >> 8;
			array[1] = port;
			return i + 1;
		}
	}
	return 0;
}

static int get_port2(const char *data, size_t dlen, u_int16_t *port)
{
	int i;

	for (i = 0; i < dlen; i++) {
		if (data[i] >= '0' && data[i] <= '9')
			*port = *port*10 + data[i] - '0';
		else /* Some other crap */
		{
			return i + 1;
		}
	}
	return 0;
}

/* Grab port: number up to delimiter */
static int get_length(const char *data, int start, size_t dlen, char delim,
		    u_int32_t array[2])
{
	u_int32_t port = 0;
	int i;

	for (i = start; i < dlen; i++) {
		if (data[i] >= '0' && data[i] <= '9')
			port = port*10 + data[i] - '0';
		else /* Some other crap */
		{
			array[0] = 0;
			array[1] = port;
			return i + 1;
		}
	}
	return 0;
}

/* Context-Data: 192.168.1.100:98978\r\n */
static int try_ipport(const char *data, size_t dlen, u_int32_t array[6],
		    char term)
{
	char delim=':';
	int length;

	/* First character is delimiter, then "1" for IPv4, then
           delimiter again. */
	if (dlen < 9) return 0;		/* 1.1.1.1:1 */

	/* Now we have IP address. */
	length = try_number(data, dlen, array, 4, '.', delim);
	if (length == 0)
		return 0;

	/* ip address trailing delimiter ':' */
	//return get_port(data, length + 1, dlen-length-1, array+4);
	length += get_port(data, length + 1, dlen-1, array+4);
	return length;
}

/* <sip:192.168.1.1:3934> */
static int try_sipipport(const char *data, size_t dlen, u_int32_t array[6],
		    char term)
{
	char delim=':';
	int length;

	/* First character is delimiter, then "1" for IPv4, then
           delimiter again. */
	if (dlen < 9) return 0;		/* 1.1.1.1:1 */

	/* Now we have IP address. */
	length = try_number(data, dlen, array, 4, '.', delim);
	if (length == 0)
		return 0;

	/* ip address trailing delimiter ':' */
	return get_port(data, length + 1, dlen-1, array+4);
}

/* IP-Address: 192.168.1.100\r\n */
static int try_ipaddr(const char *data, size_t dlen, u_int32_t array[6],
		    char term)
{
	char delim='\r';

	/* First character is delimiter, then "1" for IPv4, then
           delimiter again. */
	if (dlen < 7) return 0;		/* 1.1.1.1 */

	/* Now we have IP address. */
	return try_number(data, dlen, array, 4, '.', delim);

}

/* Content-Length: 192\r\n */
static int try_length(const char *data, size_t dlen, u_int32_t array[6],
		    char term)
{
	return get_length(data, 0, dlen, '\r', array+4);
}


/* Application-GUID: {.....}\r\n */
static int try_guid(const char *data, size_t dlen, u_int32_t array[6],
		    char term)
{
	char buf[255];
	
	memset(buf,0,sizeof(buf));
	strncpy(buf,data, strlen(MSN_WHITEBOARD_GUID) );
	if(msn_debug) 
		printk("guid=%s\n", buf);
	if( msn_whiteboard && strncmp(data, MSN_WHITEBOARD_GUID, strlen(MSN_WHITEBOARD_GUID))== 0 )
	{
		array[4] = 0;
		array[5] = MSN_WHITEBOARD_INVITE;
	}
	else if( msn_winftp && strncmp(data, MSN_WINMSG_FTP_GUID, strlen(MSN_WINMSG_FTP_GUID))== 0 )
	{
		array[4] = 0;
		array[5] = MSN_WINMSG_FTP_INVITE;	
	}
	else if( msn_video && strncmp(data, MSN_VIDEO_GUID, strlen(MSN_VIDEO_GUID))== 0 )
	{
		array[4] = 0;
		array[5] = MSN_VIDEO_INVITE;	
	}
	else
	{
		array[4] = 0;
		array[5] = 0;		
	}
	return 1;
}


static int try_content_type(const char *data)
{
	if( strncmp(data,"text/x-msmsgsinvite",strlen("text/x-msmsgsinvite"))==0 )
		return CTX_TYPE_MSMSGSINVITE;
	else if( strncmp(data,"text/plain",strlen("text/plain"))==0 )
		return CTX_TYPE_PLAINTEXT;
	else
		return CTX_TYPE_UNKNOWN;		
}

static int try_callid(const char *data, size_t dlen, u_int32_t array[6],
		    char term)
{
	u_int32_t callid = 0;
	int i;

	for (i = 0; i < dlen; i++) {
		if (data[i] != term)
			callid += data[i];
		else /* Some other crap */
		{
			array[4] = 0;
			array[5] = callid;
			return i + 1;
		}
	}	
	return 0;
}


/* Return 1 for match, 0 for accept, -1 for partial. */
static int find_pattern(int contenttype, const char *data, size_t dlen,
			const char *pattern, size_t plen,
			char skip, char term,
			unsigned int *numoff,
			unsigned int *numlen,
			u_int32_t array[6],
			int (*getnum)(const char *, size_t, u_int32_t[], char))
{
	size_t i;
	const char *p = NULL;
	int match=0;
	int ctx_type = CTX_TYPE_NONE;

	if ( dlen <= plen)
		return 0;

	if( data == NULL || data[0] == '\0' )
		return 0;

	for(i=0; i<dlen-plen; i++)
	{
		if ( contenttype != CTX_TYPE_NONE && 
			strncmp( data +i, CONTENT_TYPE_OPT, strlen(CONTENT_TYPE_OPT) ) == 0 )/* search only header part */
		{
			ctx_type = try_content_type(data+i+strlen(CONTENT_TYPE_OPT));
		}
		if ( ctx_type != CTX_TYPE_NONE && ctx_type != contenttype )
			break;
		
		if ( strncmp( data +i, pattern, plen) == 0 ) {
			match = 1;
			p = data + i;
			break;
		}
	}
	if( match == 0 || p == NULL )
		return 0;

	*numoff = i+plen;
	if( getnum )
	{
		*numlen = getnum( p+plen, dlen - i- plen, array, term);
		if (!*numlen)
		{
			if(msn_debug)
				printk("getnum failed:(size=%d)|%s|\n", dlen-i-plen,p);
			return -1;
		}
	}
	//printk("\tmsnms addr =%u.%u.%u.%u : %u \n", array[0], array[1], array[2], array[3], (array[4] << 8 | array[5]) );

	return 1;
}


/* Return 1 for match, 0 for accept, -1 for partial. */
static int find_media_port(const char *data, size_t dlen, u_int16_t array[2] )
{
	size_t i;
	int match=0;


	if( data == NULL || data[0] == '\0' )
		return 0;

	for(i=0; i<dlen; i++)
	{
		if ( strncmp( data +i, MSN_HEADER_END, strlen(MSN_HEADER_END) ) == 0 )/* search only header part */
			break;
		if ( strncmp( data +i, "m=audio", sizeof("m=audio")) == 0 ) {
			get_port2( data+i+strlen("m=audio")+1, 6, &array[0] );
			match++;
		}
		else if ( strncmp( data +i, "m=video", sizeof("m=video")) == 0 ) {
			get_port2( data+i+strlen("m=video")+1, 6, &array[1] );
			match++;
		}
		if( match >= 2 ) break;
	}
	if( match == 0 )
		return 0;

	return 1;
}

#ifdef XX

static int sip_status_expect(struct ip_conntrack *ct)
{
	WRITE_LOCK(&ip_conntrack_lock);
	ct->helper = &sip_whiteboard_helper;
	DEBUGP("sip_status_expect: Add sip status helper for whiteboard port %p\n", ct);
	WRITE_UNLOCK(&ip_conntrack_lock);
	
	return NF_ACCEPT;	/* unused */
}
#endif

/* FIXME: This should be in userspace.  Later. */
static int sip_status_help(const struct iphdr *iph, size_t len,
		     struct ip_conntrack *ct,
		     enum ip_conntrack_info ctinfo)
{
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	const char *data = (void *)udph + 8;
	unsigned int udplen = len - iph->ihl * 4;
	unsigned int datalen = udplen - sizeof(struct udphdr);
	u_int32_t array[6] = { 0 };
	int dir = CTINFO2DIR(ctinfo);
	unsigned int matchlen=0, matchoff=0;
//	struct ip_conntrack *master = master_ct(ct);
	struct ip_ct_msnms_master *ct_msnms_info = &ct->help.ct_msnms_info;
//	struct ip_ct_msnms_master *master_ct_msnms_info = &master->help.ct_msnms_info;
	struct ip_conntrack_expect expect, *exp = &expect;
	struct ip_ct_msnms_expect *exp_msnms_info = &exp->help.exp_msnms_info;
	u_int32_t newsrcip=0, newdstip=0, callid = 0;
	u_int16_t newdstport=0;
	struct ip_ct_msnms_master new_rec, *cur_rec=&new_rec;
	unsigned int i;
	int found = 0;
	int newproto = 0;
	
	ct->is_spec = 1; /* skip NATA */
	
	DEBUGP("\nsip_status_help: help entered %u.%u.%u.%u:%u->%u.%u.%u.%u:%u\n",
		NIPQUAD(iph->saddr), ntohs(udph->source),
		NIPQUAD(iph->daddr), ntohs(udph->dest));

	/* Can't track connections formed before we registered */
	if( !ct_msnms_info )
	{
		if(msn_debug)
			printk("sip_status_help: no ct_msnms_info\n");
		return NF_ACCEPT;
	}
	/* Not whole UDP header */
	if (udplen < sizeof(struct udphdr)) {
		DEBUGP("sip_status_help: udp: udplen = %u\n", (unsigned)udplen);
		return NF_ACCEPT;
	}

#ifndef SESSION_STRUC
	/* session rec */
/*
	found = find_pattern(data, datalen,
		     sip_callid[0].pattern,
		     sip_callid[0].plen,
		     sip_callid[0].skip,
		     sip_callid[0].term,
		     &matchoff, &matchlen,
		     array,
		     sip_callid[0].getnum);
	if( found != -1 && found != 0 )
		callid = array[4] << 8 | array[5] ;
*/
#ifdef LAN_NET
	cur_rec = session_find_by_ip( 0, iph->protocol, lan_net(ct->tuplehash[dir].tuple.src.ip),
		iph->saddr,iph->daddr,ntohs(udph->source),ntohs(udph->dest), callid );
#else
	cur_rec = session_find_by_ip( 0, iph->protocol, from_lan(ct),
		iph->saddr,iph->daddr,ntohs(udph->source),ntohs(udph->dest), callid );
#endif

	if( cur_rec == NULL || 
		(
			cur_rec->invite_type != MSN_WHITEBOARD_INVITE &&
			cur_rec->invite_type != MSN_WINMSG_FTP_INVITE && // no ftp invite
			cur_rec->invite_type != MSN_VIDEO_INVITE
		)
	)
	{
		if(msn_debug)
			printk("sip_request_help: session_find_by_ip() failed\n");
		return NF_ACCEPT;
	}
	exp_msnms_info->cookie = cur_rec->cookie;
	ct_msnms_info = cur_rec;
	ct_msnms_info->callid = callid;
#endif
	
	if( ct_msnms_info->bye )
	{
		if(msn_debug)
			printk("sip_status_help: refresh timeout to %d\n", MSN_SIP_CLOSING);
		ip_ct_refresh(ct, MSN_SIP_CLOSING );
		session_refresh(ct_msnms_info, MSN_SIP_CLOSING);
		rdr_clean(ct);
	}
	else
	{
		session_refresh(ct_msnms_info, MSN_SIP_TIMEOUT);
	}

	/* Initialize IP array to expected address (it's not mentioned
           in EPSV responses) */
	array[0] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 24) & 0xFF;
	array[1] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 16) & 0xFF;
	array[2] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 8) & 0xFF;
	array[3] = ntohl(ct->tuplehash[dir].tuple.src.ip) & 0xFF;

	for (i = 0; i < sizeof(sip_status) / sizeof(sip_status[0]); i++) {
	
		if (sip_status[i].dir != dir)
		{
			if(msn_debug)
				printk("sip_status[i].dir != %s \n", i==IP_CT_DIR_ORIGINAL ? "orig" : "reply" );
			continue;
		}

		//search for "Contact: <sip:a.b.c.d>\r\n" 
		found = find_pattern(CTX_TYPE_NONE, data, datalen,
				     sip_status[0].pattern,
				     sip_status[0].plen,
				     sip_status[0].skip,
				     sip_status[0].term,
				     &matchoff, &matchlen,
				     array,
				     sip_status[0].getnum);
		if( found )
		{
			if( cur_rec->invite_type == MSN_VIDEO_INVITE )
				find_media_port(data, datalen, &ct_msnms_info->media_port_callee[0]);
			break;
		}

	}
	if (found == -1) {
		if(msn_debug)
			printk("sip_status_help: pattern not found -1\n");
		/* We don't usually drop packets.  After all, this is
		   connection tracking, not packet filtering.
		   However, it is neccessary for accurate tracking in
		   this case. */
		return NF_DROP;
		
	} 
	else if (found == 0) /* No match */
	{
		if(msn_debug)
			printk("sip_status_help: pattern not found \n");
		return NF_ACCEPT;
	}
	/* do following action only if keyword matched */

	memset(&expect, 0, sizeof(expect));
	exp_msnms_info->cookie = cur_rec->cookie;
	
	/* Update the msnms info */
	LOCK_BH(&ip_msnms_lock);

	if( sip_status[i].msnmstype == IP_CT_SIP_STATUS ) {

		if(msn_debug)
			printk("sip_status_help: msnmstype=STATUS ");
		exp_msnms_info->msnmstype = sip_status[i].msnmstype;
		ct_msnms_info->is_msnp = 0;

#ifdef LAN_NET
		if( lan_net(ct->tuplehash[dir].tuple.src.ip)==0 ) //WAN reply STATUS : LAN INVITE
#else
		if( from_lan(ct) == 0 )
#endif

		{
#ifdef NO_REASSIGN
			ct_msnms_info->peer_addr = ct->tuplehash[dir].tuple.src.ip;
			ct_msnms_info->pub_addr = ct->tuplehash[dir].tuple.dst.ip;
#endif
			ct_msnms_info->peer_port = array[4] << 8 | array[5];
			//ct_msnms_info->priv_addr = master_ct_msnms_info->priv_addr;

			newsrcip = ct_msnms_info->priv_addr;
			newdstip = ct_msnms_info->peer_addr;
			newdstport = ct_msnms_info->peer_port;
		}
		else
		{
#ifdef NO_REASSIGN
			ct_msnms_info->peer_addr = ct->tuplehash[dir].tuple.dst.ip;
			ct_msnms_info->priv_addr = ct->tuplehash[dir].tuple.src.ip;
			ct_msnms_info->pub_addr = master_ct_msnms_info->pub_addr;
			ct_msnms_info->peer_port = 0;
#endif
			ct_msnms_info->pub_port = array[4] << 8 | array[5];
			ct_msnms_info->priv_port = array[4] << 8 | array[5];

			newsrcip = ct_msnms_info->peer_addr;
			newdstip = ct_msnms_info->priv_addr;
			newdstport = ct_msnms_info->priv_port;
		}
		if( cur_rec->invite_type == MSN_WHITEBOARD_INVITE )
		{
			newdstport = T120_PORT;
			newproto = IPPROTO_TCP;
		}
		else
		{
			newdstport = ct_msnms_info->media_port_callee[0];
			newproto = IPPROTO_UDP;
		}

		found = find_pattern(CTX_TYPE_NONE, data, datalen,
				     sip_statuslength[0].pattern,
				     sip_statuslength[0].plen,
				     sip_statuslength[0].skip,
				     sip_statuslength[0].term,
				     &matchoff, &matchlen,
				     array,
				     sip_statuslength[0].getnum);
		if (found )
			exp_msnms_info->ctx_len = array[4] << 8 | array[5];
		else exp_msnms_info->ctx_len = 0;
		
		exp_msnms_info->msnmstype = sip_statuslength[0].msnmstype;
				
	} 
	else {
		DEBUGP("conntrack_msnms: NOT RECORDING: %u,%u,%u,%u != %u.%u.%u.%u\n",
			array[0], array[1], array[2], array[3],
			NIPQUAD(ct->tuplehash[dir].tuple.src.ip));

		if (!loose) goto out;
	}


	/* T.120 , port 1503 TCP connection */
	ct_msnms_info->is_msnp = 0;
#if defined(CONFIG_IP_NF_PPTP) && defined(USE_RX3042_PPTP_PATCH)
	exp->tuple.src.ip = newsrcip;
	exp->tuple.src.u.all = 0;
	exp->tuple.dst.ip = newdstip;
	exp->tuple.dst.u.tcp.port = htons(newdstport);
	exp->tuple.dst.protonum = newproto;
	exp->mask.src.ip = 0xFFFFFFFF;
	exp->mask.src.u.all = 0;
	exp->mask.dst.ip = 0xFFFFFFFF;
	exp->mask.dst.u.tcp.port = 0xFFFF;
	exp->mask.dst.protonum = 0xFFFF;
	
#else
	exp->tuple = ((struct ip_conntrack_tuple)
		{ { newsrcip, { 0 } },		/* let this expect will not be matched */
		  { newdstip, { htons(newdstport) }, newproto }});
	exp->mask = ((struct ip_conntrack_tuple)
		{ { 0xFFFFFFFF, { 0 } },		/* let this expect will not be matched */
		  { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF }});
#endif

	exp->expectfn = NULL;
	
	/* Ignore failure; should only happen with NAT */
	if( ip_conntrack_expect_related(ct, exp) != 0 )
	{
		if(msn_debug)
			printk("sip_status_help: ip_conntrack_expect_related failed \n");
	}
 out:
 	UNLOCK_BH(&ip_msnms_lock);

	return NF_ACCEPT;

}

/* SIP helper is not registered! */
static struct ip_conntrack_helper sip_status_helper = 
	{ { NULL, NULL },
          "SIP Status helper",				/* name */
          IP_CT_HELPER_F_REUSE_EXPECT,		/* flags */
          NULL,					/* module */
          30,					/* max_ expected */
          30,					/* timeout */
          { { 0, { 0 } },			/* tuple */
            { 0, { 0 }, IPPROTO_TCP } },
          { { 0, { 0 } },			/* mask */
            { 0, { 0xFFFF }, 0xFFFF } },
          sip_status_help				/* helper */
	};

static int sip_request_expect(struct ip_conntrack *ct)
{
	struct ip_conntrack *master = master_ct(ct);
	struct ip_ct_msnms_master *ct_msnms_info = &ct->help.ct_msnms_info;
	struct ip_ct_msnms_master *master_ct_msnms_info = &master->help.ct_msnms_info;

	WRITE_LOCK(&ip_conntrack_lock);
	ct->helper = &sip_status_helper;
	DEBUGP("sip_request_expect: Add sip status helper for sip status port %p\n", ct);
	memcpy( ct_msnms_info, master_ct_msnms_info, sizeof(struct ip_ct_msnms_master));	
	WRITE_UNLOCK(&ip_conntrack_lock);
	
	return NF_ACCEPT;	/* unused */
}

/* SIP helper is not registered! */
static struct ip_conntrack_helper sip_request_helper = 
	{ { NULL, NULL },
          "SIP helper",				/* name */
          IP_CT_HELPER_F_REUSE_EXPECT,		/* flags */
          NULL,					/* module */
          0,					/* max_ expected */
          30,					/* timeout */
          { { 0, { 0 } },			/* tuple */
            { 0, { 1 }, IPPROTO_UDP } },
          { { 0, { 0 } },			/* mask */
            { 0, { 0xFFFF }, 0xFFFF } },
          sip_request_help				/* helper */
	};


/* FIXME: This should be in userspace.  Later. */
static int sip_request_help(const struct iphdr *iph, size_t len,
		     struct ip_conntrack *ct,
		     enum ip_conntrack_info ctinfo)
{
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	const char *data = (void *)udph + 8;
	unsigned int udplen = len - iph->ihl * 4;
	unsigned int datalen = udplen - sizeof(struct udphdr);
	u_int32_t array[6] = { 0 }, media_port[4] = { 0 };
	int dir = CTINFO2DIR(ctinfo);
	unsigned int matchlen=0, matchoff=0;
//	struct ip_conntrack *master = master_ct(ct);
	struct ip_ct_msnms_master *ct_msnms_info = &ct->help.ct_msnms_info;
//	struct ip_ct_msnms_master *master_ct_msnms_info = &master->help.ct_msnms_info;
	struct ip_conntrack_expect expect, *exp = &expect;
	struct ip_ct_msnms_expect *exp_msnms_info = &exp->help.exp_msnms_info;
	struct ip_ct_msnms_master new_rec, *cur_rec=&new_rec;
	u_int32_t newsrcip=0, newdstip=0, callid = 0;
	u_int16_t newdstport=0;
	unsigned int i;
	int found = 0;
	
	
	ct->is_spec = 1; /* skip NATA */
			
	DEBUGP("\nsip_request_help: help entered %u.%u.%u.%u:%u->%u.%u.%u.%u:%u\n",
		NIPQUAD(iph->saddr), ntohs(udph->source),
		NIPQUAD(iph->daddr), ntohs(udph->dest));

	/* Can't track connections formed before we registered */
	if( !ct_msnms_info )
		return NF_ACCEPT;

	/* Not whole UDP header */
	if (udplen < sizeof(struct udphdr)) {
		DEBUGP("sip_request_help: udplen = %u\n", (unsigned)udplen);
		return NF_ACCEPT;
	}

#ifndef SESSION_STRUC
	/* session rec */
/*
	found = find_pattern(data, datalen,
		     sip_callid[0].pattern,
		     sip_callid[0].plen,
		     sip_callid[0].skip,
		     sip_callid[0].term,
		     &matchoff, &matchlen,
		     array,
		     sip_callid[0].getnum);
	if( found != -1 && found != 0 )
		callid = array[4] << 8 | array[5] ;
*/
#ifdef LAN_NET
	cur_rec = session_find_by_ip( 1, iph->protocol, lan_net(ct->tuplehash[dir].tuple.src.ip),
		iph->saddr,iph->daddr,ntohs(udph->source),ntohs(udph->dest),callid );
#else
	cur_rec = session_find_by_ip( 1, iph->protocol, from_lan(ct),
		iph->saddr,iph->daddr,ntohs(udph->source),ntohs(udph->dest),callid );
#endif

	if( cur_rec == NULL || 
		(
			cur_rec->invite_type != MSN_WHITEBOARD_INVITE &&
			cur_rec->invite_type != MSN_WINMSG_FTP_INVITE &&
			cur_rec->invite_type != MSN_VIDEO_INVITE
		)
	)
	{
		if(msn_debug)
			printk("sip_request_help: session_find_by_ip() failed\n");
		return NF_ACCEPT;
	}
	exp_msnms_info->cookie = cur_rec->cookie;
	ct_msnms_info = cur_rec;
	ct_msnms_info->callid = callid;
	if(msn_debug)
		printk("sip_request_help: found cookie (%d) at index %d\n",
			cur_rec->cookie, cur_rec->index);
#endif

	if( strncmp(data, "BYE ", 4) ==0 )
	{
		if(msn_debug)
			printk("sip_request_help: --------- BYE SIP.... ------------\n");
		ct_msnms_info->bye = 1;
		//master_ct_msnms_info->bye = 1;
	}
	if( ct_msnms_info->bye )// | master_ct_msnms_info->bye )
	{
		ip_ct_refresh(ct, MSN_SIP_CLOSING );
		session_refresh( ct_msnms_info, MSN_SIP_CLOSING);
		rdr_clean(ct);
	}
	else
	{
		session_refresh( ct_msnms_info, MSN_SIP_TIMEOUT);
	}
	/* Initialize IP array to expected address (it's not mentioned
           in EPSV responses) */
	array[0] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 24) & 0xFF;
	array[1] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 16) & 0xFF;
	array[2] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 8) & 0xFF;
	array[3] = ntohl(ct->tuplehash[dir].tuple.src.ip) & 0xFF;

	for (i = 0; i < sizeof(sip_request) / sizeof(sip_request[0]); i++) {
		if (sip_request[i].dir != dir) continue;

		found = find_pattern(CTX_TYPE_NONE, data, datalen,
				     sip_request[0].pattern,
				     sip_request[0].plen,
				     sip_request[0].skip,
				     sip_request[0].term,
				     &matchoff, &matchlen,
				     array,
				     sip_request[0].getnum);
		if (found )
		{
			if( cur_rec->invite_type == MSN_VIDEO_INVITE )
				find_media_port(data, datalen, &ct_msnms_info->media_port_caller[0]);
			break;
		}

	}
	if (found == -1) {
		if(msn_debug)
			printk("sip help: pattern not found -1\n");
		/* We don't usually drop packets.  After all, this is
		   connection tracking, not packet filtering.
		   However, it is neccessary for accurate tracking in
		   this case. */
		return NF_DROP;
	} else if (found == 0) /* No match */
	{
		if(msn_debug)
			printk("sip_request_help: pattern not found \n");
		return NF_ACCEPT;
	}
	/* do following action only if keyword matched */
	
	memset(&expect, 0, sizeof(expect));
	exp_msnms_info->cookie = cur_rec->cookie;
	
	/* Update the msnms info */
	LOCK_BH(&ip_msnms_lock);

	/* don't use master_ct_msnms_info if possible becasue master_ct_msnms_info maynot contain
	   valid parameter becasue this session (SIP)'s parent may not have valid value */
	   
	if ( (sip_request[i].msnmstype == IP_CT_SIP_REQUEST ) ) {
		//&& (htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3])
		//    == ct->tuplehash[dir].tuple.src.ip) ) {
		
		exp_msnms_info->msnmstype = sip_request[i].msnmstype;
		ct_msnms_info->is_msnp = 0;

		if(msn_debug)
			printk("sip_request_help: msnmstype=REQUEST ");
#ifdef LAN_NET
		if( lan_net(ct->tuplehash[dir].tuple.src.ip) ) //LAN INVITE
#else
		if( from_lan(ct) ) //LAN INVITE
#endif

		{
#ifdef NO_REASSIGN
			ct_msnms_info->pub_addr = cur_rec->pub_addr; //ct->tuplehash[!dir].tuple.dst.ip;
			ct_msnms_info->peer_addr = ct->tuplehash[dir].tuple.dst.ip;
			ct_msnms_info->priv_addr = cur_rec->priv_addr;//ct->tuplehash[dir].tuple.src.ip;
#endif

			ct_msnms_info->pub_port = array[4] << 8 | array[5];
			ct_msnms_info->priv_port = array[4] << 8 | array[5];
	
			newsrcip = ct_msnms_info->peer_addr;
			newdstip = ct_msnms_info->pub_addr;
			newdstport = ct_msnms_info->pub_port;
		
		}
		else
		{
#ifdef NO_REASSIGN
			ct_msnms_info->pub_addr = ct->tuplehash[dir].tuple.dst.ip;
			ct_msnms_info->peer_addr = ct->tuplehash[dir].tuple.src.ip;
			ct_msnms_info->priv_addr = cur_rec->priv_addr;
			ct_msnms_info->pub_port = 0;
#endif

			ct_msnms_info->peer_port = array[4] << 8 | array[5];
	
			newsrcip = ct_msnms_info->priv_addr;
			newdstip = ct_msnms_info->peer_addr;
			newdstport = ct_msnms_info->peer_port;
		}

		ct_msnms_info->media_port_caller[0] = media_port[0] << 8 | media_port[1];
		ct_msnms_info->media_port_caller[2] = media_port[2] << 8 | media_port[3];

		found = find_pattern(CTX_TYPE_NONE, data, datalen,
				     sip_requestlength[0].pattern,
				     sip_requestlength[0].plen,
				     sip_requestlength[0].skip,
				     sip_requestlength[0].term,
				     &matchoff, &matchlen,
				     array,
				     sip_requestlength[0].getnum);
		if (found )
			exp_msnms_info->ctx_len = array[4] << 8 | array[5];
		else exp_msnms_info->ctx_len = 0;
		
		exp_msnms_info->msnmstype = sip_requestlength[0].msnmstype;


	} 
	else 
	{
		DEBUGP("sip_request_help: NOT RECORDING: %u,%u,%u,%u != %u.%u.%u.%u\n",
			array[0], array[1], array[2], array[3],
			NIPQUAD(ct->tuplehash[dir].tuple.src.ip));

		if (!loose) goto out;
	}

	/* SIP, Status: 100 Trying */
	if(msn_debug)
	{
		printk("sip_request_help: expect %u.%u.%u.%u -> %u.%u.%u.%u:%u \n",
			NIPQUAD(newsrcip), NIPQUAD(newdstip), newdstport);
	}
	ct_msnms_info->is_msnp = 0;

#if defined(CONFIG_IP_NF_PPTP) && defined(USE_RX3042_PPTP_PATCH)
	exp->tuple.src.ip = newsrcip;
	exp->tuple.src.u.all = 0;
	exp->tuple.dst.ip = newdstip;
	exp->tuple.dst.u.udp.port = htons(newdstport);
	exp->tuple.dst.protonum = IPPROTO_UDP;
	exp->mask.src.ip = 0xFFFFFFFF;
	exp->mask.src.u.all = 0;
	exp->mask.dst.ip = 0xFFFFFFFF;
	exp->mask.dst.u.udp.port = 0xFFFF;
	exp->mask.dst.protonum = 0xFFFF;
#else

	exp->tuple = ((struct ip_conntrack_tuple)
		{ { newsrcip, { 0 } },
		  { newdstip, { htons(newdstport) },  IPPROTO_UDP }});
	
	exp->mask = ((struct ip_conntrack_tuple)
		{ { 0xFFFFFFFF, { 0 } },
		  { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF }});
#endif

	exp->expectfn = sip_request_expect;
	
	/* Ignore failure; should only happen with NAT */
	if( ip_conntrack_expect_related(ct, exp) != 0 )
	{
		if(msn_debug)
			printk("sip_request_help: ip_conntrack_expect_related failed\n");
	}
 out:
 	UNLOCK_BH(&ip_msnms_lock);

	return NF_ACCEPT;

}



static int msnms_expect(struct ip_conntrack *ct)
{
	WRITE_LOCK(&ip_conntrack_lock);
	ct->helper = &sip_request_helper;
	DEBUGP("msnms_expect: Add sip request helper for sip request %p\n", ct);
	WRITE_UNLOCK(&ip_conntrack_lock);
	
	return NF_ACCEPT;	/* unused */
}


/* FIXME: This should be in userspace.  Later. */
static int msnms_help(const struct iphdr *iph, size_t len,
		struct ip_conntrack *ct,
		enum ip_conntrack_info ctinfo)
{
	/* tcplen not negative guaranteed by ip_conntrack_tcp.c */
	struct tcphdr *tcph = (void *)iph + iph->ihl * 4;
	const char *data = (const char *)tcph + tcph->doff * 4;
	unsigned int tcplen = len - iph->ihl * 4;
	unsigned int datalen = tcplen - tcph->doff * 4;
	u_int32_t array[6] = { 0 };
	int dir = CTINFO2DIR(ctinfo);
	unsigned int matchlen=0, matchoff=0;
	struct ip_ct_msnms_master *ct_msnms_info = &ct->help.ct_msnms_info;
	struct ip_conntrack_expect expect, *exp = &expect;
	struct ip_ct_msnms_expect *exp_msnms_info = &exp->help.exp_msnms_info;
	u_int32_t exp_sip = 0, exp_dip = 0;
	u_int16_t exp_sport = 0, exp_dport = 0;
	struct ip_ct_msnms_master new_rec, *cur_rec=&new_rec;
	u_int32_t cookie = 0;
	static u_int32_t count=0;
	unsigned int i;
	int found = 0;

	count++;
	if( msn_alg_enable == 0 )
		return NF_ACCEPT;

	if( !ct_msnms_info )
	{
		if(msn_debug)
			printk("msnms_help: ct_msnms_info = null \n");
		return NF_ACCEPT;
	}
		
	/* Until there's been traffic both ways, don't look in packets. */
	if (ctinfo != IP_CT_ESTABLISHED
	    && ctinfo != IP_CT_ESTABLISHED+IP_CT_IS_REPLY) {
		DEBUGP("msnms_help: Conntrackinfo = %u\n", ctinfo);
		return NF_ACCEPT;
	}

	/* Not whole TCP header? */
	if (tcplen < sizeof(struct tcphdr) || tcplen < tcph->doff*4) {
		DEBUGP("msnms: tcplen = %u\n", (unsigned)tcplen);
		return NF_ACCEPT;
	}

	/* Checksum invalid?  Ignore. */
	/* FIXME: Source route IP option packets --RR */
	if (tcp_v4_check(tcph, tcplen, iph->saddr, iph->daddr,
			 csum_partial((char *)tcph, tcplen, 0))) {
		DEBUGP("msnms_help: bad csum: %p %u %u.%u.%u.%u %u.%u.%u.%u\n",
		       tcph, tcplen, NIPQUAD(iph->saddr),
		       NIPQUAD(iph->daddr));
		return NF_ACCEPT;
	}


	if( datalen < MIN_CONTEXT_LEN )
		return NF_ACCEPT;

	if(msn_debug)
	{
		printk("\nmsnms_help: invitetype= %d %u.%u.%u.%u:%u->%u.%u.%u.%u\n", 
			ct_msnms_info->invite_type,
			NIPQUAD(iph->saddr), ntohs(tcph->source), NIPQUAD(iph->daddr) );
	}
	/* Initialize IP array to expected address (it's not mentioned
           in EPSV responses) */
	array[0] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 24) & 0xFF;
	array[1] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 16) & 0xFF;
	array[2] = (ntohl(ct->tuplehash[dir].tuple.src.ip) >> 8) & 0xFF;
	array[3] = ntohl(ct->tuplehash[dir].tuple.src.ip) & 0xFF;

	for (i = 0; i < sizeof(search) / sizeof(search[0]); i++) {
		if (search[i].dir != dir) continue;

		found = find_pattern(CTX_TYPE_MSMSGSINVITE, data, datalen,
				     search[i].pattern,
				     search[i].plen,
				     search[i].skip,
				     search[i].term,
				     &matchoff, &matchlen,
				     array,
				     search[i].getnum);
		if (found) 
		{
#ifndef SESSION_STRUC
			if (found != -1 && found != 0 )
			{
				int j,found2=0;
				u_int32_t array2[6] = { 0};
				
				for(j=0; j<sizeof(invite_cookie)/sizeof(invite_cookie[0]); j++) {
					found2 = find_pattern(CTX_TYPE_MSMSGSINVITE, data, datalen,
						invite_cookie[j].pattern,
						invite_cookie[j].plen,
						invite_cookie[j].skip,
						invite_cookie[j].term,
						&matchoff, &matchlen,
						array2,
						invite_cookie[j].getnum);
					if( found2 != -1 && found2 != 0 )
					{
						cookie = array2[4] << 8 | array2[5];// ok?
						if(msn_debug)
							printk("msnms_help: cookie found=%u \n", cookie );
					}
				}
			}
#endif
			if(msn_debug)
			{
				printk("msnms help: find_pattern %d, cookie=%u\n",
					search[i].msnmstype,
					cookie );
			}
			break;
		}
	}
	if (found == -1) {
		/* We don't usually drop packets.  After all, this is
		   connection tracking, not packet filtering.
		   However, it is neccessary for accurate tracking in
		   this case. */
		return NF_DROP;
	} else if (found == 0) /* No match */
		return NF_ACCEPT;

	memset(&expect, 0, sizeof(expect));
	memset(&new_rec, 0, sizeof(new_rec));

	if( search[i].msnmstype == IP_CT_MSNMS_GUID )
	{
		if(msn_debug)
			printk("============= initial ct_msnms_info============\n");
		//cannot, because struct contain other new value
		//: memset( ct_msnms_info,0,sizeof(struct ip_ct_msnms_master));
		ct_msnms_info->bye = 0;
		ct_msnms_info->invite_type = ( array[4] << 8 | array[5] );
#ifdef LAN_NET
		ct_msnms_info->inviter = lan_net(ct->tuplehash[dir].tuple.src.ip); /* INVITE by me if src from LAN */
#else
		ct_msnms_info->inviter = from_lan(ct); /* INVITE by me if src from LAN */
#endif
		if(msn_debug)
			printk("msnmstype = %d count=%d\n", ct_msnms_info->invite_type, count);
		//destroy_expectations(ct);
		//crash after patch to 2.4.26 remove_expectations(ct);
		
#ifndef SESSION_STRUC
		/* session rec */
		cur_rec = session_find( cookie );
		if( cur_rec == NULL )
		{
			cur_rec = &new_rec;
			cur_rec->cookie = cookie;
			cur_rec = session_add( &new_rec );
			
		}
		if( cur_rec )
		{
			cur_rec->bye = 0;
			cur_rec->invite_type = ( array[4] << 8 | array[5] );
#ifdef LAN_NET
			cur_rec->inviter = lan_net(ct->tuplehash[dir].tuple.src.ip);
#else
			cur_rec->inviter = from_lan(ct);
#endif
			cur_rec->proto = ct_msnms_info->invite_type==MSN_WINMSG_FTP_INVITE ? IPPROTO_TCP: IPPROTO_UDP;
			if(msn_debug)
			{
				printk("msnms_help: count=%d invite_type=%d cookie added to index=%d\n", 
					count,cur_rec->invite_type, cur_rec ? cur_rec->index : -1 );
			}
		}
#endif
				
	}

#ifndef SESSION_STRUC
	/* session rec */
	cur_rec = session_find( cookie );
	if( cookie == 0 || cur_rec == NULL || 
		(
			cur_rec->invite_type != MSN_WHITEBOARD_INVITE &&
			cur_rec->invite_type != MSN_WINMSG_FTP_INVITE &&
			cur_rec->invite_type != MSN_VIDEO_INVITE
		)
	)
	{
		if(msn_debug)
		{
			printk("msnms_help: count=%d cookie=%u not found, msninvite_type=%d ,ct_msnms:invite_type=%d skipped\n",cookie, 
				count, cur_rec->invite_type,
				ct_msnms_info->invite_type);
		}
		return NF_ACCEPT;
	}
	exp_msnms_info->cookie = cookie;
	ct_msnms_info = cur_rec;
	if(msn_debug)
	{
		printk("msnms_help: cookie=%u cur_rec->index=%d, invite_type=%d \n", 
			cur_rec->cookie, cur_rec->index, cur_rec->invite_type);
	}
#endif


#ifdef OLD_METHOD
	/* do following action only if keyword matched */
	if( ct_msnms_info->invite_type != MSN_WHITEBOARD_INVITE &&
		ct_msnms_info->invite_type != MSN_WINMSG_FTP_INVITE )
	{
		if(msn_debug)
			printk("msnms_help: invite_type=%d, not whiteboard/winmsg ftp, skip\n",ct_msnms_info->invite_type);
		return NF_ACCEPT; //do nothing if not whiteboard
	}
#endif
	
	/* Update the msnms info */
	LOCK_BH(&ip_msnms_lock);

	if ( search[i].msnmstype == IP_CT_MSNMS_CANCEL ||
		search[i].msnmstype == IP_CT_MSNMS_REPLYCANCEL)
	{
		session_del( cookie );
		if(msn_debug)
			printk("msnms_help: cancel \n");
		UNLOCK_BH(&ip_msnms_lock);
		return NF_ACCEPT;
	}
	/* WINDOWS MESSENGER FTP */
	else if ( search[i].msnmstype == IP_CT_MSNMS_ACCEPT &&
		ct_msnms_info->invite_type == MSN_WINMSG_FTP_INVITE &&
		 (htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3])
			== ct->tuplehash[dir].tuple.src.ip)  )
	{
		exp->seq = ntohl(tcph->seq) + matchoff;	// useful ? cause msnms_nat_help() not invoke?
		exp_msnms_info->msnmstype = search[i].msnmstype;
		exp_msnms_info->len = matchlen;

		found = find_pattern(CTX_TYPE_NONE, data, datalen,
				     winmsg_port[0].pattern,
				     winmsg_port[0].plen,
				     winmsg_port[0].skip,
				     winmsg_port[0].term,
				     &matchoff, &matchlen,
				     array,
				     winmsg_port[0].getnum);

		ct_msnms_info->is_msnp = 0;
		ct_msnms_info->priv_port = found ?  (array[4] << 8 | array[5]) : (WINMSG_FTP_PORT);
		ct_msnms_info->priv_addr = ct->tuplehash[dir].tuple.src.ip;
		ct_msnms_info->pub_addr = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;	

#ifndef SESSION_STRUC
		cur_rec->is_msnp = 0;
		cur_rec->priv_port = found ?  (array[4] << 8 | array[5]) : (WINMSG_FTP_PORT);
		cur_rec->priv_addr = ct->tuplehash[dir].tuple.src.ip;
		cur_rec->pub_addr = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;	
#endif
		
		exp_sip = 0;
		exp_dip = ct->tuplehash[dir].tuple.src.ip;
		exp_sport = 0;
		exp_dport = htons(ct_msnms_info->priv_port);
		//exp_dport = htons(WINMSG_FTP_PORT);

#if defined(CONFIG_IP_NF_PPTP) && defined(USE_RX3042_PPTP_PATCH)
	exp->tuple.src.ip = exp_sip;
	exp->tuple.src.u.all = 0;
	exp->tuple.dst.ip = exp_dip;
	exp->tuple.dst.u.udp.port = exp_dport;
	exp->tuple.dst.protonum = IPPROTO_TCP;
	exp->mask.src.ip = 0;
	exp->mask.src.u.all = 0;
	exp->mask.dst.ip = 0xFFFFFFFF;
	exp->mask.dst.u.udp.port = 0xFFFF;
	exp->mask.dst.protonum = 0xFFFF;
#else
		exp->tuple = ((struct ip_conntrack_tuple)
			{ { exp_sip, { 0 } },
			  { exp_dip, { exp_dport }, IPPROTO_TCP }});

		exp->mask = ((struct ip_conntrack_tuple)
			{ { 0, { 0 } },
			  { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF }});
#endif

		exp->expectfn = NULL;

		if(msn_debug)
		{
			printk("msnms help: cookie=%u winmsg ftp msnmstype=%d,invitetype=%d expect: %u.%u.%u.%u -> %u.%u.%u.%u:%u \n",
				exp_msnms_info->cookie, exp_msnms_info->msnmstype, ct_msnms_info->invite_type,
				NIPQUAD(exp_sip), NIPQUAD(exp_dip), ntohs(exp_dport) );
		}
		/* Ignore failure; should only happen with NAT */
		if( ip_conntrack_expect_related(ct, &expect) != 0 )
		{
			if(msn_debug)
				printk("msnms help: ftp ip_conntrack_expect_related failed \n");	
		}
		UNLOCK_BH(&ip_msnms_lock);

		/* NOTE : missing peer invite implied rule */
		
		return NF_ACCEPT;
	
	}
	else if ( (search[i].msnmstype == IP_CT_MSNMS_ACCEPT||
		search[i].msnmstype == IP_CT_MSNMS_CONTEXT) 
		//because of invert dir 
		//&& (htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3])
		//	== ct->tuplehash[dir].tuple.src.ip) 
		) {
		
		exp->seq = ntohl(tcph->seq) + matchoff;
		exp_msnms_info->len = matchlen;
		exp_msnms_info->msnmstype = search[i].msnmstype;

		ct_msnms_info->is_msnp = 1;
		ct_msnms_info->priv_addr = htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3]);
		if( ct_msnms_info->priv_port == 0 )
			ct_msnms_info->priv_port = array[4] << 8 | array[5];
		if( ct_msnms_info->pub_port == 0 ) /* set only if port==0 */
			ct_msnms_info->pub_port = array[4] << 8 | array[5];
		if( ct_msnms_info->pub_addr == 0 )
			ct_msnms_info->pub_addr = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;

#ifndef SESSION_STRUC
		cur_rec->is_msnp = 1;
		cur_rec->priv_addr = htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3]);
		if( cur_rec->priv_port == 0 )
			cur_rec->priv_port = array[4] << 8 | array[5];
		if( cur_rec->pub_port == 0 ) /* set only if port==0 */
			cur_rec->pub_port = array[4] << 8 | array[5];
		if( cur_rec->pub_addr == 0 )
			cur_rec->pub_addr = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
#endif		
		if( ct_msnms_info->inviter && 
			ct_msnms_info->invite_type != MSN_VIDEO_INVITE)
		{
	
			//should not initialized the following value 
			//ct_msnms_info->pub_addr = ct->tuplehash[dir].tuple.src.ip;
			//ct_msnms_info->peer_addr = 0; //unknown yet test
			//ct_msnms_info->peer_addr = 0;
			exp_sip = ct_msnms_info->priv_addr;
			exp_dip = ct_msnms_info->peer_addr;
			exp_sport = 0;
			exp_dport = htons(ct_msnms_info->peer_port);
			
		}
		else
		{	
			//should not initialized the following value 
			//ct_msnms_info->pub_addr = ct->tuplehash[dir].tuple.src.ip;
			//ct_msnms_info->peer_addr = 0; //unknown yet test
			//ct_msnms_info->peer_addr = 0;
			exp_sip = ct_msnms_info->peer_addr;
			exp_dip = ct_msnms_info->priv_addr;
			exp_sport = 0;
			exp_dport = htons(ct_msnms_info->priv_port);
		}
		if(msn_debug)
		{
			printk("msnms help: ACCEPT/CONTEXT matched port=%d src ip=%u.%u.%u.%u \n", 
				ct_msnms_info->priv_port,NIPQUAD(ct->tuplehash[dir].tuple.src.ip) );
		}
		
	} else if( search[i].msnmstype == IP_CT_MSNMS_REPLYACCEPT ) {

		//memset( ct_msnms_info, 0, sizeof( struct ip_ct_msnms_master ) );
		exp_msnms_info->msnmstype = search[i].msnmstype;

		ct_msnms_info->peer_addr = htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3]);
		//ct_msnms_info->pub_addr = ct->tuplehash[dir].tuple.dst.ip;
		//initial for new INVITATION
		ct_msnms_info->priv_port = 0;
		ct_msnms_info->pub_port = 0;
		ct_msnms_info->peer_port = 0;
		//if( ct_msnms_info->pub_addr == 0 )
		ct_msnms_info->pub_addr = ct->tuplehash[dir].tuple.dst.ip;

#ifndef SESSION_STRUC
		cur_rec->peer_addr = htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3]);
		//ct_msnms_info->pub_addr = ct->tuplehash[dir].tuple.dst.ip;
		//initial for new INVITATION
		cur_rec->priv_port = 0;
		cur_rec->pub_port = 0;
		cur_rec->peer_port = 0;
		//if( ct_msnms_info->pub_addr == 0 )
		cur_rec->pub_addr = ct->tuplehash[dir].tuple.dst.ip;
#endif

		if(msn_debug)
		{
			printk("msnms help: REPLYACCEPT: peer_addr = %u.%u.%u.%u \n", 
				NIPQUAD(ct_msnms_info->peer_addr) );
		}

	} else if( search[i].msnmstype == IP_CT_MSNMS_REPLYCONTEXT ) {
		/* we invite */
		exp_msnms_info->msnmstype = search[i].msnmstype;

		ct_msnms_info->peer_addr = htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3]);
		ct_msnms_info->peer_port = array[4] << 8 | array[5];
		//ct_msnms_info->from_lan = 1;
#ifndef SESSION_STRUC
		cur_rec->peer_addr = htonl((array[0] << 24) | (array[1] << 16) | (array[2] << 8) | array[3]);
		cur_rec->peer_port = array[4] << 8 | array[5];
#endif
		exp_sip = ct_msnms_info->priv_addr;
		exp_dip = ct_msnms_info->peer_addr;
		exp_sport = 0;
		exp_dport = htons(ct_msnms_info->peer_port);
	
	} else {
		/* Enrico Scholz's passive MSNMS to partially RNAT'd msnms
		   server: it really wants us to connect to a
		   different IP address.  Simply don't record it for
		   NAT. */
		DEBUGP("msnms help: NOT RECORDING: %u,%u,%u,%u != %u.%u.%u.%u\n",
		       array[0], array[1], array[2], array[3],
		       NIPQUAD(ct->tuplehash[dir].tuple.src.ip));

		/* Thanks to Cristiano Lincoln Mattos
		   <lincoln@cesar.org.br> for reporting this potential
		   problem (DMZ machines opening holes to internal
		   networks, or the packet filter itself). */
		if (!loose) goto out;
	}

	if(msn_debug)
	{
		printk("msnms help: peer_addr=%04x port=%d pub_addr=%04x port=%d, private: %04x:%d \n", 
			ct_msnms_info->peer_addr, 
			ct_msnms_info->peer_port,
			ct_msnms_info->pub_addr, 
			ct_msnms_info->pub_port,
			ct_msnms_info->priv_addr, 
			ct_msnms_info->priv_port );
	}
	if(msn_debug)
	{
	if( exp_msnms_info->msnmstype == IP_CT_MSNMS_CONTEXT )
		printk("msnms help: type: IP_CT_MSNMS_CONTEXT\n");
	else if( exp_msnms_info->msnmstype == IP_CT_MSNMS_ACCEPT )
		printk("msnms help: type: IP_CT_MSNMS_ACCEPT\n");
	else if( exp_msnms_info->msnmstype == IP_CT_MSNMS_REPLYCONTEXT )
		printk("msnms help: type: IP_CT_MSNMS_REPLYCONTEXT\n");
	}
	if( exp_msnms_info->msnmstype == IP_CT_MSNMS_CONTEXT ||
		exp_msnms_info->msnmstype == IP_CT_MSNMS_ACCEPT ||
		exp_msnms_info->msnmstype == IP_CT_MSNMS_REPLYCONTEXT )	
	{
		if(msn_debug)
		{
			printk("msnms help: expecting: %u.%u.%u.%u ->%u.%u.%u.%u:%u \n",
				NIPQUAD(exp_sip), NIPQUAD(exp_dip), ntohs(exp_dport) );
		}

#if defined(CONFIG_IP_NF_PPTP) && defined(USE_RX3042_PPTP_PATCH)
	exp->tuple.src.ip = exp_sip;
	exp->tuple.src.u.all = 0;
	exp->tuple.dst.ip = exp_dip;
	exp->tuple.dst.u.udp.port = exp_dport;
	exp->tuple.dst.protonum = IPPROTO_UDP;
	exp->mask.src.ip = 0xFFFFFFFF;
	exp->mask.src.u.all = 0;
	exp->mask.dst.ip = 0xFFFFFFFF;
	exp->mask.dst.u.udp.port = 0xFFFF;
	exp->mask.dst.protonum = 0xFFFF;
#else
		exp->tuple = ((struct ip_conntrack_tuple)
			{ { exp_sip, { 0 } },
			  { exp_dip, { exp_dport }, IPPROTO_UDP }});

		exp->mask = ((struct ip_conntrack_tuple)
			{ { 0xFFFFFFFF, { 0 } },
			  { 0xFFFFFFFF, { 0xFFFF }, 0xFFFF }});
#endif

		exp->expectfn = msnms_expect;

		/* Ignore failure; should only happen with NAT */
		if( ip_conntrack_expect_related(ct, &expect) != 0 )
		{
			if(msn_debug)
				printk("msnms help: ip_conntrack_expect_related failed \n");
		}
	}
 out:
	UNLOCK_BH(&ip_msnms_lock);

	return NF_ACCEPT;
}

static struct ip_conntrack_helper msnms[MAX_PORTS];
static char msnms_names[MAX_PORTS][10];


/* Not __exit: called from init() */
static void fini(void)
{
	int i;
	for (i = 0; i < ports_c; i++) {
		DEBUGP("ip_ct_msnms: unregistering helper for port %d\n",
				ports[i]);
		ip_conntrack_helper_unregister(&msnms[i]);
	}
	
//	t120_unregister();
}

static int __init init(void)
{
	int i, ret;
	char *tmpname;

	if (ports[0] == 0)
		ports[0] = MSNMS_PORT;//test

	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {
		memset(&msnms[i], 0, sizeof(struct ip_conntrack_helper));
		msnms[i].tuple.src.u.tcp.port = htons(ports[i]);
		msnms[i].tuple.dst.protonum = IPPROTO_TCP;
		msnms[i].mask.src.u.tcp.port = 0xFFFF;
		msnms[i].mask.dst.protonum = 0xFFFF;
		msnms[i].max_expected = 30;
		msnms[i].timeout = 180;
		msnms[i].flags = IP_CT_HELPER_F_REUSE_EXPECT;
		msnms[i].me = ip_conntrack_msnms;
		msnms[i].help = msnms_help;

		tmpname = &msnms_names[i][0];
		if (ports[i] == MSNMS_PORT)
			sprintf(tmpname, "msnms");
		else
			sprintf(tmpname, "msnms-%d", ports[i]);
		msnms[i].name = tmpname;

		DEBUGP("ip_ct_msnms: registering helper for port %d\n", 
				ports[i]);
		ret = ip_conntrack_helper_register(&msnms[i]);

		if (ret) {
			fini();
			return ret;
		}
		ports_c++;
	}
	
//	t120_register();
	memset(&msnms_session[0],0,sizeof(msnms_session));
	
	return 0;
}

#ifdef CONFIG_IP_NF_NAT_NEEDED
EXPORT_SYMBOL(ip_msnms_lock);
#endif

MODULE_LICENSE("GPL");
module_init(init);
module_exit(fini);
