#include <linux/types.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/in.h>
#include <linux/udp.h>
#include <linux/netfilter_ipv4/ip_conntrack_protocol.h>
#include <linux/ip.h>
#include <net/checksum.h>
#include <net/udp.h>

#define UDP_TIMEOUT (120*HZ)	/* for msn case ? */
#define UDP_DNS_TIMEOUT	(30*HZ)
#define UDP_STREAM_TIMEOUT (610*HZ)	/* SIP, DNS, ... */

#define DNS_PORT	53

static int udp_pkt_to_tuple(const void *datah, size_t datalen,
			    struct ip_conntrack_tuple *tuple)
{
	const struct udphdr *hdr = datah;

	tuple->src.u.udp.port = hdr->source;
	tuple->dst.u.udp.port = hdr->dest;

	return 1;
}

static int udp_invert_tuple(struct ip_conntrack_tuple *tuple,
			    const struct ip_conntrack_tuple *orig)
{
	tuple->src.u.udp.port = orig->dst.u.udp.port;
	tuple->dst.u.udp.port = orig->src.u.udp.port;
	return 1;
}

/* Print out the per-protocol part of the tuple. */
static unsigned int udp_print_tuple(char *buffer,
				    const struct ip_conntrack_tuple *tuple)
{
	return sprintf(buffer, "sport=%hu dport=%hu ",
		       ntohs(tuple->src.u.udp.port),
		       ntohs(tuple->dst.u.udp.port));
}

/* Print out the private part of the conntrack. */
static unsigned int udp_print_conntrack(char *buffer,
					const struct ip_conntrack *conntrack)
{
	return 0;
}

/* Returns verdict for packet, and may modify conntracktype */
static int udp_packet(struct ip_conntrack *conntrack,
		      struct iphdr *iph, size_t len,
		      enum ip_conntrack_info conntrackinfo)
{
	/* norton */
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	
	/* If we've seen traffic both ways, this is some kind of UDP
	   stream.  Extend timeout. */
	if (conntrack->status & IPS_SEEN_REPLY) {

		if( udph->dest == htons(DNS_PORT)|| udph->source == htons(DNS_PORT) )
			ip_ct_refresh(conntrack, UDP_DNS_TIMEOUT);
		else
			ip_ct_refresh(conntrack, UDP_STREAM_TIMEOUT);

		/* Also, more likely to be important, and not a probe */
		set_bit(IPS_ASSURED_BIT, &conntrack->status);

		/* 
		 * ip_ct_refresh(conntrack, 1*HZ ); 
		 */
		/* norton : 1 sec timeout for broadcast packet,
		 * unable to skip creation because of dhcp packet
		 * don't know how to skip dhcp packet creation
		 */
	} 
	else
	{
		if( udph->dest == htons(DNS_PORT) || udph->source == htons(DNS_PORT) )
			ip_ct_refresh(conntrack, UDP_DNS_TIMEOUT);
		else
			ip_ct_refresh(conntrack, UDP_TIMEOUT);
	}
	return NF_ACCEPT;
}

/* Called when a new connection for this protocol found. */
static int udp_new(struct ip_conntrack *conntrack,
			     struct iphdr *iph, size_t len)
{
	/* norton ip_ct_refresh(conntrack, UDP_TIMEOUT); */
	return 1;
}

struct ip_conntrack_protocol ip_conntrack_protocol_udp
= { { NULL, NULL }, IPPROTO_UDP, "udp",
    udp_pkt_to_tuple, udp_invert_tuple, udp_print_tuple, udp_print_conntrack,
    udp_packet, udp_new, NULL, NULL, NULL };
