/* IRC extension for TCP NAT alteration.
 * (C) 2000-2001 by Harald Welte <laforge@gnumonks.org>
 * based on a copy of RR's ip_nat_ftp.c
 *
 * ip_nat_irc.c,v 1.16 2001/12/06 07:42:10 laforge Exp
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version
 *      2 of the License, or (at your option) any later version.
 *
 *	Module load syntax:
 * 	insmod ip_nat_irc.o ports=port1,port2,...port<MAX_PORTS>
 *	
 * 	please give the ports of all IRC servers You wish to connect to.
 *	If You don't specify ports, the default will be port 6667
 */

#include <linux/module.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/kernel.h>
#include <net/tcp.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#define ETH0_IFINDEX	2

int from_lan(struct ip_conntrack *ct)
{
	if( ct->ct_general.recvifnum == ETH0_IFINDEX ) /* eth0 ifindex */
		return 1;
	
	return 0;
}


unsigned char is_pptp=0;

int K_Get_IP(char* interface);

//struct ip_conntrack *myct __attribute__ ((aligned (4)));
unsigned int ROUTER_OUT_IP;

extern unsigned int ROUTER_LOCAL_MASK;
extern unsigned int ROUTER_LOCAL_IP;

#define SPEC_TTL	864000
#define MAX_SPEC_CT	10

struct spec_list
{
	unsigned char  flag;
	unsigned short out_port_start ;
	unsigned short out_port_end ;
	
	unsigned short in_port_start ;
	unsigned short in_port_end ;

	unsigned int  dst_ip ;
	struct ip_conntrack *spec_ct[MAX_SPEC_CT] ;
	unsigned int  ct_count ;	//MAX is 10
	
	
	
}__attribute__ ((aligned (4)));


struct spec_list alg[MAX_SPEC_CT] __attribute__ ((aligned (4)));


static struct ip_conntrack_helper ip_conntrack_spec_helpers;
static char ip_nih_names[6];

static int spec_list_count;

/* remove trigger rule: rdr rule */
void Find_ALG_CT(struct ip_conntrack* ct)
{
	rdr_clean( ct );
	
#ifdef XX
	signed char i,j;

	for(i=0;(i<MAX_SPEC_CT)&&alg[0].flag;i++)
	{
		for(j=0;j<MAX_SPEC_CT;j++)
		{
			if(ct==alg[i].spec_ct[j])
			{
				alg[i].ct_count--;
				alg[i].spec_ct[j]=NULL;
				//return;
			}
		}
	}
#endif
	
}


signed char Find_ALG(unsigned short port,int flag,unsigned int proto)
{
	signed char i;
	
	
	if(!spec_list_count)
	{
		DEBUGP("no spec\n");
		return -1;
	}
	
	if(flag==0)	//out
	{
		for(i=0;(i<spec_list_count);i++)
		{
			if(alg[i].flag) /* active rule */
				if((port>=alg[i].out_port_start)&&(port<=alg[i].out_port_end))
				{
					return i;
				}
		}
	}
	else		//in
	{
		for(i=0;(i<spec_list_count);i++)
		{
			if(alg[i].flag) /* active rule */
				if((port>=alg[i].in_port_start)&&(port<=alg[i].in_port_end))
				{
					//printk("Found in port i==%d\n",i);
					return i;
				}
		}
	}

	return -1;
	
}



//static char first=0;

/* protects irc part of conntracks */
//DECLARE_LOCK_EXTERN(ip_spec_lock);
DECLARE_LOCK(ip_spec_lock);
struct module *ip_conntrack_spec = THIS_MODULE;

/* FIXME: Time out? --RR */

#define PROTOCOL_TCP 6
#define PROTOCOL_UDP 11
#define PROTOCOL_TCP_AND_UDP 255


static int conntrack_spec_help(const struct iphdr *iph, size_t len,
		struct ip_conntrack *ct,
		enum ip_conntrack_info ctinfo)

{
	int dir ;
	signed char count=-1 ;
	unsigned short myport=0;
	dir = CTINFO2DIR(ctinfo);

//should not do 
//	if (ct == NULL || ctinfo != IP_CT_NEW || dir != IP_CT_DIR_ORIGINAL )
//		return NF_ACCEPT;

//	printk("routeip=%u.%u.%u.%u  , %u.%u.%u.%u -> %u.%u.%u.%u \n",
//		NIPQUAD(ROUTER_OUT_IP), 
//		NIPQUAD(iph->saddr), NIPQUAD(iph->daddr));

	/* Special application : outbound */
	if( ( dir == IP_CT_DIR_ORIGINAL )&& (ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip !=ROUTER_OUT_IP )
		&& (ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip !=ROUTER_LOCAL_IP ))
	{
#ifdef LAN_NET
		if(((iph->saddr)&0x00ffffff)!=(ROUTER_LOCAL_IP&0x00ffffff))
			return NF_ACCEPT;
#else
		if( from_lan( ct ) == 0 )
			return NF_ACCEPT;		
#endif

		if( ct->is_spec > 1 ) /* return only if rdr_add() called */
			return NF_ACCEPT;

		if(ct->tuplehash[dir].tuple.dst.protonum==IPPROTO_TCP)
		{
			if(!((ct->status)&(1 << IPS_ASSURED_BIT)))
			{
				//printk("SN \n");
				return NF_ACCEPT;
			}
		}

		LOCK_BH(&ip_spec_lock);
		
		if((ct->tuplehash[dir].tuple.dst.protonum==IPPROTO_TCP)||
			(ct->tuplehash[dir].tuple.dst.protonum==IPPROTO_UDP))
		{
			myport = ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u.tcp.port);
			count=Find_ALG(myport,0 /* out port */,ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum);

			DEBUGP("Find_ALG=%d : from %u.%u.%u.%u to %u.%u.%u.%u port %d (%d) \n",
				count, NIPQUAD(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip), NIPQUAD(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip), myport ,
				ntohs(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.u.udp.port) );

			if(count>=0)
			{
				ct->is_spec=1; /* skip NATA accelerate on this session */

				/* should not skip UDP checking ASSURED bit , otherwise rdr_add() will add a rule without NATed IP address */
				if(ct->tuplehash[dir].tuple.dst.protonum==IPPROTO_UDP)
				{
					//if(!((ct->status)&(1 << IPS_ASSURED_BIT)))
					if( !(ct->status & IPS_SEEN_REPLY))
					{
						DEBUGP("No reply \n");
						UNLOCK_BH(&ip_spec_lock);
						return NF_ACCEPT;
					}
				}
				
				//if((alg[count].spec_ct==NULL)||(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip==alg[count].dst_ip))
				if(alg[count].ct_count==0)
				{
					u_int32_t dip = 0, tip = 0;

					tip = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
					dip = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
					rdr_add( ct, IPPROTO_TCP, ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip, dip, 0,0, alg[count].in_port_start, alg[count].in_port_end, 
						tip, alg[count].in_port_start, alg[count].in_port_end, SPEC_TTL );
					rdr_add( ct, IPPROTO_UDP, ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip, dip, 0,0, alg[count].in_port_start, alg[count].in_port_end, 
						tip, alg[count].in_port_start, alg[count].in_port_end, SPEC_TTL );

					DEBUGP("rdr_added from %u.%u.%u.%u to %u.%u.%u.%u \n",
						NIPQUAD(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip), NIPQUAD(dip) );
					ct->is_spec=2; /* skip NATA accelerate on this session */

					/*
					alg[count].dst_ip=ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
					//alg[count].proto=ct->tuplehash[dir].tuple.dst.protonum;
					alg[count].spec_ct[alg[count].ct_count]=ct;
					alg[count].ct_count++;
					ct->is_spec=1;
					
					DEBUGP("The Local Host  is %u.%u.%u.%u\n",NIPQUAD(alg[count].dst_ip));
					DEBUGP("The Remote Host  is %u.%u.%u.%u\n",NIPQUAD(ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip));
					*/
				}
				UNLOCK_BH(&ip_spec_lock);
			
				return NF_ACCEPT;
			}
		
		
		}
	
		UNLOCK_BH(&ip_spec_lock);
			
		return NF_ACCEPT;
		//printk("before NF_ACCEPT 000\n");
		//return NF_QUEUE;
	}


#ifdef XXXX
	
	//if((hooknum == NF_IP_PRE_ROUTING)&&( dir == IP_CT_DIR_ORIGINAL ))
	/* Special Application: inbound */
	if (( dir == IP_CT_DIR_ORIGINAL )&&((ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip) ==ROUTER_OUT_IP))
	{
		//printk("reply packet? \n");
		
		if(ct->is_spec_reply)
			return NF_ACCEPT;

#ifdef LAN_NET
		if(((iph->saddr)&0x00ffffff)==(ROUTER_LOCAL_IP&0x00ffffff))
			return NF_ACCEPT;
#else
		if( lan_net( ct )
			return NF_ACCEPT;
#endif
		LOCK_BH(&ip_spec_lock);

		
		if((ct->tuplehash[dir].tuple.dst.protonum==IPPROTO_TCP)||(ct->tuplehash[dir].tuple.dst.protonum==IPPROTO_UDP))
		{
		
			myport=ntohs(ct->tuplehash[dir].tuple.dst.u.tcp.port);
			count=Find_ALG(myport,1,ct->tuplehash[dir].tuple.dst.protonum);
		
			if(count>=0)
			{

				if((myport==1080)||(myport==3128))
				{
					UNLOCK_BH(&ip_spec_lock);
					DEBUGP(" Do not forward SOCKS or Proxy request !\n");
					return NF_ACCEPT;
				}

				if(alg[count].ct_count>0)
				{
					//printk("before ip_conntrack_expect_related <---> %d\n",count);
				
					DEBUGP(" I catch the income spec link and port == %d <---> proto: %d\n",myport,ct->tuplehash[dir].tuple.dst.protonum);
			
					//DEBUGP("BACK Local Host  is %u.%u.%u.%u\n",NIPQUAD(ct->tuplehash[dir].tuple.dst.ip));
					//DEBUGP("BACK Remote Host  is %u.%u.%u.%u\n",NIPQUAD(ct->tuplehash[dir].tuple.src.ip));
					//DEBUGP("BACK protocal num is   is %d\n",ct->tuplehash[dir].tuple.dst.protonum);
				
					struct ip_nat_multi_range mr;
					u_int32_t newdstip;
					int i;
				
					//test norton
					//if(ct->tuplehash[dir].tuple.dst.protonum==IPPROTO_TCP)
					{
					
						if(alg[count].ct_count>=MAX_SPEC_CT)
						{
							printk("Reply MAX CT !\n");
							//alg[count].ct_count=MAX_SPEC_CT;
							UNLOCK_BH(&ip_spec_lock);
							return NF_ACCEPT;
						}
						else
						{
							for(i=0;i<=alg[count].ct_count;i++)
							{
								if(alg[count].spec_ct[i]==NULL)
								{
									alg[count].spec_ct[i]=ct;
									alg[count].ct_count++;
									//printk("Reply TCP,now ct count is %d\n",alg[count].ct_count);
									break;
								}
							}
						}
					
					}
				
					newdstip = alg[count].dst_ip;
					mr.rangesize = 1;
					/* We don't want to manip the per-protocol, just the IPs. */
					mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
					mr.range[0].min_ip = mr.range[0].max_ip = newdstip;
					UNLOCK_BH(&ip_spec_lock);
					ct->is_spec_reply=1;
			
					return ip_nat_setup_info(ct, &mr, NF_IP_PRE_ROUTING);
				}
			
				UNLOCK_BH(&ip_spec_lock);
				return NF_ACCEPT;
			}
	

			if( myport == 1509 )
			{
				struct ip_nat_multi_range mr;
				u_int32_t newdstip;
				struct tcphdr *tcph = (void *)iph + iph->ihl*4;

				newdstip = 0x0701a8c0;
				mr.rangesize = 1;
				mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
				mr.range[0].min_ip = mr.range[0].max_ip = newdstip;
				
				UNLOCK_BH(&ip_spec_lock);
				//test ct->is_spec_reply=1;
			
				DEBUGP("v2b forwarding: %u.%u.%u.%u -> %u.%u.%u.%u (%u->%u) \n", 
					NIPQUAD(iph->saddr), NIPQUAD(newdstip),
					ntohs(tcph->source), ntohs(tcph->dest));

				return ip_nat_setup_info(ct, &mr, NF_IP_PRE_ROUTING);
			
			}

			UNLOCK_BH(&ip_spec_lock);
			return NF_ACCEPT;

	
		}//adjust protocal
		UNLOCK_BH(&ip_spec_lock);
		return NF_ACCEPT;
	
	}
	//UNLOCK_BH(&ip_spec_lock);
#endif

	
	return NF_ACCEPT;
	
}

//static struct ip_nat_expect spec_expect
//    = { {NULL, NULL}, spec_nat_expected };

/* This function is intentionally _NOT_ defined as  __exit, because
 * it is needed by init() */
static void fini(void)
{
	ip_conntrack_helper_unregister(&ip_conntrack_spec_helpers);

}

#if 0

modify call.S
.long	SYMBOL_NAME(sys_nat_spec_reset)//lsshi add

modify unistd.h
#define __NR_nat_spec_reset		(__NR_SYSCALL_BASE+218)//lsshi add

declare it
#include <unistd.h>
static inline _syscall2(void,nat_spec_reset,unsigned long *,args,int,count);

#include <unistd.h>
static inline _syscall1(void,set_out_ip,unsigned long *,args);
static inline _syscall1(void,set_local_ip,unsigned long *,args);
static inline _syscall1(void,set_pptp,unsigned char *,args);

#define __NR_set_pptp	 (__NR_SYSCALL_BASE+235)

#endif

asmlinkage void sys_set_local_ip(unsigned long *args)
{
	LOCK_BH(&ip_spec_lock);
	copy_from_user(&ROUTER_LOCAL_IP,args,4);
	DEBUGP("New LAN IP is %u.%u.%u.%u \n", NIPQUAD(ROUTER_LOCAL_IP));

	ROUTER_LOCAL_MASK = K_Get_Netmask("eth0");
	if( ROUTER_LOCAL_MASK == 0 )
		ROUTER_LOCAL_MASK = 0x00ffffff;

	UNLOCK_BH(&ip_spec_lock);
		
	return;
}


asmlinkage void sys_set_pptp(unsigned char *args)
{
	LOCK_BH(&ip_spec_lock);
	copy_from_user(&is_pptp,args,1);
	if(is_pptp)
		DEBUGP("PPTP is enabled... \n");
	else
		DEBUGP("PPTP is disabled... \n");
	UNLOCK_BH(&ip_spec_lock);
	
	return;
}


asmlinkage void sys_set_out_ip(unsigned long *args)
{
	LOCK_BH(&ip_spec_lock);
	copy_from_user(&ROUTER_OUT_IP,args,4);
	DEBUGP("New WAN IP is %u.%u.%u.%u \n", NIPQUAD(ROUTER_OUT_IP));
	UNLOCK_BH(&ip_spec_lock);
	
	return;
}



asmlinkage unsigned int sys_get_out_ip(void)
{
	return ROUTER_OUT_IP;
}

#define uint32 unsigned long

#if 0
#define htonl(A)  ((((uint32)(A) & 0xff000000) >> 24) | \
                   (((uint32)(A) & 0x00ff0000) >> 8)  | \
                   (((uint32)(A) & 0x0000ff00) << 8)  | \
                   (((uint32)(A) & 0x000000ff) << 24))

#endif


/* reset NATA's UDP/EST cache */
asmlinkage void sys_nat_ip_filter_reset(unsigned long *args,char count)
{
	unsigned short is_stop = 0;
	
	/* printk("\nflush NATA,contrack entries"); */
	copy_from_user(&is_stop,args,2);
	
	if( is_stop == 1 )
	{
		printk("NATA stop\n");
		NATA_Stop(); 
	}
	else if( is_stop == 2)
	{
		printk("NATA start\n");
		NATA_Stop();
		NATA_Start();
	}
	NAT_HitCount_ClearAll();
        NAT_ProcCount_ClearAll();
        NAT_HitCount_ClearAll();
        NAT_ProcCount_ClearAll();
	NAT_Entry_ClearAll();

	ip_conntrack_clean_proto( IPPROTO_UDP ); /* only flush udp & esp conntrack entries */
	ip_conntrack_clean_proto( IPPROTO_ESP );

	return;
	
}


asmlinkage void sys_nat_spec_reset(unsigned long *args,char count)
{
	
	char i;
	unsigned long a[1];
	struct spec_list* a0;
	
	LOCK_BH(&ip_spec_lock);
	spec_list_count=count;
	
	memset(&alg[0],0,sizeof(alg[0])*MAX_SPEC_CT);
	
	if (( count==0 )||(copy_from_user(a, args,4)))
	{
		UNLOCK_BH(&ip_spec_lock);
		
		return;
	}
		
	a0=(struct spec_list*)a[0];
	
	for(i=0;i<count;i++)
	{
		copy_from_user(alg+i, a0+i,sizeof(struct spec_list));
		DEBUGP("flag %d == %d \n", i, alg[i].flag);
		DEBUGP("out_port_start %d == %d \n", i, alg[i].out_port_start);
		DEBUGP("out_port_end %d == %d \n", i, alg[i].out_port_end);
		DEBUGP("in_port_start %d == %d \n", i, alg[i].in_port_start);
		DEBUGP("in_port_end %d == %d \n", i, alg[i].in_port_end);
		DEBUGP("dst_ip %d == %d \n", i, alg[i].dst_ip);
		DEBUGP("ct_count %d == %d \n", i, alg[i].ct_count);

	}
		
	UNLOCK_BH(&ip_spec_lock);
	
	return;
}



static int __init init(void)
{
	int ret;
	struct ip_conntrack_helper *hlpr;
	char *tmpname;

	//reg_count=0;
	
	memset(&alg[0],0,sizeof(alg[0])*MAX_SPEC_CT);
	
	hlpr = &ip_conntrack_spec_helpers;
	memset(hlpr, 0,sizeof(struct ip_conntrack_helper));
			       
	hlpr->tuple.src.u.tcp.port = 0;
	hlpr->tuple.dst.protonum = 0;
	hlpr->mask.src.u.tcp.port = 0;
	hlpr->mask.dst.protonum = 0;
	hlpr->max_expected = 100;
	hlpr->timeout = 0;
	hlpr->flags = IP_CT_HELPER_F_REUSE_EXPECT;
	hlpr->me = ip_conntrack_spec;
	hlpr->help = conntrack_spec_help;

	tmpname = &ip_nih_names[0];
	sprintf(tmpname, "spec0");

	hlpr->name = tmpname;
						     
	ret = ip_conntrack_helper_register(hlpr);

	if (ret) {
		DEBUGP("ip_conntrack_spec: error registering helper for spec.app.\n" );
		fini();
		return 1;
	}
	else
		DEBUGP("ip_conntrack_spec: registering helper for spec.app. v1\n" );
	
	return ret;
}

unsigned int K_Get_Netmask(char* interface)
{
	struct in_ifaddr **ifap = NULL;
	struct in_ifaddr *ifa = NULL;
	struct net_device *dev;
	struct in_device *in_dev;
	
	dev_probe_lock();
	
	if ((dev = __dev_get_by_name(interface)) == NULL) {
		dev_probe_unlock();
		return 0;
	}
	
	if ((in_dev =__in_dev_get(dev)) != NULL) 
	{
		for (ifap=&in_dev->ifa_list; (ifa=*ifap) != NULL; ifap=&ifa->ifa_next) {
			if (strcmp(interface, ifa->ifa_label) == 0)
			{
				dev_probe_unlock();
				return ifa->ifa_mask;
			}
		}
	}
	
	dev_probe_unlock();
	
	return 0;	
}

int K_Get_IP(char* interface)
{
	struct in_ifaddr **ifap = NULL;
	struct in_ifaddr *ifa = NULL;
	struct net_device *dev;
	struct in_device *in_dev;
	
	dev_probe_lock();
	
	if ((dev = __dev_get_by_name(interface)) == NULL) {
		dev_probe_unlock();
		return -1;
	}
	
	if ((in_dev =__in_dev_get(dev)) != NULL) 
	{
		for (ifap=&in_dev->ifa_list; (ifa=*ifap) != NULL; ifap=&ifa->ifa_next) {
			if (strcmp(interface, ifa->ifa_label) == 0)
			{
				dev_probe_unlock();
			    	return ifa->ifa_local;
			}
		}
	}
	
	dev_probe_unlock();
	
	return -1;
	
}

int
recv_in(struct sk_buff *skb, const char *interface)
{
        struct in_device *indev;
        int ret = 0;
        
        indev = in_dev_get(skb->dev);
        
        if (!indev) {
                return 0;
        }
        if (!indev->ifa_list) {
                in_dev_put(indev);
                return 0;
        }
        
        if( strncmp(  interface, indev->ifa_list->ifa_label, 16 ) == 0 )
		ret = 1;
        in_dev_put(indev);
        
	return ret;	
}

module_init(init);
module_exit(fini);

EXPORT_SYMBOL(rdr_add);
