/* IRC extension for TCP NAT alteration.
 * (C) 2000-2001 by Harald Welte <laforge@gnumonks.org>
 * based on a copy of RR's ip_nat_ftp.c
 *
 * ip_nat_irc.c,v 1.16 2001/12/06 07:42:10 laforge Exp
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version
 *      2 of the License, or (at your option) any later version.
 *
 *	Module load syntax:
 * 	insmod ip_nat_irc.o ports=port1,port2,...port<MAX_PORTS>
 *	
 * 	please give the ports of all IRC servers You wish to connect to.
 *	If You don't specify ports, the default will be port 6667
 */

#include <linux/module.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/kernel.h>
#include <net/tcp.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>

#if 1
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

DECLARE_LOCK_EXTERN(ip_spec_lock);

#define SPEC_CONFIG_FILE	"/usr/spec.conf"

struct spec_list
{
	unsigned char  flag;
	unsigned short out_port_start ;
	unsigned short out_port_end ;
	
	unsigned short in_port_start ;
	unsigned short in_port_end ;
	
	unsigned int  dst_ip ;
	struct ip_conntrack *spec_ct ;
	
	
	
};

#if 1 //lsshi add 2003-4-29 11:11


extern signed char Find_ALG(unsigned short port,int flag);
extern signed char Check_Filter_IP(struct ip_conntrack *ct,int dir );

extern struct spec_list alg[];

#endif

#ifdef SPEC_RDR

int rdr_map( u_int8_t proto, u_int32_t sip, u_int32_t dip, u_int16_t sport, u_int16_t dport,
        u_int32_t *tip, u_int16_t *tport );
int rdr_add( u_int8_t proto, u_int32_t sip, u_int32_t dip, u_int16_t sport, u_int16_t dport, 
	u_int32_t tip, u_int16_t tport, unsigned long ttl );
#endif

static unsigned int ROUTER_OUT_IP;
extern unsigned int ROUTER_LOCAL_IP;

struct ip_nat_helper ip_nat_spec_helpers;
static char ip_spec_names[6];


/* protects irc part of conntracks */
DECLARE_LOCK_EXTERN(ip_spec_lock);


static unsigned int
spec_nat_expected(struct sk_buff **pskb,
		 unsigned int hooknum,
		 struct ip_conntrack *ct,
		 struct ip_nat_info *info)

{
#if 0
	struct ip_nat_multi_range mr;
	u_int32_t newdstip,newsrcip, newip;
	
	struct ip_conntrack *master = master_ct(ct);
	
	
	printk("In NAT spec_nat_expected ok !!! \n");
	
	ct->is_spec=1;
	
	IP_NF_ASSERT(info);
	IP_NF_ASSERT(master);
	
	
	
	if(hooknum!=NF_IP_PRE_ROUTING)
	{
		printk("Spec Not hook in other ...\n");
		return 1;
	}
	
	LOCK_BH(&ip_spec_lock);
	
	
	
	newdstip = master->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip; 
	
	//printk("newdstip is ########### %u.%u.%u.%u \n",NIPQUAD(newdstip));
	
	//newdstip = 0xa01a8c0 //192.168.1.10
	
	
	newsrcip = master->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip;
		DEBUGP("Spec nat_expected: %u.%u.%u.%u->%u.%u.%u.%u\n",
		       NIPQUAD(newsrcip), NIPQUAD(newdstip));

	

	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC)
		newip = newsrcip;
	else
		newip = newdstip;

	DEBUGP("$$$ Nat_expected: IP to %u.%u.%u.%u\n", NIPQUAD(newip));

	mr.rangesize = 1;
	/* We don't want to manip the per-protocol, just the IPs. */
	mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
	mr.range[0].min_ip = mr.range[0].max_ip = newip;
	
	
	UNLOCK_BH(&ip_spec_lock);
	
	//printk("CT====%x\n",ct);

	return ip_nat_setup_info(ct, &mr, hooknum);
#endif
	return 1;
}

unsigned int nat_spec_help(struct ip_conntrack *ct,
		struct ip_conntrack_expect *exp1,
		struct ip_nat_info *info,
		enum ip_conntrack_info ctinfo,
		unsigned int hooknum,
		struct sk_buff **pskb)
			 
{
#ifdef SPEC_RDR
	int dir;
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
		
	dir = CTINFO2DIR(ctinfo);
	printk("nat_spec_help: %u.%u.%u.%u -> %u.%u.%u.%u\n", 
		NIPQUAD(ct->tuplehash[dir].tuple.src.ip),
		NIPQUAD(ct->tuplehash[dir].tuple.dst.ip) );

	if (!(hooknum == NF_IP_PRE_ROUTING && dir == IP_CT_DIR_ORIGINAL) ) {
		DEBUGP("sip_request_nat_help: Not touching dir %s at hook %s\n",
		       dir == IP_CT_DIR_ORIGINAL ? "ORIG" : "REPLY",
		       hooknum == NF_IP_POST_ROUTING ? "POSTROUTING"
		       : hooknum == NF_IP_PRE_ROUTING ? "PREROUTING"
		       : hooknum == NF_IP_LOCAL_OUT ? "OUTPUT" : "???");
		return NF_ACCEPT;
	}

	/* MSN T.120 protocol */
	if (( dir == IP_CT_DIR_ORIGINAL )&&((ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip) ==ROUTER_OUT_IP) &&
		iph->protocol == IPPROTO_TCP )
	{
		u_int32_t newdstip;
		u_int16_t newdstport;
		int ret = 0;
		
		printk("nat_spec_help: %u.%u.%u.%u:%u->%u.%u.%u.%u:%u\n",
			NIPQUAD(iph->saddr), ntohs(tcph->source),
			NIPQUAD(iph->daddr), ntohs(tcph->dest));
		
		if(ct->is_spec_reply)
		{
			printk("nat_spec_help: is_spec_reply \n");
			return NF_ACCEPT;
		}
#ifdef LAN_NET
		if(((iph->saddr)&0x00ffffff)==(ROUTER_LOCAL_IP&0x00ffffff))
			return NF_ACCEPT;
#else
		if( from_lan( ct ) )
			return NF_ACCEPT;
#endif
		ret = rdr_map( iph->protocol, iph->saddr, iph->daddr, ntohs(tcph->source), ntohs(tcph->dest),
			&newdstip, &newdstport );
		
		if( ret == 0 ) //no error
		{
			struct ip_nat_multi_range mr;
			struct tcphdr *tcph = (void *)iph + iph->ihl*4;

			LOCK_BH(&ip_spec_lock);
			mr.rangesize = 1;
			/* We don't want to manip the per-protocol, just the IPs. */
			mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
			mr.range[0].min_ip = mr.range[0].max_ip = newdstip;
			ct->is_spec_reply=1;
		
			printk("v2a forwarding: %u.%u.%u.%u -> %u.%u.%u.%u (%u->%u) \n", 
				NIPQUAD(iph->saddr), NIPQUAD(newdstip),
				ntohs(tcph->source), ntohs(tcph->dest));
			
			UNLOCK_BH(&ip_spec_lock);
			
			return ip_nat_setup_info(ct, &mr, NF_IP_PRE_ROUTING);			
		}
	}
#endif
	return NF_ACCEPT;
}


/* This function is intentionally _NOT_ defined as  __exit, because
 * it is needed by init() */
static void fini(void)
{
	ip_nat_helper_unregister(&ip_nat_spec_helpers);

	//ip_nat_expect_unregister(&spec_expect);
}



static int __init init(void)
{
	int ret;
	int i;
	struct ip_nat_helper *hlpr;
	char *tmpname;

	
	hlpr = &ip_nat_spec_helpers;
	memset(hlpr, 0,	sizeof(struct ip_nat_helper));
		
	hlpr->tuple.dst.protonum = 0;
	hlpr->tuple.src.u.tcp.port = 0;
	hlpr->mask.dst.protonum = 0;
	hlpr->mask.src.u.tcp.port = 0;
	hlpr->help = nat_spec_help;
	hlpr->me = THIS_MODULE;
	hlpr->flags = 0; //IP_NAT_HELPER_F_ALWAYS;
	hlpr->expect = spec_nat_expected;
			
	tmpname = &ip_spec_names[0];
	sprintf(tmpname, "spec0");

	hlpr->name = tmpname;
	ret = ip_nat_helper_register(hlpr);

	if (ret) {
		printk("ip_nat_spec: lsshi error registering helper for port\n" );
		fini();
		return 1;
	}

	printk("Register NAT spec successfully v2\n");
#ifdef SPEC_RDR
	rdr_add( IPPROTO_TCP, 0, 0, 0, 1503, 0x0701a8c0, 1503, 300 );	
#endif
	
	return ret;
}

module_init(init);
module_exit(fini);
