/*
 * iptables module to match domain name
 *   (c) 2004 NortonNg <jkng@csie.nctu.edu.tw>
 *
 * Released under the terms of GNU GPLv2.
 *
 */
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_policydomain.h>

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Norton <jkng@csie.nctu.edu.tw>");
MODULE_DESCRIPTION("iptables domain name match module");

int match_domain_ip( const char *domain, unsigned long addr );

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop)
{
	const struct ipt_policydomain_info *info = matchinfo;
	const struct iphdr *iph = skb->nh.iph;
	int match = 0;

#ifdef XX
	if (info->flags & POLICYDOMAIN_SRC) {
		if( match_domain_ip( info->src_domain, iph->saddr ) )
			match = 1;

		if( info->flags & POLICYDOMAIN_SRC_INV )
		{
			if( match == 1 )
				return 0;
		}
		else if( match == 0 )
			return 0;
	}
#endif
	if (info->flags & POLICYDOMAIN_DST) {
		DEBUGP("\nmatch %u.%u.%u.%u in %s ", NIPQUAD(iph->daddr), info->dst_domain);
		if( match_domain_ip( &info->dst_domain[0], iph->daddr ) )
			match = 1;
		if( info->flags & POLICYDOMAIN_DST_INV )
		{
			if( match == 1 )
				return 0;
		}
		else if( match == 0 )
			return 0;
		
	}
	
	return 1;
}

static int check(const char *tablename,
		 const struct ipt_ip *ip,
		 void *matchinfo,
		 unsigned int matchsize,
		 unsigned int hook_mask)
{
	const struct ipt_policydomain_info *info = matchinfo;
	
	/* verify size */
	if (matchsize != IPT_ALIGN(sizeof(struct ipt_policydomain_info)))
		return 0;

	return 1;
}

static struct ipt_match policydomain_match = 
{ 
	.list = { NULL, NULL }, 
	.name = "policydomain", 
	.match = &match, 
	.checkentry = &check, 
	.destroy = NULL, 
	.me = THIS_MODULE
};

static int __init init(void)
{
	return ipt_register_match(&policydomain_match);
}

static void __exit fini(void)
{
	ipt_unregister_match(&policydomain_match);
}

module_init(init);
module_exit(fini);
