/* Kernel module to match self parameters. */
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>

#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ipt_self.h>
#include <linux/netfilter_ipv4/ip_tables.h>


#ifdef DEBUG_CONNTRACK
#define duprintf(format, args...) printk(format , ## args)
#else
#define duprintf(format, args...) printk(format , ## args)
#endif

int matchifip_by_index( int ifindex, unsigned long ip );
struct net_device *dev_get_by_name(const char *name);

static int
match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop)
{
	const struct ipt_self *selfinfo = matchinfo;
	const struct iphdr *iph = skb->nh.iph;

	if( selfinfo->iface[0] && (selfinfo->flags & IPT_SELF_IFACE_DST) )
	{
		struct net_device *dev = NULL;

		if( strncmp(selfinfo->iface, "wan",3)==0 )
		{
			if( iph->daddr == sys_get_out_ip() ) 
				return 1;
			
			dev = dev_get_by_name( "ppp2" );
			if( dev && matchifip_by_index( dev->ifindex, iph->daddr )  ) 
				return 1;
		}
		
		dev = dev_get_by_name( selfinfo->iface );
		if( dev && matchifip_by_index( dev->ifindex, iph->daddr )  ) 
			return 1;

	}
	else if( selfinfo->iface[0] && (selfinfo->flags & IPT_SELF_IFACE_SRC) )
	{
		struct net_device *dev = NULL;

		if( strncmp(selfinfo->iface, "wan",3)==0 )
		{
			if( iph->saddr == sys_get_out_ip() )
				return 1;

			dev = dev_get_by_name( "ppp2" );
			if( dev && matchifip_by_index( dev->ifindex, iph->saddr )  ) 
				return 1;

		}
		
		dev = dev_get_by_name( selfinfo->iface );
		if( dev && matchifip_by_index( dev->ifindex, iph->saddr )  ) 
			return 1;

	}
	else if( in && (selfinfo->flags & IPT_SELF_IFACE_DST_ME) )
	{
		if( matchifip_by_index( in->ifindex, iph->daddr )  ) 
			return 1;
	}
	else if( out && (selfinfo->flags & IPT_SELF_IFACE_SRC_ME) )
	{
		if( matchifip_by_index( out->ifindex, iph->saddr )  ) 
			return 1;
	}

	return 0;
}

/* Called when user tries to insert an entry of this type. */
static int
checkentry(const char *tablename,
	   const struct ipt_ip *ip,
	   void *matchinfo,
	   unsigned int matchinfosize,
	   unsigned int hook_mask)
{
	if (matchinfosize != IPT_ALIGN(sizeof(struct ipt_self))) {
		duprintf("ipt_self: matchsize %u != %u\n",
			 matchinfosize, IPT_ALIGN(sizeof(struct ipt_self)));
		return 0;
	}
	/*
	if (selfinfo->invflags & ~IPT_SELF_INV_MASK) {
		duprintf("ipt_self: unknown flags %X\n",
			 selfinfo->invflags);
		return 0;
	}
	*/
	
	return 1;
}

static struct ipt_match self_match
= { { NULL, NULL }, "self", &match, &checkentry, NULL, THIS_MODULE };

static int __init init(void)
{
	/* NULL if ip_conntrack not a module */
	if (ip_conntrack_module)
		__MOD_INC_USE_COUNT(ip_conntrack_module);
	return ipt_register_match(&self_match);
}

static void __exit fini(void)
{
	ipt_unregister_match(&self_match);
	if (ip_conntrack_module)
		__MOD_DEC_USE_COUNT(ip_conntrack_module);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
