/* MSNMS extension for TCP NAT alteration. 
 *
 * CHANGES:
 *	20040831: recursive sip_status_nat_expect() for t.120 session
 */
#include <linux/module.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <net/tcp.h>
#include <net/udp.h>
#include <linux/netfilter_ipv4/ip_nat.h>
#include <linux/netfilter_ipv4/ip_nat_helper.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ip_conntrack_msnms.h>
#include <linux/netfilter_ipv4/ip_conntrack_helper.h>

#if 1
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#define MAX_NEW_DATA_LEN	1800
#define MIN_CONTEXT_LEN		20
#define MSN_SIP_TIMEOUT		300*HZ

#define T120_PORT		1503
#define T120_TTL		15

#define MAX_PORTS 2
static int ports[MAX_PORTS];
static int ports_c = 0;

#ifdef MODULE_PARM
MODULE_PARM(ports, "1-" __MODULE_STRING(MAX_PORTS) "i");
#endif

DECLARE_LOCK_EXTERN(ip_msnms_lock);

#define B64_CPL	76

/* Base-64 decoding.  This represents binary data as printable ASCII
** characters.  Three 8-bit binary bytes are turned into four 6-bit
** values, like so:
**
**   [11111111]  [22222222]  [33333333]
**
**   [111111] [112222] [222233] [333333]
**
** Then the 6-bit values are represented using the characters "A-Za-z0-9+/".
*/

static int b64_decode_table[256] = {
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 00-0F */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 10-1F */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,62,-1,-1,-1,63,  /* 20-2F */
    52,53,54,55,56,57,58,59,60,61,-1,-1,-1,-1,-1,-1,  /* 30-3F */
    -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,  /* 40-4F */
    15,16,17,18,19,20,21,22,23,24,25,-1,-1,-1,-1,-1,  /* 50-5F */
    -1,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,  /* 60-6F */
    41,42,43,44,45,46,47,48,49,50,51,-1,-1,-1,-1,-1,  /* 70-7F */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 80-8F */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 90-9F */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* A0-AF */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* B0-BF */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* C0-CF */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* D0-DF */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* E0-EF */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1   /* F0-FF */
    };
    
static char b64_table[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/" ;

/* Do base-64 decoding on a string.  Ignore any non-base64 bytes.
** Return the actual number of bytes generated.  The decoded size will
** be at most 3/4 the size of the encoded, and may be smaller if there
** are padding characters (blanks, newlines).
*/
int b64_decode(char *str, char *buf, int max_buf)
{
    char c;
    char* cp;
    int buf_idx, phase;
    int d, prev_d = 0;

    buf_idx = 0;
    phase = 0;
    for ( cp = str; *cp != '\0'; ++cp )
    {
	d = b64_decode_table[*cp];
        if ( d != -1 )
        {
            switch ( phase )
            {
                case 0:
                    ++phase;
                    break;
                    
                case 1:
                    c = ( ( prev_d << 2 ) | ( ( d & 0x30 ) >> 4 ) );
                    if ( buf_idx < max_buf ) buf[buf_idx++] = c;
                    ++phase;
                    break;
                    
                case 2:
                    c = ( ( ( prev_d & 0xf ) << 4 ) | ( ( d & 0x3c ) >> 2 ) );
                    if ( buf_idx < max_buf ) buf[buf_idx++] = c;
                    ++phase;
                    break;
                
                case 3:
                    c = ( ( ( prev_d & 0x03 ) << 6 ) | d );
                    if ( buf_idx < max_buf ) buf[buf_idx++] = c;
                    phase = 0;
                    break;
            }
        prev_d = d;
        }
    }
    return buf_idx;
}

/* do base64 encode
 * return strlen of encoded string, -1 if error 
 * note str is binary input, probably not null terminated
 */
int b64_encode(char *str, int str_len, char *buf, int max_buf)
{
    int i=0;            /* b64 ecnode is 76char per line */
    
    while( str_len > 0 )
    {
        /* new line needed?? */
        if ( i >= B64_CPL ) 
        {
            i = 0;
            if ( max_buf < 1 ) return -1;
            *buf = '\n';
            max_buf--;
            buf++;
        }
        else
        {
            i++;
        }

        /* enough buffer space?? */ 
        if ( max_buf >= 4 ) max_buf -= 4;
        else return -1;
        
        if ( str_len >= 3 )
        {
            str_len -= 3;

            *buf     = b64_table[(*str >> 2)];
            *(buf+1) = b64_table[((*str & 0x03) << 4) | (*(str+1) >> 4)];
            *(buf+2) = b64_table[((*(str+1) & 0x0F) << 2) | (*(str+2) >> 6)];
            *(buf+3) = b64_table[*(str+2) & 0x3F];
            
            buf += 4;
            str += 3;
        }
        else if ( str_len == 2 )
        {
            str_len -= 2;

            *buf     = b64_table[(*str >> 2)];
            *(buf+1) = b64_table[((*str & 0x03) << 4) | (*(str+1) >> 4)];
            *(buf+2) = b64_table[((*(str+1) & 0x0F) << 2)];
            *(buf+3) = '=';
            
            buf += 4;
            str += 2;
        }
        else /* if ( str_len == 1 ) */
        {
            str_len -= 1;

            *buf     = b64_table[(*str >> 2)];
            *(buf+1) = b64_table[((*str & 0x03) << 4)];
            *(buf+2) = '=';
            *(buf+3) = '=';
            
            buf += 4;
            str += 1;
        }
    }
    
    /* add end of string in buf */
    if ( max_buf < 1 ) return -1;
    *buf = '\0';
    
    return strlen(buf);
}

static unsigned int
strtou16( const char *pbuf)
{
	int n=0;
	unsigned int pval = 0;
	
	while( pbuf[n] >= '0' && pbuf[n] <= '9' )
	{
		pval = (pval *10) + (pbuf[n]-'0');
		n++;
	}
	return pval;
}


extern int rdr_add( u_int8_t proto, u_int32_t sip, u_int32_t dip, u_int16_t sport, u_int16_t dport, 
	u_int32_t tip, u_int16_t tport, unsigned long ttl );

int 
ip_nat_mangle_tcp_packet2(struct sk_buff **skb,
			 const char *newdata,
			 struct ip_conntrack *ct,
			 enum ip_conntrack_info ctinfo,
			 unsigned int datalen,
			 unsigned int newdatalen );
int 
ip_nat_mangle_udp_packet2(struct sk_buff **skb,
			 const char *newdata,
			 struct ip_conntrack *ct,
			 enum ip_conntrack_info ctinfo,
			 unsigned int datalen,
			 unsigned int newdatalen );


static unsigned int
sip_request_nat_expected(struct sk_buff **pskb,
		 unsigned int hooknum,
		 struct ip_conntrack *ct,
		 struct ip_nat_info *info);
static unsigned int
sip_status_nat_expected(struct sk_buff **pskb,
		 unsigned int hooknum,
		 struct ip_conntrack *ct,
		 struct ip_nat_info *info);

static unsigned int
msnms_nat_expected(struct sk_buff **pskb,
	unsigned int hooknum,
	struct ip_conntrack *ct,
	struct ip_nat_info *info);

int msnms_replace_str(const char *data, char *newdata, int datalen, int *newdatalen, 
	const struct ip_ct_msnms_expect *ct_msnms_info );

static unsigned int sip_request_nat_help(struct ip_conntrack *ct,
				  struct ip_conntrack_expect *exp,
			 	  struct ip_nat_info *info,
			 	  enum ip_conntrack_info ctinfo,
			 	  unsigned int hooknum,
			 	  struct sk_buff **pskb);

static unsigned int sip_status_nat_help(struct ip_conntrack *ct,
				  struct ip_conntrack_expect *exp,
			 	  struct ip_nat_info *info,
			 	  enum ip_conntrack_info ctinfo,
			 	  unsigned int hooknum,
			 	  struct sk_buff **pskb);

static int sip_data_fixup2( struct ip_ct_msnms_expect *exp_msnms_info,
			  struct ip_conntrack *ct,
			  struct sk_buff **pskb,
			  enum ip_conntrack_info ctinfo,
			  struct ip_conntrack_expect *exp)
{
	struct iphdr *iph = (*pskb)->nh.iph;
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	const char *data = (void *)udph + 8;
	u_int16_t port, newdstport =0;
	u_int32_t newsrcip=0, newdstip=0;
	size_t udplen = 0;
	u_int16_t datalen = 0;
	struct ip_conntrack_tuple newtuple;
	static char newdata[MAX_NEW_DATA_LEN];
	struct ip_ct_msnms_master *ct_msnms_info = &ct->help.ct_msnms_info;
	unsigned int newdatalen = 0;
	int i=0;
	
	MUST_BE_LOCKED(&ip_msnms_lock);
	
	ip_ct_refresh(ct, MSN_SIP_TIMEOUT );

	/* Change address inside packet to match way we're mapping
	   this connection. */
	if ( exp_msnms_info->msnmstype != IP_CT_SIP_REQUEST &&
		exp_msnms_info->msnmstype != IP_CT_SIP_STATUS ) {
		printk("sip_data_fixup2: not REQ/STATUS\n");
		return 1;
	}
	
	udplen = (*pskb)->len - iph->ihl * 4;
	datalen = udplen - sizeof(struct udphdr);
	
	if( exp_msnms_info->msnmstype == IP_CT_SIP_REQUEST )
	{
		printk("sip_data_fixup2: req\n");
		newsrcip = ct_msnms_info->priv_addr;
		newdstip = ct_msnms_info->peer_addr;
		newdstport = ct_msnms_info->peer_port;
	}
	else if( exp_msnms_info->msnmstype == IP_CT_SIP_STATUS )
	{
		printk("sip_data_fixup2: status port=%d\n",ct_msnms_info->pub_port);
		if( exp_msnms_info->ctx_len == 0 )
		{
			printk("ctx_len == 0 , skipped\n");
			return 1;
		}	
		newsrcip = ct_msnms_info->peer_addr;
		newdstip = ct_msnms_info->pub_addr;
		newdstport = T120_PORT; //ct_msnms_info->pub_port;
	}
	else printk("sip_data_fixup2: should not reached here!\n");
	
	/* Try to get same port: if not, try to change it. */
	for (port = newdstport; port != 0; port++) {

		newtuple.src.ip = newsrcip;
		newtuple.dst.ip = newdstip;
		
		if( exp_msnms_info->msnmstype == IP_CT_SIP_REQUEST )
		{
			newtuple.src.u.udp.port = 0;
			newtuple.dst.u.udp.port = htons(port);
			ct_msnms_info->peer_port = port; 
			newtuple.dst.protonum = IPPROTO_UDP;

			printk("sip_data_fix2: change_expect: %04x -> %04x:%d \n",newtuple.src.ip, newtuple.dst.ip, port  );

			/* expect for STATUS: session */
			//if (ip_conntrack_change_expect(exp, &newtuple) == 0)  //will not == 0 because unknown 
			ip_conntrack_change_expect(exp, &newtuple);
				break;
		}
		else
		{
			newtuple.src.u.tcp.port = 0;
			newtuple.dst.u.tcp.port = htons(port);
			//ct_msnms_info->pub_port = port; /* norton */
			newtuple.dst.protonum = IPPROTO_TCP;

#ifdef STATUS_EXPECT
			ip_conntrack_change_expect(exp, &newtuple);
#endif
			
			printk("rdr_add: %u.%u.%u.%u -> %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u ",
				NIPQUAD(newsrcip), NIPQUAD(newdstip), T120_PORT, NIPQUAD(ct_msnms_info->priv_addr), T120_PORT );
			rdr_add( IPPROTO_TCP, newsrcip, newdstip, 0, T120_PORT, ct_msnms_info->priv_addr, T120_PORT, T120_TTL );

			break;
		}
	}

	printk("sip_data_fix2 expect: peer_addr=%04x port=%d, pub_addr=%04x port=%d, private: %04x:%d \n", 
		ct_msnms_info->peer_addr, ct_msnms_info->peer_port,
		ct_msnms_info->pub_addr, ct_msnms_info->pub_port,
		ct_msnms_info->priv_addr, ct_msnms_info->priv_port );
	
	memset(newdata,0,sizeof(newdata));
	
#ifdef XXL
	if( exp_msnms_info->msnmstype == IP_CT_SIP_REQUEST )
	{
		char buf[24];
		char *p=NULL;
		
		i = 1;
		/* pattern 0 : IP-Address: 192.168.2.100\r\n */
		memset(&exp_msnms_info->newpattern[i][0],0,sizeof(exp_msnms_info->newpattern[i]));
		memset(&exp_msnms_info->pattern[i][0],0,sizeof(exp_msnms_info->pattern[i]));

		sprintf( &exp_msnms_info->pattern[i][0], "%u.%u.%u.%u", NIPQUAD(ct_msnms_info->priv_addr));
		exp_msnms_info->plen[i] = strlen(exp_msnms_info->pattern[i]);
		sprintf( &exp_msnms_info->newpattern[i][0], "%u.%u.%u.%u", NIPQUAD(ct_msnms_info->pub_addr));
		exp_msnms_info->newplen[i] = strlen(exp_msnms_info->newpattern[i]);

		i++;
		memset(&exp_msnms_info->newpattern[i][0],0,sizeof(exp_msnms_info->newpattern[i]));
		memset(&exp_msnms_info->pattern[i][0],0,sizeof(exp_msnms_info->pattern[i]));

		/* pattern 1 :IP-Address-Enc64: MTkyLjE2OC4yLjEwMA==\r\n*/
		sprintf( buf, "%u.%u.%u.%u", NIPQUAD(ct_msnms_info->priv_addr));
		b64_encode( buf, strlen(buf), &exp_msnms_info->pattern[i][0], sizeof(exp_msnms_info->pattern[i]) );
		exp_msnms_info->plen[i] = strlen(exp_msnms_info->pattern[i]);
		sprintf( buf, "%u.%u.%u.%u", NIPQUAD(ct_msnms_info->pub_addr));
		b64_encode( buf, strlen(buf), &exp_msnms_info->newpattern[i][0], sizeof(exp_msnms_info->newpattern[i]) );
		exp_msnms_info->newplen[i] = strlen(exp_msnms_info->newpattern[i]);
		
		i++;
		exp_msnms_info->pattern[i][0] = '\0';
		exp_msnms_info->plen[i] = 0;
		exp_msnms_info->newplen[i] = 0;

		i = 0;
		
	}
	else
#endif
	if (exp_msnms_info->msnmstype == IP_CT_SIP_STATUS )
	{
		if( exp_msnms_info->ctx_len == 0 )
			i = 0;
		else
			i = 1;

		printk("sip_data_fix2: context-length=%d \n", exp_msnms_info->ctx_len );
		
		memset(&exp_msnms_info->newpattern[i][0],0,sizeof(exp_msnms_info->newpattern[i]));
		memset(&exp_msnms_info->pattern[i][0],0,sizeof(exp_msnms_info->pattern[i]));

		sprintf( &exp_msnms_info->pattern[i][0],"%u.%u.%u.%u:%d", NIPQUAD(ct_msnms_info->priv_addr), (ct_msnms_info->priv_port) );
		exp_msnms_info->plen[i] = strlen(exp_msnms_info->pattern[i]);
		sprintf( &exp_msnms_info->newpattern[i][0],"%u.%u.%u.%u:%d", NIPQUAD(ct_msnms_info->pub_addr), (ct_msnms_info->pub_port) );
		exp_msnms_info->newplen[i] = strlen(exp_msnms_info->newpattern[i]);

		i++;
		memset(&exp_msnms_info->newpattern[i][0],0,sizeof(exp_msnms_info->newpattern[i]));
		memset(&exp_msnms_info->pattern[i][0],0,sizeof(exp_msnms_info->pattern[i]));

		sprintf( &exp_msnms_info->pattern[i][0],"%u.%u.%u.%u", NIPQUAD(ct_msnms_info->priv_addr) );
		exp_msnms_info->plen[i] = strlen(exp_msnms_info->pattern[i]);
		sprintf( &exp_msnms_info->newpattern[i][0],"%u.%u.%u.%u", NIPQUAD(ct_msnms_info->pub_addr) );
		exp_msnms_info->newplen[i] = strlen(exp_msnms_info->newpattern[i]);

		for( i++; i<4; i++ )
		{
			exp_msnms_info->pattern[i][0] = '\0';
			exp_msnms_info->plen[i] = 0;
			exp_msnms_info->newplen[i] = 0;
		}

		if( exp_msnms_info->ctx_len )
		{
			int replace_num = 2;
			i = 0;
			
			/* "Content-Length: YYY\r\n"  YYY : bytes of the following data */
			memset(&exp_msnms_info->pattern[i][0],0,sizeof(exp_msnms_info->pattern[i]));
			memset(&exp_msnms_info->newpattern[i][0],0,sizeof(exp_msnms_info->newpattern[i]));

			sprintf(&exp_msnms_info->pattern[i][0],"Content-Length: %d", exp_msnms_info->ctx_len );
			exp_msnms_info->plen[i] = strlen(exp_msnms_info->pattern[i]);

			sprintf(&exp_msnms_info->newpattern[i][0],"Content-Length: %d", 
				exp_msnms_info->ctx_len + (replace_num*(exp_msnms_info->newplen[2]-exp_msnms_info->plen[2])) );
			exp_msnms_info->newplen[i] = strlen(exp_msnms_info->newpattern[i]);

			printk("sip_data_fix2: SIP_STATUS pattern ok\n");
		}
	}
	else return 1;
	
		
	if ( msnms_replace_str(data, newdata, datalen, &newdatalen, exp_msnms_info) )
	{
		exp_msnms_info->msnmstype = 0xff;
		printk("sip_data_fix2: call ip_nat_mangle_udp_packet2\n");
		return ip_nat_mangle_udp_packet2(pskb, newdata, ct, ctinfo, datalen, newdatalen );
	}
	else
	{
		printk("sip_data_fixup2: msnms_replace_str failed\n");
	}
	return 1;
}


static unsigned int sip_request_nat_help(struct ip_conntrack *ct,
			 struct ip_conntrack_expect *exp,
			 struct ip_nat_info *info,
			 enum ip_conntrack_info ctinfo,
			 unsigned int hooknum,
			 struct sk_buff **pskb)
{
	int dir;
	struct ip_ct_msnms_expect *exp_msnms_info;
	struct ip_ct_msnms_master *ct_msnms_info = &ct->help.ct_msnms_info;
	
	/* Only mangle things once: original direction in PRE_ROUTING */
	dir = CTINFO2DIR(ctinfo);

	printk("sip_request_nat_help: %u.%u.%u.%u -> %u.%u.%u.%u\n", 
		NIPQUAD(ct->tuplehash[dir].tuple.src.ip),
		NIPQUAD(ct->tuplehash[dir].tuple.dst.ip) );
	
	if (!(hooknum == NF_IP_PRE_ROUTING && dir == IP_CT_DIR_ORIGINAL) ) {
		DEBUGP("sip_request_nat_help: Not touching dir %s at hook %s\n",
		       dir == IP_CT_DIR_ORIGINAL ? "ORIG" : "REPLY",
		       hooknum == NF_IP_POST_ROUTING ? "POSTROUTING"
		       : hooknum == NF_IP_PRE_ROUTING ? "PREROUTING"
		       : hooknum == NF_IP_LOCAL_OUT ? "OUTPUT" : "???");
		return NF_ACCEPT;
	}

	if (!exp)
	{
		DEBUGP("sip_nat_help: no exp!!\n");
		return NF_ACCEPT;
	}

	exp_msnms_info = &exp->help.exp_msnms_info;
	if( exp_msnms_info->msnmstype != IP_CT_SIP_REQUEST )
	{
		printk("sip_request_nat_help: not req/status\n");
		return NF_ACCEPT;
	}
	
//	if( datalen < MIN_CONTEXT_LEN )	/* skip unneccessary packet */
//		return NF_ACCEPT;
	
	LOCK_BH(&ip_msnms_lock);

	//ct_msnms_info->priv_addr = ct->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip;
	ct_msnms_info->pub_addr = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip;
	ct_msnms_info->peer_addr = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
	ct_msnms_info->peer_port = ntohs(exp->tuple.dst.u.udp.port); //test!!

	printk("sip_request_nat_help: peer_port = %d \n", ct_msnms_info->peer_port );

	if (!sip_data_fixup2(exp_msnms_info, ct, pskb, ctinfo, exp)) {
		UNLOCK_BH(&ip_msnms_lock);
		return NF_DROP;
	}
	UNLOCK_BH(&ip_msnms_lock);

	return NF_ACCEPT;
}



static unsigned int sip_status_nat_help(struct ip_conntrack *ct,
			 struct ip_conntrack_expect *exp,
			 struct ip_nat_info *info,
			 enum ip_conntrack_info ctinfo,
			 unsigned int hooknum,
			 struct sk_buff **pskb)
{
	int dir;
	struct ip_ct_msnms_expect *exp_msnms_info;
	struct ip_conntrack *master = master_ct(ct);
	//struct ip_ct_msnms_master *ct_msnms_info = &ct->help.ct_msnms_info;
	struct ip_ct_msnms_master *ct_msnms_info = &master->help.ct_msnms_info;
	
	/* Only mangle things once: original direction in PRE_ROUTING */
	dir = CTINFO2DIR(ctinfo);

	printk("sip_status_nat_help: %u.%u.%u.%u -> %u.%u.%u.%u\n", 
		NIPQUAD(ct->tuplehash[dir].tuple.src.ip),
		NIPQUAD(ct->tuplehash[dir].tuple.dst.ip) );
	
	if (!(hooknum == NF_IP_POST_ROUTING && dir == IP_CT_DIR_ORIGINAL) ) {
		DEBUGP("sip_status_nat_help: Not touching dir %s at hook %s\n",
		       dir == IP_CT_DIR_ORIGINAL ? "ORIG" : "REPLY",
		       hooknum == NF_IP_POST_ROUTING ? "POSTROUTING"
		       : hooknum == NF_IP_PRE_ROUTING ? "PREROUTING"
		       : hooknum == NF_IP_LOCAL_OUT ? "OUTPUT" : "???");
		return NF_ACCEPT;
	}

	if (!exp)
	{
		DEBUGP("sip_status_nat_help: no exp!!\n");
		return NF_ACCEPT;
	}

	exp_msnms_info = &exp->help.exp_msnms_info;

/* ..... */
	{ 
	/* 
	 * exp_msnms_info->msnmstype not set to IP_CT_SIP_STATUS because unknown reason 
	 * so, we search again.
	 */
	
	struct iphdr *iph = (*pskb)->nh.iph;
	struct udphdr *udph = (void *)iph + iph->ihl * 4;
	const char *data = (void *)udph + 8;
	unsigned int udplen = iph->tot_len - iph->ihl * 4;
	unsigned int datalen = udplen - sizeof(struct udphdr);
	size_t i, plen;
	const char *p = NULL;

	if( data == NULL || data[0] == '\0' )
		return NF_ACCEPT;

	plen = strlen("Content-Length: ");
	for(i=0; i<datalen-plen; i++)
	{
		if ( strncmp( data +i, "Content-Length: ", plen) == 0 ) {
			p = data + i;
			break;
		}
	}
	if(  p == NULL )
		return NF_ACCEPT;

	exp_msnms_info->ctx_len = strtou16( p+plen );

	if( exp_msnms_info->ctx_len )
		exp_msnms_info->msnmstype = IP_CT_SIP_STATUS;
	}
/* ...... */

	if( exp_msnms_info->msnmstype != IP_CT_SIP_STATUS )
	{
		printk("sip_status_nat_help: not req/status, msnmstype=%d ctx_len=%d\n",exp_msnms_info->msnmstype, exp_msnms_info->ctx_len);
		return NF_ACCEPT;
	}
	
//	if( datalen < MIN_CONTEXT_LEN )	/* skip unneccessary packet */
//		return NF_ACCEPT;
	
	LOCK_BH(&ip_msnms_lock);

	ct_msnms_info->priv_addr = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.src.ip;
	ct_msnms_info->priv_port = ntohs(exp->tuple.dst.u.tcp.port);
	ct_msnms_info->peer_addr = ct->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.ip;
	//ct_msnms_info->peer_port = ntohs(exp->tuple.dst.u.tcp.port); //test!!
	
	printk("sip_status_nat_help: pub_port = %d \n", ct_msnms_info->pub_port );

	if (!sip_data_fixup2(exp_msnms_info, ct, pskb, ctinfo, exp)) {
		UNLOCK_BH(&ip_msnms_lock);
		return NF_DROP;
	}
	UNLOCK_BH(&ip_msnms_lock);

	return NF_ACCEPT;
}


static unsigned int
sip_status_nat_expected(struct sk_buff **pskb,
		 unsigned int hooknum,
		 struct ip_conntrack *ct,
		 struct ip_nat_info *info)
{
	int ret = 0;
	struct ip_nat_multi_range mr;
	u_int32_t newdstip, newsrcip, newip;
	struct ip_ct_msnms_expect *master_exp_msnms_info,*exp_msnms_info;
	struct ip_ct_msnms_master *master_ct_msnms_info,*ct_msnms_info;
	struct ip_conntrack *master = master_ct(ct);
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
		
	IP_NF_ASSERT(info);
	IP_NF_ASSERT(master);
	IP_NF_ASSERT(!(info->initialized & (1<<HOOK2MANIP(hooknum))));

	DEBUGP("sip_status_nat_expected: %u.%u.%u.%u:%u->%u.%u.%u.%u:%u\n",
		NIPQUAD(iph->saddr), ntohs(tcph->source),
		NIPQUAD(iph->daddr), ntohs(tcph->dest));
	
	printk("---> sip_status_nat_expected, T.120 session\n");

	ct_msnms_info = &ct->help.ct_msnms_info;
	master_ct_msnms_info = &master->help.ct_msnms_info;
	exp_msnms_info = &ct->master->help.exp_msnms_info;
	master_exp_msnms_info = &master->master->help.exp_msnms_info;

	newdstip = master_ct_msnms_info->priv_addr;
	newsrcip = iph->saddr;

	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC)
		newip = newsrcip;
	else
		newip = newdstip;
	
	printk("sip_status_nat_expected: pub_addr=%04x port=%d, private: %04x:%d \n", 
		master_ct_msnms_info->pub_addr, 
		master_ct_msnms_info->pub_port,
		master_ct_msnms_info->priv_addr, 
		master_ct_msnms_info->priv_port );
	
	printk("sip_status_nat_expected: new IP %u.%u.%u.%u \n", NIPQUAD(newip) );

	mr.rangesize = 1;
	/* We don't want to manip the per-protocol, just the IPs... */
	mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
	mr.range[0].min_ip = mr.range[0].max_ip = newip;

	/* ... unless we're doing a MANIP_DST, in which case, make
	   sure we map to the correct port */
	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_DST) {
		mr.range[0].flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
		mr.range[0].min = mr.range[0].max
			= ((union ip_conntrack_manip_proto)
				{ tcph->dest });
	}
	ret = ip_nat_setup_info(ct, &mr, hooknum);

	/* can't not associate T.120 connection to T.120 connection because it
	 * will be closed soon. So, associte T.120 connection expection to STATUS connection but
	 * we must extend STATUS connection timeout.
	 */
#ifdef XXX
	 
	printk("sip_status_nat_expected: Add t120 status helper for t120 port %d\n", ntohs(tcph->dest) );
	memcpy((void*)ct_msnms_info, (void*)master_ct_msnms_info,sizeof(struct ip_ct_msnms_master)); 
	ct_msnms_info->is_msnp = 0;

	sip_t120_nat_helper.flags = IP_NAT_HELPER_F_ALWAYS;
	sip_t120_nat_helper.tuple.dst.protonum = IPPROTO_TCP;
	sip_t120_nat_helper.tuple.src.ip = (iph->saddr);
	sip_t120_nat_helper.tuple.dst.ip = (iph->daddr);
	sip_t120_nat_helper.tuple.src.u.tcp.port = 0;
	sip_t120_nat_helper.tuple.dst.u.tcp.port = tcph->dest;

	printk("sip_status_nat_expected: peer_addr=%04x port=%d, pub_addr=%04x port=%d, private: %04x:%d \n", 
		ct_msnms_info->peer_addr, ct_msnms_info->peer_port,
		ct_msnms_info->pub_addr, ct_msnms_info->pub_port,
		ct_msnms_info->priv_addr, ct_msnms_info->priv_port );
	
	info->helper = &sip_t120_nat_helper;
#endif
	
	return ret;

}


static struct ip_nat_helper sip_status_nat_helper = 
	{ { NULL, NULL },
	  "SIP Helper",				/* name */
	  IP_NAT_HELPER_F_ALWAYS,		/* flags */
	  NULL,					/* module */
	  { { 0, { 0 } },			/* tuple */
	    { 0, { 0 }, IPPROTO_TCP } },
	  { { 0, { 0 } },			/* mask */
	    { 0xFFFF, { 0xFFFF }, 0xFFFF } },
	  sip_status_nat_help,			/* helper */
	  sip_status_nat_expected		/* expectfn */
	}; //test


static unsigned int
sip_request_nat_expected(struct sk_buff **pskb,
		 unsigned int hooknum,
		 struct ip_conntrack *ct,
		 struct ip_nat_info *info)
{
	struct ip_nat_multi_range mr;
	u_int32_t newdstip, newsrcip, newip;
	struct ip_ct_msnms_expect *master_exp_msnms_info,*exp_msnms_info;
	struct ip_ct_msnms_master *master_ct_msnms_info,*ct_msnms_info;
	int ret = 0;
	struct ip_conntrack *master = master_ct(ct);
	struct iphdr *iph = (*pskb)->nh.iph;
	struct udphdr *udph;
	
	udph = (void *)iph + iph->ihl * 4;
	
	IP_NF_ASSERT(info);
	IP_NF_ASSERT(master);
	IP_NF_ASSERT(!(info->initialized & (1<<HOOK2MANIP(hooknum))));

	DEBUGP("sip_request_nat_expected: %u.%u.%u.%u:%u->%u.%u.%u.%u:%u\n",
		NIPQUAD(iph->saddr), ntohs(udph->source),
		NIPQUAD(iph->daddr), ntohs(udph->dest));

	ct_msnms_info = &ct->help.ct_msnms_info;
	master_ct_msnms_info = &master->help.ct_msnms_info;

#ifdef TEST
	exp_msnms_info = &ct->master->help.exp_msnms_info;
	//master_exp_msnms_info = &master->master->help.exp_msnms_info;


	if( exp_msnms_info->msnmstype != IP_CT_SIP_STATUS &&
		master_exp_msnms_info->msnmstype != IP_CT_SIP_STATUS )
	{
		printk("not status, skipped, ctx_len=%d , master_ctx_len=%d\n",exp_msnms_info->ctx_len,
			master_exp_msnms_info->ctx_len);
		return NF_ACCEPT;
	}
#endif
	
	//test
	newdstip = iph->daddr;
	newsrcip = master_ct_msnms_info->pub_addr;

	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC)
		newip = newsrcip;
	else
		newip = newdstip;
	
	printk("msnms_nat_expected: pub_addr=%04x port=%d, private: %04x:%d \n", 
		master_ct_msnms_info->pub_addr, 
		master_ct_msnms_info->pub_port,
		master_ct_msnms_info->priv_addr, 
		master_ct_msnms_info->priv_port );
	
	printk("sip_request_nat_expected: new IP %u.%u.%u.%u \n",
		NIPQUAD(newip) );

	mr.rangesize = 1;
	/* We don't want to manip the per-protocol, just the IPs... */
	mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
	mr.range[0].min_ip = mr.range[0].max_ip = newip;

#ifdef XX
	/* ... unless we're doing a MANIP_DST, in which case, make
	   sure we map to the correct port */
	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_DST) {
		mr.range[0].flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
		mr.range[0].min = mr.range[0].max
			= ((union ip_conntrack_manip_proto)
				{ htons(master_ct_msnms_info->priv_port) });
	}
#endif
	ret = ip_nat_setup_info(ct, &mr, hooknum);
	
	printk("sip_request_nat_expected: Add sip status helper for sip port %d\n", ntohs(udph->dest) );
	memcpy((void*)ct_msnms_info, (void*)master_ct_msnms_info,sizeof(struct ip_ct_msnms_master)); 
	ct_msnms_info->is_msnp = 0;

	sip_status_nat_helper.flags = IP_NAT_HELPER_F_ALWAYS;
	sip_status_nat_helper.tuple.dst.protonum = IPPROTO_UDP;
	sip_status_nat_helper.tuple.src.ip = (iph->saddr);
	sip_status_nat_helper.tuple.dst.ip = (iph->daddr);
	sip_status_nat_helper.tuple.src.u.udp.port = 0; //test udph->source;
	sip_status_nat_helper.tuple.dst.u.udp.port = udph->dest;

	ct_msnms_info->peer_addr  = iph->daddr;
	ct_msnms_info->peer_port  = ntohs(udph->dest);
	ct_msnms_info->priv_addr  = iph->saddr;
	ct_msnms_info->priv_port  = ntohs(udph->source);
	
	printk("sip_request_nat_expected: peer_addr=%04x port=%d, pub_addr=%04x port=%d, private: %04x:%d \n", 
		ct_msnms_info->peer_addr, ct_msnms_info->peer_port,
		ct_msnms_info->pub_addr, ct_msnms_info->pub_port,
		ct_msnms_info->priv_addr, ct_msnms_info->priv_port );
	
	info->helper = &sip_status_nat_helper;

	return ret;
}

static struct ip_nat_helper sip_request_nat_helper = 
	{ { NULL, NULL },
          "SIP Helper",				/* name */
	  IP_NAT_HELPER_F_ALWAYS,		/* flags */
	  NULL,					/* module */
	  { { 0, { 0 } },			/* tuple */
	    { 0, { 0 }, IPPROTO_UDP } },
	  { { 0, { 0 } },			/* mask */
	    { 0xFFFF, { 0xFF }, 0xFFFF } },
	  sip_request_nat_help,			/* helper */
	  sip_request_nat_expected		/* expectfn */
	}; //test

/* expect for SIP session */

static unsigned int
msnms_nat_expected(struct sk_buff **pskb,
		 unsigned int hooknum,
		 struct ip_conntrack *ct,
		 struct ip_nat_info *info)
{
	struct ip_nat_multi_range mr;
	u_int32_t newdstip, newsrcip, newip;
	struct ip_ct_msnms_expect *exp_msnms_info;
	struct ip_ct_msnms_master *master_ct_msnms_info,*ct_msnms_info;
	int ret = 0;
	struct iphdr *iph = (*pskb)->nh.iph;
	struct ip_conntrack *master = master_ct(ct);
	
	IP_NF_ASSERT(info);
	IP_NF_ASSERT(master);
	IP_NF_ASSERT(!(info->initialized & (1<<HOOK2MANIP(hooknum))));
	
	ct_msnms_info = &ct->help.ct_msnms_info;
	master_ct_msnms_info = &master->help.ct_msnms_info;
	exp_msnms_info = &ct->master->help.exp_msnms_info;

	//test
	newdstip = master_ct_msnms_info->priv_addr;
	newsrcip = iph->saddr; //master_ct_msnms_info->pub_addr;

	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC)
		newip = newsrcip;
	else
		newip = newdstip;
	
	printk("msnms_nat_expected: pub_addr=%04x port=%d, private: %04x:%d \n", 
		master_ct_msnms_info->pub_addr, 
		master_ct_msnms_info->pub_port,
		master_ct_msnms_info->priv_addr, 
		master_ct_msnms_info->priv_port );
	
	printk("msnms_nat_expected: new IP %u.%u.%u.%u\n", NIPQUAD(newip) );

	mr.rangesize = 1;
	/* We don't want to manip the per-protocol, just the IPs... */
	mr.range[0].flags = IP_NAT_RANGE_MAP_IPS;
	mr.range[0].min_ip = mr.range[0].max_ip = newip;

	/* ... unless we're doing a MANIP_DST, in which case, make
	   sure we map to the correct port */
	if (HOOK2MANIP(hooknum) == IP_NAT_MANIP_DST) {
		mr.range[0].flags |= IP_NAT_RANGE_PROTO_SPECIFIED;
		mr.range[0].min = mr.range[0].max
			= ((union ip_conntrack_manip_proto)
				{ htons(master_ct_msnms_info->priv_port) });
	}
	ret = ip_nat_setup_info(ct, &mr, hooknum);
	
	/* test */
	if( master_ct_msnms_info->is_msnp )
	{
		printk("msnms_nat_expected: Add sip request helper for sip port %d\n", (master_ct_msnms_info->priv_port));
		memcpy((void*)ct_msnms_info, (void*)master_ct_msnms_info,sizeof(struct ip_ct_msnms_master)); 
		ct_msnms_info->is_msnp = 0;
		info->helper = &sip_request_nat_helper;
	}

	return ret;
}

/* only able to match ASCII string !! */
int msnms_replace_str(const char *data, char *newdata, int datalen, int *newdatalen, 
	const struct ip_ct_msnms_expect *ct_msnms_info )
{
	size_t i,j;
	int match=0;
	int dataset = 1;
	
	
	*newdatalen = 0;

	if( data == NULL || data[0] == '\0' || datalen > MAX_NEW_DATA_LEN || newdata == NULL )
		return 0;

	for(i=0; i<4; i++)
	{
		printk("i=%d msnms_replace_str(): len=%d pattern %s \n",  
			i, ct_msnms_info->plen[i],
			&ct_msnms_info->pattern[i][0] );
	}
	
	for(i=0; i<datalen; i++)
	{
		dataset = 1;
		for( j=0 ; j < sizeof( ct_msnms_info->pattern )/sizeof(ct_msnms_info->pattern[0]); j++ )
		{
			if( ct_msnms_info->pattern[j][0] == '\0' || ct_msnms_info->plen[j] == 0 )
				break;
			
			if ( strncmp( data +i, &ct_msnms_info->pattern[j][0], ct_msnms_info->plen[j]) == 0 ) 
			{
				printk("replace: %s - %s \n",ct_msnms_info->pattern[j], ct_msnms_info->newpattern[j]);
				match++;
				dataset = 0;
				if( ct_msnms_info->newplen[j] )
					strncpy( &newdata[ *newdatalen ], &ct_msnms_info->newpattern[j][0], ct_msnms_info->newplen[j] );
				*newdatalen += ct_msnms_info->newplen[j];
				i+= ct_msnms_info->plen[j]-1; /* minus 1 because of i++ in for loop */
				
				break;
			}
		}
		if( dataset == 1 )
			newdata[ (*newdatalen)++] = data[i];

	}
	
	return match;
}

static int msnms_data_fixup2( struct ip_ct_msnms_expect *exp_msnms_info,
			  struct ip_conntrack *ct,
			  struct sk_buff **pskb,
			  enum ip_conntrack_info ctinfo,
			  struct ip_conntrack_expect *expect)
{
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
	u_int16_t port;
	struct ip_conntrack_tuple newtuple;
	const char *data = (const char *)tcph + tcph->doff * 4;
	static char newdata[MAX_NEW_DATA_LEN];
	unsigned int datalen = (*pskb)->len - iph->ihl * 4 - tcph->doff * 4;
	struct ip_ct_msnms_master *ct_msnms_info = &ct->help.ct_msnms_info;
	unsigned int newdatalen = 0;
	int i=0;
	
	MUST_BE_LOCKED(&ip_msnms_lock);
	
	/* Change address inside packet to match way we're mapping
	   this connection. */
	if ( exp_msnms_info->msnmstype != IP_CT_MSNMS_ACCEPT &&
		 exp_msnms_info->msnmstype != IP_CT_MSNMS_CONTEXT ) {

		return 0; //test 
		
	}
	
	//ct_msnms_info->pub_addr = ct->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip;
	
	/* Try to get same port: if not, try to change it. */
	for (port = ct_msnms_info->priv_port; port != 0; port++) {
		/*-- norton */
		if( ct_msnms_info->peer_addr == 0 )
		{
			printk("peer addr not found\n");
			break;
		}
		newtuple.dst.protonum = IPPROTO_UDP;
		newtuple.src.ip = ct_msnms_info->peer_addr;
		newtuple.dst.ip = ct_msnms_info->pub_addr;
		newtuple.src.u.udp.port = 0;
		newtuple.dst.u.udp.port = htons(port);
		ct_msnms_info->pub_port = port; /* norton */

		printk("msnms_data_fixup2: change_expect: %u.%u.%u.%u -> %u.%u.%u.%u:%d \n",
			NIPQUAD(newtuple.src.ip), 
			NIPQUAD(newtuple.dst.ip), 
			port  );
		/* norton -- */

		if (ip_conntrack_change_expect(expect, &newtuple) == 0)
			break;
	}

	printk("msnms_data_fixup2: pub_addr=%04x port=%d, private: %04x:%d \n", 
		ct_msnms_info->pub_addr, ct_msnms_info->pub_port,
		ct_msnms_info->priv_addr, ct_msnms_info->priv_port );
	
	//test
	//return 1;
	
	memset(newdata,0,sizeof(newdata));
	
	if( exp_msnms_info->msnmstype == IP_CT_MSNMS_ACCEPT )
	{
		char buf[24];
		char *p=NULL;
		
		i = 1;
		/* pattern 0 : IP-Address: 192.168.2.100\r\n */
		memset(&exp_msnms_info->newpattern[i][0],0,sizeof(exp_msnms_info->newpattern[i]));
		memset(&exp_msnms_info->pattern[i][0],0,sizeof(exp_msnms_info->pattern[i]));

		sprintf( &exp_msnms_info->pattern[i][0], "%u.%u.%u.%u", NIPQUAD(ct_msnms_info->priv_addr));
		exp_msnms_info->plen[i] = strlen(exp_msnms_info->pattern[i]);
		sprintf( &exp_msnms_info->newpattern[i][0], "%u.%u.%u.%u", NIPQUAD(ct_msnms_info->pub_addr));
		exp_msnms_info->newplen[i] = strlen(exp_msnms_info->newpattern[i]);

		i++;
		memset(&exp_msnms_info->newpattern[i][0],0,sizeof(exp_msnms_info->newpattern[i]));
		memset(&exp_msnms_info->pattern[i][0],0,sizeof(exp_msnms_info->pattern[i]));

		/* pattern 1 :IP-Address-Enc64: MTkyLjE2OC4yLjEwMA==\r\n*/
		sprintf( buf, "%u.%u.%u.%u", NIPQUAD(ct_msnms_info->priv_addr));
		b64_encode( buf, strlen(buf), &exp_msnms_info->pattern[i][0], sizeof(exp_msnms_info->pattern[i]) );
		exp_msnms_info->plen[i] = strlen(exp_msnms_info->pattern[i]);
		sprintf( buf, "%u.%u.%u.%u", NIPQUAD(ct_msnms_info->pub_addr));
		b64_encode( buf, strlen(buf), &exp_msnms_info->newpattern[i][0], sizeof(exp_msnms_info->newpattern[i]) );
		exp_msnms_info->newplen[i] = strlen(exp_msnms_info->newpattern[i]);
		
		i++;
		exp_msnms_info->pattern[i][0] = '\0';
		exp_msnms_info->plen[i] = 0;
		exp_msnms_info->newplen[i] = 0;

		i = 0;
		/* "MSG XX A YYY\r\n"  YYY : bytes of the following data */
		if( strncmp(data,"MSG ",4)==0)
		{
			p = strchr(data, '\r');
			if( p && strchr(data,'A') )
			{
				unsigned int size=0;
				
				memset(&exp_msnms_info->pattern[i][0],0,sizeof(exp_msnms_info->pattern[i]));
				memset(&exp_msnms_info->newpattern[i][0],0,sizeof(exp_msnms_info->newpattern[i]));

				strncpy( &exp_msnms_info->pattern[i][0], data, p-data );
				exp_msnms_info->plen[i] = strlen(exp_msnms_info->pattern[i]);
				p = strchr(data, 'A');
				p++; p++;
				size = strtou16(p);

				strncpy( &exp_msnms_info->newpattern[i][0], data, p-data );
				sprintf( &exp_msnms_info->newpattern[i][0] + strlen(exp_msnms_info->newpattern[i]),
					"%d",
					size+(exp_msnms_info->newplen[1]-exp_msnms_info->plen[1] )+
					(exp_msnms_info->newplen[2]-exp_msnms_info->plen[2])+
					(exp_msnms_info->newplen[3]-exp_msnms_info->plen[3]) );
				exp_msnms_info->newplen[i] = strlen(exp_msnms_info->newpattern[i]);
				i++;
			}
		}
		
	}
	else if (exp_msnms_info->msnmstype == IP_CT_MSNMS_CONTEXT )
	{
		char *p=NULL;
		
		i = 1;
			
		memset(&exp_msnms_info->newpattern[i][0],0,sizeof(exp_msnms_info->newpattern[i]));
		memset(&exp_msnms_info->pattern[i][0],0,sizeof(exp_msnms_info->pattern[i]));

		sprintf( &exp_msnms_info->pattern[i][0],"%u.%u.%u.%u:%d", NIPQUAD(ct_msnms_info->priv_addr), (ct_msnms_info->priv_port) );
		exp_msnms_info->plen[i] = strlen(exp_msnms_info->pattern[i]);
		sprintf( &exp_msnms_info->newpattern[i][0],"%u.%u.%u.%u:%d", NIPQUAD(ct_msnms_info->pub_addr), (ct_msnms_info->pub_port) );
		exp_msnms_info->newplen[i] = strlen(exp_msnms_info->newpattern[i]);

		for( i++; i<4; i++ )
		{
			exp_msnms_info->pattern[i][0] = '\0';
			exp_msnms_info->plen[i] = 0;
			exp_msnms_info->newplen[i] = 0;
		}

		i = 0;
		/* "MSG XX A YYY\r\n"  YYY : bytes of the following data */
		if( strncmp(data,"MSG ",4)==0)
		{
			p = strchr(data, '\r');
			if( p && strchr(data,'A') )
			{
				unsigned int size=0;
				
				memset(&exp_msnms_info->newpattern[i][0],0,sizeof(exp_msnms_info->newpattern[i]));
				memset(&exp_msnms_info->pattern[i][0],0,sizeof(exp_msnms_info->pattern[i]));

				strncpy( &exp_msnms_info->pattern[i][0], data, p-data );
				exp_msnms_info->plen[i] = strlen(exp_msnms_info->pattern[i]);
				p = strchr(data, 'A');
				p++; p++;
				size = strtou16(p);

				strncpy( &exp_msnms_info->newpattern[i][0], data, p-data );
				sprintf( &exp_msnms_info->newpattern[i][0] + strlen(exp_msnms_info->newpattern[i]),
					"%d",
					size+(exp_msnms_info->newplen[1]-exp_msnms_info->plen[1] )+
					(exp_msnms_info->newplen[2]-exp_msnms_info->plen[2])+
					(exp_msnms_info->newplen[3]-exp_msnms_info->plen[3]) );
				exp_msnms_info->newplen[i] = strlen(exp_msnms_info->newpattern[i]);
				i++;
			}
		}

	}
	else return 1;
	
		
	if ( msnms_replace_str(data, newdata, datalen, &newdatalen, exp_msnms_info) )
	{
		exp_msnms_info->msnmstype = 0xff;
		return ip_nat_mangle_tcp_packet2(pskb, newdata, ct, ctinfo, datalen, newdatalen );
	}
	return 1;
}

static unsigned int msnms_nat_help(struct ip_conntrack *ct,
			 struct ip_conntrack_expect *exp,
			 struct ip_nat_info *info,
			 enum ip_conntrack_info ctinfo,
			 unsigned int hooknum,
			 struct sk_buff **pskb)
{
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
	unsigned int datalen;
	int dir;
	struct ip_ct_msnms_expect *exp_msnms_info;

	if (!exp)
		DEBUGP("ip_nat_msnms: no exp!!");

	printk("nat help\n");
	
	exp_msnms_info = &exp->help.exp_msnms_info;

	/* Only mangle things once: original direction in POST_ROUTING
	   and reply direction on PRE_ROUTING. */
	dir = CTINFO2DIR(ctinfo);
	if (!((hooknum == NF_IP_POST_ROUTING && dir == IP_CT_DIR_ORIGINAL)
	      || (hooknum == NF_IP_PRE_ROUTING && dir == IP_CT_DIR_REPLY))) {
		DEBUGP("nat_msnms: Not touching dir %s at hook %s\n",
		       dir == IP_CT_DIR_ORIGINAL ? "ORIG" : "REPLY",
		       hooknum == NF_IP_POST_ROUTING ? "POSTROUTING"
		       : hooknum == NF_IP_PRE_ROUTING ? "PREROUTING"
		       : hooknum == NF_IP_LOCAL_OUT ? "OUTPUT" : "???");
		return NF_ACCEPT;
	}

	datalen = (*pskb)->len - iph->ihl * 4 - tcph->doff * 4;
//	if( datalen < MIN_CONTEXT_LEN )	/* skip unneccessary packet */
//		return NF_ACCEPT;
	
	LOCK_BH(&ip_msnms_lock);
	/* If it's in the right range... */
	if (between(exp->seq + exp_msnms_info->len,
		ntohl(tcph->seq),
		ntohl(tcph->seq) + datalen)) {
		if (!msnms_data_fixup2(exp_msnms_info, ct, pskb, ctinfo, exp)) {
			UNLOCK_BH(&ip_msnms_lock);
			return NF_DROP;
		}
	} else {
		/* Half a match?  This means a partial retransmisison.
		   It's a cracker being funky. */
		if (net_ratelimit()) {
			printk("MSNMS_NAT: partial packet %u/%u in %u/%u\n",
			       exp->seq, exp_msnms_info->len,
			       ntohl(tcph->seq),
			       ntohl(tcph->seq) + datalen);
		}
		UNLOCK_BH(&ip_msnms_lock);
		return NF_DROP;
	}
	UNLOCK_BH(&ip_msnms_lock);

	return NF_ACCEPT;
}


static struct ip_nat_helper msnms[MAX_PORTS];
static struct ip_nat_helper hlpr;
static char msnms_names[MAX_PORTS][10];


/* Not __exit: called from init() */
static void fini(void)
{
	int i;

	for (i = 0; i < ports_c; i++) {
		DEBUGP("ip_nat_msnms: unregistering port %d\n", ports[i]);
		ip_nat_helper_unregister(&msnms[i]);
	}	
}

static int __init init(void)
{
	int i, ret = 0;
	char *tmpname;

	if (ports[0] == 0)
		ports[0] = MSNMS_PORT;

	for (i = 0; (i < MAX_PORTS) && ports[i]; i++) {

		memset(&msnms[i], 0, sizeof(struct ip_nat_helper));

		msnms[i].tuple.dst.protonum = IPPROTO_TCP;
		msnms[i].tuple.src.u.tcp.port = htons(ports[i]);
		msnms[i].mask.dst.protonum = 0xFFFF;
		msnms[i].mask.src.u.tcp.port = 0xFFFF;
		msnms[i].me = THIS_MODULE;
		msnms[i].flags = 0;//system crash if set to IP_NAT_HELPER_F_ALWAYS;
		msnms[i].help = msnms_nat_help;
		msnms[i].expect = msnms_nat_expected;

		tmpname = &msnms_names[i][0];
		if (ports[i] == MSNMS_PORT)
			sprintf(tmpname, "msnms");
		else
			sprintf(tmpname, "msnms-%d", i);
		msnms[i].name = tmpname;

		DEBUGP("ip_nat_msnms: register to port %d\n",ports[i]);
		ret = ip_nat_helper_register(&msnms[i]);

		if (ret) {
			printk("ip_nat_msnms: error registering "
			       "helper for port %d\n", ports[i]);
			fini();
			return ret;
		}
		ports_c++;
	}

	return ret;
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
